Package: tidyfun
Title: Tidy Functional Data Wrangling and Visualization
Version: 0.1.2
Authors@R: c(
    person("Fabian", "Scheipl", , "fabian.scheipl@googlemail.com", role = c("aut", "cre", "cph"),
           comment = c(ORCID = "0000-0001-8172-3603")),
    person("Jeff", "Goldsmith", role = "aut"),
    person("Julia", "Wrobel", role = "aut",
           comment = c(ORCID = "0000-0001-6783-1421")),
    person("Maximilian", "Mücke", role = "ctb",
           comment = c(ORCID = "0009-0000-9432-9795"))
  )
Description: Represent, visualize, describe and wrangle functional data
    in tidy data frames, building on the 'tf' package. Provides data types
    for functional observations that work as columns in data frames,
    enabling manipulation with 'dplyr' verbs and visualization with
    'ggplot2' geoms designed for functional data.
License: MIT + file LICENSE
URL: https://github.com/tidyfun/tidyfun,
        https://tidyfun.github.io/tidyfun/
BugReports: https://github.com/tidyfun/tidyfun/issues
Depends: R (>= 4.1), tf (>= 0.4.0)
Imports: cli, dplyr, GGally, ggplot2, grid, pillar, purrr, rlang,
        tibble, tidyr (>= 1.0.0), tidyselect (>= 1.0.0)
Suggests: covr, fda, fdasrvf, gridExtra, janitor, knitr, lme4, maps,
        mgcv, refund, rmarkdown, scales, testthat (>= 3.0.0),
        tidyverse, viridisLite
VignetteBuilder: knitr
Config/testthat/edition: 3
Encoding: UTF-8
LazyData: true
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-04-24 10:40:31 UTC; fabians
Author: Fabian Scheipl [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0001-8172-3603>),
  Jeff Goldsmith [aut],
  Julia Wrobel [aut] (ORCID: <https://orcid.org/0000-0001-6783-1421>),
  Maximilian Mücke [ctb] (ORCID: <https://orcid.org/0009-0000-9432-9795>)
Maintainer: Fabian Scheipl <fabian.scheipl@googlemail.com>
Repository: CRAN
Date/Publication: 2026-04-24 15:00:02 UTC
Built: R 4.4.3; ; 2026-04-24 16:46:40 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method("+",tf_ggplot)
S3method(autolayer,tf)
S3method(autoplot,tf)
S3method(ggplot_build,tf_ggplot)
S3method(is.finite,tf)
S3method(print,tf_ggplot)
S3method(scale_type,tf)
S3method(tf_evaluate,data.frame)
S3method(tf_unnest,data.frame)
S3method(tf_unnest,tf)
export(GeomCapellini)
export(GeomErrorband)
export(GeomFboxplot)
export(GeomMeatballs)
export(GeomSpaghetti)
export(StatCapellini)
export(StatErrorband)
export(StatFboxplot)
export(StatTf)
export(autolayer)
export(autoplot)
export(geom_capelini)
export(geom_capellini)
export(geom_cappelini)
export(geom_cappellini)
export(geom_errorband)
export(geom_fboxplot)
export(geom_meatballs)
export(geom_spaghetti)
export(gglasagna)
export(is_tf_ggplot)
export(parse_tf_aesthetics)
export(stat_capellini)
export(stat_errorband)
export(stat_fboxplot)
export(stat_tf)
export(tf_gather)
export(tf_ggplot)
export(tf_nest)
export(tf_spread)
export(tf_unnest)
import(dplyr)
import(ggplot2)
import(rlang)
import(tf)
importFrom(GGally,rescale01)
importFrom(GGally,rescale11)
importFrom(ggplot2,autolayer)
importFrom(ggplot2,autoplot)
importFrom(grid,gList)
importFrom(grid,gpar)
importFrom(grid,grobTree)
importFrom(grid,textGrob)
importFrom(grid,unit)
importFrom(pillar,obj_sum)
importFrom(pillar,pillar_shaft)
importFrom(pillar,type_sum)
importFrom(purrr,map)
importFrom(purrr,map_lgl)
importFrom(purrr,pmap)
importFrom(tibble,tibble)
importFrom(tidyr,unnest)
importFrom(tidyselect,eval_select)
importFrom(tidyselect,vars_pull)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# tidyfun <a href="https://tidyfun.github.io/tidyfun/"><img src="man/figures/logo.gif" align="right" height="150" alt="tidyfun website" /></a>

<!-- badges: start -->

[![Lifecycle:
experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental)
[![R-CMD-check](https://github.com/tidyfun/tidyfun/actions/workflows/full-check.yaml/badge.svg)](https://github.com/tidyfun/tidyfun/actions/workflows/full-check.yaml)
[![R-universe
version](https://tidyfun.r-universe.dev/tidyfun/badges/version)](https://tidyfun.r-universe.dev/tidyfun)
[![MIT
license](https://img.shields.io/badge/license-MIT-brightgreen.svg)](https://opensource.org/licenses/MIT)
[![codecov](https://codecov.io/gh/tidyfun/tidyfun/graph/badge.svg)](https://app.codecov.io/gh/tidyfun/tidyfun)
<!-- badges: end -->

**`tidyfun`** is a **`tidyverse`**-oriented interface layer around the
core **`tf`** package. It makes functional data analysis in `R` easier
by focusing on data wrangling, exploratory analysis, and graphics for
functional data stored in ordinary data frames.

The underlying **`tf`** package defines the `tf` vector type and the
core functional-data methods. **`tidyfun`** builds on those objects with
helpers for working inside data frames, plus plotting and workflow tools
that fit naturally into the **`tidyverse`**.

**Crucially**, vectors of class `tf` can be included in data frames
alongside other variables, enabling data manipulation with standard
**`tidyverse`** tools. This matches the functional-data view that
*complete functions* are the unit of observation; with **`tidyfun`**,
full curves sit alongside numeric, factor, and other observations on the
same subject.

## Installation

Install the released version from CRAN:

``` r
install.packages("tidyfun")
```

Or install the current development version from GitHub:

``` r
install.packages("pak")
pak::pak("tidyfun/tidyfun")
```

## Overview

**`tidyfun`** provides:

- **`tidyverse`**-friendly **wrangling** tools for functional data
  inside data frames
- **`tidyfun`**-level **graphics** helpers for `tf` vectors and tidy
  functional data frames
- documentation and examples that sit on top of the core **`tf`** data
  structures and methods

For detailed information on the features of **`tidyfun`**, check out
articles on the following topics:

- [Representing](https://tidyfun.github.io/tidyfun/articles/x01_tf_Vectors.html)
  functional data as `tf` vectors, and operating on those vectors with
  core **`tf`** methods and **`tidyfun`** conveniences
- [Converting](https://tidyfun.github.io/tidyfun/articles/x02_Conversion.html)
  non-tidy functional data (matrices, “long” and “wide” data frames) to
  tidy functional data using **`tidyfun`** workflows built on **`tf`**
- [Wrangling](https://tidyfun.github.io/tidyfun/articles/x03_Data_Wrangling.html)
  data frames that include functional data using **`tidyverse`** and
  **`tidyfun`** tools
- [Visualizing](https://tidyfun.github.io/tidyfun/articles/x04_Visualization.html)
  tidy functional data with **`tidyfun`** graphics helpers
- [Registering](https://tidyfun.github.io/tidyfun/articles/x06_Registration.html)
  functional data with the **`tf`** registration methods documented for
  use in the **`tidyfun`** ecosystem

The result is a package that enables exploratory data analysis like the
following, which computes group-specific mean curves in the `dti_df`
dataset, and plots the result:

``` r
library("tidyfun")
data(dti_df, package = "tidyfun")

dti_df |>
  group_by(case, sex) |>
  summarize(mean_cca = mean(cca, na.rm = TRUE)) |>
  tf_ggplot(aes(tf = mean_cca, color = case)) +
  geom_line(linewidth = 2) +
  facet_grid(~sex)
```

<img src="man/figures/README-dti-ggplot-example-1.png" alt="" width="100%" />

## What does it do?

#### Core `tf` data types and methods

The underlying [**`{tf}`**](https://github.com/tidyfun/tf) package
defines the `tf` vector classes and related methods, including raw
functional data (`tfd`) and basis representations (`tfb`). For an
overview, see the [`tf` vectors
article](https://tidyfun.github.io/tidyfun/articles/x01_tf_Vectors.html)
and the [reference index](https://tidyfun.github.io/tidyfun/reference/).

Such `tf`-objects can be subsetted or subassigned, computed on and
summarized.

Almost all

- operators like `==`, `+` or `*`
- math functions like `sum`, `log` or `abs`
- and statistics functions like `mean` or `sd`

are defined for these `tf` data structures, as illustrated in the [`tf`
vectors
article](https://tidyfun.github.io/tidyfun/articles/x01_tf_Vectors.html)
and the [reference index](https://tidyfun.github.io/tidyfun/reference/).

The `tf` objects are basically glorified lists, so they work well as
columns in data frames. That makes it a lot easier to keep your other
data and functional measurements together in one object for
preprocessing, exploratory analysis and description. At the same time,
these objects actually behave like vectors of *functions* to some
extent, i.e., they can be evaluated on any point in their domain, they
can be integrated or differentiated, etc.

[See
here](https://tidyfun.github.io/tidyfun/articles/x01_tf_Vectors.html)
for more information on the operations defined for `tf` vectors.

#### Methods for converting existing data to `tf`

The underlying **`tf`** package provides the `tfd` and `tfb`
constructors for converting matrices, data frames, etc. to `tf` vectors.
**`tidyfun`** builds on those objects with `tf_gather` & `tf_nest` to
reshape tables with functional data, by going from wide to narrow or
from long to short; functions like `as.matrix`, `tf_spread` &
`tf_unnest` can reverse these data conversions.

[See
here](https://tidyfun.github.io/tidyfun/articles/x02_Conversion.html)
for details on getting data into (and out of) the `tf` format.

#### **`tidyverse`** verbs for dealing with functional data inside data frames

All **`dplyr`** verbs work on `tf`-columns, so you can `filter`,
`mutate`, `group_by` & `summarize`, etc., functional data pretty much
like conventional “tidy” data. **`tidyfun`** workflows also work
naturally with **`tf`** helpers such as `tf_anywhere` and `tf_smooth`.

[See
here](https://tidyfun.github.io/tidyfun/articles/x03_Data_Wrangling.html)
to see how you can wrangle functional data.

#### New **`ggplot2`** `geoms` and `stats` for functional data

**`tidyfun`** provides **`tf_ggplot()`**, which lets you use standard
**`ggplot2`** geoms (`geom_line`, `geom_point`, `geom_ribbon`, …) with
functional data via the `tf` aesthetic. It also includes specialized
**pasta-themed** geoms:

- **`gglasagna`** for [lasagna
  plots](https://asset.jmir.pub/assets/76aeec48564abf0e6f6da8e9cd06346d.png)
  (heatmaps), with an **`order`**-aesthetic to sort the lasagna layers,
- **`geom_capellini`** for
  [glyphs](http://ggobi.github.io/ggally/articles/glyph_files/figure-html/glyphs-basic-usage-1.png)
  plots (i.e., sparklines),
- **`geom_fboxplot`** for functional boxplots based on data depth,

as well as methods for base R graphics functions `plot`, `lines` and
`points` for quick and easy visualizations of functional data.

[See
here](https://tidyfun.github.io/tidyfun/articles/x04_Visualization.html)
for the documentation of the visualization approaches, or browse the
[reference index](https://tidyfun.github.io/tidyfun/reference/).

#### Curve registration

**`tidyfun`** documents multiple **`tf`** methods for aligning curves in
time via `tf_register`: elastic SRVF alignment, continuous registration
(CC), affine shift/scale models, and landmark registration. All methods
return `tf_registration` objects with a consistent interface for
extracting aligned curves, inverse warping functions, and diagnostic
summaries.

[See
here](https://tidyfun.github.io/tidyfun/articles/x06_Registration.html)
for a practical guide covering method selection, diagnostics, and worked
examples on real data.

------------------------------------------------------------------------

Found a bug? Got a question? Missing some functionality?  
Please let us know so we can make it better.

# Visual Testing for tf_ggplot

This directory contains comprehensive visual testing tools for the tf_ggplot functionality. These tools help ensure that tf_ggplot produces sensible, correct, and visually appealing plots.

## Overview

The visual testing system provides three levels of testing:

1. **Unit Tests with Visual Validation** - Standard testthat tests that also save plots
2. **Interactive Development Testing** - Quick tests for development workflow  
3. **Comprehensive Visual Reports** - Automated HTML reports with all test plots

## Files

### Test Files
- `test-tf-ggplot-visual.R` - Main visual tests with plot generation
- `test-tf-ggplot-basic.R` - Basic functionality tests
- `test-tf-ggplot-advanced.R` - Advanced feature tests
- `test-tf-ggplot-aesthetic-equivalence.R` - Aesthetic specification equivalence tests
- `helper-tf-ggplot.R` - Test utilities and helper functions

### Development Tools
- `visual-test-runner.R` - Comprehensive test runner with HTML report generation
- `quick-visual-check.R` - Interactive development testing tool
- `README-visual-testing.md` - This documentation

## Usage

### 1. Quick Development Testing

For rapid development and debugging:

```r
# Load the quick testing tools
source("tests/quick-visual-check.R")

# Run interactive test menu
run_interactive_tests()

# Or run specific tests
test_basic_line()          # Test basic line plots
test_ribbon_plot()         # Test confidence bands
test_faceting()           # Test faceted plots
test_multiple_aesthetics() # Test complex aesthetic mappings
```

### 2. Comprehensive Visual Testing

For thorough testing with HTML reports:

```r
# Load the test runner
source("tests/visual-test-runner.R")

# Run all tests and create HTML report
run_tf_ggplot_visual_tests()

# Options:
run_tf_ggplot_visual_tests(
  interactive_display = TRUE,    # Show plots as they're created
  create_report = TRUE,          # Generate HTML report (default)
  open_report = TRUE             # Auto-open report in browser (default)
)

# Just create report from existing plots
create_visual_report()
```

### 3. Standard Unit Testing

Run through testthat:

```r
# Run specific visual test file
testthat::test_file("tests/testthat/test-tf-ggplot-visual.R")

# Run all tf_ggplot tests
testthat::test_dir("tests/testthat", filter = "tf-ggplot")
```

## Output Structure

The visual testing system creates organized output:

```
tests/
├── visual-output/
│   ├── plots/                    # Individual plot PNG files
│   │   ├── 01_basic_line_constructor.png
│   │   ├── 02_basic_line_geom.png
│   │   ├── ...
│   │   └── 17_wide_range.png
│   ├── reports/                  # HTML reports
│   │   └── visual_test_report.html
│   └── comparisons/              # Side-by-side comparisons
└── test-plots/                   # Basic test output (legacy)
```

## Test Categories

### Basic Functionality
- Constructor with different aesthetic specifications
- Simple line plots with grouping
- Basic geom integration

### Advanced Features
- Multiple tf aesthetics (tf_x, tf_y)
- Ribbon plots with confidence bands
- Faceting with tf data
- Complex aesthetic mappings
- Different geom types (line, point, step, area)

### Comparisons
- tf_ggplot vs current geom_spaghetti
- Constructor vs geom-level aesthetic specification
- Different evaluation grids

### Edge Cases
- Single functions
- Very sparse data (few evaluation points)
- Many functions (visual clutter)
- Wide evaluation ranges
- Irregular tf objects
- NA/empty functions

## Visual Validation Checklist

When reviewing plots, check for:

### ✅ Correctness
- [ ] Functions are plotted with correct x/y values
- [ ] Grouping is preserved (separate lines per function)
- [ ] Colors/aesthetics are applied consistently within groups
- [ ] Faceting splits data correctly
- [ ] Evaluation grids are respected

### ✅ Visual Quality
- [ ] Lines are smooth and continuous
- [ ] Points are clearly visible when used
- [ ] Ribbons show appropriate confidence regions
- [ ] Colors are distinguishable
- [ ] Transparency/alpha works correctly
- [ ] Text and labels are readable

### ✅ Equivalence
- [ ] tf_ggplot plots match geom_spaghetti reference plots
- [ ] Constructor aesthetics ≡ geom-level aesthetics
- [ ] Different geom types produce expected shapes
- [ ] Custom evaluation grids work correctly

### ✅ Edge Cases
- [ ] Single functions render without errors
- [ ] Sparse data shows appropriate point gaps
- [ ] Many functions are readable with transparency
- [ ] Wide ranges don't cause axis problems
- [ ] Irregular data handles missing points gracefully

## Development Workflow

### 1. Initial Implementation
```r
# Quick check during development
source("tests/quick-visual-check.R")
test_basic_line()  # Will show current geom_spaghetti for reference
```

### 2. Feature Testing
```r
# Test specific features as you implement them
test_ribbon_plot()         # For ribbon geom support
test_multiple_aesthetics() # For complex aesthetic handling
```

### 3. Comprehensive Validation
```r
# Full test suite with report
source("tests/visual-test-runner.R")
run_tf_ggplot_visual_tests()
# Review HTML report for all plots
```

### 4. Regression Testing
```r
# Standard unit tests
testthat::test_dir("tests/testthat")
```

## Customization

### Adding New Visual Tests

Add tests to `test-tf-ggplot-visual.R`:

```r
test_that("my new feature looks correct", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()
  
  # Create test data
  data <- create_test_tf_data(...)
  
  # Create plot
  p <- tf_ggplot(data, aes(...)) + geom_xxx(...)
  
  # Save for visual inspection
  save_test_plot(p, "my_new_feature")
  
  # Basic validation
  expect_s3_class(p, "ggplot")
  # ... other checks
})
```

### Custom Plot Categories

Modify `save_visual_test_plot()` in `visual-test-runner.R` to add new categories.

### Integration with CI/CD

For automated testing, disable interactive features:

```r
run_tf_ggplot_visual_tests(
  interactive_display = FALSE,
  create_report = TRUE,
  open_report = FALSE
)
```

## Troubleshooting

### Common Issues

1. **Plots not saving**: Check write permissions in tests/ directory
2. **HTML report not opening**: Set `open_report = FALSE` and open manually
3. **Missing tf functions**: Load tf/tidyfun packages first
4. **Memory issues**: Test with smaller datasets during development

### Performance

For large test suites:
- Use coarser evaluation grids during development
- Limit number of functions in test data
- Use `ggsave()` with lower DPI for faster saving

### Platform-Specific Notes

- **Windows**: HTML reports open with default browser
- **macOS/Linux**: Uses `open`/`xdg-open` commands
- **Server/headless**: Set `interactive_display = FALSE`

## Contributing

When adding new tf_ggplot features:

1. Add corresponding visual tests
2. Update this README if new test categories are added
3. Ensure visual tests pass on your development platform
4. Include example plots in PR descriptions when possible

The visual testing system is designed to grow with the tf_ggplot functionality and provide confidence that the implementation produces correct, beautiful functional data visualizations.

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "##",
  fig.width = 6,
  fig.height = 4,
  dpi = 72,
  fig.retina = 1,
  out.width = "90%"
)

library(tidyverse)
library(viridisLite)

theme_set(theme_minimal() + theme(legend.position = "bottom"))

options(
  ggplot2.continuous.colour = "viridis",
  ggplot2.continuous.fill = "viridis",
  ggplot2.discrete.colour = "viridis_d",
  ggplot2.discrete.fill = "viridis_d"
)

library("tidyfun")
data(chf_df, package = "tidyfun")
data(dti_df, package = "tidyfun")

pal_5 <- viridis(7)[-(1:2)]
set.seed(1221)

## ----plot_chf-----------------------------------------------------------------
dti_df[1:10,] |>
  tf_ggplot(aes(tf = cca)) + geom_line(alpha = .3)

## -----------------------------------------------------------------------------
dti_df[1:3,] |>
  tf_ggplot(aes(tf = rcst)) + geom_line(alpha = .3) + geom_point(alpha= .3)

## -----------------------------------------------------------------------------
chf_df |>
  filter(id %in% 1:5) |>
  tf_ggplot(
    aes(tf = tf_smooth(activity, f = .05), # smoothed inputs for clearer viz
        color = gender)) +
  geom_line(alpha = 0.3)

## -----------------------------------------------------------------------------
chf_df |>
  filter((id %in% 1:10) & (day %in% c("Mon", "Sun"))) |>
  tf_ggplot(aes(tf = tf_smooth(activity, f = .05), color = gender)) +
  geom_line(alpha = 0.3, lwd = 1) +
  facet_grid(~day)

## ----dti-fig1-----------------------------------------------------------------
dti_df |>
  tf_ggplot(aes(tf = cca, col = case, alpha = 0.2 + 0.4 * (case == "control"))) +
  geom_line() + facet_wrap(~sex) +
  scale_alpha(guide = "none", range = c(0.2, 0.4))

## -----------------------------------------------------------------------------
chf_df |>
  group_by(gender, day) |>
  summarize(mean_act = mean(activity),
            .groups = "drop_last") |>
  mutate(smooth_mean = tfb(mean_act, verbose = FALSE)) |>
  filter(day %in% c("Mon", "Sun")) |>
  tf_ggplot(aes(color = gender)) +
  geom_line(aes(tf = smooth_mean), linewidth = 1.25) +
  geom_line(aes(tf = mean_act), alpha = 0.1) +
  geom_point(aes(tf = mean_act), alpha = 0.1, size = .1) +
  facet_grid(day~.)

## -----------------------------------------------------------------------------
dti_df |>
  group_by(sex, case) |>
  summarize(
    mean_cca = mean(tfb(cca, verbose = FALSE)), #pointwise mean function
    sd_cca = sd(tfb(cca, verbose = FALSE)), # pointwise sd function
    .groups = "drop_last"
  ) |>
  group_by(sex, case) |>
  mutate(
    upper_cca = mean_cca + 2 * sd_cca,
    lower_cca = mean_cca - 2 * sd_cca
  ) |>
  tf_ggplot() +
  geom_line(aes(tf = mean_cca, color = sex)) +
  geom_ribbon(aes(tf_ymin = lower_cca, tf_ymax = upper_cca, fill = sex), alpha = 0.3) +
  facet_grid(sex ~ case)

## -----------------------------------------------------------------------------
dti_df |> 
  tf_ggplot(aes(tf = cca, fill = case)) +
  geom_fboxplot(alpha = 0.35) +
  facet_grid(~ sex) + labs(title="MBD-based boxplot")

## -----------------------------------------------------------------------------
dti_df |>
  tf_ggplot(aes(tf = cca, colour = case)) +
  geom_fboxplot(depth = "FM", alpha = 0.3) +
  facet_grid(~ sex) + labs(title="FM-based boxplot")

## -----------------------------------------------------------------------------
dti_df |>
  tf_ggplot(aes(tf = cca, colour = case)) +
  geom_fboxplot(depth = "RPD", alpha = 0.3) +
  facet_grid(~ sex) + labs(title="RPD-based boxplot")

## -----------------------------------------------------------------------------
tf_ggplot(dti_df, aes(tf = rcst)) + geom_fboxplot()

## -----------------------------------------------------------------------------
tf_ggplot(dti_df, aes(tf = rcst)) + 
  geom_fboxplot(alpha = .5)
tf_ggplot(dti_df, aes(tf = rcst)) + 
  geom_fboxplot(alpha = .5, central = .2)
tf_ggplot(dti_df, aes(tf = rcst)) + 
  geom_fboxplot(alpha = .5, central = .2, outliers = FALSE)
tf_ggplot(dti_df, aes(tf = rcst)) +
  geom_fboxplot(orientation = "y", alpha = .3)

## -----------------------------------------------------------------------------
chf_df |>
  filter(day %in% c("Mon", "Sun")) |>
  gglasagna(activity)

## ----dti-fig2-----------------------------------------------------------------
dti_df |>
  gglasagna(
    tf = cca,
    order = tf_integrate(cca, definite = TRUE), #order by area under the curve
    arg = seq(0, 1, length.out = 101)
  ) +
  theme(axis.text.y = element_text(size = 6)) +
  facet_wrap(~ case:sex, ncol = 2, scales = "free")

## -----------------------------------------------------------------------------
canada <- data.frame(
  place = fda::CanadianWeather$place,
  region = fda::CanadianWeather$region,
  lat = fda::CanadianWeather$coordinates[, 1],
  lon = -fda::CanadianWeather$coordinates[, 2]
)

canada$temp <- tfd(t(fda::CanadianWeather$dailyAv[, , 1]), arg = 1:365)
canada$precipl10 <- tfd(t(fda::CanadianWeather$dailyAv[, , 3]), arg = 1:365) |>
  tf_smooth()

canada_map <-
  data.frame(maps::map("world", "Canada", plot = FALSE)[c("x", "y")])

## -----------------------------------------------------------------------------
ggplot(canada, aes(x = lon, y = lat)) +
  geom_capellini(aes(tf = precipl10),
    width = 4, height = 5, colour = "blue",
    line.linetype = 1
  ) +
  geom_capellini(aes(tf = temp),
    width = 4, height = 5, colour = "red",
    line.linetype = 1
  ) +
  geom_path(data = canada_map, aes(x = x, y = y), alpha = 0.1) +
  coord_quickmap()

## ----warning=FALSE------------------------------------------------------------
cca_fpc_tbl <- tibble(
  cca = dti_df$cca[1:30],
  cca_fpc = tfb_fpc(cca, pve = .8), 
  fpc_1 = map(coef(cca_fpc), 2) |> unlist(), # 1st PC loading
  fpc_2 = map(coef(cca_fpc), 3) |> unlist() # 2nd PC loading
) 
# rescale FPCs by sqrt of eigenvalues for visualization
cca_fpcs_1_2 <- 
  tf_basis(cca_fpc_tbl$cca_fpc, as_tfd = TRUE)[2:3] * 
    sqrt(attr(cca_fpc_tbl$cca_fpc, "score_variance")[1:2]) 
# scaled eigenfunctions look like this:
tibble(
   eigenfunction = cca_fpcs_1_2,
   FPC = factor(1:2)
) |> tf_ggplot() + 
  geom_line(aes(tf = eigenfunction, col = FPC)) + 
  geom_hline(yintercept = 0)

## ----warning=FALSE------------------------------------------------------------
ggplot(cca_fpc_tbl[1:40,], aes(x =  fpc_1, y = fpc_2)) +
  geom_point(size = .5, col = viridis(3)[2]) +
  geom_capellini(aes(tf =cca_fpc),width = .01, height = .01, line.linetype = 1) +
  labs(x = "FPC1 score", y = "FPC2 score")

## -----------------------------------------------------------------------------
cca <- dti_df$cca |>
  tfd(arg = seq(0, 1, length.out = 93), interpolate = TRUE)

layout(t(1:2))

plot(cca, type = "spaghetti")
lines(c(median(cca), mean = mean(cca)), col = viridis(3)[c(1, 3)])

plot(cca, type = "lasagna", col = viridis(50))

## ----ex-fig2------------------------------------------------------------------
cca_five <- cca[1:5]

cca_five |> plot(xlim = c(-0.15, 1), col = pal_5, lwd = 2)

text(
  x = -0.1, y = cca_five[, 0.07], labels = names(cca_five), col = pal_5, cex = 1
)

median(cca_five) |> lines(col = pal_5[3], lwd = 4)

## -----------------------------------------------------------------------------
pinch_reg <- tf::pinch |> tfb() |> #smooth before registration for better results
  tf_register() 
pinch_reg
summary(pinch_reg)
plot(pinch_reg)

## -----------------------------------------------------------------------------
layout(t(1:3))
plot(tf::pinch[1:5], col = pal_5, lwd = 2, points = FALSE)

plot(tf_inv_warps(pinch_reg)[1:5], col = pal_5, lwd = 2, points = FALSE)
abline(c(0, 1), col = "grey", lty = 2)

plot(tf_aligned(pinch_reg)[1:5], col = pal_5, lwd = 2)
lines(tf_template(pinch_reg), col = "black", lwd = 3, lty= 3)

set.seed(17711L)

smoo <- tf_rgp(10, nugget = 0)
rough <- tf_rgp(10, arg = 121L, nugget = 0.2, scale = 0.005)
narrow <- tf_jiggle(tf_rgp(10, arg = 11L, nugget = 0))
irr <- tf_sparsify(tf_jiggle(smoo))
sparse <- tf_sparsify(smoo)

smoo_list <- tf_evaluations(smoo)
smoo_matrix <- as.matrix(smoo)
smoo_df <- tf_unnest(smoo)

irr_list <- tf_evaluations(irr)
irr_matrix <- suppressWarnings(as.matrix(irr))
irr_df <- tf_unnest(irr)
narrow_df <- tf_unnest(narrow)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  cache = TRUE,
  collapse = TRUE,
  comment = "#>",
  dpi = 72,
  fig.retina = 1
)
library(tf)
oldpar <- par(no.readonly = TRUE)
par(las = 1)
alpha_palette <- c(
  "#00000080", "#DF536B80", "#61D04F80", "#2297E680", "#28E2E580",
  "#CD0BBC80", "#F5C71080", "#9E9E9E80", "#00000080", "#DF536B80"
)
have_srvf <- rlang::is_installed("fdasrvf")
have_cc <- TRUE

## ----eval=FALSE, echo = TRUE, class.source = "fold-show"----------------------
# # One-shot registration (returns tf_registration object):
# reg <- tf_register(x, method = "...")
# tf_aligned(reg)   # registered/aligned curves
# tf_inv_warps(reg)     # estimated inverse warping functions (observed -> aligned time)
# tf_template(reg)  # template used
# summary(reg)      # alignment diagnostics
# plot(reg)          # 3-panel diagnostic plot
# 
# # Or step by step:
# warps <- tf_estimate_warps(x, method = "...")
# x_registered <- tf_align(x, warps)

## ----warp-illustration, echo = FALSE, fig.width = 6, fig.height = 6-----------
# Synthetic illustration of different warp types
s <- seq(0, 1, length.out = 201)

# Template: 
template_synth <- tfd(exp(-((s - 0.45)^2) / (2 * 0.08^2)) + cos(2*pi*s) + s, arg = s)

# Warp 1: affine shift + scale (not domain-preserving -- extends beyond [0,1])
h_shift <- tfd(0.8 * s + 0.1, arg = s)

# Warp 2: smooth diffeomorphism (domain-preserving)
h_smooth_raw <- s - 0.25 * sin(pi * s)
h_smooth <- tfd(
  (h_smooth_raw - h_smooth_raw[1]) / (h_smooth_raw[length(s)] - h_smooth_raw[1]),
  arg = s)

# Warp 3: wiggly diffeomorphism (domain-preserving)
h_wiggly_raw <- .1 * pbeta(s, .1, .1) + .5 * pbeta(s, 25, 25) + 
  .2 * pbeta(s, 4, 20) + .2 * pbeta(s, 100, 10)
h_wiggly <- tfd(h_wiggly_raw, arg = s)  

# Compose template with warps: m(h(s))
warped_shift <- tf_warp(template_synth, h_shift) |> suppressWarnings()
warped_smooth <- tf_warp(template_synth, h_smooth)
warped_wiggly <- tf_warp(template_synth, h_wiggly)

warp_cols <- c("#DF536B", "#2297E6", "#61D04F")

layout(matrix(1:9, 3, 3))
par(mar = c(4, 4, 3, 1))

warp_names <- c("Affine (shift & scale)", "Smooth elastic", "Wiggly elastic")
warps <- list(h_shift, h_smooth, h_wiggly)
warped <- list(warped_shift, warped_smooth, warped_wiggly)

# Column 1: blank, template, blank
plot.new()
plot(template_synth, lwd = 2.5,
     main = "Template m(s)", xlab = "s", ylab = "m(s)")
plot.new()

# Column 2: one warp per row
for (i in 1:3) {
  plot(s, s, type = "l", lty = 3, lwd = 1,
       main = paste0("h(s): ", warp_names[[i]]),
       xlab = "s", ylab = "h(s)", ylim = c(-0.05, 1.15))
  do.call(graphics::lines, list(warps[[i]], col = warp_cols[[i]], lwd = 2.5))
}

# Column 3: template (dashed) + warped curve per row
for (i in 1:3) {
  plot(template_synth, lwd = 2, lty = 3,
       main = paste0("x(t) = m(h(s)): ", warp_names[[i]]),
       xlab = "t", ylab = "x(t)")
  do.call(graphics::lines, list(warped[[i]], col = warp_cols[[i]], lwd = 2.5))
}

## ----quickstart, class.source = "fold-show"-----------------------------------
pinch_small <- pinch[1:10]
template_affine <- pinch[7] |> tf_smooth(f= .2)

reg_shift <- tf_register(
  pinch_small,
  method = "affine",
  template = template_affine,
  type = "shift"
)
reg_shift
summary(reg_shift)
pinch_inv_warp_shift <- tf_inv_warps(reg_shift)
pinch_small_reg <- tf_aligned(reg_shift)
pinch_template_shift <- tf_template(reg_shift)

## ----quickstart-plot, echo = FALSE, class.source = "fold-show", fig.width = 6, fig.height = 3----
# plot(reg_shift) gives a compact diagnostic; below reuses the extracted
# components for a customized version of the same view:
layout(t(1:3))
plot(
  pinch_small,
  main = "Original Curves",
  xlab = "Observed Time t [sec]",
  ylab = "Force [N]",
  lwd = 1.5,
  col = alpha_palette
)
lines(pinch_template_shift, lwd = 2, lty = 3, col = rgb(0,0,.5))
plot(
  pinch_inv_warp_shift,
  points = FALSE,
  main = "Inverse Warping Functions",
  xlab = "Observed Time t [sec]",
  ylab = "Registered Time s [sec]",
  lwd = 1.5,
  col = alpha_palette
)
abline(0, 1, lty = 3)
plot(
  pinch_small_reg,
  main = "Registered Curves",
  xlab = "Registered Time s [sec]",
  ylab = "Force [N]",
  lwd = 1.5,
  col = alpha_palette, 
  points = FALSE
)
lines(pinch_template_shift, lwd = 2, lty = 3, col = rgb(0,0,.5))

## ----template-choice, fig.width = 6, fig.height = 4---------------------------
# Gaussian bumps with large shifts:
s <- seq(-4, 6, length.out = 201)
mus <- c(-2, -1, 0, 1, 2)
bumps <- tfd(t(sapply(mus, \(mu) dnorm(s, mu, sd = 0.5))), arg = s)

# Pointwise mean is smeared -- not a good template:
bumps_mean <- mean(bumps)
# MBD median picks the most central observed curve:
bumps_median <- median(bumps, depth = "MBD")

# Register with default (mean) vs MBD median template:
reg_mean <- tf_register(bumps, method = "affine", type = "shift",
                        template = bumps_mean)
reg_median <- tf_register(bumps, method = "affine", type = "shift",
                          template = bumps_median)

layout(matrix(1:6, 2, 3, byrow = TRUE))
plot(bumps, col = alpha_palette, lwd = 1.5,
     main = "Original + mean template")
lines(bumps_mean, lwd = 3, lty = 2)
plot(tf_inv_warps(reg_mean), col = alpha_palette, lwd = 1.5, points = FALSE,
     main = "Inverse warps (mean template)",
     xlab = "Observed time t", ylab = "Aligned time s")
abline(0, 1, lty = 3)
plot(tf_aligned(reg_mean), col = alpha_palette, lwd = 1.5, points = FALSE,
     main = "Aligned to mean")
lines(bumps_mean, lwd = 3, lty = 2)


plot(bumps, col = alpha_palette, lwd = 1.5,
     main = "Original + MBD median template")
lines(bumps_median, lwd = 3, lty = 2, col = "red3")
plot(tf_inv_warps(reg_median), col = alpha_palette, lwd = 1.5, points = FALSE,
     main = "Inverse warps (median template)",
     xlab = "Observed time t", ylab = "Aligned time s")
abline(0, 1, lty = 3)
plot(tf_aligned(reg_median), col = alpha_palette, lwd = 1.5, points = FALSE,
     main = "Aligned to MBD median")
lines(bumps_median, lwd = 3, lty = 2, col = "red3")

## ----diagnostics, class.source = "fold-show"----------------------------------
x <- pinch[1:10]

# Register with affine shift warps:
reg_aff <- tf_register(x, method = "affine", type = "shift_scale")

# summary() gives a quick quantitative overview:
summary(reg_aff)

## ----diagnostics-plot, fig.width = 6, fig.height = 3, class.source = "fold-show"----
# plot() provides a 3-panel diagnostic:
plot(reg_aff)

## ----diagnostics-features, class.source = "fold-show"-------------------------
# Are global peak locations more concentrated after alignment?
peak_before <- tf_where(x, value == max(value)) |> as.numeric()
peak_after_aff <- tf_where(tf_aligned(reg_aff), value == max(value)) |> as.numeric()

data.frame(
  metric = c("sd_peak_before", "sd_peak_after_affine"),
  value = c(sd(peak_before), sd(peak_after_aff))
)

## ----diagnostics-optional-srvf, eval = have_srvf, include = have_srvf, class.source = "fold-show"----
# Compare with SRVF (non-linear warps):
reg_srvf <- tf_register(x, method = "srvf")

# Side-by-side summaries:
summary(reg_srvf)

## ----diagnostics-optional-srvf-compare, eval = have_srvf, include = have_srvf, class.source = "fold-show"----
peak_after_srvf <- tf_where(tf_aligned(reg_srvf), value == max(value)) |> as.numeric()

data.frame(
  metric = c("sd_peak_before", "sd_peak_after_affine", "sd_peak_after_srvf"),
  value = c(sd(peak_before), sd(peak_after_aff), sd(peak_after_srvf))
)

## ----diagnostics-optional-srvf-plot, eval = have_srvf, include = have_srvf, fig.width = 6, fig.height = 4----
layout(matrix(1:6, 2, 3, byrow = TRUE))
# Affine:
plot(x, main = "Original", col = alpha_palette, lwd = 1.5)
lines(tf_template(reg_aff), lwd = 2, lty = 2)
plot(tf_inv_warps(reg_aff), main = "Affine Inverse Warps",
     col = alpha_palette, lwd = 1.5,  points = FALSE,
     xlab = "Observed time t", ylab = "Aligned time s")
abline(0, 1, lty = 3)
plot(tf_aligned(reg_aff), main = "Affine Aligned",
     col = alpha_palette, lwd = 1.5, points = FALSE)
lines(tf_template(reg_aff), lwd = 2, lty = 2)
# SRVF:
plot(x, main = "Original", col = alpha_palette, lwd = 1.5)
lines(tf_template(reg_srvf), lwd = 2, lty = 2)
plot(tf_inv_warps(reg_srvf), main = "SRVF Inverse Warps",
     col = alpha_palette, lwd = 1.5,   points = FALSE,
     xlab = "Observed time t", ylab = "Aligned time s")
abline(0, 1, lty = 3)
plot(tf_aligned(reg_srvf), main = "SRVF Aligned",
     col = alpha_palette, lwd = 1.5, points = FALSE)
lines(tf_template(reg_srvf), lwd = 2, lty = 2)

## ----pinch-compare, class.source = "fold-show"--------------------------------
pinch_small <- pinch[1:10]

reg_aff <- tf_register(pinch_small, method = "affine", type = "shift_scale")
inv_warp_aff <- tf_inv_warps(reg_aff)

# tf_landmarks_extrema needs smoothed inputs,
#     otherwise it tends to detect lots of spurious features:
pinch_small |> tf_landmarks_extrema(which = "max") |> head()
#     ... so in this case, we simply use the global maximum for each curve:
(peak_locs <- pinch_small |> tf_where(value == max(value)) |> unlist() |> as.matrix())
reg_lm <- tf_register(pinch_small, method = "landmark", landmarks = peak_locs)
inv_warp_lm <- tf_inv_warps(reg_lm)

# ... but using more than one peak location as the only landmark will get us
# better alignment here - use times where curve first & last exceeds 3 as well:
pinch_landmarks  <- cbind(
  start = pinch_small |> tf_where(value > 3, return = "first") |> unlist(),
  peak  = peak_locs,
  end   = pinch_small |> tf_where(value > 3, return = "last") |> unlist()
)
pinch_landmarks |> head()
reg_lm2 <- tf_register(pinch_small, method = "landmark", landmarks = pinch_landmarks)
inv_warp_lm2 <- tf_inv_warps(reg_lm2)

## ----pinch-compare-plot, fig.width = 6, fig.height = 6------------------------
layout(t(matrix(1:12, 4, 3)))
par(cex.main = 0.8)
plot.new()
plot(inv_warp_aff, main = "Affine Inverse Warps", ylab = "",
     points = FALSE, col = alpha_palette, lwd = 1.5)
abline(0, 1, lty = 3)
plot(tf_aligned(reg_aff), main = "Affine Registered",
     col = alpha_palette, lwd = 1.5, points = FALSE)
plot(tf_aligned(reg_aff), main = "Affine Registered",
     type = "lasagna", col = hcl.colors(12, rev = TRUE))

plot(pinch_small, main = "Original",
     col = alpha_palette, lwd = 1.5)
plot(inv_warp_lm, main = "Landmark Inverse Warps \n (Peak only)", ylab = "",
     points = FALSE, col = alpha_palette, lwd = 1.5)
abline(0, 1, lty = 3)
plot(tf_aligned(reg_lm), main = "Landmark Registered\n (Peak only)", col = alpha_palette, lwd = 1.5)
plot(tf_aligned(reg_lm), main = "Landmark Registered\n (Peak only)", type = "lasagna", col = hcl.colors(12, rev = TRUE))

plot.new()
plot(inv_warp_lm2, main = "Landmark Inverse Warps \n (Start + Peak + End)", ylab = "",
     points = FALSE, col = alpha_palette, lwd = 1.5)
abline(0, 1, lty = 3)
plot(tf_aligned(reg_lm2), main = "Landmark Registered\n (Start + Peak + End)", col = alpha_palette, lwd = 1.5)
plot(tf_aligned(reg_lm2), main = "Landmark Registered\n (Start + Peak + End)", type = "lasagna", col = hcl.colors(12, rev = TRUE))

## ----pinch-compare-srvf-cc, eval = have_srvf && have_cc, class.source = "fold-show"----
reg_srvf <- tf_register(pinch_small, method = "srvf")
inv_warp_srvf <- tf_inv_warps(reg_srvf)
reg_cc_unpen <- tf_register(pinch_small, method = "cc", max_iter = 10, nbasis = 10, crit = 1)
inv_warp_cc_unpen <- tf_inv_warps(reg_cc_unpen)
reg_cc_pen <- tf_register(pinch_small, method = "cc", lambda = 1e-4, max_iter = 20)
inv_warp_cc_pen <- tf_inv_warps(reg_cc_pen)

## ----pinch-compare-srvf-cc-summary, eval = have_srvf && have_cc, class.source = "fold-show"----
summary(reg_cc_unpen)
# ... ouch! max slope almost 100 and only 40% amplitude variance reduction ....

summary(reg_srvf)
summary(reg_cc_pen)

## ----pinch-compare-srvf-cc-plot, eval = have_srvf && have_cc, fig.width = 6, fig.height = 6----
layout(t(matrix(1:12, 4, 3)))
par(cex.main = 0.8)
plot.new()
plot(inv_warp_srvf, main = "SRVF Inverse Warps", ylab = "",
     points = FALSE, col = alpha_palette, lwd = 1.5)
abline(0, 1, lty = 3)
plot(tf_aligned(reg_srvf), main = "SRVF Registered",
     col = alpha_palette, lwd = 1.5, points = FALSE)
plot(tf_aligned(reg_srvf), main = "SRVF Registered",
     type = "lasagna", col = hcl.colors(12, rev = TRUE))

plot(pinch_small, main = "Original",
     col = alpha_palette, lwd = 1.5)
plot(inv_warp_cc_unpen, main = "CC Inverse Warps (unpen.)", ylab = "",
     points = FALSE, col = alpha_palette, lwd = 1.5)
abline(0, 1, lty = 3)
plot(tf_aligned(reg_cc_unpen), main = "CC Registered (unpen.)",
     col = alpha_palette, lwd = 1.5, points = FALSE)
plot(tf_aligned(reg_cc_unpen), main = "CC Registered (unpen.)",
     type = "lasagna", col = hcl.colors(12, rev = TRUE))

plot.new()
plot(inv_warp_cc_pen, main = expression("CC Inverse Warps (" * lambda * " = 1e-4)"),
     ylab = "", points = FALSE, col = alpha_palette, lwd = 1.5)
abline(0, 1, lty = 3)
plot(tf_aligned(reg_cc_pen), main = expression("CC Registered (" * lambda * " = 1e-4)"),
     col = alpha_palette, lwd = 1.5, points = FALSE)
plot(tf_aligned(reg_cc_pen), main = expression("CC Registered (" * lambda * " = 1e-4)"),
     type = "lasagna", col = hcl.colors(12, rev = TRUE))

## ----growth-prep, class.source = "fold-show"----------------------------------
growth <- tf::growth |> dplyr::filter(gender == "female")

# Raw velocity via finite differences -- noisy, only 30 midpoints from 31 measurements:
growth$raw_vel <- tf_derive(growth$height)

# Smooth velocity via spline representation on a much denser grid, then derive analytically:
growth$smooth_vel <- tfb(
  growth$height,
  k = 15,
  bs = "tp",
  arg = seq(1, 18, l = 80),
  global = TRUE
  # family = gaussian(link = "log") # ensures positivity of velocity
) |>
  tf_derive()

## ----growth-raw-vs-smooth, eval = have_srvf, include = have_srvf, class.source = "fold-show"----
# SRVF on raw (noisy) velocity:
reg_raw_obj <- tf_register(growth$raw_vel, method = "srvf")
inv_warp_raw <- tf_inv_warps(reg_raw_obj)
reg_raw <- tf_aligned(reg_raw_obj)

# SRVF on smooth velocity:
reg_smooth_obj <- tf_register(growth$smooth_vel, method = "srvf")
inv_warp_smooth <- tf_inv_warps(reg_smooth_obj)
reg_smooth <- tf_aligned(reg_smooth_obj)

## ----growth-raw-vs-smooth-plot, eval = have_srvf, include = have_srvf, fig.width = 6, fig.height = 4----
reg_brks <- range(c(tf_evaluations(reg_raw), tf_evaluations(reg_smooth))) |>
  (\(x) seq(x[1], x[2], l = 13))()

layout(t(matrix(1:8, 4, 2)))
par(cex.main = 0.8)

plot(growth$raw_vel, main = "Raw Velocity (finite diff.)",
     xlab = "Age [years]", ylab = "cm/year", lwd = 0.8, ylim = c(0, 30))
plot(inv_warp_raw, points = FALSE, main = "SRVF Inverse Warps (raw)",
     xlab = "Chronological Age", ylab = "Registered Age", lwd = 0.8)
abline(0, 1, lty = 3)
plot(reg_raw, main = "SRVF Registered (raw)",
     xlab = "Registered Age [years]", ylab = "cm/year", lwd = 0.8, ylim = c(0, 30))
plot(reg_raw, main = "SRVF Registered (raw)",
     type = "lasagna", col = hcl.colors(12, rev = TRUE), breaks = reg_brks)

plot(growth$smooth_vel, main = "Smooth Velocity (tfb deriv.)",
     xlab = "Age [years]", ylab = "cm/year", lwd = 0.8, ylim = c(0, 30))
plot(inv_warp_smooth, points = FALSE, main = "SRVF Inverse Warps (smooth)",
     xlab = "Chronological Age", ylab = "Registered Age", lwd = 0.8)
abline(0, 1, lty = 3)
plot(reg_smooth, main = "SRVF Registered (smooth)",
     xlab = "Registered Age [years]", ylab = "cm/year", lwd = 0.8, ylim = c(0, 30))
plot(reg_smooth, main = "SRVF Registered (smooth)",
     type = "lasagna", col = hcl.colors(12, rev = TRUE), breaks = reg_brks)

## ----growth-penalization, eval = have_srvf, include = have_srvf, class.source = "fold-show"----
# Penalized SRVF on the raw velocity:
reg_raw_pen_obj <- tf_register(growth$raw_vel, method = "srvf", lambda = 0.1)
inv_warp_raw_pen <- tf_inv_warps(reg_raw_pen_obj)
reg_raw_pen <- tf_aligned(reg_raw_pen_obj)

## ----growth-penalization-plot, eval = have_srvf, include = have_srvf, fig.width = 6, fig.height = 3----
layout(t(matrix(1:4, 4, 1)))
par(cex.main = 0.8)
plot(growth$raw_vel, main = "Raw Velocity",
     xlab = "Age [years]", ylab = "cm/year", lwd = 0.8)
plot(inv_warp_raw_pen, points = FALSE,
     main = expression("Penalized Warps (raw, " * lambda * " = 0.1)"),
     xlab = "Chronological Age", ylab = "Registered Age", lwd = 0.8)
abline(0, 1, lty = 3)
plot(reg_raw_pen,
     main = expression("Penalized Registered (raw, " * lambda * " = 0.1)"),
     xlab = "Registered Age [years]", ylab = "cm/year", lwd = 0.8)
plot(reg_raw_pen, main = "Penalized Registered (raw)",
     type = "lasagna", col = hcl.colors(12, rev = TRUE), breaks = reg_brks)

## ----growth-landmarks, class.source = "fold-show"-----------------------------
# End of rapid infant growth: 
#    less than 2/3 of max early childhood growth (1-3) velocity before age 5
growth_slows <- growth$smooth_vel |> tf_zoom(begin = 1, end = 5) |> 
  tf_where(value < 0.66 * max(value[1:10]), return = "first") |> unlist()

# Pubertal peak: maximum velocity in age 8--17
growth_peaks <- growth$smooth_vel |> tf_zoom(begin = 8, end = 17) |> 
  tf_where(value == max(value)) |> unlist()

# Pre-pubertal trough: minimum velocity between age 5 and 1 year before the peak
growth_troughs <- growth$smooth_vel |> tf_zoom(begin = 5, end = growth_peaks - 1) |> 
  tf_where(value == min(value)) |> unlist() 

# Build landmark matrix and register:
growth_lm <- cbind(slowdown = growth_slows, trough = growth_troughs, peak = growth_peaks)
reg_lm_obj <- tf_register(growth$smooth_vel, method = "landmark", landmarks = growth_lm)
inv_warp_lm <- tf_inv_warps(reg_lm_obj)
reg_lm <- tf_aligned(reg_lm_obj)

## ----growth-landmarks-plot, fig.width = 6, fig.height = 3---------------------
if (!exists("reg_brks")) {
  reg_brks <- range(c(tf_evaluations(growth$smooth_vel), tf_evaluations(reg_lm))) |>
    (\(x) seq(x[1], x[2], l = 13))()
}
layout(t(1:4))
plot(growth$smooth_vel, main = "Smooth Velocity",
     xlab = "Age [years]", ylab = "cm/year", lwd = 0.8)
plot(inv_warp_lm, points = FALSE, main = "Landmark Inverse Warps",
     xlab = "Chronological Age", ylab = "Registered Age", lwd = 0.8)
abline(0, 1, lty = 3)
plot(reg_lm, main = "Landmark Registered",
     xlab = "Registered Age [years]", ylab = "cm/year", lwd = 0.8)
plot(reg_lm, main = "Landmark Registered",
     type = "lasagna", col = hcl.colors(12, rev = TRUE), breaks = reg_brks)

## ----restore-par, include = FALSE---------------------------------------------
par(oldpar)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "##",
  fig.width = 6,
  fig.height = 4,
  dpi = 72,
  fig.retina = 1,
  out.width = "90%"
)

library(tidyverse)
library(viridisLite)

theme_set(theme_minimal() + theme(legend.position = "bottom"))

options(
  ggplot2.continuous.colour = "viridis",
  ggplot2.continuous.fill = "viridis"
)
scale_colour_discrete <- scale_colour_viridis_d
scale_fill_discrete <- scale_fill_viridis_d

library("tidyfun")
data(chf_df, package = "tidyfun")
data(dti_df, package = "tidyfun")

pal_5 <- viridis(7)[-(1:2)]
set.seed(1221)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "##",
  fig.width = 6,
  fig.height = 4,
  dpi = 72,
  fig.retina = 1,
  out.width = "90%"
)

library(tidyverse)
library(viridisLite)
library(lme4)
library(gridExtra)

theme_set(theme_minimal() + theme(legend.position = "bottom"))

options(
  ggplot2.continuous.colour = "viridis",
  ggplot2.continuous.fill = "viridis"
)
scale_colour_discrete <- scale_colour_viridis_d
scale_fill_discrete <- scale_fill_viridis_d

library("tidyfun")

pal_5 <- viridis(7)[-(1:2)]
set.seed(1221)

## -----------------------------------------------------------------------------
dti_df <- tibble(
  id = refund::DTI$ID,
  visit = refund::DTI$visit,
  sex = refund::DTI$sex,
  case = factor(ifelse(refund::DTI$case, "MS", "control"))
)

dti_df$cca <- tfd(refund::DTI$cca, arg = seq(0, 1, length.out = 93))
dti_df$rcst <- tfd(refund::DTI$rcst, arg = seq(0, 1, length.out = 55))

## -----------------------------------------------------------------------------
dti_df

## -----------------------------------------------------------------------------
dti_df |>
  tf_ggplot(aes(tf = cca, col = case, alpha = 0.2 + 0.4 * (case == "control"))) +
  geom_line() +
  facet_wrap(~sex) +
  scale_alpha(guide = "none", range = c(0.2, 0.4))

## -----------------------------------------------------------------------------
canada <- tibble(
  place = fda::CanadianWeather$place,
  region = fda::CanadianWeather$region,
  lat = fda::CanadianWeather$coordinates[, 1],
  lon = -fda::CanadianWeather$coordinates[, 2]
) |>
  mutate(
    temp = t(fda::CanadianWeather$dailyAv[, , 1]) |>
      tfd(arg = 1:365),
    precipl10 = t(fda::CanadianWeather$dailyAv[, , 3]) |>
      tfd(arg = 1:365) |>
      tf_smooth()
  )

## -----------------------------------------------------------------------------
canada

## -----------------------------------------------------------------------------
temp_panel <- canada |>
  tf_ggplot(aes(tf = temp, color = region)) +
  geom_line()

precip_panel <- canada |>
  tf_ggplot(aes(tf = precipl10, color = region)) +
  geom_line()

gridExtra::grid.arrange(temp_panel, precip_panel, nrow = 1)

## -----------------------------------------------------------------------------
data("sleepstudy", package = "lme4")
sleepstudy <- as_tibble(sleepstudy)

sleepstudy

## -----------------------------------------------------------------------------
sleepstudy_tf <- sleepstudy |>
  tf_nest(Reaction, .id = Subject, .arg = Days)

sleepstudy_tf

## -----------------------------------------------------------------------------
sleepstudy_tf |>
  tf_ggplot(aes(tf = Reaction)) +
  geom_line()

## -----------------------------------------------------------------------------
tibble(
  Subject = unique(sleepstudy$Subject), 
  Reaction = tfd(sleepstudy, id = "Subject", arg = "Days", value = "Reaction")
)

## ----eval = FALSE-------------------------------------------------------------
# ALA::fev1 |>
#   group_by(id) |>
#   mutate(n_obs = n()) |>
#   filter(n_obs > 1) |>
#   ungroup() |>
#   tf_nest(logFEV1, height, .arg = age) |>
#   glimpse()

## -----------------------------------------------------------------------------
dti_df <- refund::DTI |>
  janitor::clean_names() |>
  select(-starts_with("rcst")) |>
  glimpse()

dti_df |>
  tf_gather(starts_with("cca")) |>
  glimpse()

## -----------------------------------------------------------------------------
# reload the tidyfun version of the DTI data
data(dti_df, package = "tidyfun")

# raw functional data
cca_raw <- dti_df$cca[1:5]
cca_raw

# represent in a spline basis
cca_basis <- tfb(dti_df$cca[1:5], k = 25)
cca_basis

# re-express the raw data in the same basis representation
cca_rebased <- tf_rebase(cca_raw, basis_from = cca_basis)
cca_rebased

# or convert a spline-based representation to a grid-based one for a specific grid:
tf_rebase(cca_basis, basis_from = cca_raw)

## -----------------------------------------------------------------------------
# split CCA profiles at their midpoint
cca_halves <- tf_split(dti_df$cca[1:10], splits = 0.5)

# result is a list of tf vectors, one per segment
cca_halves[[1]]
cca_halves[[2]]

# recombine
cca_recombined <- tf_combine(cca_halves[[1]], cca_halves[[2]])
cca_recombined 

## -----------------------------------------------------------------------------
# create an fd object from the Canadian weather data
weather_basis <- fda::create.fourier.basis(c(0, 365), nbasis = 65)
weather_fd <- fda::smooth.basis(
  argvals = 1:365,
  y = fda::CanadianWeather$dailyAv[, , 1],
  fdParobj = weather_basis
)

# convert fdSmooth to tfb
weather_tf <- tfb_spline(weather_fd)
weather_tf[1:3]

## -----------------------------------------------------------------------------
tibble(
  place = fda::CanadianWeather$place,
  region = fda::CanadianWeather$region,
  temp = weather_tf
) |>
  tf_ggplot(aes(tf = temp, color = region)) +
  geom_line(alpha = 0.5)

## -----------------------------------------------------------------------------
sleepstudy_tf |>
  tf_unnest(cols = Reaction) |>
  glimpse()

## -----------------------------------------------------------------------------
sleepstudy_tf |>
  tf_spread() |>
  glimpse()

## -----------------------------------------------------------------------------
reaction_matrix <- sleepstudy_tf |> pull(Reaction) |> as.matrix() 
 
head(reaction_matrix)

# argument values of input data saved in `arg`-attribute:
attr(reaction_matrix, "arg")

## -----------------------------------------------------------------------------
sleepstudy_tf |> pull(Reaction) |> 
  as.data.frame(unnest = TRUE) |>
  head()

# from https://github.com/r-lib/scales/blob/master/R/scale-discrete.r
is.discrete <- function(x) {
  is.factor(x) || is.character(x) || is.logical(x)
}

#' Lasagna plots for `tf`s using `ggplot2`
#'
#' Lasagna plots show one color bar for each function.
#'
#' The vertical order of the lasagna layers is **increasing** in
#'
#' - `order` (if provided),
#' - the values returned by `order_by` (if provided),
#' - and the row number of the observations.
#'
#' i.e., lowest values are on top so that by default the first layer is the first
#' observation in `data` and the vertical order of the layers is the
#' ordering of observations obtained by `dplyr::arrange(data, order, order_by(value), row_number())`.
#'
#' @param data A data frame containing the `tf` column to visualize.
#' @param tf bare name of the `tf` column to visualize
#' @param order (optional) bare name of a column in `data` to define vertical
#'   order of lasagna layers.
#' @param label (optional) bare name of a column in `data` to define labels for
#'   lasagna layers. Defaults to names of `y`, if present, or row numbers.
#' @param arg `arg` to evaluate `y` on
#' @param order_by a function applied to each row in `y[, arg]` that must
#'   return a scalar value to define the order of lasagna layers.
#' @param order_ticks add horizontal lines indicating borders between levels of
#'   `order` (if it is a discrete variable) and labels for its levels?
#'   Defaults to TRUE. Supply a named list to override tick appearance, including
#'   label styling, line type, alpha, rotation, and label placement.
#'   Disable this when faceting; the tick annotations are not designed for faceted layouts.
#' @returns a `ggplot2` object
#' @export
#' @importFrom grid unit grobTree textGrob gpar
#' @family tidyfun visualization
#' @examples
#' \donttest{
#' library(ggplot2)
#' set.seed(1221)
#' data <- expand.grid(group = factor(1:5), rep = 1:10)
#' data <- dplyr::mutate(data,
#'   id = paste(group, rep, sep = "-"),
#'   f = tf_rgp(50),
#'   fb = tfb(f)
#' )
#'
#' gglasagna(data, f, label = id)
#' gglasagna(data, fb, label = id, order = group)
#' # order is lowest first / on top by default
#' gglasagna(data, f, label = id, order = tf_depth(f))
#' gglasagna(data, f, label = id, order_by = dplyr::first) +
#'   facet_wrap(~group, scales = "free")
#' # order of layers is by "order_by" within "order":
#' gglasagna(data, fb, label = id, order = group, order_by = dplyr::first)
#' }
gglasagna <- function(
  data,
  tf,
  order = NULL,
  label = NULL,
  arg = NULL,
  order_by = NULL,
  order_ticks = TRUE
) {
  order_ticks_args <- list(
    color = "black",
    linetype = 2,
    alpha = 0.5,
    fontsize = theme_get()$text$size * 0.8,
    rot = 90,
    label_offset = grid::unit(0.02, "npc")
  )
  if (is.list(order_ticks)) {
    order_ticks_args <- utils::modifyList(order_ticks_args, order_ticks)
    order_ticks <- TRUE
  }

  # FIXME: render errors for weird arg lengths (e.g. 93)
  if (!is_tf(pull(data, {{ tf }}))) {
    cli::cli_abort(
      "{.arg tf} must be a {.cls tf} object, not {.obj_type_friendly {pull(data, {{ tf }})}}."
    )
  }
  order <- enexpr(order)
  has_order <- !is.null(order)
  order_by_label <- enexpr(order_by)
  has_order_by <- !is.null(order_by_label)
  if (has_order) {
    order_label <- as_label(order)
  } else {
    order_label <- NULL
    order <- expr(..row)
    order_ticks <- FALSE
  }
  label <- enexpr(label)
  has_label <- !is.null(label)
  if (!has_label) {
    label <- expr(names(!!enexpr(tf)) %||% row_number())
    labelname <- ""
  } else {
    labelname <- as_label(label)
  }
  y_name <- as_name(enexpr(tf))
  data <- mutate(
    data,
    ..label = !!label,
    ..row = row_number(),
    ..order = !!order
  )
  tf_eval <- data |>
    # TODO: add .preserve for all list columns not being plotted
    mutate(..y = !!label) |> # vertical position variable
    tf_unnest(y_name, .arg = arg, names_sep = "___", try_dropping = FALSE) |>
    rename(..x = matches("___arg"), ..fill = matches("___value"))
  if (has_order_by) {
    order_by_label <- as_label(order_by_label)
    if (!is.function(order_by)) {
      cli::cli_abort(
        "{.arg order_by} must be a function, not {.obj_type_friendly {order_by}}."
      )
    }
    order_by_value <- tf_eval |>
      group_by(..y) |>
      summarise(..order_by_value = order_by(..fill)) |>
      ungroup() |>
      mutate(..order_by_value = rank(..order_by_value))
    tf_eval <- left_join(tf_eval, order_by_value, by = "..y")
    # override previous ordering
    if (!has_order) {
      tf_eval <- tf_eval |> mutate(..order = ..order_by_value)
    }
  } else {
    order_by_label <- NULL
    tf_eval <- tf_eval |> mutate(..order_by_value = ..row)
  }
  # create order of rows
  tf_eval <- tf_eval |>
    arrange(..order, ..order_by_value, ..row) |>
    mutate(..y = ordered(..y, levels = rev(unique(..y)))) |>
    # mutate(..y = as.numeric(..y)) |>
    rename(!!y_name := ..fill)
  labeldata <- with(
    tf_eval,
    tibble(
      breaks = unique(..y),
      labels = ..label[!duplicated(..y)]
    )
  )

  p <- ggplot(tf_eval) +
    geom_tile(aes(
      y = ..y,
      x = ..x,
      fill = .data[[y_name]],
      colour = .data[[y_name]]
    )) +
    labs(x = "", y = "")
  if (!is.null(order_label) || !is.null(order_by_label)) {
    p <- p +
      labs(
        caption = paste(
          "ordered by:",
          paste0(
            order_label,
            ifelse(has_order_by & has_order, "; ", ""),
            order_by_label
          )
        )
      )
  }
  if (!is.null(order_ticks) && is.discrete(pull(tf_eval, ..order))) {
    order_ticks_data <- data |>
      arrange(desc(!!order)) |>
      mutate(ticks = row_number()) |>
      group_by(!!order) |>
      summarize(tick_hi = max(ticks), tick_lo = min(ticks)) |>
      mutate(
        label_pos = (tick_hi + tick_lo) / 2,
        tick_pos = lead(tick_hi) + 0.5
      )
    p <- p +
      geom_hline(
        data = order_ticks_data,
        aes(yintercept = tick_pos),
        col = order_ticks_args$color,
        alpha = order_ticks_args$alpha,
        linetype = order_ticks_args$linetype
      )
    # adding a secondary axis with labels for the order-variable only works
    # if we make the y-axis continuous -- but a) it isn't,really and
    # b)scales="free" is needed for good looking facets but won't
    # close gaps for pseudo-continuous y-scales,
    # so we use this annotation_custom weirdness instead
    order_ticks_labelpos <- with(
      order_ticks_data,
      grid::unit(label_pos / max(tick_hi), "npc")
    )
    order_ticks_labels <- grid::grobTree(
      grid::textGrob(
        order_ticks_data |> pull(!!order),
        x = order_ticks_args$label_offset,
        y = order_ticks_labelpos,
        rot = order_ticks_args$rot,
        gp = grid::gpar(fontsize = order_ticks_args$fontsize)
      )
    )
    p <- p + annotation_custom(order_ticks_labels)
  }
  p <- p +
    scale_y_discrete(
      labelname,
      breaks = labeldata$breaks,
      labels = labeldata$labels
    )
  p
}

#-------------------------------------------------------------------------------

## tibble methods:
# adapted from https://github.com/r-spatial/sf/blob/master/R/tidyverse.R

#' Format tidy functional data for tibbles
#'
#' Summarize tidy functional data for tibble
#' @param x object of class tf
#' @param ... ignored
#' @rdname tftibble
#' @family tidyfun print
#' @importFrom pillar type_sum obj_sum pillar_shaft
#' @details see [pillar::type_sum()]
type_sum.tf <- function(x, ...) {
  class(x)[2]
}

#' @rdname tftibble
obj_sum.tf <- function(x) {
  paste0(type_sum.tf(x), "[", length(x), "]")
}

#' @rdname tftibble
pillar_shaft.tf <- function(x, ...) {
  digits <- options("pillar.sigfig")$pillar.sigfig
  if (is.null(digits)) {
    digits <- options("digits")$digits
  }
  out <- format(x, width = 30L, digits = digits, prefix = FALSE, ...)
  pillar::new_pillar_shaft_simple(out, align = "right", min_width = NULL)
}

#' Glyph plots for `tf` objects
#'
#' Plots a miniature glyph / sparkline for each entry of a `tf`-object.
#' (Capellini are tiny spaghetti -- *angel hair* pasta.) Aesthetics `x` and `y`
#' specify the location of the glyphs, the `tf` aesthetic defines their shapes.
#' The aliases `geom_cappelini`, `geom_cappellini`, and `geom_capelini` are
#' also accepted.
#'
#' @section `tf` aesthetic:
#'   Mandatory. Used to designate a column of class `tf` to be visualized as glyphs.
#' @returns A [ggplot2::layer()] object for use in a ggplot.
#' @examplesIf rlang::is_installed(c("fda", "maps"))
#' library(ggplot2)
#' weather <- fda::CanadianWeather
#' canada <- data.frame(
#'   place = weather$place,
#'   region = weather$region,
#'   lat = weather$coordinates[, 1],
#'   lon = -weather$coordinates[, 2],
#'   region = weather$region
#' )
#' canada$temp <- tfd(t(weather$dailyAv[, , 1]), arg = 1:365)
#' canada$precipl10 <- tfd(t(weather$dailyAv[, , 3]), arg = 1:365) |> tf_smooth()
#' canada_map <-
#'   data.frame(maps::map("world", "Canada", plot = FALSE)[c("x", "y")])
#' # map of canada with annual temperature averages in red, precipitation in blue:
#' ggplot(canada, aes(x = lon, y = lat)) +
#'   geom_capellini(aes(tf = precipl10), width = 3, height = 5, colour = "blue") +
#'   geom_capellini(aes(tf = temp), width = 3, height = 5, colour = "red") +
#'   geom_path(data = canada_map, aes(x = x, y = y), alpha = 0.1) +
#'   coord_quickmap()
#'
#' ggplot(canada, aes(x = lon, y = lat, colour = region)) +
#'   geom_capellini(aes(tf = precipl10),
#'     width = 5, height = 3,
#'     line.linetype = 1, box.fill = "white", box.alpha = 0.5, box.colour = NA
#'   )
#' @name ggcapellini
#' @family tidyfun visualization
#' @seealso [GGally::glyphs()]
NULL

#' @export
#' @importFrom GGally rescale11 rescale01
#' @rdname ggcapellini
#' @usage NULL
#' @format NULL
StatCapellini <- ggplot2::ggproto(
  "StatCapellini",
  ggplot2::Stat,
  required_aes = c("x", "y", "tf"),
  setup_params = function(data, params) {
    if (is.null(params$arg)) {
      params$arg <- list(tf_arg(pull(data, tf)))
    }
    if (is.null(params$width)) {
      params$width <- ggplot2::resolution(data$x) / 1.5
    }
    if (is.null(params$height)) {
      params$height <- ggplot2::resolution(data$y) / 1.5
    }
    params
  },
  compute_layer = function(self, data, params, layout) {
    if (!is_tf(data$tf)) {
      cli::cli_abort(
        "{.arg tf} must be a {.cls tf} object, not {.obj_type_friendly {data$tf}}."
      )
    }
    tf_eval <- suppressMessages(
      data |>
        mutate(tf___id = names(tf) %||% seq_along(tf)) |>
        tf_unnest(tf, arg = params$arg, names_sep = "___")
    ) |>
      select(-group) |>
      rename(group = tf___id, arg = tf___arg, value = tf___value) |>
      mutate(
        xgrid = x,
        ygrid = y,
        x = x + rescale11(arg) * params$width / 2,
        y = y + rescale11(value) * params$height / 2,
        width = params$width,
        height = params$height,
        lines = params$add_lines,
        boxes = params$add_boxes
      )
    tf_eval
  },
  # need this so arg etc get recognized as valid parameters
  # because layer() only checks compute_panel & compute_group
  compute_panel = function(
    self,
    data,
    scales,
    arg,
    add_lines,
    add_boxes,
    width,
    height
  ) {
    ggplot2::Stat$compute_panel(self, data, scales)
  }
)


#' @export
#' @rdname ggcapellini
#' @inheritParams ggplot2::stat_identity
#' @param na.rm remove NAs? defaults to `TRUE`
stat_capellini <- function(
  mapping = NULL,
  data = NULL,
  geom = "capellini",
  position = "identity",
  na.rm = TRUE,
  show.legend = NA,
  inherit.aes = TRUE,
  arg = NULL,
  add_lines = FALSE,
  add_boxes = TRUE,
  width = NULL,
  height = NULL,
  ...
) {
  ggplot2::layer(
    stat = StatCapellini,
    data = data,
    mapping = mapping,
    geom = geom,
    position = position,
    show.legend = show.legend,
    inherit.aes = inherit.aes,
    params = list(
      na.rm = na.rm,
      arg = arg,
      add_lines = add_lines,
      add_boxes = add_boxes,
      width = width,
      height = height,
      ...
    )
  )
}

# geom --------------------------------------------------------------------

#' @export
#' @rdname ggcapellini
#' @format NULL
#' @param stat the stat to use; defaults to `"capellini"`.
#' @param arg where to evaluate `tf`; defaults to the object's default
#'   evaluation grid.
#' @param add_lines should a reference line in the middle of the range of the
#'    functions' values be added to each glyph? defaults to TRUE
#' @param add_boxes should a box be added to frame each glyph? defaults to TRUE
#' @param width the width of the glyphs. Defaults to 2/3 of the [ggplot2::resolution()]
#'   of the variable for the `x`-aesthetic, this will be too small if any values
#'   are close together.
#' @param height the height of the glyphs. Defaults to 2/3 of the [ggplot2::resolution()]
#'   of the variable for the `y`-aesthetic, this will be too small if any values
#'   are close together.
#' @param box.colour aesthetic property of the box
#' @param box.linetype  aesthetic property of the box
#' @param box.fill  aesthetic property of the box
#' @param box.linewidth  aesthetic property of the box
#' @param box.alpha  aesthetic property of the box
#' @param line.colour aesthetic property of the reference line
#' @param line.linetype aesthetic property of the reference line
#' @param line.linewidth aesthetic property of the reference line
#' @param line.alpha aesthetic property of the reference line
geom_capellini <-
  function(
    mapping = NULL,
    data = NULL,
    stat = "capellini",
    position = "identity",
    ...,
    na.rm = TRUE,
    show.legend = NA,
    inherit.aes = TRUE,
    arg = NULL,
    add_lines = TRUE,
    add_boxes = TRUE,
    width = NULL,
    height = NULL,
    box.colour = "#0000001A",
    box.linetype = 1,
    box.fill = NA,
    box.linewidth = 0.1,
    box.alpha = 0.1,
    line.colour = "black",
    line.linetype = 2,
    line.linewidth = 0.3,
    line.alpha = 0.5
  ) {
    ggplot2::layer(
      stat = StatCapellini,
      data = data,
      mapping = mapping,
      geom = "capellini",
      position = position,
      show.legend = show.legend,
      inherit.aes = inherit.aes,
      params = list(
        na.rm = na.rm,
        arg = arg,
        add_lines = add_lines,
        add_boxes = add_boxes,
        width = width,
        height = height,
        box.colour = box.colour,
        box.linetype = box.linetype,
        box.fill = box.fill,
        box.linewidth = box.linewidth,
        box.alpha = box.alpha,
        line.colour = line.colour,
        line.linetype = line.linetype,
        line.linewidth = line.linewidth,
        line.alpha = line.alpha,
        ...
      )
    )
  }
#' @export
#' @rdname ggcapellini
#' @usage NULL
geom_cappellini <- geom_capellini
#' @export
#' @usage NULL
#' @rdname ggcapellini
geom_capelini <- geom_capellini
#' @export
#' @usage NULL
#' @rdname ggcapellini
geom_cappelini <- geom_capellini


#' @export
#' @rdname ggcapellini
#' @usage NULL
#' @format NULL
GeomCapellini <- ggplot2::ggproto(
  "GeomCapellini",
  ggplot2::Geom,
  setup_data = function(data, params) {
    GeomPath$setup_data(data, params)
  },
  draw_group = function(
    data,
    panel_params,
    coord,
    box.colour = "#0000001A",
    box.linetype = 1,
    box.fill = NA,
    box.linewidth = 0.1,
    box.alpha = 0.1,
    line.colour = "black",
    line.linetype = 2,
    line.linewidth = 0.3,
    line.alpha = 0.5
  ) {
    glyph_grob <- GeomPath$draw_panel(data, panel_params, coord)
    if (data$lines[1]) {
      lines <- data.frame(
        x = data$xgrid[1] - data$width[1] / 2,
        xend = data$xgrid[1] + data$width[1] / 2,
        y = data$ygrid[1],
        yend = data$ygrid[1],
        colour = line.colour,
        linetype = line.linetype,
        linewidth = line.linewidth,
        alpha = line.alpha
      )
      lines_grob <- GeomSegment$draw_panel(lines, panel_params, coord)
    } else {
      lines_grob <- NULL
    }
    if (data$boxes[1]) {
      boxes <- data.frame(
        xmin = data$xgrid[1] - data$width[1] / 2,
        xmax = data$xgrid[1] + data$width[1] / 2,
        ymin = data$ygrid[1] - data$height[1] / 2,
        ymax = data$ygrid[1] + data$height[1] / 2,
        colour = box.colour,
        linetype = box.linetype,
        fill = box.fill,
        alpha = box.alpha,
        linewidth = box.linewidth
      )
      boxes_grob <- GeomRect$draw_panel(boxes, panel_params, coord)
    } else {
      boxes_grob <- NULL
    }
    grid::gList(
      lines_grob,
      boxes_grob,
      glyph_grob
    )
  },
  default_aes = ggplot2::aes(
    colour = "black",
    linewidth = 0.5,
    linetype = 1,
    alpha = 0.5
  ),
  draw_key = ggplot2::GeomPath$draw_key
)

# Documentation for tidyfun datasets

#' Congestive heart failure accelerometry data
#'
#' @description
#' Activity data from a study of congestive heart failure (CHF) patients. Data
#' were originally presented in Huang et al. (2019); these data are publicly
#' available, with download information in the paper.
#'
#' @format A tibble with 329 rows and 8 columns:
#' \describe{
#'   \item{id}{(numeric) Subject identifier.}
#'   \item{gender}{(character) `"Male"` or `"Female"`.}
#'   \item{age}{(numeric) Age in years.}
#'   \item{bmi}{(numeric) Body mass index.}
#'   \item{event_week}{(numeric) Week of cardiac event.}
#'   \item{event_type}{(character) Type of cardiac event.}
#'   \item{day}{(ordered factor) Day of the week (`Mon` < `Tue` < ... < `Sun`).}
#'   \item{activity}{(`tfd_reg`) Minute-by-minute activity counts over
#'     a 24-hour period (arg domain 1--1440).}
#' }
#'
#' @source
#' Data are from a study of physical activity in CHF patients conducted by
#' Huang et al. The original data are publicly available; see the paper for
#' download details.
#'
#' @references
#' Huang, L., Bai, J., Ivanescu, A., Harris, T., Maurer, M., Green, P., and
#' Zipunnikov, V. (2019). Multilevel Matrix-Variate Analysis and its
#' Application to Accelerometry-Measured Physical Activity in Clinical
#' Populations. *Journal of the American Statistical Association*, 114(526),
#' 553--564. \doi{10.1080/01621459.2018.1482750}
#'
#' @seealso [dti_df] for another example dataset,
#'   `vignette("x04_Visualization", package = "tidyfun")` for usage examples.
#' @keywords data
#' @family tidyfun datasets
#' @examples
#' chf_df
#'
#' library(ggplot2)
#' chf_df |>
#'   dplyr::filter(id %in% 1:5) |>
#'   gglasagna(activity, order_by = mean)
"chf_df"

#' Diffusion tensor imaging data
#'
#' @description
#' Fractional anisotropy (FA) tract profiles for the corpus callosum (`cca`)
#' and the right corticospinal tract (`rcst`) from a diffusion tensor imaging
#' (DTI) study of multiple sclerosis patients and healthy controls.
#'
#' The original data in [refund::DTI] include additional variables (`pasat`,
#' `Nscans`, `visit.time`) that were not carried over here.
#'
#' @format A tibble with 382 rows and 6 columns:
#' \describe{
#'   \item{id}{(numeric) Subject identifier.}
#'   \item{visit}{(integer) Visit number.}
#'   \item{sex}{(factor) `"male"` or `"female"`.}
#'   \item{case}{(factor) `"control"` or `"MS"` (multiple sclerosis status).}
#'   \item{cca}{(`tfd_irreg`) FA tract profiles for the corpus callosum
#'     (up to 93 evaluation points, domain 0--1).}
#'   \item{rcst}{(`tfd_irreg`) FA tract profiles for the right corticospinal
#'     tract (up to 55 evaluation points, domain 0--1).}
#' }
#'
#' @details
#' If you use this data as an example in written work, please include the
#' following acknowledgment: *"The MRI/DTI data were collected at Johns Hopkins
#' University and the Kennedy-Krieger Institute."*
#'
#' @source
#' Data are from the Johns Hopkins University and the Kennedy-Krieger Institute.
#' Also available in a different format as [refund::DTI].
#'
#' @references
#' Goldsmith, J., Bobb, J., Crainiceanu, C., Caffo, B., and Reich, D. (2011).
#' Penalized Functional Regression. *Journal of Computational and Graphical
#' Statistics*, 20(4), 830--851. \doi{10.1198/jcgs.2010.10007}
#'
#' Goldsmith, J., Crainiceanu, C., Caffo, B., and Reich, D. (2012).
#' Longitudinal Penalized Functional Regression for Cognitive Outcomes on
#' Neuronal Tract Measurements. *Journal of the Royal Statistical Society:
#' Series C*, 61(3), 453--469. \doi{10.1111/j.1467-9876.2011.01031.x}
#'
#' @seealso [chf_df] for another example dataset,
#'   `vignette("x04_Visualization", package = "tidyfun")` for usage examples.
#' @keywords data
#' @family tidyfun datasets
#' @examples
#' dti_df
#'
#' library(ggplot2)
#' dti_df |>
#'   dplyr::filter(visit == 1) |>
#'   tf_ggplot(aes(tf = cca, color = case)) +
#'   geom_line(alpha = 0.3)
"dti_df"

#' Gather all columns representing functional measurements into a `tfd`-object
#'
#' Similar in spirit to [tidyr::gather()], but does NOT put the values in the
#' gathered columns into one very long "value"-column while labeling the different
#' original columns in a very long "key"-column -- instead it creates a `tfd`-column
#' containing the functional measurements of the columns given in `...`.
#'
#' @param data a data frame -- note that `dplyr` does not handle matrix columns well,
#'    if `data` contains more than one of those, `tf_gather` will fail...
#' @param ... A selection of columns to collect as a `tfd` object. Each column
#'   represents measurements of a functional variable at a specific `arg`-val.
#'   Can also be the name of a matrix-valued column, but see above.
#'   If empty, all variables are selected. You can supply bare variable names,
#'   select all variables between x and z with x:z, exclude y with -y. For more
#'   options, see the [dplyr::select()] documentation.
#' @param key the name of the created `tfd`-column. Defaults to `".tfd"`, but
#'   the function will try to guess the name based on the column names of the
#'   gathered columns in this case. If a common prefix of all column names
#'   is found, this is used instead. You also get a message about this.
#' @param arg optional. Argument values for the functions. If not provided, will be guessed from the column names as well.
#'   See also [tf::tfd()].
#' @param evaluator optional. A function accepting arguments x, arg, evaluations. See [tf::tfd()] for details.
#' @param domain optional. Range of possible `arg`-values. See [tf::tfd()] for details.
#' @returns a modified `data.frame` with a `tfd` column replacing the `...`.
#' @import dplyr
#' @export
#' @seealso [dplyr::select()]
#' @family tidyfun data wrangling functions
#' @examples
#' d <- tf_spread(growth[1:5, ]) |> dplyr::mutate(id = 1:dplyr::n())
#' dplyr::glimpse(d)
#' # tidyselect syntax for column selection:
#' tf_gather(d, starts_with("height"))
#' tf_gather(d, height_1:height_18)
#' tf_gather(d, -gender, -id)
#' # custom key name and arg values:
#' tf_gather(d, starts_with("height"), key = "height")
#' tf_gather(d, starts_with("height"), arg = seq(0, 1, length.out = 31))
tf_gather <- function(
  data,
  ...,
  key = ".tfd",
  arg = NULL,
  domain = NULL,
  evaluator = tf_approx_linear
) {
  key_var <- as_name(enexpr(key))
  search_key <- isTRUE(key == ".tfd")
  if (...length() == 0) {
    gather_vars <- names(data)
  } else {
    gather_vars <- names(eval_select(expr(c(...)), data))
  }
  if (is_empty(gather_vars)) {
    return(data)
  }
  # turn matrix column into regular columns:
  if (
    length(gather_vars) == 1 && is.matrix(data[[gather_vars]]) && search_key
  ) {
    key_var <- gather_vars
    search_key <- FALSE
    cli::cli_inform("creating new {.cls tfd}-column {.val {key_var}}")
  }

  tfd_data <- data |>
    select(all_of(gather_vars)) |>
    as.matrix()
  if (search_key) {
    # see also find_arg: will interpret separating-dashes as minus-signs
    # regex adapted from https://www.regular-expressions.info/floatingpoint.html
    found_key <- unique(sub(
      "[-+]?(0|(0\\.[0-9]+)|([1-9][0-9]*\\.?[0-9]*))([eE][-+]?[0-9]+)?$",
      "",
      colnames(tfd_data)
    ))
    # assume trailing 0's are padding:
    found_key <- sub("[0]+$", "", found_key)
    # assume trailing punctuation is separator:
    found_key <- sub("[[:punct:]]$", "", found_key)
    # check again for uniqueness of resulting value ...
    found_key <- unique(found_key)
    if (length(found_key) == 1 && all(found_key != "")) {
      key_var <- found_key
      cli::cli_inform("creating new {.cls tfd}-column {.val {key_var}}")
    }
  }

  data |>
    select(-all_of(gather_vars)) |>
    mutate(
      !!key_var := tfd(
        tfd_data,
        arg = arg,
        domain = domain,
        evaluator = {{ evaluator }}
      )
    )
}

#' Spread a `tf`-column into many columns representing the
#' function evaluations.
#'
#' Similar in spirit to [tidyr::spread()], but does NOT shorten,
#' just widens the data frame  -- a `tf`-column is spread out into many columns
#' containing the functional measurements.
#'
#' @param data a data frame with at least one `tf`-column
#' @param value the name of the `tf`-column to 'spread'/evaluate.
#'   You can supply bare variable names etc., see the [dplyr::select()]
#'   documentation. Also works without this if there's only one `tf` in `data`,
#'   see examples.
#' @param arg (Semi-)optional. A vector of `arg`-values on which to evaluate the
#'   functions. If not provided, uses the default `arg`s. Should be
#'   specified for `tf_irreg`, otherwise *all* observed gridpoints are used for
#'   *every* function.
#' @param sep separating character used to create column names for the new columns,
#'   defaults to `"_"` for column names "<name of the `tf`>_<`arg`-value>".
#'   Set to NULL to get column names that only contain the `arg`-value.
#' @param interpolate `interpolate`-argument for evaluating the functional data.
#'   Defaults to FALSE, i.e., `tfd`s are *not* inter/extrapolated on unobserved
#'    `arg`-values.
#' @returns a wider dataframe with the `tf`-column spread out into many columns
#'   each containing the functional measurements for one `arg`-value.
#' @importFrom tidyselect vars_pull
#' @export
#' @family tidyfun data wrangling functions
#' @examples
#' d <- dplyr::tibble(g = 1:3)
#' d$f <- tf_rgp(3, 11L)
#' tf_spread(d, f)
#' tf_spread(d, -g)
#' tf_spread(d)
tf_spread <- function(data, value, arg, sep = "_", interpolate = FALSE) {
  if (missing(value)) {
    tf_cols <- which(map_lgl(data, is_tf))
    if (length(tf_cols) == 0) {
      cli::cli_warn(
        "{.arg value} {.val {tf_var}} is not a column of class {.cls tf}. Nothing's happening here."
      )
      return(data)
    }
    if (length(tf_cols) == 1) {
      value <- tf_cols
    } else {
      cli::cli_abort(
        "More than one {.cls tf} found, specify which one to spread in {.arg value}."
      )
    }
  }
  tf_var <- vars_pull(names(data), !!enquo(value))
  tf <- data[[tf_var]]
  if (!is_tf(tf)) {
    cli::cli_warn(
      "{.arg value} {.val {tf_var}} is not a column of class {.cls tf}. Nothing's happening here."
    )
    return(data)
  }
  if (missing(arg)) {
    arg <- tf_arg(tf)
    if (is_irreg(tf)) {
      arg <- unlist(arg, use.names = TRUE) |> unique() |> sort()
      cli::cli_warn(
        "no explicit {.arg arg} for irregular {.val {tf_var}} provided -- using all {length(arg)}, distinct observed argument values.",
      )
    }
  }
  tf_eval <- tf[, arg, matrix = TRUE, interpolate = interpolate] |>
    as.data.frame()
  if (!is.null(sep)) {
    colnames(tf_eval) <- paste0(tf_var, sep, arg)
  }
  data |>
    select(-all_of(tf_var)) |>
    bind_cols(tf_eval)
}

# ------------------------------------------------------------------------------

#' Turn "long" tables into tidy data frames with `tf`-objects
#'
#' Similar in spirit to [tidyr::nest()]. This turns tables in "long" format,
#' where one column (`.id`) defines the unit of observation, one column (`.arg`)
#' defines the evaluation grids of the functional observations, and other columns (`...`)
#' define the values of the functions at those points into a (much shorter) table containing
#' `tfd`-objects. All other variables are checked for constancy over `.id` and
#' appended as well.
#'
#' `domain` and `evaluator` can be specified as lists or vectors
#' if you are nesting multiple functional data columns with different properties.
#' Because this interface captures evaluator names as text, supply the evaluator
#' as a string rather than a bare function name.
#'
#' @param data a data frame
#' @param ... A selection of columns. If empty, all variables except the
#'   `.id` and `.arg` columns are selected. You can supply bare variable names,
#'   select all variables between `x` and `z` with `x:z`, exclude `y` with `-y`.
#'   For more options, see the [dplyr::select()] documentation.
#' @param .id the (bare or quoted) name of the column defining the different
#'   observations. Defaults to "id".
#' @param .arg the (bare or quoted) name of the column defining the `arg`-values
#'   of the observed functions. Defaults to "arg".
#' @inheritParams tf_gather
#' @returns a data frame with (at least) `.id` and `tfd` columns
#' @export
#' @family tidyfun data wrangling functions
#' @seealso [tfd()] for details on `domain` and `evaluator`.
#' @examples
#' d <- dplyr::tibble(id = rep(1:3, each = 5), arg = rep(1:5, 3), value = rnorm(15))
#' tf_nest(d, .id = id, .arg = arg)
tf_nest <- function(
  data,
  ...,
  .id = "id",
  .arg = "arg",
  domain = NULL,
  evaluator = "tf_approx_linear"
) {
  if (!is.data.frame(data)) {
    cli::cli_abort(
      "{.arg data} must be data frame, not {.obj_type_friendly {data}}."
    )
  }
  if (inherits(data, "grouped_df")) {
    cli::cli_abort(c(
      "{.fn tf_nest} does not work for {.cls grouped_df}.",
      i = "{.fn ungroup} your data before nesting."
    ))
  }
  arg_var <- as_name(enexpr(.arg))
  id_var <- as_name(enexpr(.id))
  if (...length() == 0) {
    value_vars <- setdiff(names(data), c(id_var, arg_var))
  } else {
    value_vars <- names(eval_select(expr(c(...)), data))
  }
  n_value_vars <- length(value_vars)
  if (n_value_vars == 0) {
    return(data)
  }
  # homogenize inputs:
  if (!length(evaluator) %in% c(1, n_value_vars)) {
    cli::cli_abort(
      "{.arg evaluator} length must be 1 or {n_value_vars}, not {length(evaluator)}."
    )
  }
  if (!is.list(domain)) {
    domain <- replicate(n_value_vars, domain, simplify = FALSE)
  } else {
    if (!length(domain) %in% c(1, n_value_vars)) {
      cli::cli_abort(
        "{.arg domain} length must be 1 or {n_value_vars}, not {length(domain)}."
      )
    }
  }

  evaluator <- as.list(evaluator)
  if (!length(evaluator) %in% c(1, n_value_vars)) {
    cli::cli_abort(
      "{.arg evaluator} length must be 1 or {n_value_vars}, not {length(evaluator)}."
    )
  }

  remaining <- setdiff(names(data), c(id_var, arg_var, value_vars))
  # check that nesting is possible without information loss
  ret <- data |>
    select(all_of(c(id_var, remaining))) |>
    group_by(across(all_of(id_var)))
  not_constant <- ret |>
    summarise(across(everything(), n_distinct)) |>
    select(-all_of(id_var)) |>
    summarise(across(everything(), \(x) !all(x == 1L))) |>
    select(where(isTRUE))
  if (ncol(not_constant) > 0) {
    cli::cli_abort(
      "Can't nest - columns {.val {names(not_constant)}} are not constant for all levels of {.arg id}"
    )
  }

  # keep first line of every id-level:
  ret <- ret |> slice(1) |> ungroup()

  tfd_list <- pmap(
    list(value_vars, evaluator, domain),
    function(x, y, z) {
      data |>
        select(all_of(c(id_var, arg_var, x))) |>
        tfd(evaluator = !!y, domain = z)
    }
  )
  names(tfd_list) <- value_vars
  # re-index to make sure order is correct,
  # use <character> index so that numeric ids are interpreted correctly
  # as names (f["1000",]), not index positions (f[1000, ])
  id_index <- ret |> pull(id_var) |> as.character()
  for (v in value_vars) {
    ret[[v]] <- tfd_list[[v]][id_index, ]
  }
  ret
}

#-------------------------------------------------------------------------------

#' Turn (data frames with) `tf`-objects / list columns into "long" tables.
#'
#' Similar in spirit to [tidyr::unnest()], the reverse of [tf_nest()].
#' The `tf`-method simply turns a single `tfd` or `tfb` vector into a "long" [tibble::tibble()].
#'
#' - Caution: this uses slightly different defaults for names of unnested columns
#'   than `tidyr::unnest()`.
#' - For `data.frames`, include an ID column with a unique row identifier before
#'   unnesting. Without it, arg-value pairs cannot be matched back to their
#'   original functions after unnesting.
#'
#' @param data a data.frame or a `tf`-object
#' @param arg optional values for the `arg` argument of
#'   [tf::tf_evaluate()]
#' @param interpolate return function values for `arg`-values not on original grid?
#'   Defaults to `TRUE`.
#' @param ... not used currently
#' @inheritParams tidyr::unnest
#' @returns a "long" data frame with `tf`-columns expanded into `arg, value`-
#'   columns.
#' @seealso [tf_evaluate.data.frame()]
#' @export
#' @family tidyfun data wrangling functions
#' @examples
#' d <- dplyr::tibble(id = 1:3)
#' d$f <- tf_rgp(3, 11L)
#' tf_unnest(d, f)
tf_unnest <- function(data, cols, arg, interpolate = TRUE, ...) {
  UseMethod("tf_unnest")
}

#' @export
#' @importFrom tibble tibble
#' @importFrom tidyr unnest
#' @rdname tf_unnest
tf_unnest.tf <- function(data, cols, arg, interpolate = TRUE, ...) {
  if (missing(arg)) {
    arg <- tf::ensure_list(tf_arg(data))
  }
  tmp <- data[, arg, matrix = FALSE, interpolate = interpolate]
  id <- unique_id(names(data)) %||% seq_along(data)
  id <- ordered(id, levels = id) # don't reshuffle
  tidyr::unnest(tibble::tibble(id = id, data = tmp), cols = data)
}

#' @export
#' @rdname tf_unnest
tf_unnest.data.frame <- function(
  data,
  cols,
  arg,
  interpolate = TRUE,
  keep_empty = FALSE,
  ptype = NULL,
  names_sep = "_",
  names_repair = "check_unique",
  ...
) {
  if (missing(cols)) {
    tf_cols <- names(data)[map_lgl(data, is_tf)]
    cols <- expr(!!!syms(tf_cols))
    cli::cli_warn(
      "{.arg cols} is now required. Please use {.code cols = {expr_text(cols)}}"
    )
  }

  ret <- tf_evaluate.data.frame(data, {{ cols }}, arg = arg) |>
    tidyr::unnest(
      cols = {{ cols }},
      keep_empty = keep_empty,
      ptype = ptype,
      names_sep = names_sep,
      names_repair = names_repair
    )
  ret
}

data <- refund::DTI[1:5, ]
d1 <- data.frame(data$cca)
d2 <- data.frame(id = data$ID, data$cca)
cca <- tfd(data$cca)

test_that("tf_gather works", {
  expect_s3_class(tf_gather(d1)$cca |> suppressMessages(), "tfd")
  expect_message(tf_gather(d1)$cca, "cca")
  expect_identical(tf_gather(d1)$cca |> suppressMessages(), cca)
  expect_equal(
    tf_gather(d1)$cca |> suppressMessages(),
    tf_gather(data[, 1:8], cca)$cca |> suppressMessages(),
    ignore_attr = TRUE
  )

  expect_identical(tf_gather(d2, -1)$cca |> suppressMessages(), cca)
  expect_identical(tf_gather(d2, -1)$id |> suppressMessages(), d2$id)
  expect_identical(
    tf_gather(d2, -1, key = "nuhnuh")$nuhnuh |>
      suppressMessages(),
    tf_gather(d2, -1)$cca |> suppressMessages()
  )

  expect_identical(
    tf_gather(d2, -id)$cca |> suppressMessages(),
    tf_gather(d2, -1)$cca |> suppressMessages()
  )
  expect_identical(
    tf_gather(d2, starts_with("cca"))$cca |> suppressMessages(),
    tf_gather(d2, -1)$cca |> suppressMessages()
  )

  expect_identical(
    attr(tf_gather(d1, evaluator = tf_approx_spline)$cca, "evaluator_name") |>
      suppressMessages(),
    "tf_approx_spline"
  )

  expect_named(tf_gather(d1) |> suppressMessages(), "cca")
})

test_that("tf_spread works", {
  d <- tibble(g = 1:3, f = tf_rgp(3, 11L))
  expect_equal(
    tf_spread(d, f, sep = NULL)[, -1],
    as.data.frame(as.matrix(d$f)),
    ignore_attr = TRUE
  )
  expect_identical(tf_spread(d, f), tf_spread(d, -g))
  expect_identical(tf_spread(d, f), tf_spread(d))
  expect_warning(
    tf_spread(d, f, arg = seq(0, 1, length.out = 20), sep = NULL),
    "interpolate = FALSE"
  )
  expect_equal(
    suppressWarnings(
      tf_spread(d, f, arg = seq(0, 1, length.out = 20), sep = NULL)[, -1]
    ),
    suppressWarnings(
      as.data.frame(d$f[, seq(0, 1, length.out = 20), interpolate = FALSE])
    ),
    ignore_attr = TRUE
  )
  d$fb <- tfb(tf_rgp(3, 11L), verbose = FALSE) |> suppressWarnings()
  expect_error(tf_spread(d), "More than one")
  expect_equal(
    tf_spread(d, fb, sep = NULL)[, -(1:2)] |> suppressMessages(),
    as.data.frame(as.matrix(d$fb)),
    ignore_attr = TRUE
  )
  set.seed(1312)
  d$fi <- tf_jiggle(tf_rgp(3, 11L))
  tf_spread(d, fi) |>
    expect_warning("no explicit `arg` for irregular") |>
    suppressWarnings()
  tf_spread(d, fi) |>
    expect_warning("interpolate = FALSE") |>
    suppressWarnings()
  expect_true(suppressWarnings(ncol(tf_spread(d, fi)) == 36))
  expect_equal(
    tf_spread(
      d,
      fi,
      arg = seq(0, 1, length.out = 20),
      sep = NULL,
      interpolate = TRUE
    )[, -(1:3)],
    as.data.frame(
      as.matrix(d$fi, arg = seq(0, 1, length.out = 20), interpolate = TRUE)
    ),
    ignore_attr = TRUE
  )
})

test_that("tf_nest works", {
  f1 <- tf_rgp(3, 11L)
  f2 <- tf_rgp(3, 11L)
  data <- dplyr::inner_join(tf_unnest(f1), tf_unnest(f2), by = c("id", "arg"))
  expect_equal(tf_nest(data)$value.x, f1, ignore_attr = TRUE)
  expect_equal(tf_nest(data)$value.y, f2, ignore_attr = TRUE)
  expect_named(tf_nest(data, value.x:value.y), names(tf_nest(data)))
  expect_named(tf_nest(data, -(1:2)), names(tf_nest(data)))

  g <- rnorm(3)
  data <- bind_cols(data, g = rep(g, e = tf_count(f1)))
  expect_error(tf_nest(data |> group_by(g)), "grouped_df")
  expect_identical(tf_nest(data, value.x:value.y)$g, g)
  data <- bind_cols(data, f = rep(rnorm(nrow(data))))
  expect_error(tf_nest(data, value.x:value.y), "Can't nest")
})

# weird scoping problem going on -- fixed by assigning
# data to tf_evaluate as evaluated object instead of quoted for now,
# otherwise tfdata is not found inside test environments (related to testthat/#266?)
test_that("tf_unnest works", {
  set.seed(121211)
  f1 <- tf_rgp(3, 11L)
  f2 <- tf_rgp(3, 11L)
  data <- dplyr::inner_join(tf_unnest(f1), tf_unnest(f2), by = c("id", "arg"))
  tfdata <- tf_nest(data)
  expect_identical(NCOL(tf_unnest(tfdata, cols = c(value.x, value.y))), 5L)
  expect_equal(
    as.matrix(tf_unnest(tfdata, cols = c(value.x, value.y))[-c(1, 4)]),
    as.matrix(data[, 2:4]),
    ignore_attr = TRUE
  )
  expect_s3_class(tf_unnest(tfdata, value.x)$value.y, "tfd")
})

# tidyfun#109
test_that("tf_nest / tf_unnest work with numeric id-variables", {
  d <- tf_rgp(10) |> tf_unnest()
  d$id <- as.numeric(d$id) * 10
  nested <- tf_nest(d)
  unnested <- tf_unnest(nested, cols = value)
  expect_equal(d, unnested, ignore_attr = TRUE)
})

#' Evaluate `tf`s inside a `data.frame`
#'
#' @details The `arg`-argument of `tf_evaluate.data.frame` method can be a
#'   list of `arg`-vectors or -lists used as the `arg` argument for the
#'   [tf::tf_evaluate()]-method for the respective `tf`-columns in `object`.
#'   `...` is not used for a `tf`-`object`, but a second unnamed argument to these
#'   methods will be interpreted as `arg`.
#' @param object a `data.frame`-like object with `tf` columns.
#' @param ... optional: a selection of `tf`-columns. If empty, all `tf`-variables
#'   in the data frame are selected. You can supply bare variable names,
#'   select all variables between `x` and `z` with `x:z`, exclude `y` with `-y`.
#'   For more options, see the [dplyr::select()] documentation.
#' @param arg optional evaluation grid (vector or list of vectors).
#'   Defaults to `tf_arg(object)`.
#' @returns Replaces `tf`-columns with list columns of
#'   smaller `data.frames` containing the functions' arguments (`arg`) and
#'   evaluations (`value`) and returns the modified nested dataframe.
#' @export
#' @import tf
#' @importFrom tidyselect eval_select
#' @importFrom purrr map map_lgl pmap
#' @family tidyfun data wrangling functions
#' @examples
#' d <- dplyr::tibble(id = 1:3)
#' d$f <- tf_rgp(3, 11L)
#' str(tf_evaluate(d))
tf_evaluate.data.frame <- function(object, ..., arg) {
  # figure out which tf columns to evaluate:
  tf_cols <- names(object)[map_lgl(object, is_tf)]
  if (...length() > 0) {
    tf_to_evaluate <- names(eval_select(expr(c(...)), object))
    tf_cols <- intersect(tf_cols, tf_to_evaluate)
  }
  if (!length(tf_cols)) {
    cli::cli_warn("Nothing to be done for {.fn tf_evaluate}.")
    return(object)
  }
  if (!missing(arg) && !is.null(arg)) {
    arg <- tf::ensure_list(arg)
    if (length(arg) == 1 && length(tf_cols) > 1) {
      arg <- replicate(length(tf_cols), arg, simplify = FALSE)
    }
  } else {
    arg <- map(object[tf_cols], \(x) tf::ensure_list(tf_arg(x)))
  }
  if (length(arg) != length(tf_cols)) {
    cli::cli_abort("{.arg arg} length must match number of {.cls tf} columns.")
  }
  names(arg) <- tf_cols
  # convert them to list-columns of data.frames
  for (f in tf_cols) {
    object[[f]] <- object[[f]][, arg[[f]], matrix = FALSE]
  }
  object
}

#' Error bands using `tf` objects as bounds
#'
#' Plots a shaded region between `tf`-objects `ymax` and `ymin`.
#' This is primarily intended to help with plotting confidence bands
#' although other purposes are possible.
#'
#' @examples
#' set.seed(1221)
#' data <- data.frame(id = factor(1:2))
#' data$f <- tf_rgp(2)
#' data$ymax <- data$f + 1
#' data$ymin <- data$f - 1
#' library(ggplot2)
#' ggplot(data, aes(y = f, color = id)) +
#'   geom_spaghetti() +
#'   geom_errorband(aes(ymax = ymax, ymin = ymin, fill = id)) +
#'   facet_wrap(~id)
#' @returns A [ggplot2::layer()] object for use in a ggplot.
#' @name ggerrorband
NULL

#' @export
#' @rdname ggerrorband
#' @usage NULL
#' @format NULL
StatErrorband <- ggproto(
  "StatErrorband",
  Stat,
  required_aes = c("ymax", "ymin"),
  setup_params = function(data, params) {
    if (is.null(params$arg)) {
      params$arg <- list(tf_arg(pull(data, ymax)))
    }
    params
  },
  compute_layer = function(self, data, params, layout) {
    if (!is_tf(data$ymin) || !is_tf(data$ymax)) {
      cli::cli_abort(
        "{.arg ymax} and {.arg ymin} must be {.cls tf} objects, not {.obj_type_friendly {data$ymin}} and {.obj_type_friendly {data$ymax}}."
      )
    }
    tf_eval <-
      suppressMessages(
        data |>
          mutate(id = names(ymax) %||% seq_along(ymax)) |>
          tf_unnest(c(ymax, ymin), .arg = params$arg, names_sep = "___")
      ) |>
      select(-group, -ymin___arg) |>
      rename(
        group = id,
        x = ymax___arg,
        ymin = ymin___value,
        ymax = ymax___value
      )
    tf_eval
  },
  compute_panel = function(self, data, scales, arg, errorband) {
    Stat$compute_panel(self, data, scales)
  }
)

#' @export
#' @rdname ggerrorband
#' @inheritParams ggplot2::stat_identity
#' @param na.rm remove NAs? defaults to `TRUE`
stat_errorband <- function(
  mapping = NULL,
  data = NULL,
  geom = "errorband",
  position = "identity",
  na.rm = TRUE,
  show.legend = NA,
  inherit.aes = TRUE,
  arg = NULL,
  ...
) {
  layer(
    stat = StatErrorband,
    data = data,
    mapping = mapping,
    geom = geom,
    position = position,
    show.legend = show.legend,
    inherit.aes = inherit.aes,
    params = list(na.rm = na.rm, arg = arg, ...)
  )
}

# geom --------------------------------------------------------------------

#' @export
#' @rdname ggerrorband
#' @format NULL
#' @param arg where to evaluate `tf`; defaults to the object's default
#'   evaluation grid.
geom_errorband <- function(
  mapping = NULL,
  data = NULL,
  position = "identity",
  na.rm = TRUE,
  show.legend = NA,
  inherit.aes = TRUE,
  arg = NULL,
  ...
) {
  layer(
    stat = StatErrorband,
    data = data,
    mapping = mapping,
    geom = "errorband",
    position = position,
    show.legend = show.legend,
    inherit.aes = inherit.aes,
    params = list(na.rm = na.rm, arg = arg, ...)
  )
}
#' @export
#' @rdname ggerrorband
#' @usage NULL
#' @format NULL
GeomErrorband <- ggproto(
  "GeomErrorband",
  Geom,
  setup_params = function(data, params) {
    # TODO: implement proper "orientation" - see extending ggplot vignette
    params$flipped_aes <- FALSE
    params
  },
  setup_data = function(data, params) {
    GeomRibbon$setup_data(data, params)
  },
  draw_panel = function(data, panel_params, coord) {
    GeomRibbon$draw_panel(data, panel_params, coord)
  },
  default_aes = aes(
    fill = "grey70",
    linetype = 0,
    alpha = 0.3,
    linewidth = 0.1
  ),
  draw_key = GeomRibbon$draw_key,
  required_aes = c("ymin", "ymax")
)

# Helper functions for tf_ggplot testing

# Create simple test tf data
create_test_tf_data <- function(n_funcs = 3, n_points = 11, seed = 123) {
  set.seed(seed)
  data <- data.frame(
    id = seq_len(n_funcs),
    group = factor(sample(c("A", "B"), n_funcs, replace = TRUE))
  )
  data$func <- tf_rgp(n_funcs, arg = seq(0, 1, length.out = n_points))
  data
}

# Create test data with multiple tf columns
create_multi_tf_data <- function(n_funcs = 2, n_points = 11, seed = 123) {
  set.seed(seed)
  data <- data.frame(id = seq_len(n_funcs))
  data$func1 <- tf_rgp(n_funcs, arg = seq(0, 1, length.out = n_points))
  data$func2 <- tf_rgp(n_funcs, arg = seq(0, 1, length.out = n_points))
  data$func3 <- tf_rgp(n_funcs, arg = seq(0, 1, length.out = n_points))
  data
}

# Create test data with confidence bands
create_band_tf_data <- function(n_funcs = 2, n_points = 11, seed = 123) {
  set.seed(seed)
  data <- data.frame(id = seq_len(n_funcs))

  # Create mean function
  data$mean_func <- tf_rgp(n_funcs, arg = seq(0, 1, length.out = n_points))

  # Create confidence bands by adding/subtracting noise
  mean_vals <- tf_evaluations(data$mean_func)
  lower_vals <- map(mean_vals, ~ .x - abs(rnorm(length(.x), 0, 0.1)))
  upper_vals <- map(mean_vals, ~ .x + abs(rnorm(length(.x), 0, 0.1)))

  data$lower_func <- tfd(lower_vals, arg = seq(0, 1, length.out = n_points))
  data$upper_func <- tfd(upper_vals, arg = seq(0, 1, length.out = n_points))

  data
}

# Check if a ggplot object has expected aesthetic mappings
check_plot_mappings <- function(plot, expected_aes) {
  built <- ggplot_build(plot)
  actual_aes <- names(built$data[[1]])
  all(expected_aes %in% actual_aes)
}

# Count unique groups in plot data
count_plot_groups <- function(plot) {
  built <- ggplot_build(plot)
  length(unique(built$data[[1]]$group))
}

# Extract x values from plot data
get_plot_x_values <- function(plot) {
  built <- ggplot_build(plot)
  sort(unique(built$data[[1]]$x))
}

# Check if plot data has expected number of rows
check_plot_nrows <- function(plot, expected_rows) {
  built <- ggplot_build(plot)
  nrow(built$data[[1]]) == expected_rows
}

# Test if tf_ggplot functionality is available
skip_if_no_tf_ggplot <- function() {
  skip_if_not(exists("tf_ggplot"), "tf_ggplot function not available")
}

capture_warnings_silently <- function(expr) {
  warnings <- character(0)
  value <- withCallingHandlers(
    expr,
    warning = function(w) {
      warnings <<- c(warnings, conditionMessage(w))
      invokeRestart("muffleWarning")
    }
  )

  list(value = value, warnings = warnings)
}

# Tests for complex tf_ggplot use cases
# Tests complex expressions in tf aesthetics and derived quantities

test_that("tf aesthetic with function transformations works", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  # Create test data with positive and negative values
  set.seed(123)
  data <- data.frame(id = 1:3)
  arg_vals <- seq(-2, 2, length.out = 11)
  func_vals <- list(
    sin(arg_vals), # Has negative values
    cos(arg_vals), # Has negative values
    arg_vals^2 - 1 # Has negative values
  )
  data$curves <- tfd(func_vals, arg = arg_vals)

  # Test abs() transformation
  expect_no_error({
    p_abs <- tf_ggplot(data, aes(tf = abs(curves))) +
      geom_line()
  })

  # Check that plot builds correctly
  built_abs <- ggplot_build(p_abs)
  plot_data_abs <- built_abs$data[[1]]

  # All y values should be non-negative after abs()
  expect_true(all(plot_data_abs$y >= 0))

  # Should have 3 groups (one per function)
  expect_equal(length(unique(plot_data_abs$group)), 3)

  # Should have correct number of rows (3 functions × 11 points)
  expect_equal(nrow(plot_data_abs), 33)
})

test_that("tf aesthetic with mathematical operations works", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  set.seed(456)
  data <- data.frame(id = 1:2)
  data$func1 <- tf_rgp(2, arg = seq(0, 1, length.out = 6))
  data$func2 <- tf_rgp(2, arg = seq(0, 1, length.out = 6))

  # Test arithmetic operations
  expect_no_error({
    p_sum <- tf_ggplot(data, aes(tf = func1 + func2)) +
      geom_line()
    p_scaled <- tf_ggplot(data, aes(tf = 2 * func1)) +
      geom_line()
  })

  # Check that plots build correctly
  built_sum <- ggplot_build(p_sum)
  built_scaled <- ggplot_build(p_scaled)

  expect_equal(length(unique(built_sum$data[[1]]$group)), 2)
  expect_equal(length(unique(built_scaled$data[[1]]$group)), 2)
  expect_equal(nrow(built_sum$data[[1]]), 12) # 2 functions × 6 points
  expect_equal(nrow(built_scaled$data[[1]]), 12)
})

test_that("color aesthetic with tf-derived quantities works", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  set.seed(789)
  data <- data.frame(id = 1:4)
  data$curves <- tf_rgp(4, arg = seq(0, 1, length.out = 11))

  # Test color by tf_depth
  expect_no_error({
    p_depth <- tf_ggplot(data, aes(tf = curves, color = tf_depth(curves))) +
      geom_line()
  })

  built_depth <- ggplot_build(p_depth)
  plot_data_depth <- built_depth$data[[1]]

  # Should have 4 groups (one per function)
  expect_equal(length(unique(plot_data_depth$group)), 4)

  # Color should vary based on depth
  expect_true(length(unique(plot_data_depth$colour)) > 1)

  # Each group should have consistent color within group
  color_by_group <- split(plot_data_depth$colour, plot_data_depth$group)
  consistent_colors <- all(sapply(
    color_by_group,
    function(x) length(unique(x)) == 1
  ))
  expect_true(consistent_colors)
})

test_that("multiple derived aesthetics work together", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  set.seed(101112)
  data <- data.frame(id = 1:5)
  data$curves <- tf_rgp(5, arg = seq(0, 1, length.out = 11))

  # Test multiple tf-derived aesthetics
  expect_no_error({
    p_multi <- tf_ggplot(
      data,
      aes(
        tf = curves,
        color = tf_depth(curves),
        alpha = tf_fmin(curves)
      )
    ) +
      geom_line()
  })
  expect_no_error(print(p_multi))
  built_multi <- ggplot_build(p_multi)
  plot_data_multi <- built_multi$data[[1]]

  # Should have all aesthetic mappings
  expect_true("colour" %in% names(plot_data_multi))
  expect_true("alpha" %in% names(plot_data_multi))

  # Should have 5 groups
  expect_equal(length(unique(plot_data_multi$group)), 5)

  # Alpha should vary based on function range
  expect_true(length(unique(plot_data_multi$alpha)) > 1)
})

test_that("complex tf expressions in constructor vs geom are equivalent", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  set.seed(131415)
  data <- data.frame(id = 1:3)
  data$curves <- tf_rgp(3, arg = seq(-1, 1, length.out = 9))

  # Method 1: Complex expression in constructor
  p1 <- tf_ggplot(data, aes(tf = abs(curves), color = tf_depth(curves))) +
    geom_line()

  # Method 2: Complex expression in geom
  p2 <- tf_ggplot(data) +
    geom_line(aes(tf = abs(curves), color = tf_depth(curves)))

  # Both should build successfully
  built1 <- ggplot_build(p1)
  built2 <- ggplot_build(p2)

  # Should have same dimensions
  expect_equal(dim(built1$data[[1]]), dim(built2$data[[1]]))

  # Should have same number of groups
  expect_equal(
    length(unique(built1$data[[1]]$group)),
    length(unique(built2$data[[1]]$group))
  )

  # Both should have only non-negative y values (due to abs)
  expect_true(all(built1$data[[1]]$y >= 0))
  expect_true(all(built2$data[[1]]$y >= 0))

  # Colors should vary in both (due to tf_depth)
  expect_true(length(unique(built1$data[[1]]$colour)) > 1)
  expect_true(length(unique(built2$data[[1]]$colour)) > 1)
})

test_that("nested function calls in tf aesthetics work", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  set.seed(161718)
  data <- data.frame(id = 1:2)
  data$curves <- tf_rgp(2, arg = seq(-1, 1, length.out = 7))

  # Test nested function calls
  expect_no_error({
    p_nested <- tf_ggplot(data, aes(tf = sqrt(abs(curves)))) +
      geom_line()
  })

  built_nested <- ggplot_build(p_nested)
  plot_data_nested <- built_nested$data[[1]]

  # All y values should be non-negative (due to sqrt(abs()))
  expect_true(all(plot_data_nested$y >= 0))
  expect_true(all(is.finite(plot_data_nested$y)))

  # Should have 2 groups
  expect_equal(length(unique(plot_data_nested$group)), 2)
})

test_that("tf aesthetics with scalars and derived quantities", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  set.seed(192021)
  data <- data.frame(
    id = 1:4,
    treatment = factor(c("A", "A", "B", "B")),
    baseline = rnorm(4)
  )
  data$curves <- tf_rgp(4, arg = seq(0, 1, length.out = 8))

  # Mix tf aesthetics with scalar summaries and regular variables
  expect_no_error({
    p_mixed <- tf_ggplot(
      data,
      aes(
        tf = curves,
        color = treatment, # Regular factor variable
        linewidth = tf_fmean(curves), # Scalar summary of tf
        alpha = abs(baseline) # Transformation of scalar
      )
    ) +
      geom_line()
  })

  built_mixed <- ggplot_build(p_mixed)
  plot_data_mixed <- built_mixed$data[[1]]

  # Should have all aesthetics
  expect_true("colour" %in% names(plot_data_mixed))
  expect_true("linewidth" %in% names(plot_data_mixed))
  expect_true("alpha" %in% names(plot_data_mixed))

  # Should have 4 groups (one per function)
  expect_equal(length(unique(plot_data_mixed$group)), 4)

  # Linewidth should vary based on function means
  expect_true(length(unique(plot_data_mixed$linewidth)) > 1)

  # Alpha should vary based on baseline
  expect_true(length(unique(plot_data_mixed$alpha)) > 1)

  # Color should reflect treatment levels (2 levels)
  expect_equal(length(unique(plot_data_mixed$colour)), 2)
})

test_that("tf transformations preserve grouping correctly", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  set.seed(222324)
  data <- data.frame(
    id = 1:6,
    group = factor(rep(c("X", "Y", "Z"), each = 2))
  )
  data$curves <- tf_rgp(6, arg = seq(0, 1, length.out = 5))

  # Test transformation with grouping variable
  p_transform <- tf_ggplot(
    data,
    aes(
      tf = abs(curves - tf_fmean(curves)), # Center around mean then abs
      color = group
    )
  ) +
    geom_line()

  built_transform <- ggplot_build(p_transform)
  plot_data_transform <- built_transform$data[[1]]

  # Should have 6 groups (one per function)
  expect_equal(length(unique(plot_data_transform$group)), 6)

  # Should have 3 colors (for groups X, Y, Z)
  expect_equal(length(unique(plot_data_transform$colour)), 3)

  # All y values should be non-negative (due to abs)
  expect_true(all(plot_data_transform$y >= 0))

  # Should have 30 rows (6 functions × 5 points)
  expect_equal(nrow(plot_data_transform), 30)
})

test_that("error handling for invalid tf expressions", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  set.seed(252627)
  data <- data.frame(id = 1:2)
  data$curves <- tf_rgp(2, arg = seq(0, 1, length.out = 5))
  data$not_tf <- rnorm(2)

  # Should error when applying tf functions to non-tf objects
  expect_error(
    {
      p <- tf_ggplot(data, aes(tf = abs(not_tf))) +
        geom_line()
      ggplot_build(p) # Trigger validation
    },
    "tf.*object|must be.*tf"
  )

  # Should error when tf aesthetic references non-existent column
  expect_error(
    {
      p <- tf_ggplot(data, aes(tf = nonexistent_func)) +
        geom_line()
      ggplot_build(p) # Trigger validation
    },
    "not found|object.*not found"
  )
})

test_that("performance with complex tf expressions", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  # Test with moderately large dataset that triggers warning
  set.seed(282930)
  data <- data.frame(id = 1:100) # 100 functions
  data$curves <- tf_rgp(100, arg = seq(0, 1, length.out = 101)) # 101 points each (100 * 101 = 10100 > 10000)

  # Create large plot (warning logic needs to be re-added to finalization process)
  p_large <- tf_ggplot(
    data,
    aes(
      tf = abs(curves),
      color = tf_depth(curves)
    )
  ) +
    geom_line()

  # But should still build correctly
  built_large <- ggplot_build(p_large)
  plot_data_large <- built_large$data[[1]]

  # Should have 100 groups
  expect_equal(length(unique(plot_data_large$group)), 100)

  # Should have 10100 rows (100 functions × 101 points)
  expect_equal(nrow(plot_data_large), 10100)

  # All y values should be non-negative
  expect_true(all(plot_data_large$y >= 0))
})

test_that("complex expressions work with faceting", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  set.seed(313233)
  data <- data.frame(
    id = 1:8,
    condition = factor(rep(c("Control", "Treatment"), each = 4)),
    replicate = factor(rep(1:4, 2))
  )
  data$curves <- tf_rgp(8, arg = seq(0, 1, length.out = 6))

  # Test complex expressions with faceting
  expect_no_error({
    p_facet <- tf_ggplot(
      data,
      aes(
        tf = abs(curves),
        color = tf_depth(curves)
      )
    ) +
      geom_line() +
      facet_wrap(~condition)
  })

  built_facet <- ggplot_build(p_facet)

  # Should have 2 panels (Control, Treatment)
  expect_equal(length(unique(built_facet$layout$layout$PANEL)), 2)

  # Should have 8 groups total
  plot_data_facet <- built_facet$data[[1]]
  expect_equal(length(unique(plot_data_facet$group)), 8)

  # All y values should be non-negative
  expect_true(all(plot_data_facet$y >= 0))
})

test_that("tf expressions work with custom evaluation grids", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  set.seed(343536)
  data <- data.frame(id = 1:3)
  data$curves <- tf_rgp(3, arg = seq(0, 1, length.out = 21)) # Dense original grid

  # Test with custom coarser grid
  custom_arg <- seq(0, 1, length.out = 6) # Coarser grid

  expect_no_error({
    p_custom <- tf_ggplot(data, aes(tf = abs(curves)), arg = custom_arg) +
      geom_line()
  })

  built_custom <- ggplot_build(p_custom)
  plot_data_custom <- built_custom$data[[1]]

  # Should use custom grid (6 points)
  expect_equal(length(unique(plot_data_custom$x)), 6)
  expect_equal(sort(unique(plot_data_custom$x)), custom_arg)

  # Should have 18 rows (3 functions × 6 points)
  expect_equal(nrow(plot_data_custom), 18)

  # All y values should be non-negative
  expect_true(all(plot_data_custom$y >= 0))
})

# Advanced tf_ggplot functionality tests
# These test more complex scenarios and edge cases

test_that("multiple tf aesthetics work correctly", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  data <- create_multi_tf_data(n_funcs = 2, n_points = 5)

  # Test tf_x and tf_y aesthetics
  p <- tf_ggplot(data) +
    geom_point(aes(tf_x = func1, tf_y = func2))

  built <- ggplot_build(p)
  plot_data <- built$data[[1]]

  # Should have both x and y from tf objects
  expect_true(all(c("x", "y") %in% names(plot_data)))

  # Should have correct number of points (2 functions × 5 evaluations)
  expect_equal(nrow(plot_data), 10)

  # Should have 2 groups (one per function pair)
  expect_equal(length(unique(plot_data$group)), 2)
})

test_that("ribbon geoms work with tf_ymin and tf_ymax", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  data <- create_band_tf_data(n_funcs = 2, n_points = 5)

  # Test ribbon with tf confidence bands
  p <- tf_ggplot(data) +
    geom_ribbon(
      aes(tf_ymin = lower_func, tf_ymax = upper_func),
      alpha = 0.3
    )

  built <- ggplot_build(p)
  plot_data <- built$data[[1]]

  # Should have ymin and ymax columns
  expect_true(all(c("ymin", "ymax") %in% names(plot_data)))

  # Should have correct structure
  expect_equal(nrow(plot_data), 10) # 2 functions × 5 points
  expect_equal(length(unique(plot_data$group)), 2)

  # ymax should be >= ymin for all points
  expect_true(all(plot_data$ymax >= plot_data$ymin))
})

test_that("summary ribbon tf aesthetics broadcast across plot-level tf data", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  set.seed(123)
  data <- data.frame(g = gl(2, 5))
  data$f <- tf_rgp(10) + 5

  expect_no_warning({
    built <- ggplot_build(
      tf_ggplot(data, aes(tf = f, tf_ymin = min(f), tf_ymax = max(f))) +
        geom_ribbon(alpha = 0.01)
    )
  })

  plot_data <- built$data[[1]]
  expect_true(all(c("ymin", "ymax") %in% names(plot_data)))
  expect_true(all(is.finite(plot_data$ymin)))
  expect_true(all(is.finite(plot_data$ymax)))
})

test_that("summary ribbon and summary line layers work without row-size errors", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  set.seed(123)
  data <- data.frame(id = 1:10)
  data$f <- tf_rgp(10) + 5

  expect_no_error({
    built <- ggplot_build(
      tf_ggplot(data) +
        geom_ribbon(
          aes(tf_ymin = mean(f) - sd(f), tf_ymax = mean(f) + sd(f)),
          alpha = 0.3
        ) +
        geom_line(aes(tf = mean(f)))
    )
  })

  expect_equal(length(built$data), 2)
  expect_true(all(c("ymin", "ymax") %in% names(built$data[[1]])))
  expect_true(all(c("x", "y") %in% names(built$data[[2]])))
})

test_that("combining tf and regular geoms works", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  data <- create_test_tf_data(n_funcs = 3, n_points = 5)
  data$mean_val <- c(1, 2, 3)

  # This should work - different aesthetics
  captured <- capture_warnings_silently(
    tf_ggplot(data) +
      geom_line(aes(tf = func, color = group)) +
      geom_point(aes(x = 0.5, y = mean_val), size = 3) # Regular geom
  )
  p <- captured$value
  expect_true(any(grepl(
    "Potential.*conflict|scale.*conflict",
    captured$warnings
  )))

  built <- ggplot_build(p)

  # Should have 2 layers
  expect_equal(length(built$data), 2)

  # First layer should be tf data (15 rows: 3 functions × 5 points)
  expect_equal(nrow(built$data[[1]]), 15)

  # Second layer should be regular data (3 points)
  expect_equal(nrow(built$data[[2]]), 3)
})

test_that("faceting works with tf data", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  data <- create_test_tf_data(n_funcs = 6, n_points = 5)
  data$treatment <- factor(rep(c("A", "B"), each = 3))

  # Test faceting by treatment
  p <- tf_ggplot(data, aes(tf = func)) +
    geom_line() +
    facet_wrap(~treatment)

  built <- ggplot_build(p)

  # Should have 2 panels (one per treatment)
  expect_equal(length(unique(built$layout$layout$PANEL)), 2)

  # Plot should build successfully and have correct structure
  plot_data <- built$data[[1]]
  expect_equal(length(unique(plot_data$group)), 6) # 6 functions
  expect_equal(nrow(plot_data), 30) # 6 functions × 5 points
})

test_that("irregular tf objects are handled correctly", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  # Create irregular tf data by sparsifying
  data <- data.frame(id = 1:2)
  regular_tf <- tf_rgp(2, arg = seq(0, 1, length.out = 11))
  data$irreg_func <- tf_sparsify(regular_tf, dropout = 0.5)

  # Should still work with irregular data
  p <- tf_ggplot(data, aes(tf = irreg_func)) + geom_line()

  built <- ggplot_build(p)
  plot_data <- built$data[[1]]

  # Should have data for both functions
  expect_equal(length(unique(plot_data$group)), 2)
  expect_true(nrow(plot_data) > 0)

  # Should have fewer points than regular data due to sparsification
  expect_true(nrow(plot_data) <= 2 * 11)
})

test_that("different geom types work with tf aesthetics", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  data <- create_test_tf_data(n_funcs = 2, n_points = 5)

  # Test different geoms
  p_line <- tf_ggplot(data, aes(tf = func)) + geom_line()
  p_point <- tf_ggplot(data, aes(tf = func)) + geom_point()
  p_step <- tf_ggplot(data, aes(tf = func)) + geom_step()
  p_area <- tf_ggplot(data, aes(tf = func)) + geom_area()

  # All should build successfully
  expect_s3_class(ggplot_build(p_line), "ggplot_built")
  expect_s3_class(ggplot_build(p_point), "ggplot_built")
  expect_s3_class(ggplot_build(p_step), "ggplot_built")
  expect_s3_class(ggplot_build(p_area), "ggplot_built")

  # All should have same basic structure
  expect_equal(count_plot_groups(p_line), 2)
  expect_equal(count_plot_groups(p_point), 2)
  expect_equal(count_plot_groups(p_step), 2)
  expect_equal(count_plot_groups(p_area), 2)
})

test_that("complex aesthetic mappings work", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  data <- create_test_tf_data(n_funcs = 4, n_points = 5)
  data$treatment <- factor(rep(c("A", "B"), each = 2))
  data$subject <- factor(1:4)
  data$baseline <- rnorm(4)

  # Complex mapping with multiple aesthetics
  p <- tf_ggplot(
    data,
    aes(tf = func, color = treatment, linetype = subject, alpha = baseline)
  ) +
    geom_line()

  built <- ggplot_build(p)
  plot_data <- built$data[[1]]

  # Should preserve all aesthetic mappings
  expected_aes <- c("colour", "linetype", "alpha")
  expect_true(all(expected_aes %in% names(plot_data)))

  # Should have correct grouping
  expect_equal(length(unique(plot_data$group)), 4)

  # Aesthetics should be consistent within groups
  aesthetics_by_group <- split(plot_data[expected_aes], plot_data$group)
  consistent <- all(sapply(aesthetics_by_group, function(x) {
    sapply(x, function(col) length(unique(col)) == 1)
  }))
  expect_true(consistent)
})

test_that("performance warnings are triggered appropriately", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  # Warning triggers only for > 200 functions AND > 100 grid points
  data <- data.frame(id = 1:201)
  data$func <- tf_rgp(201, arg = seq(0, 1, length.out = 101))

  captured <- capture_warnings_silently({
    p <- tf_ggplot(data, aes(tf = func)) + geom_line()
    ggplot_build(p)
  })
  expect_true(any(grepl(
    "Large data expansion",
    captured$warnings
  )))

  # Below threshold: no expansion warning
  data_small <- data.frame(id = 1:100)
  data_small$func <- tf_rgp(100, arg = seq(0, 1, length.out = 101))

  captured_small <- capture_warnings_silently({
    p2 <- tf_ggplot(data_small, aes(tf = func)) + geom_line()
    ggplot_build(p2)
  })
  expect_false(any(grepl(
    "Large data expansion",
    captured_small$warnings
  )))
})

test_that("scale conflicts are detected", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  data <- create_test_tf_data(n_funcs = 2, n_points = 5)
  data$scalar_y <- rnorm(2)

  # This creates a scale conflict (tf values and scalar values on same y-scale)
  expect_warning(
    {
      p <- tf_ggplot(data) +
        geom_line(aes(tf = func)) + # tf data on y-scale
        geom_point(aes(x = 0.5, y = scalar_y)) # scalar data on y-scale
    },
    "scale.*conflict|mixed.*aesthetic"
  )
})

test_that("empty or NA tf objects are handled gracefully", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  # Create data with some NA functions
  data <- data.frame(id = 1:3)
  data$func <- tf_rgp(3, arg = seq(0, 1, length.out = 5))

  # Set one function to NA
  data$func[2] <- NA

  # Should handle NA functions gracefully
  p <- tf_ggplot(data, aes(tf = func)) + geom_line()

  built <- ggplot_build(p)
  plot_data <- built$data[[1]]

  # Should only have data for non-NA functions
  expect_equal(length(unique(plot_data$group)), 2) # Only 2 non-NA functions
  expect_equal(nrow(plot_data), 2 * 5) # 2 functions × 5 points
})

test_that("theme and scale customization works with tf_ggplot", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  data <- create_test_tf_data(n_funcs = 2, n_points = 5)

  # Test that standard ggplot2 customization works
  p <- tf_ggplot(data, aes(tf = func, color = group)) +
    geom_line() +
    scale_color_manual(values = c("A" = "red", "B" = "blue")) +
    theme_minimal() +
    labs(title = "Test Plot", x = "Time", y = "Value")

  # Should build successfully with customizations
  built <- ggplot_build(p)
  expect_s3_class(built, "ggplot_built")

  # Check that labels are preserved
  expect_equal(p$labels$title, "Test Plot")
  expect_equal(p$labels$x, "Time")
  expect_equal(p$labels$y, "Value")
})

test_that("multiple tf layers work correctly", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  data <- create_test_tf_data(n_funcs = 3, n_points = 5)

  # Create plot with multiple tf layers
  p <- tf_ggplot(data) +
    geom_line(aes(tf = func, color = group), alpha = 0.7) +
    geom_point(aes(tf = func, color = group), size = 2) +
    geom_line(
      aes(tf = func),
      color = "black",
      linetype = "dashed",
      alpha = 0.3
    )

  built <- ggplot_build(p)

  # Should have 3 layers
  expect_equal(length(built$data), 3)

  # All layers should have same number of groups (3 functions)
  expect_equal(length(unique(built$data[[1]]$group)), 3) # colored lines
  expect_equal(length(unique(built$data[[2]]$group)), 3) # colored points
  expect_equal(length(unique(built$data[[3]]$group)), 3) # black dashed lines

  # All layers should have same number of rows (3 functions × 5 points)
  expect_equal(nrow(built$data[[1]]), 15)
  expect_equal(nrow(built$data[[2]]), 15)
  expect_equal(nrow(built$data[[3]]), 15)
})

test_that("mixing multiple tf and regular layers works", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  data <- create_test_tf_data(n_funcs = 3, n_points = 5)
  data$mean_val <- c(1, 2, 3)
  data$max_val <- c(1.5, 2.5, 3.5)

  # Create plot mixing tf and regular layers
  captured <- capture_warnings_silently(
    tf_ggplot(data) +
      geom_line(aes(tf = func, color = group)) + # tf layer 1
      geom_point(aes(x = 0.5, y = mean_val, color = group), size = 3) + # regular layer 1
      geom_point(aes(tf = func), alpha = 0.5) + # tf layer 2
      geom_point(aes(x = 0.8, y = max_val, color = group), size = 2) + # regular layer 2
      geom_line(aes(tf = func), linetype = "dotted") # tf layer 3
  )
  expect_true(any(grepl(
    "scale.*conflict|Potential.*conflict",
    captured$warnings
  )))
  p <- captured$value

  built <- ggplot_build(p)

  # Should have 5 layers
  expect_equal(length(built$data), 5)

  # tf layers should have expanded data (15 rows each)
  expect_equal(nrow(built$data[[1]]), 15) # tf lines
  expect_equal(nrow(built$data[[3]]), 15) # tf points
  expect_equal(nrow(built$data[[5]]), 15) # tf dotted lines

  # Regular layers should have original data (3 rows each)
  expect_equal(nrow(built$data[[2]]), 3) # regular points 1
  expect_equal(nrow(built$data[[4]]), 3) # regular points 2
})

test_that("complex multi-layer plots with different aesthetics work", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  # Create more complex data
  data <- create_band_tf_data(n_funcs = 2, n_points = 5)
  data$summary_stat <- c(0.5, -0.5)

  # Create complex multi-layer plot
  expect_warning(
    {
      p <- tf_ggplot(data, aes(color = factor(id))) +
        geom_ribbon(
          aes(tf_ymin = lower_func, tf_ymax = upper_func, fill = factor(id)),
          alpha = 0.2
        ) + # tf ribbon
        geom_line(aes(tf = mean_func), linewidth = 1) + # tf line
        geom_hline(
          aes(yintercept = summary_stat, color = factor(id)),
          linetype = "dashed"
        ) + # regular hlines
        geom_point(aes(tf = mean_func), size = 2) + # tf points
        geom_point(
          aes(x = 0.5, y = summary_stat, color = factor(id)),
          size = 4,
          shape = "diamond"
        ) # regular points
    },
    "scale.*conflict|Potential.*conflict"
  )

  built <- ggplot_build(p)

  # Should have 5 layers
  expect_equal(length(built$data), 5)

  # tf layers should have expanded data (10 rows each: 2 functions × 5 points)
  expect_equal(nrow(built$data[[1]]), 10) # ribbon
  expect_equal(nrow(built$data[[2]]), 10) # line
  expect_equal(nrow(built$data[[4]]), 10) # points

  # Regular layers should have original data (2 rows each)
  expect_equal(nrow(built$data[[3]]), 2) # hlines
  expect_equal(nrow(built$data[[5]]), 2) # points
})

test_that("many layers with different tf expressions work", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  data <- create_test_tf_data(n_funcs = 2, n_points = 5)

  # Create plot with many different tf expressions
  p <- tf_ggplot(data, aes(color = group)) +
    geom_line(aes(tf = func), alpha = 0.8) + # original functions
    geom_line(aes(tf = func + 0.5), linetype = "dashed") + # shifted up
    geom_line(aes(tf = func - 0.5), linetype = "dotted") + # shifted down
    geom_point(aes(tf = func, size = tf_fmean(func))) + # points with size by mean
    geom_point(aes(tf = func + 1, alpha = tf_depth(func))) # shifted points with alpha by depth

  built <- ggplot_build(p)

  # Should have 5 layers
  expect_equal(length(built$data), 5)

  # All layers should have same number of rows (2 functions × 5 points = 10)
  for (i in 1:5) {
    expect_equal(
      nrow(built$data[[i]]),
      10,
      info = sprintf("Layer %d should have 10 rows", i)
    )
    expect_equal(
      length(unique(built$data[[i]]$group)),
      2,
      info = sprintf("Layer %d should have 2 groups", i)
    )
  }

  # Check that transformations worked correctly
  # Layer 2 should be shifted up by 0.5 compared to layer 1
  layer1_y <- built$data[[1]]$y[order(built$data[[1]]$x, built$data[[1]]$group)]
  layer2_y <- built$data[[2]]$y[order(built$data[[2]]$x, built$data[[2]]$group)]
  expect_true(
    all(abs((layer2_y - layer1_y) - 0.5) < 1e-10),
    "Layer 2 should be shifted up by 0.5"
  )

  # Layer 3 should be shifted down by 0.5 compared to layer 1
  layer3_y <- built$data[[3]]$y[order(built$data[[3]]$x, built$data[[3]]$group)]
  expect_true(
    all(abs((layer3_y - layer1_y) + 0.5) < 1e-10),
    "Layer 3 should be shifted down by 0.5"
  )
})

# Tests for aesthetic specification equivalence in tf_ggplot
# Verifies that aesthetics can be specified in constructor or geom layer with identical results

test_that("tf aesthetics in constructor vs geom layer are equivalent - basic case", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  data <- create_test_tf_data(n_funcs = 3, n_points = 5)

  # Method 1: tf aesthetic in constructor
  p1 <- tf_ggplot(data, aes(tf = func)) + geom_line()

  # Method 2: tf aesthetic in geom layer
  p2 <- tf_ggplot(data) + geom_line(aes(tf = func))

  # Build both plots
  built1 <- ggplot_build(p1)
  built2 <- ggplot_build(p2)

  # Should have identical data structure
  expect_equal(nrow(built1$data[[1]]), nrow(built2$data[[1]]))
  expect_equal(ncol(built1$data[[1]]), ncol(built2$data[[1]]))
  expect_equal(names(built1$data[[1]]), names(built2$data[[1]]))

  # Should have same number of groups
  expect_equal(
    length(unique(built1$data[[1]]$group)),
    length(unique(built2$data[[1]]$group))
  )

  # x and y values should be identical (allowing for floating point differences)
  expect_equal(sort(built1$data[[1]]$x), sort(built2$data[[1]]$x))
  expect_equal(
    sort(built1$data[[1]]$y),
    sort(built2$data[[1]]$y),
    tolerance = 1e-10
  )
})

test_that("tf aesthetics with color mapping are equivalent", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  data <- create_test_tf_data(n_funcs = 4, n_points = 6)

  # Method 1: aesthetics in constructor
  p1 <- tf_ggplot(data, aes(tf = func, color = group)) + geom_line()

  # Method 2: aesthetics in geom layer
  p2 <- tf_ggplot(data) +
    geom_line(aes(tf = func, color = group))

  built1 <- ggplot_build(p1)
  built2 <- ggplot_build(p2)

  # Should have identical structure
  expect_equal(dim(built1$data[[1]]), dim(built2$data[[1]]))

  # Color mappings should be identical
  expect_equal(sort(built1$data[[1]]$colour), sort(built2$data[[1]]$colour))

  # Groups should be identical
  expect_equal(sort(built1$data[[1]]$group), sort(built2$data[[1]]$group))

  # Check that color is consistent within groups for both methods
  check_color_consistency <- function(plot_data) {
    color_by_group <- split(plot_data$colour, plot_data$group)
    all(sapply(color_by_group, function(x) length(unique(x)) == 1))
  }

  expect_true(check_color_consistency(built1$data[[1]]))
  expect_true(check_color_consistency(built2$data[[1]]))
})

test_that("mixed tf and regular aesthetics are equivalent", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  data <- create_test_tf_data(n_funcs = 3, n_points = 5)
  data$lw_var <- c(1, 2, 3)
  data$alpha_var <- c(0.3, 0.7, 0.9)

  # Method 1: mixed aesthetics in constructor
  p1 <- tf_ggplot(
    data,
    aes(tf = func, color = group, linewidth = lw_var, alpha = alpha_var)
  ) +
    geom_line()

  # Method 2: mixed aesthetics in geom layer
  p2 <- tf_ggplot(data) +
    geom_line(aes(
      tf = func,
      color = group,
      linewidth = lw_var,
      alpha = alpha_var
    ))

  built1 <- ggplot_build(p1)
  built2 <- ggplot_build(p2)

  # Should have identical dimensions and structure
  expect_equal(dim(built1$data[[1]]), dim(built2$data[[1]]))
  expect_setequal(names(built1$data[[1]]), names(built2$data[[1]]))

  # All aesthetic mappings should be identical
  aesthetic_cols <- c("colour", "linewidth", "alpha")
  for (col in aesthetic_cols) {
    expect_equal(
      sort(built1$data[[1]][[col]]),
      sort(built2$data[[1]][[col]]),
      info = paste("Mismatch in", col, "aesthetic")
    )
  }
})

test_that("multiple tf aesthetics are equivalent", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  data <- create_multi_tf_data(n_funcs = 2, n_points = 4)

  # Method 1: multiple tf aesthetics in constructor
  p1 <- tf_ggplot(data, aes(tf_x = func1, tf_y = func2)) + geom_point()

  # Method 2: multiple tf aesthetics in geom layer
  p2 <- tf_ggplot(data) +
    geom_point(aes(tf_x = func1, tf_y = func2))

  built1 <- ggplot_build(p1)
  built2 <- ggplot_build(p2)

  # Should have identical structure
  expect_equal(dim(built1$data[[1]]), dim(built2$data[[1]]))

  # x and y coordinates should be identical
  expect_equal(sort(built1$data[[1]]$x), sort(built2$data[[1]]$x))
  expect_equal(sort(built1$data[[1]]$y), sort(built2$data[[1]]$y))

  # Groups should be identical
  expect_equal(sort(built1$data[[1]]$group), sort(built2$data[[1]]$group))
})

test_that("ribbon tf aesthetics are equivalent", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  data <- create_band_tf_data(n_funcs = 2, n_points = 5)

  # Method 1: ribbon aesthetics in constructor
  p1 <- tf_ggplot(data, aes(tf_ymin = lower_func, tf_ymax = upper_func)) +
    geom_ribbon(alpha = 0.3)

  # Method 2: ribbon aesthetics in geom layer
  p2 <- tf_ggplot(data) +
    geom_ribbon(
      aes(tf_ymin = lower_func, tf_ymax = upper_func),
      alpha = 0.3
    )

  built1 <- ggplot_build(p1)
  built2 <- ggplot_build(p2)

  # Should have identical structure
  expect_equal(dim(built1$data[[1]]), dim(built2$data[[1]]))

  # ymin and ymax should be identical
  expect_equal(sort(built1$data[[1]]$ymin), sort(built2$data[[1]]$ymin))
  expect_equal(sort(built1$data[[1]]$ymax), sort(built2$data[[1]]$ymax))

  # Alpha should be identical
  expect_equal(built1$data[[1]]$alpha, built2$data[[1]]$alpha)
})

test_that("aesthetic inheritance and override work correctly", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  data <- create_test_tf_data(n_funcs = 3, n_points = 5)

  # Base aesthetics in constructor, additional in geom
  p1 <- tf_ggplot(data, aes(tf = func)) +
    geom_line(aes(color = group))

  # All aesthetics in geom layer
  p2 <- tf_ggplot(data) +
    geom_line(aes(tf = func, color = group))

  built1 <- ggplot_build(p1)
  built2 <- ggplot_build(p2)

  # Should produce identical results
  expect_equal(dim(built1$data[[1]]), dim(built2$data[[1]]))
  expect_equal(sort(built1$data[[1]]$colour), sort(built2$data[[1]]$colour))
  expect_equal(sort(built1$data[[1]]$x), sort(built2$data[[1]]$x))
  expect_equal(sort(built1$data[[1]]$y), sort(built2$data[[1]]$y))
})

test_that("aesthetic override in geom layer works", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  data <- create_test_tf_data(n_funcs = 3, n_points = 5)
  data$alt_group <- factor(c("X", "Y", "Z"))

  # Constructor aesthetic overridden in geom
  p1 <- tf_ggplot(data, aes(tf = func, color = group)) +
    geom_line(aes(color = alt_group)) # Override color aesthetic

  # Direct specification in geom
  p2 <- tf_ggplot(data, aes(tf = func)) +
    geom_line(aes(color = alt_group))

  built1 <- ggplot_build(p1)
  built2 <- ggplot_build(p2)

  # Should use alt_group for coloring in both cases
  expect_equal(sort(built1$data[[1]]$colour), sort(built2$data[[1]]$colour))

  # Should have 3 unique colors (X, Y, Z)
  expect_equal(length(unique(built1$data[[1]]$colour)), 3)
  expect_equal(length(unique(built2$data[[1]]$colour)), 3)
})

test_that("faceting works equivalently with different aesthetic specifications", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  data <- create_test_tf_data(n_funcs = 6, n_points = 4)
  data$treatment <- factor(rep(c("A", "B"), each = 3))

  # Method 1: aesthetics in constructor
  p1 <- tf_ggplot(data, aes(tf = func)) +
    geom_line() +
    facet_wrap(~treatment)

  # Method 2: aesthetics in geom layer
  p2 <- tf_ggplot(data) +
    geom_line(aes(tf = func)) +
    facet_wrap(~treatment)

  built1 <- ggplot_build(p1)
  built2 <- ggplot_build(p2)

  # Should have identical panel structure
  expect_equal(dim(built1$layout$layout), dim(built2$layout$layout))
  expect_equal(built1$layout$layout$PANEL, built2$layout$layout$PANEL)

  # Plot data should be identical
  expect_equal(dim(built1$data[[1]]), dim(built2$data[[1]]))
  expect_equal(sort(built1$data[[1]]$PANEL), sort(built2$data[[1]]$PANEL))
})

test_that("custom arg parameter works equivalently", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  data <- create_test_tf_data(n_funcs = 2, n_points = 11)
  custom_arg <- seq(0, 1, length.out = 6)

  # Method 1: aesthetics in constructor with custom arg
  p1 <- tf_ggplot(data, aes(tf = func), arg = custom_arg) + geom_line()

  # Method 2: aesthetics in geom layer with custom arg
  p2 <- tf_ggplot(data, arg = custom_arg) +
    geom_line(aes(tf = func))

  built1 <- ggplot_build(p1)
  built2 <- ggplot_build(p2)

  # Should use custom arg in both cases
  expect_equal(sort(unique(built1$data[[1]]$x)), custom_arg)
  expect_equal(sort(unique(built2$data[[1]]$x)), custom_arg)

  # Should have identical structure
  expect_equal(dim(built1$data[[1]]), dim(built2$data[[1]]))
  expect_equal(sort(built1$data[[1]]$x), sort(built2$data[[1]]$x))
  expect_equal(sort(built1$data[[1]]$y), sort(built2$data[[1]]$y))
})

test_that("multiple layers with different aesthetic specifications work", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  data <- create_test_tf_data(n_funcs = 2, n_points = 5)

  # Mix of constructor and geom-level aesthetics across multiple layers
  p1 <- tf_ggplot(data, aes(tf = func)) +
    geom_line(aes(color = group)) +
    geom_point(aes(color = group), size = 2)

  # All aesthetics specified at geom level
  p2 <- tf_ggplot(data) +
    geom_line(aes(tf = func, color = group)) +
    geom_point(aes(tf = func, color = group), size = 2)

  built1 <- ggplot_build(p1)
  built2 <- ggplot_build(p2)

  # Should have same number of layers
  expect_equal(length(built1$data), length(built2$data))

  # Each layer should be identical
  for (i in seq_along(built1$data)) {
    expect_equal(
      dim(built1$data[[i]]),
      dim(built2$data[[i]]),
      info = paste("Layer", i, "dimensions differ")
    )
    expect_equal(
      sort(built1$data[[i]]$x),
      sort(built2$data[[i]]$x),
      info = paste("Layer", i, "x values differ")
    )
    expect_equal(
      sort(built1$data[[i]]$y),
      sort(built2$data[[i]]$y),
      info = paste("Layer", i, "y values differ")
    )
    expect_equal(
      sort(built1$data[[i]]$colour),
      sort(built2$data[[i]]$colour),
      info = paste("Layer", i, "color values differ")
    )
  }
})

test_that("data transformation is identical regardless of aesthetic specification method", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  data <- create_test_tf_data(n_funcs = 3, n_points = 7)
  data$extra_var <- letters[1:3]

  # Complex aesthetic mapping
  aesthetic_mapping <- aes(tf = func, color = group, linetype = extra_var)

  # Method 1: in constructor
  p1 <- tf_ggplot(data, aesthetic_mapping) + geom_line()

  # Method 2: in geom layer
  p2 <- tf_ggplot(data) + geom_line(aesthetic_mapping)

  built1 <- ggplot_build(p1)
  built2 <- ggplot_build(p2)

  # Extract and sort all data for comparison
  data1 <- built1$data[[1]][order(built1$data[[1]]$group, built1$data[[1]]$x), ]
  data2 <- built2$data[[1]][order(built2$data[[1]]$group, built2$data[[1]]$x), ]

  # Reset row names for comparison
  rownames(data1) <- NULL
  rownames(data2) <- NULL

  # Should be completely identical
  common_cols <- sort(intersect(names(data1), names(data2)))
  expect_equal(data1[common_cols], data2[common_cols])
})

# Visual tests for tf_ggplot functionality
# These tests create actual plots that can be inspected visually during development

# Set up visual testing environment
setup_visual_tests <- function() {
  # Create a temporary directory for test plots if it doesn't exist
  test_dir <- file.path(getwd(), "test-plots")
  if (!dir.exists(test_dir)) {
    dir.create(test_dir, recursive = TRUE)
  }
  return(test_dir)
}

# Helper function to save and optionally display plots
save_test_plot <- function(
  plot,
  filename,
  width = 8,
  height = 6,
  display = interactive(),
  save_plot = TRUE
) {
  if (save_plot) {
    test_dir <- setup_visual_tests()
    full_path <- file.path(test_dir, paste0(filename, ".png"))

    tryCatch(
      {
        ggsave(full_path, plot, width = width, height = height, dpi = 150)
        # cat("Plot saved to:", full_path, "\n")
      },
      error = function(e) {
        warning("Could not save plot: ", e$message)
      }
    )
  }

  if (display) {
    print(plot)
  }

  invisible(plot)
}

test_that("basic tf_ggplot line plots look sensible", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  # Create test data with recognizable patterns
  set.seed(123)
  data <- data.frame(
    id = 1:4,
    group = factor(c("Sine", "Cosine", "Linear", "Quadratic"))
  )

  # Create functions with different patterns for visual verification
  arg_vals <- seq(0, 2 * pi, length.out = 51)
  data$func <- tfd(
    list(
      sin(arg_vals), # Sine wave
      cos(arg_vals), # Cosine wave
      arg_vals / pi, # Linear function
      (arg_vals - pi)^2 / pi^2 # Quadratic function
    ),
    arg = arg_vals
  )

  # Test both aesthetic specification methods
  p1 <- tf_ggplot(data, aes(tf = func, color = group)) +
    geom_line(linewidth = 1) +
    labs(
      title = "tf_ggplot: Basic Line Plot (Constructor Aesthetics)",
      subtitle = "Should show: sine, cosine, linear, and quadratic functions"
    ) +
    theme_minimal()

  p2 <- tf_ggplot(data) +
    geom_line(aes(tf = func, color = group), linewidth = 1) +
    labs(
      title = "tf_ggplot: Basic Line Plot (Geom Aesthetics)",
      subtitle = "Should be identical to constructor version"
    ) +
    theme_minimal()

  # Save plots for visual inspection
  save_test_plot(p1, "01_basic_line_constructor")
  save_test_plot(p2, "02_basic_line_geom")

  # Basic validation that plots were created
  expect_s3_class(p1, "ggplot")
  expect_s3_class(p2, "ggplot")

  # Verify they have the expected structure
  expect_equal(count_plot_groups(p1), 4)
  expect_equal(count_plot_groups(p2), 4)
})

test_that("tf_ggplot with points and lines looks correct", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  # Create sparse irregular data to test points + lines
  set.seed(456)
  data <- data.frame(
    subject = 1:3,
    treatment = factor(c("Control", "Treatment A", "Treatment B"))
  )

  # Create irregular functions (sparse data)
  arg_vals <- sort(runif(15, 0, 10)) # Irregular time points
  data$func <- tfd(
    list(
      exp(-arg_vals / 5) + rnorm(15, 0, 0.1), # Decay + noise
      2 * exp(-arg_vals / 3) + rnorm(15, 0, 0.15), # Different decay
      1.5 * exp(-arg_vals / 4) + rnorm(15, 0, 0.12) # Third decay
    ),
    arg = list(arg_vals, arg_vals, arg_vals)
  )

  p <- tf_ggplot(data, aes(tf = func, color = treatment)) +
    geom_line(alpha = 0.7, linewidth = 1) +
    geom_point(size = 2) +
    labs(
      title = "tf_ggplot: Lines + Points with Irregular Data",
      subtitle = "Should show: 3 decay curves with visible data points"
    ) +
    theme_minimal()

  save_test_plot(p, "03_lines_points_irregular")

  expect_s3_class(p, "ggplot")
  expect_equal(count_plot_groups(p), 3)
})

test_that("tf_ggplot ribbon plots look sensible", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  # Create confidence band data
  set.seed(789)
  data <- data.frame(method = c("Method A", "Method B"))

  arg_vals <- seq(0, 5, length.out = 26)

  # Create mean functions
  mean_vals1 <- 2 * exp(-arg_vals / 2) * cos(arg_vals)
  mean_vals2 <- 1.5 * exp(-arg_vals / 3) * sin(arg_vals * 1.2)

  # Create confidence bands
  sd1 <- 0.2 + 0.1 * arg_vals / 5 # Increasing uncertainty
  sd2 <- 0.15 + 0.05 * arg_vals / 5

  data$mean_func <- tfd(list(mean_vals1, mean_vals2), arg = arg_vals)
  data$lower_func <- tfd(
    list(mean_vals1 - 1.96 * sd1, mean_vals2 - 1.96 * sd2),
    arg = arg_vals
  )
  data$upper_func <- tfd(
    list(mean_vals1 + 1.96 * sd1, mean_vals2 + 1.96 * sd2),
    arg = arg_vals
  )

  p <- tf_ggplot(data, aes(color = method, fill = method)) +
    geom_ribbon(
      aes(tf_ymin = lower_func, tf_ymax = upper_func),
      alpha = 0.3
    ) +
    geom_line(aes(tf = mean_func), linewidth = 1) +
    labs(
      title = "tf_ggplot: Confidence Bands",
      subtitle = "Should show: 2 oscillating functions with confidence ribbons"
    ) +
    theme_minimal()

  save_test_plot(p, "04_confidence_ribbons")

  expect_s3_class(p, "ggplot")
  expect_equal(count_plot_groups(p), 2) # Should have 2 methods
})

test_that("tf_ggplot faceting looks correct", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  # Create data for faceted plot
  set.seed(101112)
  data <- expand.grid(
    subject = 1:4,
    condition = c("Baseline", "Treatment"),
    session = c("Pre", "Post")
  )

  # Create different function patterns for each condition
  arg_vals <- seq(0, 10, length.out = 21)
  n_funcs <- nrow(data)

  func_vals <- map(1:n_funcs, function(i) {
    baseline <- sin(arg_vals * 0.5) + rnorm(21, 0, 0.1)
    if (data$condition[i] == "Treatment") {
      # Treatment effect: amplify and shift
      baseline <- baseline * 1.5 + 0.5
    }
    if (data$session[i] == "Post") {
      # Post effect: add upward trend
      baseline <- baseline + arg_vals * 0.05
    }
    baseline
  })

  data$func <- tfd(func_vals, arg = arg_vals)

  p <- tf_ggplot(data, aes(tf = func, color = condition)) +
    geom_line(alpha = 0.7) +
    facet_grid(session ~ condition) +
    labs(
      title = "tf_ggplot: Faceted Plot",
      subtitle = "Should show: Treatment effects across Pre/Post sessions"
    ) +
    theme_minimal()

  save_test_plot(p, "05_faceted_plot", width = 10, height = 8)

  expect_s3_class(p, "ggplot")

  # Check faceting structure
  built <- ggplot_build(p)
  expect_equal(length(unique(built$layout$layout$PANEL)), 4) # 2×2 facets
})

test_that("tf_ggplot with multiple aesthetics looks good", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  # Create data with multiple grouping variables
  set.seed(131415)
  data <- expand.grid(
    method = c("A", "B", "C"),
    replicate = 1:3
  )
  data$difficulty <- factor(rep(c("Easy", "Medium", "Hard"), 3))
  data$performance <- runif(9, 0.5, 1.0)

  # Create performance curves over time
  arg_vals <- seq(0, 12, length.out = 25)
  func_vals <- map(1:nrow(data), function(i) {
    # Base learning curve
    base <- data$performance[i] * (1 - exp(-arg_vals / 3))

    # Method effects
    if (data$method[i] == "B") base <- base * 1.1
    if (data$method[i] == "C") base <- base * 0.9

    # Difficulty effects
    if (data$difficulty[i] == "Medium") base <- base * 0.8
    if (data$difficulty[i] == "Hard") base <- base * 0.6

    # Add noise
    base + rnorm(25, 0, 0.05)
  })

  data$learning_curve <- tfd(func_vals, arg = arg_vals)

  p <- tf_ggplot(
    data,
    aes(
      tf = learning_curve,
      color = method,
      linetype = difficulty,
      alpha = performance
    )
  ) +
    geom_line(linewidth = 1) +
    scale_alpha_continuous(range = c(0.4, 1.0)) +
    labs(
      title = "tf_ggplot: Multiple Aesthetics",
      subtitle = "Should show: Learning curves by method, difficulty, and performance"
    ) +
    theme_minimal() +
    theme(legend.position = "bottom")

  save_test_plot(p, "06_multiple_aesthetics", width = 10, height = 8)

  expect_s3_class(p, "ggplot")
  expect_equal(count_plot_groups(p), 9) # 3 methods × 3 replicates
})

test_that("tf_ggplot comparison with current geom_spaghetti", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  # Create comparison data
  set.seed(161718)
  data <- data.frame(
    id = 1:6,
    group = factor(rep(c("Group A", "Group B"), each = 3))
  )
  data$func <- tf_rgp(6, arg = seq(0, 1, length.out = 31))

  # Current approach with geom_spaghetti
  p_current <- ggplot(data, aes(y = func, color = group)) +
    geom_spaghetti(alpha = 0.7) +
    labs(
      title = "Current: geom_spaghetti",
      subtitle = "6 random Gaussian process functions"
    ) +
    theme_minimal()

  # New approach with tf_ggplot
  p_new <- tf_ggplot(data, aes(tf = func, color = group)) +
    geom_line(alpha = 0.7) +
    labs(
      title = "New: tf_ggplot + geom_line",
      subtitle = "Same data, should look identical"
    ) +
    theme_minimal()

  save_test_plot(p_current, "07_current_spaghetti")
  save_test_plot(p_new, "08_new_tf_ggplot")

  # Both should work
  expect_s3_class(p_current, "ggplot")
  expect_s3_class(p_new, "ggplot")

  # Should have same number of groups
  expect_equal(count_plot_groups(p_current), count_plot_groups(p_new))
})

test_that("tf_ggplot with different geom types", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  # Create test data with distinctive patterns
  set.seed(192021)
  data <- data.frame(id = 1:2)

  arg_vals <- seq(0, 4 * pi, length.out = 41)
  data$func <- tfd(
    list(
      sin(arg_vals) + 0.1 * arg_vals, # Sine with trend
      cos(arg_vals) - 0.1 * arg_vals # Cosine with trend
    ),
    arg = arg_vals
  )

  # Test different geom types
  p_line <- tf_ggplot(data) +
    geom_line(aes(tf = func), color = "blue", linewidth = 1) +
    labs(title = "geom_line", subtitle = "Smooth continuous lines") +
    theme_minimal()

  p_point <- tf_ggplot(data) +
    geom_point(
      aes(tf = func),
      color = "red",
      size = 1.5,
      alpha = 0.7
    ) +
    labs(title = "geom_point", subtitle = "Individual evaluation points") +
    theme_minimal()

  p_step <- tf_ggplot(data, arg = seq(0, 4 * pi, length.out = 21)) +
    geom_step(
      aes(tf = func),
      color = "green",
      linewidth = 1
    ) +
    labs(title = "geom_step", subtitle = "Step function approximation") +
    theme_minimal()

  p_area <- tf_ggplot(data) +
    geom_area(aes(tf = func), fill = "purple", alpha = 0.5) +
    labs(title = "geom_area", subtitle = "Filled area under curves") +
    theme_minimal()

  # Combined plot
  p_combined <- tf_ggplot(data, aes(tf = func, color = factor(id))) +
    geom_line(linewidth = 1, alpha = 0.8) +
    geom_point(size = 2) +
    labs(
      title = "Combined: Lines + Points",
      subtitle = "Should show both continuous lines and data points",
      color = "Function"
    ) +
    theme_minimal()

  save_test_plot(p_line, "09_geom_line")
  save_test_plot(p_point, "10_geom_point")
  save_test_plot(p_step, "11_geom_step")
  save_test_plot(p_area, "12_geom_area")
  save_test_plot(p_combined, "13_combined_geoms")

  # All should build successfully
  expect_s3_class(p_line, "ggplot")
  expect_s3_class(p_point, "ggplot")
  expect_s3_class(p_step, "ggplot")
  expect_s3_class(p_area, "ggplot")
  expect_s3_class(p_combined, "ggplot")
})

test_that("tf_ggplot edge cases render properly", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  # Test various edge cases
  set.seed(222324)

  # 1. Single function
  data1 <- data.frame(id = 1)
  data1$func <- tf_rgp(1, arg = seq(0, 1, length.out = 21))

  p1 <- tf_ggplot(data1, aes(tf = func)) +
    geom_line(color = "blue", linewidth = 1) +
    labs(title = "Edge Case: Single Function") +
    theme_minimal()

  # 2. Very sparse data (few evaluation points)
  data2 <- data.frame(id = 1:3)
  data2$func <- tf_rgp(3, arg = c(0, 0.5, 1)) # Only 3 points

  p2 <- tf_ggplot(data2, aes(tf = func, color = factor(id))) +
    geom_line(linewidth = 1) +
    geom_point(size = 3) +
    labs(title = "Edge Case: Very Sparse Data (3 points)", color = "Function") +
    theme_minimal()

  # 3. Many functions (visual clutter test)
  data3 <- data.frame(id = 1:20)
  data3$group <- factor(rep(c("A", "B"), each = 10))
  data3$func <- tf_rgp(20, arg = seq(0, 1, length.out = 51))

  p3 <- tf_ggplot(data3, aes(tf = func, color = group)) +
    geom_line(alpha = 0.3, linewidth = 0.5) +
    labs(
      title = "Edge Case: Many Functions (20)",
      subtitle = "Should be readable with transparency"
    ) +
    theme_minimal()

  # 4. Wide evaluation range
  wide_arg <- seq(-10, 10, length.out = 101)
  data4 <- data.frame(id = 1:2)
  data4$func <- tfd(
    list(
      exp(-wide_arg^2 / 10), # Gaussian
      sin(wide_arg) * exp(-abs(wide_arg) / 5) # Damped oscillation
    ),
    arg = wide_arg
  )

  p4 <- tf_ggplot(data4, aes(tf = func, color = factor(id))) +
    geom_line(linewidth = 1) +
    labs(
      title = "Edge Case: Wide Evaluation Range (-10 to 10)",
      color = "Function"
    ) +
    theme_minimal()

  save_test_plot(p1, "14_single_function")
  save_test_plot(p2, "15_sparse_data")
  save_test_plot(p3, "16_many_functions")
  save_test_plot(p4, "17_wide_range")

  # All should render without errors
  expect_s3_class(p1, "ggplot")
  expect_s3_class(p2, "ggplot")
  expect_s3_class(p3, "ggplot")
  expect_s3_class(p4, "ggplot")
})

# Helper function to run all visual tests at once
run_visual_tests <- function(display = interactive(), save_plots = TRUE) {
  cat("Running visual tests for tf_ggplot...\n")
  cat("Plots will be saved to:", setup_visual_tests(), "\n")

  # Set global options for this test run
  old_display <- getOption("tf_ggplot_test_display", interactive())
  old_save <- getOption("tf_ggplot_test_save", TRUE)

  options(tf_ggplot_test_display = display)
  options(tf_ggplot_test_save = save_plots)

  tryCatch(
    {
      # Run the visual tests
      test_file("test-tf-ggplot-visual.R")
      cat("Visual tests completed successfully!\n")
    },
    error = function(e) {
      cat("Error in visual tests:", e$message, "\n")
    },
    finally = {
      # Restore options
      options(tf_ggplot_test_display = old_display)
      options(tf_ggplot_test_save = old_save)
    }
  )
}

# Print helper message
if (interactive()) {
  cat("Visual testing functions available:\n")
  cat("- run_visual_tests()  # Run all visual tests\n")
  cat("- Test plots saved to: test-plots/\n")
  cat("- Set display=TRUE to show plots interactively\n")
}

# Basic tf_ggplot functionality tests
# These tests focus on core features that must work before more advanced functionality

test_that("tf_ggplot basic constructor works", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  data <- create_test_tf_data(n_funcs = 2, n_points = 5)

  # Basic constructor without aesthetics
  p1 <- tf_ggplot(data)
  expect_s3_class(p1, c("tf_ggplot", "ggplot"))
  expect_identical(p1$data, data)

  # Constructor with aesthetics
  p2 <- tf_ggplot(data, aes(tf = func, color = group))
  expect_s3_class(p2, c("tf_ggplot", "ggplot"))
  expect_equal(length(p2$mapping), 2)

  # Constructor with custom arg
  custom_arg <- c(0, 0.5, 1)
  p3 <- tf_ggplot(data, arg = custom_arg)
  expect_equal(attr(p3, "tf_arg"), custom_arg)
})

test_that("basic tf aesthetic transformation works", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  data <- create_test_tf_data(n_funcs = 2, n_points = 5)

  # Single tf aesthetic
  p <- tf_ggplot(data, aes(tf = func)) + geom_line()

  # Should still be tf_ggplot until finalized
  expect_s3_class(p, "ggplot")
  expect_true(inherits(p, "tf_ggplot"))

  # Check plot builds successfully
  built <- ggplot_build(p)
  expect_s3_class(built, "ggplot_built")

  # Should have correct number of groups (one per function)
  expect_equal(count_plot_groups(p), 2)

  # Should have x and y mappings
  expect_true(check_plot_mappings(p, c("x", "y", "group")))
})

test_that("tf aesthetic parsing identifies tf columns correctly", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  # Test different tf aesthetic patterns
  mapping1 <- aes(tf = func)
  parsed1 <- parse_tf_aesthetics(mapping1)
  expect_equal(names(parsed1$tf_aes), "tf")
  expect_equal(length(parsed1$regular_aes), 0)

  mapping2 <- aes(tf = func, color = group)
  parsed2 <- parse_tf_aesthetics(mapping2)
  expect_equal(names(parsed2$tf_aes), "tf")
  expect_equal(length(parsed2$regular_aes), 1)

  mapping3 <- aes(tf_x = func1, tf_y = func2)
  parsed3 <- parse_tf_aesthetics(mapping3)
  expect_equal(sort(names(parsed3$tf_aes)), c("tf_x", "tf_y"))
  expect_equal(length(parsed3$regular_aes), 0)

  # Test no tf aesthetics
  mapping4 <- aes(x = time, y = value)
  parsed4 <- parse_tf_aesthetics(mapping4)
  expect_equal(length(parsed4$tf_aes), 0)
  expect_equal(length(parsed4$regular_aes), 2)
})

test_that("data transformation creates correct structure via ggplot_build", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  data <- data.frame(
    id = 1:2,
    group = factor(c("A", "B"))
  )
  data$func <- tf_rgp(2, arg = seq(0, 1, length.out = 5))

  p <- tf_ggplot(data, aes(tf = func)) + geom_line()
  built <- ggplot_build(p)
  plot_data <- built$data[[1]]

  # Check dimensions (2 functions × 5 points)
  expect_equal(nrow(plot_data), 10)
  expect_equal(length(unique(plot_data$group)), 2)
  expect_equal(length(unique(plot_data$x)), 5)
})

test_that("geom_line integration produces correct plot structure", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  data <- create_test_tf_data(n_funcs = 3, n_points = 6)

  p <- tf_ggplot(data, aes(tf = func, color = group)) +
    geom_line()

  # Build plot and check structure
  built <- ggplot_build(p)
  plot_data <- built$data[[1]]

  # Should have 3 groups (one per function)
  expect_equal(length(unique(plot_data$group)), 3)

  # Should have 6 unique x values (evaluation points)
  expect_equal(length(unique(plot_data$x)), 6)

  # Should have 18 total rows (3 functions × 6 points)
  expect_equal(nrow(plot_data), 18)

  # Should have color aesthetic preserved
  expect_true("colour" %in% names(plot_data))
})

test_that("error handling for invalid tf aesthetics", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  # Data without tf columns
  data <- data.frame(
    id = 1:3,
    regular_col = rnorm(3)
  )

  # Should error when tf aesthetic references non-tf column
  expect_error(
    {
      p <- tf_ggplot(data, aes(tf = regular_col)) +
        geom_line()
      ggplot_build(p) # Trigger the validation
    },
    "tf.*object|must be.*tf"
  )

  # Should error when tf aesthetic references non-existent column
  expect_error(
    {
      p <- tf_ggplot(data, aes(tf = nonexistent)) +
        geom_line()
      ggplot_build(p) # Trigger the validation
    },
    "object.*not found|not found"
  )
})

test_that("custom arg parameter works correctly", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  data <- create_test_tf_data(n_funcs = 2, n_points = 11)

  # Use coarser evaluation grid (numeric vector)
  custom_arg <- seq(0, 1, length.out = 5)
  p <- tf_ggplot(data, aes(tf = func), arg = custom_arg) +
    geom_line()

  # Check that custom arg is used
  x_values <- get_plot_x_values(p)
  expect_equal(x_values, custom_arg)
  expect_equal(length(x_values), 5) # Not original 11

  # Check total rows (2 functions × 5 points)
  expect_true(check_plot_nrows(p, 10))
})

test_that("integer arg sets grid length", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  data <- create_test_tf_data(n_funcs = 2, n_points = 21)

  # Integer arg = grid length
  p <- tf_ggplot(data, aes(tf = func), arg = 5L) + geom_line()
  x_values <- get_plot_x_values(p)
  expect_equal(length(x_values), 5)
  expect_equal(range(x_values), c(0, 1)) # Should span the tf domain

  # Check total rows (2 functions × 5 points)
  expect_true(check_plot_nrows(p, 10))

  # Validation: arg < 2 should error
  expect_error(tf_ggplot(data, arg = 1L), "grid length must be >= 2")
  expect_error(tf_ggplot(data, arg = NA_real_), "numeric")
})

test_that("grouping variables are preserved in transformation", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  data <- create_test_tf_data(n_funcs = 4, n_points = 5)
  data$treatment <- factor(c("A", "A", "B", "B"))
  data$subject <- factor(1:4)

  p <- tf_ggplot(data, aes(tf = func, color = treatment)) +
    geom_line()
  built <- ggplot_build(p)
  plot_data <- built$data[[1]]

  # Note: unmapped variables are not preserved in final plot data (this is normal ggplot2 behavior)
  # But the color aesthetic should reflect treatment mapping
  expect_true("colour" %in% names(plot_data))

  # Should have correct number of colors (2 treatment levels)
  expect_equal(length(unique(plot_data$colour)), 2)

  # Should have consistent color within each function
  color_by_group <- split(plot_data$colour, plot_data$group)
  consistent_colors <- all(sapply(
    color_by_group,
    function(x) length(unique(x)) == 1
  ))
  expect_true(consistent_colors)
})

# Tests for tf_ggplot axis labeling

test_that("simple tf aesthetic has meaningful axis labels", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  data <- data.frame(id = 1:2)
  data$curves <- tf_rgp(2, arg = seq(0, 1, length.out = 5))

  p <- tf_ggplot(data, aes(tf = curves)) + geom_line()
  built <- ggplot_build(p)

  # For simple column references, y-label should be the column name
  expect_equal(built$plot$labels$y, "curves")
  expect_equal(built$plot$labels$x, "curves.arg")
})

test_that("complex tf expression has meaningful axis labels", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  data <- data.frame(id = 1:2)
  data$func1 <- tf_rgp(2, arg = seq(0, 1, length.out = 5))
  data$func2 <- tf_rgp(2, arg = seq(0, 1, length.out = 5))

  p <- tf_ggplot(data, aes(tf = func1 + func2)) + geom_line()
  built <- ggplot_build(p)

  # For complex expressions, y-label should be the expression text
  expect_equal(built$plot$labels$y, "func1 + func2")
  expect_equal(built$plot$labels$x, "func1 + func2.arg")
})

test_that("nested tf expression has meaningful axis labels", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  data <- data.frame(id = 1:2)
  data$curves <- tf_rgp(2, arg = seq(-1, 1, length.out = 5))

  p <- tf_ggplot(data, aes(tf = abs(curves))) + geom_line()
  built <- ggplot_build(p)

  # For nested expressions, y-label should be the expression text
  expect_equal(built$plot$labels$y, "abs(curves)")
  expect_equal(built$plot$labels$x, "abs(curves).arg")
})

test_that("multiple tf expressions have distinct meaningful labels", {
  skip_if_not_installed("ggplot2")
  skip_if_no_tf_ggplot()

  data <- data.frame(id = 1:2)
  data$func1 <- tf_rgp(2, arg = seq(0, 1, length.out = 5))
  data$func2 <- tf_rgp(2, arg = seq(0, 1, length.out = 5))

  p <- tf_ggplot(data, aes(tf = func1 + func2)) +
    geom_line() +
    geom_line(aes(tf = func1), color = "red") +
    geom_line(aes(tf = func2), color = "blue")

  built <- ggplot_build(p)

  # Base plot should have the sum expression label
  expect_equal(built$plot$labels$y, "func1 + func2")

  # The transformed data should have columns for each expression with meaningful names
  data_cols <- names(built$data[[1]])
  # Should not have ugly .tf_expr_ names in the actual plot data
  expect_false(any(grepl("^\\.tf_expr_", data_cols)))
})

#' Create a tf-aware ggplot
#'
#' `tf_ggplot()` creates a ggplot object that can handle tf (functional data) aesthetics.
#' It works similarly to `ggplot()` but automatically transforms tf objects into
#' long-format data suitable for standard ggplot2 geoms.
#'
#' @param data Default dataset to use for plot. If not provided, must be supplied
#'   in each layer added to the plot.
#' @param mapping Default list of aesthetic mappings to use for plot. Can include
#'   tf-specific aesthetics like `tf`, `tf_x`, `tf_y`, `tf_ymin`, `tf_ymax`.
#' @param arg Optional. Evaluation grid for tf objects. A numeric vector of arg
#'   values, or a single integer specifying the desired grid length (resolved to
#'   an equidistant grid over the tf domain). If `NULL` (default), uses the
#'   natural grid of the tf objects.
#' @param interpolate Logical. Should tf objects be interpolated to the evaluation
#'   grid? Defaults to TRUE.
#' @param ... Other arguments passed to ggplot2 functions.
#'
#' @details
#' `tf_ggplot` supports the following tf-specific aesthetics:
#' - `tf`: Maps a `tf` object to `y` aesthetic (shorthand for `tf_y`)
#' - `tf_x`: Maps a `tf` object to `x` aesthetic
#' - `tf_y`: Maps a `tf` object to `y` aesthetic
#' - `tf_ymin`: Maps a tf object to ymin aesthetic (for ribbons)
#' - `tf_ymax`: Maps a tf object to ymax aesthetic (for ribbons)
#'
#' When tf aesthetics are used, the data is automatically transformed:
#' - tf objects are evaluated on a common grid
#' - Each function becomes multiple rows (one per evaluation point)
#' - Group identifiers are created to maintain function identity
#' - Non-tf columns are replicated appropriately
#'
#' @return A tf_ggplot object that inherits from ggplot
#'
#' @examples
#' # Basic usage
#' data <- data.frame(
#' id = 1:10,
#' group = sample(c("A", "B"), 10, replace = TRUE)
#' )
#' data$f <- tf_rgp(10)
#'
#' # Method 1: tf aesthetic in constructor
#' tf_ggplot(data, ggplot2::aes(tf = f, color = group)) + ggplot2::geom_line()
#'
#' # Method 2: tf aesthetic in geom (equivalent)
#' tf_ggplot(data) + ggplot2::geom_line(ggplot2::aes(tf = f, color = group))
#'
#' # Confidence bands
#' tf_ggplot(data) +
#'   ggplot2::geom_ribbon(
#'     ggplot2::aes(tf_ymin = mean(f) - sd(f), tf_ymax = mean(f) + sd(f)),
#'     alpha = 0.3
#'   ) +
#'   ggplot2::geom_line(ggplot2::aes(tf = mean(f)))
#'
#' @export
tf_ggplot <- function(
  data = NULL,
  mapping = aes(),
  ...,
  arg = NULL,
  interpolate = TRUE
) {
  # Validate inputs
  if (!is.null(data) && !is.data.frame(data)) {
    cli::cli_abort("data must be a data.frame, not {.obj_type_friendly {data}}")
  }

  if (!is_mapping(mapping)) {
    cli::cli_abort(
      "mapping must be created with aes(), not {.obj_type_friendly {mapping}}"
    )
  }

  if (!is.null(arg)) {
    if (!is.numeric(arg) || anyNA(arg)) {
      cli::cli_abort(
        "{.arg arg} must be a numeric vector or a single integer grid length"
      )
    }
    if (length(arg) == 1L && arg < 2) {
      cli::cli_abort("{.arg arg} as grid length must be >= 2, not {arg}")
    }
  }

  # Create base ggplot object
  p <- ggplot(data = data, mapping = mapping, ...)

  # Convert to tf_ggplot class and add tf-specific attributes
  class(p) <- c("tf_ggplot", class(p))
  attr(p, "tf_arg") <- arg
  attr(p, "tf_interpolate") <- interpolate
  attr(p, "tf_original_data") <- data
  attr(p, "tf_original_mapping") <- mapping
  attr(p, "all_layers") <- list() # Store all layers in order with metadata
  attr(p, "tf_expression_counter") <- 0 # Counter for unique expression names

  return(p)
}

#' Check if object is a tf_ggplot
#' @param x Object to test
#' @returns `TRUE` if `x` inherits from `"tf_ggplot"`, `FALSE` otherwise.
#' @examples
#' p <- tf_ggplot(data.frame(x = 1))
#' is_tf_ggplot(p)
#' is_tf_ggplot(ggplot2::ggplot())
#' @export
is_tf_ggplot <- function(x) {
  inherits(x, "tf_ggplot")
}

#' Parse aesthetic mappings to separate tf and regular aesthetics
#'
#' @param mapping An aesthetic mapping created with [ggplot2::aes()].
#' @param data Data frame to evaluate expressions against.
#' @returns A list with components `tf_aes`, `scalar_tf_aes`, and `regular_aes`.
#' @examples
#' parse_tf_aesthetics(ggplot2::aes(tf = f, color = group))
#' parse_tf_aesthetics(ggplot2::aes(x = x, y = y))
#' @export
parse_tf_aesthetics <- function(mapping, data = NULL) {
  if (length(mapping) == 0) {
    return(list(tf_aes = list(), scalar_tf_aes = list(), regular_aes = aes()))
  }

  # Get aesthetic names
  aes_names <- names(mapping)

  # Identify tf-specific aesthetics (prefixed with tf_)
  tf_pattern <- "^tf(_.*)?$"
  is_direct_tf_aes <- grepl(tf_pattern, aes_names)

  # For non-tf-prefixed aesthetics, check if they contain tf functions
  is_scalar_tf_aes <- rep(FALSE, length(mapping))
  if (!is.null(data)) {
    for (i in seq_along(mapping)) {
      if (!is_direct_tf_aes[i]) {
        # Only try to extract expression if it's a quosure
        if (rlang::is_quosure(mapping[[i]])) {
          expr <- rlang::quo_get_expr(mapping[[i]])
          # Check if expression contains tf function calls
          expr_text <- rlang::expr_deparse(expr)
          if (grepl("tf_", expr_text)) {
            is_scalar_tf_aes[i] <- TRUE
          }
        }
        # Non-quosure aesthetics (constants) cannot contain tf functions
      }
    }
  }

  # Split aesthetics into three categories
  tf_aes <- mapping[is_direct_tf_aes]
  scalar_tf_aes <- mapping[is_scalar_tf_aes]
  regular_aes <- mapping[!is_direct_tf_aes & !is_scalar_tf_aes]

  # Convert regular_aes back to aes object if not empty
  if (length(regular_aes) > 0) {
    class(regular_aes) <- "uneval"
  } else {
    regular_aes <- aes()
  }

  return(list(
    tf_aes = tf_aes,
    scalar_tf_aes = scalar_tf_aes,
    regular_aes = regular_aes
  ))
}

#' Add layers to tf_ggplot objects
#'
#' @param e1 A tf_ggplot object
#' @param e2 A ggplot2 layer, scale, theme, etc.
#' @returns A modified `tf_ggplot` object.
#' @export
`+.tf_ggplot` <- function(e1, e2) {
  # If e2 is a layer, check if it has tf aesthetics
  if (inherits(e2, "LayerInstance") || inherits(e2, "Layer")) {
    # Reject geom_capellini: fundamentally incompatible (glyph-plot semantics)
    if (inherits(e2$stat, "StatCapellini")) {
      cli::cli_abort(c(
        "{.fn geom_capellini} is not compatible with {.fn tf_ggplot}",
        "i" = "{.fn geom_capellini} uses {.code x}/{.code y} for position and {.code tf} for sparkline shape",
        "i" = "Use {.fn geom_capellini} with plain {.fn ggplot} instead"
      ))
    }

    # Translate old-style tf geoms (geom_spaghetti, geom_meatballs, geom_errorband)
    if (inherits(e2$stat, "StatTf") || inherits(e2$stat, "StatErrorband")) {
      result <- translate_old_tf_layer(e2, e1)
      e1 <- result$plot_obj
      for (new_layer in result$layers) {
        e1 <- e1 + new_layer
      }
      return(e1)
    }

    # Promote tf objects passed as constant params (e.g. geom_line(tf = tf_rgp(5)))
    # into proper layer data + mapping so the normal tf-layer path can handle them.
    .promoted_data <- NULL
    .tf_aes_names <- c("tf", "tf_x", "tf_y", "tf_ymin", "tf_ymax")
    .tf_in_params <- intersect(names(e2$aes_params), .tf_aes_names)
    if (length(.tf_in_params) > 0) {
      .tf_col_list <- list()
      .new_mapping <- e2$mapping %||% aes()
      for (.p in .tf_in_params) {
        .val <- e2$aes_params[[.p]]
        if (is_tf(.val)) {
          .col <- paste0(".", .p, "_const_")
          .tf_col_list[[.col]] <- .val
          .new_mapping[[.p]] <- rlang::sym(.col)
          e2$aes_params[[.p]] <- NULL
        }
      }
      if (length(.tf_col_list) > 0) {
        .promoted_data <- structure(
          .tf_col_list,
          class = "data.frame",
          row.names = seq_len(length(.tf_col_list[[1]]))
        )
        # Do NOT set layer$data here (ggproto reference semantics would mutate
        # the layer permanently, breaking subsequent finalize calls on the same
        # tf_ggplot object). Store promoted data in layer_info instead.
        e2$mapping <- .new_mapping
      }
    }

    # Extract layer mapping
    layer_mapping <- e2$mapping %||% aes()

    # Check if this layer or the plot has tf aesthetics
    parsed_layer_aes <- parse_tf_aesthetics(
      layer_mapping,
      .promoted_data %||% e1$data
    )
    parsed_plot_aes <- parse_tf_aesthetics(e1$mapping, e1$data)

    layer_has_tf <- length(parsed_layer_aes$tf_aes) > 0 ||
      length(parsed_layer_aes$scalar_tf_aes) > 0
    plot_has_tf <- length(parsed_plot_aes$tf_aes) > 0 ||
      length(parsed_plot_aes$scalar_tf_aes) > 0

    if (layer_has_tf || plot_has_tf) {
      # Store this layer with tf aesthetics
      all_layers <- attr(e1, "all_layers")
      if (is.null(all_layers)) all_layers <- list()
      all_layers[[length(all_layers) + 1]] <- list(
        layer = e2,
        layer_mapping = layer_mapping,
        parsed_aes = parsed_layer_aes,
        is_tf_layer = TRUE,
        # Promoted data from tf-param expansion lives here (not in layer$data)
        # so that repeated finalize calls don't see mutated layer state.
        promoted_data = .promoted_data,
        # Snapshot layer-level data at add-time (before any finalize mutation).
        # Used when a layer carries its own data (e.g. from autolayer.tf).
        layer_data_snapshot = if (!inherits(e2$data, "waiver")) e2$data else
          NULL
      )
      attr(e1, "all_layers") <- all_layers

      # Return the tf_ggplot object to continue accumulating layers
      return(e1)
    }
  }

  # Handle regular layers (non-tf layers)
  if (inherits(e2, "LayerInstance") || inherits(e2, "Layer")) {
    layer_mapping <- e2$mapping %||% aes()
    parsed_layer_aes <- parse_tf_aesthetics(layer_mapping, e1$data)
    layer_has_tf <- length(parsed_layer_aes$tf_aes) > 0 ||
      length(parsed_layer_aes$scalar_tf_aes) > 0

    if (!layer_has_tf) {
      # This is a regular (non-tf) layer - store it instead of finalizing immediately

      # Set the layer to use original data
      e2$data <- e1$data

      # Check for potential scale conflicts
      all_layers <- attr(e1, "all_layers")
      tf_layers_exist <- (!is.null(all_layers) &&
        any(sapply(all_layers, function(x) x$is_tf_layer))) ||
        length(parse_tf_aesthetics(e1$mapping, e1$data)$tf_aes) > 0

      if (tf_layers_exist) {
        # Check if this regular layer uses aesthetics that might conflict with tf layers
        regular_aes_names <- names(layer_mapping)
        tf_conflicting_aes <- c("x", "y", "ymin", "ymax") # aesthetics that tf layers commonly use

        potential_conflicts <- intersect(regular_aes_names, tf_conflicting_aes)
        if (length(potential_conflicts) > 0) {
          cli::cli_warn(c(
            "Potential scale conflict detected: regular layer uses {.field {potential_conflicts}} aesthetic(s) while tf layers also modify these scales",
            "i" = "This may result in unexpected scale ranges or behavior",
            "i" = "Consider using different aesthetics or ensuring data ranges are compatible"
          ))
        }
      }

      # Store the regular layer in order
      all_layers <- attr(e1, "all_layers")
      if (is.null(all_layers)) all_layers <- list()
      all_layers[[length(all_layers) + 1]] <- list(
        layer = e2,
        layer_mapping = layer_mapping,
        parsed_aes = parsed_layer_aes,
        is_tf_layer = FALSE
      )
      attr(e1, "all_layers") <- all_layers

      # Return the tf_ggplot object to continue accumulating layers
      return(e1)
    }
  }

  # Non-layer object (like themes, scales, etc.) - need to finalize tf_ggplot
  all_layers <- attr(e1, "all_layers")
  if (
    (!is.null(all_layers) && length(all_layers) > 0) ||
      length(parse_tf_aesthetics(e1$mapping, e1$data)$tf_aes) > 0 ||
      length(parse_tf_aesthetics(e1$mapping, e1$data)$scalar_tf_aes) > 0
  ) {
    # Convert tf_ggplot with all accumulated layers
    regular_plot <- finalize_tf_ggplot(e1)
    return(regular_plot + e2)
  }

  # No tf aesthetics at all - convert to regular ggplot
  regular_plot <- ggplot(data = e1$data, mapping = e1$mapping)
  regular_plot$theme <- e1$theme
  regular_plot$coordinates <- e1$coordinates
  regular_plot$facet <- e1$facet
  regular_plot$labels <- e1$labels

  return(regular_plot + e2)
}

#' Finalize tf_ggplot by processing all tf layers independently
#'
#' Each layer is evaluated on its own natural argument grid (Option B architecture).
#' The base ggplot has `data = NULL`; each layer receives its own long-format data.
#'
#' @param tf_plot A tf_ggplot object with accumulated layers
#' @return Regular ggplot object with all layers properly transformed
#' @keywords internal
finalize_tf_ggplot <- function(tf_plot) {
  user_arg <- attr(tf_plot, "tf_arg")
  interpolate <- attr(tf_plot, "tf_interpolate") %||% TRUE
  original_data <- tf_plot$data
  all_layers <- attr(tf_plot, "all_layers") %||% list()

  # Parse plot-level aesthetics
  parsed_plot_aes <- parse_tf_aesthetics(tf_plot$mapping, original_data)

  # Pre-evaluate all scalar tf aesthetics (e.g., color = tf_depth(f)) into enriched_data.
  # These produce one scalar per original row and are replicated automatically via join.
  enriched_data <- original_data
  scalar_col_map <- list() # key -> col_name in enriched_data

  n_data_rows <- nrow(enriched_data)
  eval_scalar_tf_aes <- function(aes_name, quo) {
    result <- tryCatch(
      rlang::eval_tidy(quo, data = enriched_data),
      error = function(e)
        cli::cli_abort(
          "Error evaluating scalar tf aesthetic {.code {aes_name}}: {e$message}"
        )
    )
    if (!length(result) %in% c(1L, n_data_rows)) {
      cli::cli_abort(c(
        "Scalar tf aesthetic {.code {aes_name}} returned {length(result)} value(s)",
        "i" = "Expected 1 or {n_data_rows} (one per row of plot data)"
      ))
    }
    result
  }

  for (aes_name in names(parsed_plot_aes$scalar_tf_aes)) {
    quo <- parsed_plot_aes$scalar_tf_aes[[aes_name]]
    result <- eval_scalar_tf_aes(aes_name, quo)
    col_nm <- make_safe_column_name(
      paste0(".s.", aes_name),
      names(enriched_data)
    )
    enriched_data[[col_nm]] <- result
    scalar_col_map[[aes_name]] <- col_nm
  }
  for (i in seq_along(all_layers)) {
    if (!all_layers[[i]]$is_tf_layer) next
    for (aes_name in names(all_layers[[i]]$parsed_aes$scalar_tf_aes)) {
      key <- paste0(aes_name, ".layer.", i)
      quo <- all_layers[[i]]$parsed_aes$scalar_tf_aes[[aes_name]]
      result <- eval_scalar_tf_aes(aes_name, quo)
      col_nm <- make_safe_column_name(paste0(".s.", key), names(enriched_data))
      enriched_data[[col_nm]] <- result
      scalar_col_map[[key]] <- col_nm
    }
  }

  # Base mapping: regular plot-level aes + scalar tf aes remapped to column names.
  # Does NOT include x/y/group from tf aes -- those are per-layer.
  base_mapping <- parsed_plot_aes$regular_aes
  for (aes_name in names(parsed_plot_aes$scalar_tf_aes)) {
    base_mapping[[aes_name]] <- rlang::sym(scalar_col_map[[aes_name]])
  }

  # Create base plot with data = NULL; each layer brings its own data
  regular_plot <- ggplot(data = NULL, mapping = base_mapping)
  raw_tf_axis_labels <- NULL

  for (i in seq_along(all_layers)) {
    layer_info <- all_layers[[i]]
    layer <- layer_info$layer

    if (layer_info$is_tf_layer) {
      # Use promoted data (from tf-param expansion) when present;
      # fall back to plot-level enriched_data.  Never use layer$data here --
      # ggproto reference semantics mean layer$data gets mutated by a previous
      # finalize call, breaking repeated ggplot_build() calls on the same object.
      effective_data <- layer_info$promoted_data %||%
        layer_info$layer_data_snapshot %||%
        enriched_data
      result <- if (inherits(layer$stat, "StatFboxplot")) {
        build_raw_tf_layer_data(
          layer_info = layer_info,
          plot_mapping = tf_plot$mapping,
          layer_idx = i,
          enriched_data = effective_data
        )
      } else {
        build_tf_layer_data(
          layer_info = layer_info,
          plot_tf_aes = parsed_plot_aes$tf_aes,
          scalar_col_map = scalar_col_map,
          layer_idx = i,
          enriched_data = effective_data,
          user_arg = user_arg,
          interpolate = interpolate
        )
      }
      if (!is.null(result)) {
        layer$data <- result$long_data
        layer$mapping <- result$new_mapping
        if (!is.null(result$stat_params)) {
          layer$stat_params <- utils::modifyList(
            layer$stat_params,
            result$stat_params
          )
        }
        if (is.null(raw_tf_axis_labels) && !is.null(result$axis_labels)) {
          raw_tf_axis_labels <- result$axis_labels
        }
      } else if (is.null(layer$data)) {
        layer$data <- enriched_data
      }
    } else if (is.null(layer$data)) {
      layer$data <- enriched_data
    }

    regular_plot <- regular_plot + layer
  }

  regular_plot$theme <- tf_plot$theme
  regular_plot$coordinates <- tf_plot$coordinates
  regular_plot$facet <- tf_plot$facet
  regular_plot$labels <- tf_plot$labels

  # Fix legend titles for scalar tf aesthetics (e.g. linewidth = tf_fmean(f)):
  # the internal column name (.s.linewidth) leaks into the legend; replace with
  # the original expression text.
  for (aes_name in names(parsed_plot_aes$scalar_tf_aes)) {
    if (is.null(regular_plot$labels[[aes_name]])) {
      quo <- parsed_plot_aes$scalar_tf_aes[[aes_name]]
      expr_text <- paste(
        rlang::expr_deparse(rlang::quo_get_expr(quo)),
        collapse = ""
      )
      regular_plot$labels[[aes_name]] <- expr_text
    }
  }

  if (!is.null(raw_tf_axis_labels)) {
    if (is.null(regular_plot$labels$x)) {
      regular_plot$labels$x <- raw_tf_axis_labels$x
    }
    if (is.null(regular_plot$labels$y)) {
      regular_plot$labels$y <- raw_tf_axis_labels$y
    }
  }

  regular_plot
}

combine_layer_mappings <- function(
  plot_mapping,
  layer_mapping,
  inherit.aes = TRUE
) {
  plot_mapping <- plot_mapping %||% aes()
  layer_mapping <- layer_mapping %||% aes()

  combined <- if (isTRUE(inherit.aes)) {
    c(
      plot_mapping[!names(plot_mapping) %in% names(layer_mapping)],
      layer_mapping
    )
  } else {
    layer_mapping
  }
  class(combined) <- "uneval"
  combined
}

build_raw_tf_layer_data <- function(
  layer_info,
  plot_mapping,
  layer_idx,
  enriched_data
) {
  layer <- layer_info$layer
  combined_mapping <- combine_layer_mappings(
    plot_mapping = plot_mapping,
    layer_mapping = layer$mapping,
    inherit.aes = layer$inherit.aes
  )

  if (inherits(layer$stat, "StatFboxplot")) {
    normalized <- normalize_fboxplot_mapping(combined_mapping)
    combined_mapping <- normalized$mapping
    stat_params <- list(use_group_aes = normalized$use_group_aes)
    axis_labels <- infer_tf_axis_labels(
      combined_mapping,
      orientation = layer$stat_params$orientation %||% NA
    )
  } else {
    stat_params <- NULL
    axis_labels <- NULL
  }

  list(
    long_data = enriched_data,
    new_mapping = combined_mapping,
    stat_params = stat_params,
    axis_labels = axis_labels
  )
}

infer_tf_axis_labels <- function(mapping, orientation = NA) {
  tf_name <- intersect(c("tf", "tf_y"), names(mapping))[1]
  if (is.na(tf_name) || is.null(tf_name)) {
    return(NULL)
  }

  tf_expr <- mapping[[tf_name]]
  if (!rlang::is_quosure(tf_expr)) {
    return(NULL)
  }

  expr_text <- paste(
    rlang::expr_deparse(rlang::quo_get_expr(tf_expr)),
    collapse = ""
  )

  orientation <- if (is.null(orientation) || is.na(orientation)) {
    "x"
  } else {
    match.arg(orientation, c("x", "y"))
  }

  if (orientation == "y") {
    list(
      x = expr_text,
      y = paste0(expr_text, ".arg")
    )
  } else {
    list(
      x = paste0(expr_text, ".arg"),
      y = expr_text
    )
  }
}

#' Build long-format data and new mapping for a single tf layer
#'
#' Evaluates the layer's tf aesthetics independently on their own natural grid
#' (or the user-specified grid). Called by [finalize_tf_ggplot()] per layer.
#'
#' @param layer_info List from `all_layers` with `layer`, `parsed_aes`, `is_tf_layer`
#' @param plot_tf_aes tf aesthetics from the plot level (used when `inherit.aes = TRUE`)
#' @param scalar_col_map Named list: key -> column name in enriched_data
#' @param layer_idx Integer index of this layer (for keying layer-level scalar aes)
#' @param enriched_data Data frame with original data + pre-evaluated scalar tf columns
#' @param user_arg Optional evaluation grid (overrides natural grid)
#' @param interpolate Whether to interpolate tf objects to `arg`
#' @return `NULL` when no effective tf aes, otherwise `list(long_data, new_mapping)`
#' @keywords internal
build_tf_layer_data <- function(
  layer_info,
  plot_tf_aes,
  scalar_col_map,
  layer_idx,
  enriched_data,
  user_arg,
  interpolate
) {
  layer <- layer_info$layer
  parsed_aes <- layer_info$parsed_aes

  # Effective tf aes: plot-level + layer-level, with layer overriding (respects inherit.aes)
  effective_tf_aes <- if (isTRUE(layer$inherit.aes)) {
    c(
      plot_tf_aes[!names(plot_tf_aes) %in% names(parsed_aes$tf_aes)],
      parsed_aes$tf_aes
    )
  } else {
    parsed_aes$tf_aes
  }

  if (length(effective_tf_aes) == 0) {
    return(NULL)
  }

  # Evaluate all tf expressions
  tf_objects <- vector("list", length(effective_tf_aes))
  names(tf_objects) <- names(effective_tf_aes)
  for (aes_name in names(effective_tf_aes)) {
    quo <- effective_tf_aes[[aes_name]]
    tf_obj <- tryCatch(
      rlang::eval_tidy(quo, data = enriched_data),
      error = function(e) {
        expr_text <- paste(
          rlang::expr_deparse(rlang::quo_get_expr(quo)),
          collapse = ""
        )
        cli::cli_abort(
          "Error evaluating tf aesthetic {.code {expr_text}}: {e$message}"
        )
      }
    )
    if (!is_tf(tf_obj)) {
      expr_text <- paste(
        rlang::expr_deparse(rlang::quo_get_expr(quo)),
        collapse = ""
      )
      cli::cli_abort(
        "tf aesthetic {.code {aes_name} = {expr_text}} must evaluate to a tf object, not {.obj_type_friendly {tf_obj}}"
      )
    }
    tf_objects[[aes_name]] <- tf_obj
  }

  tf_lengths <- vapply(tf_objects, length, integer(1))
  common_n_funcs <- max(tf_lengths)
  if (!all(tf_lengths %in% c(1L, common_n_funcs))) {
    cli::cli_abort(c(
      "All tf aesthetics in a layer must have matching lengths or length 1",
      "i" = "Observed lengths: {paste(tf_lengths, collapse = ', ')}"
    ))
  }

  primary_idx <- which(tf_lengths == common_n_funcs)[1]
  if (!identical(primary_idx, 1L)) {
    tf_objects <- tf_objects[c(
      primary_idx,
      setdiff(seq_along(tf_objects), primary_idx)
    )]
    effective_tf_aes <- effective_tf_aes[c(
      primary_idx,
      setdiff(seq_along(effective_tf_aes), primary_idx)
    )]
    tf_lengths <- tf_lengths[c(
      primary_idx,
      setdiff(seq_along(tf_lengths), primary_idx)
    )]
  }

  # This layer's evaluation grid: user-specified or natural grid of first tf object.
  # user_arg can be NULL, a numeric vector (explicit points), or a single integer
  # (desired grid length, resolved to an equidistant grid over the tf domain).
  arg <- user_arg
  if (is.null(arg)) {
    first_arg <- tf_arg(tf_objects[[1]])
    arg <- if (is.list(first_arg)) first_arg[[1]] else first_arg
  } else if (length(arg) == 1L) {
    domain <- tf_domain(tf_objects[[1]])
    arg <- seq(domain[1], domain[2], length.out = as.integer(arg))
  }

  # If there is no plot-level data (e.g. tf_ggplot() + geom_line(aes(tf = tf_rgp(5)))),
  # synthesise a one-row-per-function data frame so joins and indexing work correctly.
  n_funcs <- common_n_funcs
  .n_enriched <- nrow(enriched_data) # NULL for NULL / waiver / non-data-frames
  if (is.null(.n_enriched) || .n_enriched == 0) {
    enriched_data <- structure(
      list(),
      class = "data.frame",
      row.names = seq_len(n_funcs)
    )
  }

  n_rows <- nrow(enriched_data)
  if (n_rows != n_funcs) {
    if (n_funcs == 1L && n_rows > 0) {
      enriched_data <- enriched_data[1, , drop = FALSE]
      n_rows <- 1L
    } else if (n_rows == 1L && n_funcs > 1L) {
      enriched_data <- enriched_data[rep(1L, n_funcs), , drop = FALSE]
      n_rows <- n_funcs
    } else {
      cli::cli_abort(c(
        "Layer data cannot be aligned with the evaluated tf aesthetics",
        "i" = "Data has {n_rows} row(s), but tf aesthetics evaluate to {n_funcs} function(s)."
      ))
    }
  }
  n_grid <- length(arg)
  if (n_rows > 200 && n_grid > 100) {
    cli::cli_warn(
      c(
        "Large data expansion: {n_rows} functions \u00d7 {n_grid} grid points
       = {n_rows * n_grid} rows",
        "i" = "Use {.arg arg} in {.fn tf_ggplot} to specify a coarser
       evaluation grid (integer for grid length, numeric vector for
       specific points)"
      ),
      .frequency = "regularly",
      .frequency_id = "tidyfun_large_expansion"
    )
  }

  # Unnest first tf aesthetic to long format
  first_aes_name <- names(tf_objects)[1]
  first_tf <- tf_objects[[1]]

  # Use expression text for column names (produces meaningful axis labels).
  # For simple column refs like aes(tf = curves), val_col = "curves".
  # For complex expressions like aes(tf = func1 + func2), val_col = "func1 + func2".
  first_quo <- effective_tf_aes[[first_aes_name]]
  first_expr <- rlang::quo_get_expr(first_quo)
  expr_text <- paste(rlang::expr_deparse(first_expr), collapse = "")
  # Exclude the source column name from conflict check (it will be removed before join)
  excl <- if (is.symbol(first_expr)) as.character(first_expr) else character(0)
  safe_name <- make_safe_column_name(
    expr_text,
    existing_names = setdiff(names(enriched_data), excl)
  )
  val_col <- safe_name
  arg_col <- paste0(safe_name, ".arg")
  id_col <- paste0(safe_name, ".id")

  tf_long <- suppressMessages(tf_unnest(
    first_tf,
    arg = arg,
    interpolate = interpolate
  ))
  # tf_long has columns: id (ordered factor), arg, value
  # Rename immediately to avoid conflicts with user data columns of the same name
  names(tf_long)[names(tf_long) == "id"] <- id_col
  names(tf_long)[names(tf_long) == "arg"] <- arg_col
  names(tf_long)[names(tf_long) == "value"] <- val_col

  tf_long$.row_id_ <- rep(seq_len(n_rows), each = n_grid)
  tf_long <- tf_long[!is.na(tf_long[[val_col]]), ]

  # Join with enriched_data to replicate covariates + scalar columns
  work_data <- enriched_data
  work_data$.row_id_ <- seq_len(n_rows)
  # Remove original tf columns to avoid conflicts during join
  for (aes_name in names(effective_tf_aes)) {
    expr <- rlang::quo_get_expr(effective_tf_aes[[aes_name]])
    if (is.symbol(expr)) {
      col_nm <- as.character(expr)
      if (col_nm %in% names(work_data)) work_data[[col_nm]] <- NULL
    }
  }

  long_data <- left_join(tf_long, work_data, by = ".row_id_") |>
    select(-.row_id_)

  # Build the layer mapping
  new_mapping <- parsed_aes$regular_aes

  for (i in seq_along(tf_objects)) {
    aes_name_i <- names(tf_objects)[i]
    if (i == 1) {
      v_col <- val_col
      a_col <- arg_col
      g_col <- id_col
    } else {
      # Evaluate additional tf aes on the FULL arg grid (not NA-filtered subset).
      # Using curr_arg (from long_data after NA filtering) would silently drop points
      # if the primary tf had NAs, misaligning the secondary aesthetic.
      tf_vals <- tf_evaluate(tf_objects[[i]], arg = arg)
      # Use expression text for meaningful column names
      quo_i <- effective_tf_aes[[aes_name_i]]
      expr_i <- rlang::quo_get_expr(quo_i)
      expr_text_i <- paste(rlang::expr_deparse(expr_i), collapse = "")
      excl_i <- if (is.symbol(expr_i)) as.character(expr_i) else character(0)
      safe_i <- make_safe_column_name(
        expr_text_i,
        existing_names = setdiff(
          c(names(long_data), names(enriched_data)),
          excl_i
        )
      )
      v_col <- safe_i
      a_col <- paste0(safe_i, ".arg")
      g_col <- paste0(safe_i, ".id")
      n_tf_i <- length(tf_vals)
      if (n_tf_i == 1L && n_funcs > 1L) {
        long_data[[v_col]] <- unlist(tf_vals)[
          match(long_data[[arg_col]], arg)
        ]
        long_data[[g_col]] <- long_data[[id_col]]
      } else {
        # Build full secondary long-form, then left-join on (id, arg) to align with primary
        sec_long <- data.frame(
          .row_id_i_ = rep(seq_len(n_tf_i), each = length(arg)),
          .arg_i_ = rep(arg, n_tf_i),
          .val_i_ = unlist(tf_vals)
        )
        # Attach secondary values to long_data by matching row-id and arg position
        long_data[[v_col]] <- sec_long$.val_i_[
          match(
            paste(as.integer(long_data[[id_col]]), long_data[[arg_col]]),
            paste(sec_long$.row_id_i_, sec_long$.arg_i_)
          )
        ]
        long_data[[g_col]] <- long_data[[id_col]]
      }
      long_data[[a_col]] <- long_data[[arg_col]]
    }
    new_mapping <- add_tf_aes_to_mapping(
      new_mapping,
      aes_name_i,
      v_col,
      a_col,
      g_col
    )
  }

  # Layer-level scalar tf aes: column is already in long_data (joined from enriched_data)
  for (aes_name in names(parsed_aes$scalar_tf_aes)) {
    key <- paste0(aes_name, ".layer.", layer_idx)
    if (key %in% names(scalar_col_map)) {
      new_mapping[[aes_name]] <- rlang::sym(scalar_col_map[[key]])
    }
  }

  list(long_data = long_data, new_mapping = new_mapping)
}

#' Add a tf aesthetic to an ggplot2 mapping object
#' @keywords internal
add_tf_aes_to_mapping <- function(mapping, aes_name, val_col, arg_col, id_col) {
  if (aes_name %in% c("tf", "tf_y")) {
    mapping$y <- rlang::sym(val_col)
    mapping$x <- rlang::sym(arg_col)
    mapping$group <- rlang::sym(id_col)
  } else if (aes_name == "tf_x") {
    mapping$x <- rlang::sym(val_col)
    if (is.null(mapping$group)) mapping$group <- rlang::sym(id_col)
  } else if (aes_name == "tf_ymin") {
    mapping$ymin <- rlang::sym(val_col)
    if (is.null(mapping$x)) mapping$x <- rlang::sym(arg_col)
    if (is.null(mapping$group)) mapping$group <- rlang::sym(id_col)
  } else if (aes_name == "tf_ymax") {
    mapping$ymax <- rlang::sym(val_col)
    if (is.null(mapping$x)) mapping$x <- rlang::sym(arg_col)
    if (is.null(mapping$group)) mapping$group <- rlang::sym(id_col)
  }
  mapping
}

# Translate old-style tf geoms to tf_ggplot-compatible layers ----------------

#' Translate a geom_spaghetti/geom_meatballs/geom_errorband layer for tf_ggplot
#'
#' Remaps old-style y/ymin/ymax tf aesthetics to tf/tf_ymin/tf_ymax and
#' substitutes the appropriate standard ggplot2 geom.
#'
#' @param layer A LayerInstance using StatTf or StatErrorband
#' @param e1 The tf_ggplot object (may be modified to clean up plot-level y)
#' @return A list with `plot_obj` (possibly modified e1) and `layers` (list of
#'   new layers to add)
#' @keywords internal
translate_old_tf_layer <- function(layer, e1) {
  layer_mapping <- layer$mapping %||% aes()
  plot_mapping <- e1$mapping %||% aes()
  extra_params <- layer$aes_params

  make_layer <- function(geom_fn, mapping) {
    layer_data <- if (!inherits(layer$data, "waiver")) layer$data else NULL
    do.call(
      geom_fn,
      c(
        list(
          mapping = mapping,
          data = layer_data,
          position = layer$position,
          show.legend = layer$show.legend,
          inherit.aes = layer$inherit.aes
        ),
        extra_params
      )
    )
  }

  if (inherits(layer$stat, "StatTf")) {
    if ("y" %in% names(layer_mapping)) {
      new_mapping <- layer_mapping
      new_mapping[["tf"]] <- new_mapping[["y"]]
      new_mapping[["y"]] <- NULL
    } else if ("y" %in% names(plot_mapping)) {
      new_mapping <- layer_mapping
      new_mapping[["tf"]] <- plot_mapping[["y"]]
      e1$mapping[["y"]] <- NULL # remove from plot mapping to avoid conflicts
    } else {
      cli::cli_warn(
        "Cannot translate {.fn geom_spaghetti}/{.fn geom_meatballs}: no {.code y} aesthetic found"
      )
      return(list(plot_obj = e1, layers = list(layer)))
    }

    cli::cli_inform(
      c(
        "i" = "{.fn geom_spaghetti} layer automatically translated for {.fn tf_ggplot}",
        "*" = "Use {.fn geom_line} with {.code aes(tf = f)} directly to silence this"
      ),
      .frequency = "regularly",
      .frequency_id = "translate_spaghetti"
    )

    layers <- if (inherits(layer$geom, "GeomMeatball")) {
      list(
        make_layer(geom_line, new_mapping),
        make_layer(geom_point, new_mapping)
      )
    } else {
      list(make_layer(geom_line, new_mapping))
    }
    list(plot_obj = e1, layers = layers)
  } else if (inherits(layer$stat, "StatErrorband")) {
    new_mapping <- layer_mapping
    if ("ymin" %in% names(new_mapping)) {
      new_mapping[["tf_ymin"]] <- new_mapping[["ymin"]]
      new_mapping[["ymin"]] <- NULL
    }
    if ("ymax" %in% names(new_mapping)) {
      new_mapping[["tf_ymax"]] <- new_mapping[["ymax"]]
      new_mapping[["ymax"]] <- NULL
    }

    cli::cli_inform(
      c(
        "i" = "{.fn geom_errorband} layer automatically translated for {.fn tf_ggplot}",
        "*" = "Use {.fn geom_ribbon} with {.code aes(tf_ymin = lo, tf_ymax = hi)} directly to silence this"
      ),
      .frequency = "regularly",
      .frequency_id = "translate_errorband"
    )

    list(plot_obj = e1, layers = list(make_layer(geom_ribbon, new_mapping)))
  }
}

# Helper function to check if mapping is an aes object
is_mapping <- function(x) {
  inherits(x, "uneval")
}

# Helper function to create safe column names from expression text
make_safe_column_name <- function(expr_text, existing_names = character(0)) {
  # For simple names (just letters, numbers, underscore), use as-is if valid
  if (
    grepl("^[a-zA-Z][a-zA-Z0-9_]*$", expr_text) &&
      !expr_text %in% existing_names
  ) {
    return(expr_text)
  }

  # For complex expressions, use the original text as column name
  # R allows any string as a column name if accessed properly
  safe_name <- expr_text

  # If the resulting name conflicts with existing names, add suffix
  original_safe_name <- safe_name
  counter <- 1
  while (safe_name %in% existing_names) {
    safe_name <- paste0(original_safe_name, "_", counter)
    counter <- counter + 1
  }

  safe_name
}

#' Print method for tf_ggplot
#' @param x A tf_ggplot object
#' @param ... Additional arguments
#' @returns `x`, invisibly. Called for its side effect of printing the plot.
#' @export
print.tf_ggplot <- function(x, ...) {
  # If there are tf layers or tf aesthetics, finalize before printing
  all_layers <- attr(x, "all_layers")
  if (
    (!is.null(all_layers) && length(all_layers) > 0) ||
      length(parse_tf_aesthetics(x$mapping, x$data)$tf_aes) > 0 ||
      length(parse_tf_aesthetics(x$mapping, x$data)$scalar_tf_aes) > 0
  ) {
    regular_plot <- finalize_tf_ggplot(x)
    print(regular_plot)
  } else {
    # No tf aesthetics, print as regular ggplot
    class(x) <- setdiff(class(x), "tf_ggplot")
    print(x)
  }
}

#' ggplot_build method for tf_ggplot
#' @param plot A tf_ggplot object
#' @param ... Additional arguments passed through from [ggplot2::ggplot_build()].
#' @returns A built ggplot object (class `ggplot_built`).
#' @export
ggplot_build.tf_ggplot <- function(plot, ...) {
  # Finalize tf_ggplot before building
  all_layers <- attr(plot, "all_layers")
  if (
    (!is.null(all_layers) && length(all_layers) > 0) ||
      length(parse_tf_aesthetics(plot$mapping, plot$data)$tf_aes) > 0 ||
      length(parse_tf_aesthetics(plot$mapping, plot$data)$scalar_tf_aes) > 0
  ) {
    regular_plot <- finalize_tf_ggplot(plot)
    return(ggplot_build(regular_plot))
  } else {
    # No tf aesthetics, build as regular ggplot
    class(plot) <- setdiff(class(plot), "tf_ggplot")
    return(ggplot_build(plot))
  }
}

# Register tf aesthetics as optional on ggplot2 geoms so that
# geom_line(aes(tf = ...)) etc. do not warn "Ignoring unknown aesthetics".
# Appending to optional_aes is the lightest-touch approach: it prevents the
# warning at the source (inside ggplot2::layer()) and works in all contexts
# (interactive, knitr, testthat). Cleaned up in .onUnload().
.tf_optional_aes <- c("tf", "tf_x", "tf_y", "tf_ymin", "tf_ymax")

.onLoad <- function(libname, pkgname) {
  geoms <- list(
    ggplot2::GeomLine,
    ggplot2::GeomPath,
    ggplot2::GeomPoint,
    ggplot2::GeomRibbon,
    ggplot2::GeomArea,
    ggplot2::GeomStep
  )
  for (g in geoms) {
    g$optional_aes <- union(g$optional_aes, .tf_optional_aes)
  }
}

.onUnload <- function(libpath) {
  geoms <- list(
    ggplot2::GeomLine,
    ggplot2::GeomPath,
    ggplot2::GeomPoint,
    ggplot2::GeomRibbon,
    ggplot2::GeomArea,
    ggplot2::GeomStep
  )
  for (g in geoms) {
    g$optional_aes <- setdiff(g$optional_aes, .tf_optional_aes)
  }
}

make_fboxplot_data <- function() {
  arg <- seq(0, 1, length.out = 21)
  curves <- rbind(
    1.0 + 0.10 * sin(2 * pi * arg),
    1.0 + 0.12 * sin(2 * pi * arg + 0.1),
    0.98 + 0.09 * sin(2 * pi * arg - 0.1),
    1.55 + 0.10 * sin(2 * pi * arg),
    1.2 + 0.08 * cos(2 * pi * arg),
    1.22 + 0.09 * cos(2 * pi * arg + 0.1),
    1.18 + 0.08 * cos(2 * pi * arg - 0.1),
    1.75 + 0.10 * cos(2 * pi * arg)
  )

  data.frame(
    id = seq_len(nrow(curves)),
    grp = factor(rep(c("A", "B"), each = 4)),
    facet_grp = factor(rep(c("left", "right"), times = 4))
  ) |>
    dplyr::mutate(func = tfd(curves, arg = arg))
}

test_that("geom_fboxplot builds on tf_ggplot", {
  data <- make_fboxplot_data()

  p <- tf_ggplot(data, aes(tf = func)) + geom_fboxplot()
  built <- ggplot_build(p)
  plot_data <- built$data[[1]]

  expect_true(all(
    c(".fbox_component", ".fbox_stat_group") %in% names(plot_data)
  ))
  expect_true(all(
    c("box", "median", "whisker_lower", "whisker_upper") %in%
      unique(plot_data$.fbox_component)
  ))
})

test_that("geom_fboxplot sets tf axis labels like standard tf geoms", {
  data <- make_fboxplot_data()

  p <- tf_ggplot(data, aes(tf = func)) + geom_fboxplot()
  built <- ggplot_build(p)

  expect_equal(built$plot$labels$y, "func")
  expect_equal(built$plot$labels$x, "func.arg")
})

test_that("geom_fboxplot swaps default axis labels for horizontal orientation", {
  data <- make_fboxplot_data()

  p <- tf_ggplot(data, aes(tf = func)) + geom_fboxplot(orientation = "y")
  built <- ggplot_build(p)

  expect_equal(built$plot$labels$x, "func")
  expect_equal(built$plot$labels$y, "func.arg")
})

test_that("geom_fboxplot groups within a panel by fill and colour", {
  data <- make_fboxplot_data()

  p <- tf_ggplot(data, aes(tf = func, fill = grp, colour = grp)) +
    geom_fboxplot()
  built <- ggplot_build(p)

  expect_equal(length(unique(built$data[[1]]$.fbox_stat_group)), 2)
})

test_that("geom_fboxplot uses mapped colour for ribbons when fill is absent", {
  data <- make_fboxplot_data()

  p <- tf_ggplot(data, aes(tf = func, colour = grp)) + geom_fboxplot()
  panel_grob <- ggplotGrob(p)$grobs[[which(
    ggplotGrob(p)$layout$name == "panel"
  )]]
  layer_grob <- panel_grob$children[[3]]
  polygon_idx <- grepl("^geom_polygon", names(layer_grob$children))
  fills <- vapply(
    layer_grob$children[polygon_idx],
    \(g) g$gp$fill,
    character(1)
  )

  expect_equal(length(unique(fills)), 2)
})

test_that("geom_fboxplot is compatible with facetting", {
  data <- make_fboxplot_data()

  p <- tf_ggplot(data, aes(tf = func, fill = grp, colour = grp)) +
    geom_fboxplot() +
    facet_wrap(~facet_grp)
  built <- ggplot_build(p)

  expect_equal(as.character(sort(unique(built$data[[1]]$PANEL))), c("1", "2"))
})

test_that("geom_fboxplot handles irregular tf data on common support", {
  p <- tf_ggplot(tidyfun::dti_df, aes(tf = rcst)) + geom_fboxplot()

  expect_no_warning(built <- ggplot_build(p))
  expect_true(nrow(built$data[[1]]) > 0)
})

test_that("geom_fboxplot is compatible with scaled axes", {
  data <- make_fboxplot_data()

  p <- tf_ggplot(data, aes(tf = func)) +
    geom_fboxplot() +
    scale_y_log10()
  built <- ggplot_build(p)
  numeric_cols <- intersect(
    c("x", "y", "ymin", "ymax", "xmin", "xmax"),
    names(built$data[[1]])
  )

  expect_true(length(numeric_cols) > 0)
  expect_true(all(vapply(
    built$data[[1]][numeric_cols],
    \(x) all(is.finite(x[!is.na(x)])),
    logical(1)
  )))
})

test_that("geom_fboxplot stat output is retransformed by y scales", {
  data <- make_fboxplot_data()
  data <- dplyr::mutate(data, pos_func = exp(func) + 100)

  p <- tf_ggplot(data, aes(tf = pos_func, fill = grp)) +
    geom_fboxplot() +
    scale_y_log10()
  built <- ggplot_build(p)
  numeric_cols <- intersect(
    c("y", "ymin", "ymax", "xmin", "xmax"),
    names(built$data[[1]])
  )
  vals <- unlist(built$data[[1]][numeric_cols])
  vals <- vals[is.finite(vals)]

  expect_true(max(vals) < 3)
  expect_true(min(vals) > 1)
})

test_that("geom_fboxplot supports flipped orientation", {
  data <- make_fboxplot_data()

  p <- tf_ggplot(data, aes(tf = func)) +
    geom_fboxplot(orientation = "y")
  built <- ggplot_build(p)
  plot_data <- built$data[[1]]

  expect_true(all(c("x", "y") %in% names(plot_data)))
  expect_true(all(
    c("xmin", "xmax") %in%
      names(plot_data[plot_data$.fbox_component == "box", ])
  ))
})

test_that("geom_fboxplot accepts custom depth and fence functions", {
  data <- make_fboxplot_data()

  depth_fn <- function(x, arg = NULL) {
    rev(seq_along(x))
  }
  fence_fn <- function(
    tf_vec,
    arg,
    depth,
    median,
    central_lower,
    central_upper,
    coef,
    central
  ) {
    list(
      lower = central_lower - 0.01,
      upper = central_upper + 0.01,
      outliers = c(rep(FALSE, length(tf_vec) - 1), TRUE)
    )
  }

  p <- tf_ggplot(data, aes(tf = func)) +
    geom_fboxplot(depth_fn = depth_fn, fence_fn = fence_fn)
  built <- ggplot_build(p)

  expect_true(any(built$data[[1]]$.fbox_component == "outlier"))
})

test_that("tf_ggplot constructor creates proper object", {
  skip_if_not_installed("ggplot2")

  # Create test data
  set.seed(123)
  data <- data.frame(
    id = 1:3,
    group = factor(c("A", "A", "B"))
  )
  data$func <- tf_rgp(3, arg = seq(0, 1, length.out = 11))

  # Test basic constructor
  p <- tf_ggplot(data)
  expect_s3_class(p, "tf_ggplot")
  expect_s3_class(p, "ggplot") # Should inherit from ggplot

  # Test with aesthetic mapping
  p2 <- tf_ggplot(data, aes(tf = func, color = group))
  expect_s3_class(p2, "tf_ggplot")
  expect_equal(length(p2$mapping), 2)

  # Test with arg parameter
  p3 <- tf_ggplot(data, arg = seq(0, 1, length.out = 21))
  expect_s3_class(p3, "tf_ggplot")
  expect_equal(attr(p3, "tf_arg"), seq(0, 1, length.out = 21))
})

test_that("parse_tf_aesthetics correctly identifies tf aesthetics", {
  skip_if_not_installed("ggplot2")

  # Test simple tf aesthetic
  mapping1 <- aes(tf = func, color = group)
  parsed1 <- parse_tf_aesthetics(mapping1)
  expect_named(parsed1, c("tf_aes", "scalar_tf_aes", "regular_aes"))
  expect_equal(length(parsed1$tf_aes), 1)
  expect_equal(names(parsed1$tf_aes), "tf")
  expect_equal(length(parsed1$regular_aes), 1)

  # Test multiple tf aesthetics
  mapping2 <- aes(tf_x = func1, tf_y = func2, color = group)
  parsed2 <- parse_tf_aesthetics(mapping2)
  expect_equal(length(parsed2$tf_aes), 2)
  expect_equal(names(parsed2$tf_aes), c("tf_x", "tf_y"))

  # Test ribbon aesthetics
  mapping3 <- aes(tf_ymin = lower, tf_ymax = upper, fill = group)
  parsed3 <- parse_tf_aesthetics(mapping3)
  expect_equal(length(parsed3$tf_aes), 2)
  expect_equal(names(parsed3$tf_aes), c("tf_ymin", "tf_ymax"))

  # Test no tf aesthetics
  mapping4 <- aes(x = time, y = value, color = group)
  parsed4 <- parse_tf_aesthetics(mapping4)
  expect_equal(length(parsed4$tf_aes), 0)
  expect_equal(length(parsed4$regular_aes), 3)
})

test_that("data transformation via ggplot_build works correctly", {
  skip_if_not_installed("ggplot2")

  set.seed(123)
  data <- data.frame(
    id = 1:3,
    group = factor(c("A", "A", "B"))
  )
  data$func <- tf_rgp(3, arg = seq(0, 1, length.out = 11))

  p <- tf_ggplot(data, aes(tf = func, color = group)) + geom_line()
  built <- ggplot_build(p)
  plot_data <- built$data[[1]]

  # 3 functions × 11 points = 33 rows
  expect_equal(nrow(plot_data), 33)
  expect_equal(length(unique(plot_data$group)), 3)
  expect_equal(length(unique(plot_data$x)), 11)
  expect_equal(length(unique(plot_data$colour)), 2) # A and B
})

test_that("tf_ggplot + geom_line creates correct plot", {
  skip_if_not_installed("ggplot2")

  # Create test data
  set.seed(123)
  data <- data.frame(
    id = 1:3,
    group = factor(c("A", "A", "B"))
  )
  data$func <- tf_rgp(3, arg = seq(0, 1, length.out = 11))

  # Test basic line plot
  p <- tf_ggplot(data, aes(tf = func))
  p_with_geom <- p + geom_line()

  # Should remain tf_ggplot until finalized
  expect_s3_class(p_with_geom, "ggplot")
  expect_true(inherits(p_with_geom, "tf_ggplot")) # Should stay tf_ggplot

  # Check that the plot has the right data structure
  built <- ggplot_build(p_with_geom)
  plot_data <- built$data[[1]]

  # Should have the right number of groups (one per function)
  expect_equal(length(unique(plot_data$group)), 3)

  # Should have x and y columns (transformed from tf)
  expect_true("x" %in% names(plot_data))
  expect_true("y" %in% names(plot_data))
})

test_that("tf_ggplot handles grouping aesthetics correctly", {
  skip_if_not_installed("ggplot2")

  # Create test data
  set.seed(123)
  data <- data.frame(
    id = 1:4,
    treatment = factor(c("A", "A", "B", "B")),
    subject = factor(1:4)
  )
  data$func <- tf_rgp(4, arg = seq(0, 1, length.out = 11))

  # Test with color grouping
  p <- tf_ggplot(data, aes(tf = func, color = treatment)) + geom_line()
  built <- ggplot_build(p)
  plot_data <- built$data[[1]]

  # Should have 4 groups (one per function) but 2 colors (treatments)
  expect_equal(length(unique(plot_data$group)), 4)
  expect_equal(length(unique(plot_data$colour)), 2)

  # Color should be consistent within each function
  color_by_group <- split(plot_data$colour, plot_data$group)
  expect_true(all(sapply(color_by_group, function(x) length(unique(x)) == 1)))
})

test_that("tf_ggplot + geom_ribbon works with tf_ymin/tf_ymax", {
  skip_if_not_installed("ggplot2")

  # Create test data with confidence bands
  set.seed(123)
  data <- data.frame(id = 1:2)
  data$mean_func <- tf_rgp(2, arg = seq(0, 1, length.out = 11))
  data$lower_func <- data$mean_func +
    abs(tf_rgp(2, arg = seq(0, 1, length.out = 11)))
  data$upper_func <- data$mean_func +
    abs(tf_rgp(2, arg = seq(0, 1, length.out = 11)))

  # Test ribbon plot
  p <- tf_ggplot(data) +
    geom_ribbon(
      aes(tf_ymin = lower_func, tf_ymax = upper_func),
      alpha = 0.3
    )

  expect_s3_class(p, "ggplot")
  built <- ggplot_build(p)
  plot_data <- built$data[[1]]

  # Should have ymin and ymax columns
  expect_true("ymin" %in% names(plot_data))
  expect_true("ymax" %in% names(plot_data))
})

test_that("tf_ggplot validates input correctly", {
  skip_if_not_installed("ggplot2")

  # Test with non-tf data
  data <- data.frame(
    id = 1:3,
    regular_col = rnorm(3)
  )

  # Should work fine if no tf aesthetics are used
  p1 <- tf_ggplot(data, aes(x = id, y = regular_col))
  expect_s3_class(p1, "tf_ggplot")

  # Should error if tf aesthetic references non-tf column
  expect_error(
    {
      p <- tf_ggplot(data, aes(tf = regular_col)) + geom_line()
      ggplot_build(p) # Trigger validation
    },
    "tf aesthetic.*must evaluate to a tf object"
  )

  # Should error if tf aesthetic references non-existent column
  expect_error(
    {
      p <- tf_ggplot(data, aes(tf = nonexistent_col)) + geom_line()
      ggplot_build(p) # Trigger validation
    },
    "object.*not found|not found"
  )
})

test_that("tf_ggplot handles irregular tf objects", {
  skip_if_not_installed("ggplot2")

  # Create irregular tf data
  set.seed(123)
  data <- data.frame(id = 1:2)
  # Create irregular tf by jiggling and sparsifying
  regular_tf <- tf_rgp(2, arg = seq(0, 1, length.out = 21))
  data$irreg_func <- tf_sparsify(regular_tf, dropout = 0.7) # Make sparse/irregular

  # Should still work
  p <- tf_ggplot(data, aes(tf = irreg_func)) + geom_line()
  expect_s3_class(p, "ggplot")

  built <- ggplot_build(p)
  plot_data <- built$data[[1]]

  # Should have data for both functions
  expect_equal(length(unique(plot_data$group)), 2)
  expect_true(nrow(plot_data) > 0)
})

test_that("tf_ggplot preserves faceting variables", {
  skip_if_not_installed("ggplot2")

  # Create test data with faceting variable
  set.seed(123)
  data <- data.frame(
    id = 1:6,
    treatment = factor(rep(c("A", "B"), each = 3)),
    subject = factor(1:6)
  )
  data$func <- tf_rgp(6, arg = seq(0, 1, length.out = 11))

  # Test with faceting
  p <- tf_ggplot(data, aes(tf = func)) +
    geom_line() +
    facet_wrap(~treatment)

  expect_s3_class(p, "ggplot")
  built <- ggplot_build(p)

  # Should have panels for each treatment level
  expect_equal(length(unique(built$layout$layout$PANEL)), 2)

  # Check that plot data has correct structure
  plot_data <- built$data[[1]]
  expect_equal(length(unique(plot_data$group)), 6) # 6 functions
  expect_true(nrow(plot_data) > 0) # Has data
})

test_that("tf_ggplot handles custom arg specification", {
  skip_if_not_installed("ggplot2")

  # Create test data
  set.seed(123)
  data <- data.frame(id = 1:2)
  data$func <- tf_rgp(2, arg = seq(0, 1, length.out = 21))

  # Test with custom arg
  custom_arg <- seq(0, 1, length.out = 11) # Coarser grid
  p <- tf_ggplot(data, aes(tf = func), arg = custom_arg) + geom_line()

  built <- ggplot_build(p)
  plot_data <- built$data[[1]]

  # Should use the custom arg values
  unique_x <- sort(unique(plot_data$x))
  expect_equal(unique_x, custom_arg)
  expect_equal(length(unique_x), 11) # Not the original 21
})

test_that("tf_ggplot error handling for scale conflicts", {
  skip_if_not_installed("ggplot2")

  # Create mixed data
  set.seed(123)
  data <- data.frame(
    id = 1:3,
    scalar_y = rnorm(3)
  )
  data$func <- tf_rgp(3, arg = seq(0, 1, length.out = 11))

  expect_warning(
    {
      p <- tf_ggplot(data) +
        geom_line(aes(tf = func)) + # tf data on y-scale
        geom_point(aes(x = id, y = scalar_y)) # scalar data on y-scale
    },
    "scale conflict"
  )

  # This should work without error; scale conflict warnings are acceptable
  expect_no_error({
    p <- suppressWarnings(
      tf_ggplot(data) +
        geom_line(aes(tf = func)) + # tf data on y-scale
        geom_point(aes(x = id, y = scalar_y))
    ) # scalar data on y-scale

    # Should build successfully
    built <- ggplot_build(p)
    expect_s3_class(built, "ggplot_built")
  })
})

test_that("tf_ggplot handles large datasets", {
  skip_if_not_installed("ggplot2")

  # Large dataset should build without error
  set.seed(123)
  data <- data.frame(id = 1:201)
  data$func <- tf_rgp(201, arg = seq(0, 1, length.out = 101))

  expect_no_error({
    captured <- capture_warnings_silently({
      p <- tf_ggplot(data, aes(tf = func)) + geom_line()
      ggplot_build(p)
    })
    built <- captured$value

    expect_s3_class(built, "ggplot_built")
    expect_equal(length(unique(built$data[[1]]$group)), 201)
  })
})

#' @description
#' `tidyfun` makes data wrangling and exploratory analysis for functional data
#' in `R` easier.\cr\cr
#' The `tf` package defines the core functional data classes and methods.
#' `tidyfun` builds on those objects with:
#'
#'   - data-frame workflows for functional data: [tf_gather()] etc.
#'   - graphics helpers for `tf` objects and tidy functional data frames: [geom_spaghetti()], [gglasagna()]
#'   - documentation and examples that show how to use `tf` objects in tidy workflows
#' \cr\cr
#' \cr
#' Start with `vignette(package = "tidyfun")` for the article index.
#'
#' @keywords internal
#' @import rlang
"_PACKAGE"

#' Functional boxplots for `tf` objects
#'
#' Draw functional boxplots based on functional depth rankings.
#'
#' `stat_fboxplot()` computes a median curve (thick line), a central region
#' envelope (filled ribbon), functional fences (dashed lines), and optional
#' outlying curves (defined as "exceeds the fences somewhere", plotted as solid lines)
#' from a `tf` column. `geom_fboxplot()` draws these summaries as a band plus
#' line layers.
#'
#' The interface intentionally follows [ggplot2::stat_boxplot()] and
#' [ggplot2::geom_boxplot()] where this is meaningful for functional data.
#' Use `aes(tf = f)` to map a `tf` column. Separate functional boxplots inside a
#' panel are defined by `group`, `colour`/`color`, or `fill`; separate panels are
#' handled through facetting. Note that only `color` or `fill` need to be
#' specified explicitly, as the other will be automatically mapped to the same
#' variable if not provided.
#'
#' @param mapping,data,position,show.legend,inherit.aes,na.rm,orientation,...
#'   Passed on to [ggplot2::layer()].
#' @param stat,geom Use the functional boxplot stat/geom.
#' @param coef Inflation factor for the central envelope used to define outer fences, defaults to 1.5.
#' @param depth Character scalar naming the built-in depth measure passed to
#'   [tf::tf_depth()] when `depth_fn` is `NULL`.
#' @param depth_fn Optional custom depth function. Must return one numeric depth
#'   value per function.
#' @param fence_fn Optional custom fence function. Must return a list with
#'   elements `lower`, `upper`, and `outliers`.
#' @param central Fraction of deepest curves used to construct the central
#'   envelope. Defaults to `0.5`.
#' @param arg Optional evaluation grid used for depth calculation, envelopes, and
#'   drawing. Defaults to the natural grid of the mapped `tf`.
#' @param outliers Should outlying curves be drawn?
#' @param outlier.colour,outlier.color,outlier.fill,outlier.shape,outlier.size
#'   Styling parameters for outlier curves.
#' @param outlier.stroke Accepted for interface compatibility with
#'   [ggplot2::geom_boxplot()] but ignored because functional outliers are drawn
#'   as curves.
#' @param outlier.alpha Alpha used for outlier curves.
#' @param whisker.colour,whisker.color,whisker.linetype,whisker.linewidth Styling
#'   parameters for fence lines.
#' @param staple.colour,staple.color,staple.linetype,staple.linewidth Accepted for
#'   interface compatibility with [ggplot2::geom_boxplot()] but currently unused.
#' @param median.colour,median.color,median.linetype,median.linewidth Styling
#'   parameters for the median curve.
#' @param box.colour,box.color,box.linetype,box.linewidth Styling parameters for
#'   the central band outline.
#' @param notch,notchwidth,staplewidth,varwidth Accepted for interface
#'   compatibility with [ggplot2::geom_boxplot()] but currently unused.
#' @returns A ggplot2 layer.
#' @examples
#' \donttest{
#' library(ggplot2)
#' set.seed(1312)
#' data <- data.frame(id = 1:50, grp = rep(c("A", "B"), each = 25))
#' data$f <- tf_rgp(50) + 5 * (data$grp == "A")
#'
#' tf_ggplot(data, aes(tf = f, fill = grp)) + # same as `colour = grp` here!
#'   geom_fboxplot(alpha = 0.3)
#'
#' tf_ggplot(data, aes(tf = f)) +
#'   geom_fboxplot(orientation = "y")
#' }
#' @name ggfboxplot
NULL

normalize_fboxplot_mapping <- function(mapping) {
  mapping <- mapping %||% aes()
  use_group_aes <- "group" %in% names(mapping)

  xy_aes <- intersect(c("x", "y"), names(mapping))
  if (length(xy_aes) > 0) {
    if (any(c("colour", "color", "fill") %in% names(mapping))) {
      cli::cli_abort(c(
        "{.fn geom_fboxplot}/{.fn stat_fboxplot} do not use {.code x}/{.code y} for grouping",
        "i" = "Use {.code aes(tf = f, colour = grp, fill = grp)} and/or facets instead."
      ))
    }

    group_expr <- mapping[[xy_aes[1]]]
    mapping$group <- group_expr
    mapping$fill <- group_expr
    mapping$colour <- group_expr
    mapping[xy_aes] <- NULL
    use_group_aes <- TRUE

    cli::cli_warn(c(
      "{.fn geom_fboxplot}/{.fn stat_fboxplot} ignore {.code x}/{.code y} grouping semantics",
      "i" = "Mapped {.code {xy_aes[1]}} to {.code group}, {.code fill}, and {.code colour} instead."
    ))
  }

  list(mapping = mapping, use_group_aes = use_group_aes)
}

resolve_fboxplot_orientation <- function(orientation) {
  if (is.na(orientation) || is.null(orientation)) {
    return("x")
  }

  orientation <- match.arg(orientation, c("x", "y"))
  orientation
}

build_fboxplot_groups <- function(data, use_group_aes = FALSE) {
  keys <- list()

  if (isTRUE(use_group_aes) && "group" %in% names(data)) {
    keys[["group"]] <- data$group
  }
  if ("colour" %in% names(data)) {
    keys[["colour"]] <- data$colour
  }
  if ("fill" %in% names(data)) {
    keys[["fill"]] <- data$fill
  }

  keys <- Filter(\(x) !all(is.na(x)), keys)
  if (length(keys) == 0) {
    return(factor(rep("all", nrow(data))))
  }

  do.call(interaction, c(keys, list(drop = TRUE, lex.order = TRUE)))
}

compute_fboxplot_depth <- function(x, arg, depth, depth_fn = NULL) {
  depth_values <- if (is.null(depth_fn)) {
    tf::tf_depth(x, arg = arg, depth = depth)
  } else {
    depth_fn(x, arg = arg)
  }

  if (!is.numeric(depth_values) || length(depth_values) != length(x)) {
    cli::cli_abort(c(
      "{.arg depth_fn} must return one numeric value per function",
      "i" = "Expected {length(x)} values, got {length(depth_values)}."
    ))
  }

  depth_values
}

default_fboxplot_arg <- function(tf_vec) {
  tf_arg0 <- tf_arg(tf_vec)
  if (!is.list(tf_arg0)) {
    return(tf_arg0)
  }

  arg_starts <- vapply(tf_arg0, min, numeric(1))
  arg_ends <- vapply(tf_arg0, max, numeric(1))
  common_lower <- max(arg_starts)
  common_upper <- min(arg_ends)

  if (common_lower > common_upper) {
    cli::cli_abort(c(
      "Cannot infer a default evaluation grid for irregular {.cls tf} data with no common support.",
      "i" = "Provide {.arg arg} explicitly."
    ))
  }

  common_arg <- sort(unique(unlist(tf_arg0)))
  common_arg <- common_arg[
    common_arg >= common_lower & common_arg <= common_upper
  ]

  if (length(common_arg) == 0) {
    if (identical(common_lower, common_upper)) {
      return(common_lower)
    }

    cli::cli_abort(c(
      "Cannot infer a default evaluation grid for irregular {.cls tf} data.",
      "i" = "Provide {.arg arg} explicitly."
    ))
  }

  common_arg
}

prepare_fboxplot_curves <- function(tf_vec, data, arg) {
  curve_mat <- as.matrix(tf_vec, arg = arg, interpolate = TRUE)
  keep <- stats::complete.cases(curve_mat)

  if (!all(keep)) {
    cli::cli_warn(c(
      "Dropped {sum(!keep)} function{?s} from {.fn stat_fboxplot} because they cannot be evaluated on the selected grid.",
      "i" = "Provide {.arg arg} explicitly to control the evaluation support."
    ))
  }

  if (!any(keep)) {
    return(NULL)
  }

  list(
    tf_vec = tf_vec[keep],
    data = data[keep, , drop = FALSE],
    curve_mat = curve_mat[keep, , drop = FALSE]
  )
}

as_fboxplot_curve <- function(x, arg, what) {
  if (is_tf(x)) {
    curve_mat <- as.matrix(x, arg = arg, interpolate = TRUE)
    if (nrow(curve_mat) != 1) {
      cli::cli_abort("{.arg {what}} must evaluate to exactly one function.")
    }
    return(as.numeric(curve_mat[1, ]))
  }

  if (is.numeric(x) && length(x) == length(arg)) {
    return(as.numeric(x))
  }

  cli::cli_abort(c(
    "{.arg {what}} must be a numeric vector with one value per evaluation point or a length-1 {.cls tf}.",
    "i" = "Expected length {length(arg)}, got {.obj_type_friendly {x}}."
  ))
}

default_fboxplot_fence <- function(
  tf_vec,
  arg,
  depth,
  median,
  central_lower,
  central_upper,
  coef,
  central
) {
  curve_mat <- as.matrix(tf_vec, arg = arg, interpolate = TRUE)
  spread <- central_upper - central_lower
  lower <- central_lower - coef * spread
  upper <- central_upper + coef * spread
  outliers <- apply(
    curve_mat < rep(lower, each = nrow(curve_mat)) |
      curve_mat > rep(upper, each = nrow(curve_mat)),
    1,
    any,
    na.rm = TRUE
  )

  list(lower = lower, upper = upper, outliers = outliers)
}

compute_fboxplot_fence <- function(
  tf_vec,
  arg,
  depth,
  median,
  central_lower,
  central_upper,
  coef,
  central,
  fence_fn = NULL
) {
  fence <- if (is.null(fence_fn)) {
    default_fboxplot_fence(
      tf_vec = tf_vec,
      arg = arg,
      depth = depth,
      median = median,
      central_lower = central_lower,
      central_upper = central_upper,
      coef = coef,
      central = central
    )
  } else {
    fence_fn(
      tf_vec = tf_vec,
      arg = arg,
      depth = depth,
      median = median,
      central_lower = central_lower,
      central_upper = central_upper,
      coef = coef,
      central = central
    )
  }

  if (
    !is.list(fence) || !all(c("lower", "upper", "outliers") %in% names(fence))
  ) {
    cli::cli_abort(
      "{.arg fence_fn} must return a list with components {.code lower}, {.code upper}, and {.code outliers}."
    )
  }

  lower <- as_fboxplot_curve(fence$lower, arg = arg, what = "lower")
  upper <- as_fboxplot_curve(fence$upper, arg = arg, what = "upper")
  outliers <- fence$outliers

  if (!is.logical(outliers) || length(outliers) != length(tf_vec)) {
    cli::cli_abort(c(
      "{.arg fence_fn} must return one logical outlier flag per function",
      "i" = "Expected {length(tf_vec)} values, got {length(outliers)}."
    ))
  }

  list(lower = lower, upper = upper, outliers = outliers)
}

curve_df <- function(x, y, component, stat_group, path_group, orientation) {
  if (orientation == "x") {
    data.frame(
      x = x,
      y = y,
      .fbox_component = component,
      .fbox_stat_group = stat_group,
      group = path_group
    )
  } else {
    data.frame(
      x = y,
      y = x,
      .fbox_component = component,
      .fbox_stat_group = stat_group,
      group = path_group
    )
  }
}

band_df <- function(x, lower, upper, stat_group, orientation) {
  if (orientation == "x") {
    data.frame(
      x = x,
      ymin = lower,
      ymax = upper,
      .fbox_component = "box",
      .fbox_stat_group = stat_group,
      group = stat_group
    )
  } else {
    data.frame(
      y = x,
      xmin = lower,
      xmax = upper,
      .fbox_component = "box",
      .fbox_stat_group = stat_group,
      group = stat_group
    )
  }
}

make_box_polygon <- function(data) {
  if ("ymin" %in% names(data)) {
    data <- data[order(data$x), ]
    data.frame(
      x = c(data$x, rev(data$x)),
      y = c(data$ymin, rev(data$ymax)),
      group = data$group[1]
    )
  } else {
    data <- data[order(data$y), ]
    data.frame(
      x = c(data$xmin, rev(data$xmax)),
      y = c(data$y, rev(data$y)),
      group = data$group[1]
    )
  }
}

prepare_component_defaults <- function(
  data,
  colour = NULL,
  fill = NULL,
  linewidth = NULL,
  linetype = NULL,
  alpha = NULL
) {
  if (!is.null(colour)) data$colour <- colour
  if (!is.null(fill)) data$fill <- fill
  if (!is.null(linewidth)) data$linewidth <- linewidth
  if (!is.null(linetype)) data$linetype <- linetype
  if (!is.null(alpha)) data$alpha <- alpha
  data
}

default_fboxplot_colour <- function(data) {
  colour_vals <- unique(stats::na.omit(data$colour))
  fill_vals <- unique(stats::na.omit(data$fill))
  default_colour <- as.character(GeomFboxplot$default_aes$colour)

  if (
    length(colour_vals) > 0 &&
      (colour_vals[1] != default_colour || length(fill_vals) == 0)
  ) {
    return(colour_vals[1])
  }
  if (length(fill_vals) > 0) {
    return(fill_vals[1])
  }
  if (length(colour_vals) > 0) {
    return(colour_vals[1])
  }

  default_colour
}

default_fboxplot_fill <- function(data, fallback_colour = NULL) {
  default_fill <- as.character(GeomFboxplot$default_aes$fill)
  fill_vals <- unique(stats::na.omit(data$fill))

  if (length(fill_vals) > 0 && any(fill_vals != default_fill)) {
    return(fill_vals[1])
  }

  fallback_colour %||% as.character(GeomFboxplot$default_aes$fill)
}

has_user_fboxplot_colour <- function(data) {
  default_colour <- as.character(GeomFboxplot$default_aes$colour)
  default_fill <- as.character(GeomFboxplot$default_aes$fill)
  colour_vals <- unique(stats::na.omit(data$colour))
  fill_vals <- unique(stats::na.omit(data$fill))

  any(colour_vals != default_colour) || any(fill_vals != default_fill)
}

canonicalise_fboxplot_colour_args <- function(args) {
  for (prefix in c("outlier", "whisker", "staple", "median", "box")) {
    color_name <- paste0(prefix, ".color")
    colour_name <- paste0(prefix, ".colour")
    colour_value <- args[[color_name]] %||% args[[colour_name]]
    args[[color_name]] <- NULL
    args[[colour_name]] <- colour_value
  }
  args
}

get_fboxplot_scale_transform <- function(scales, orientation) {
  scale_obj <- if (orientation == "x") scales$y else scales$x
  if (is.null(scale_obj)) {
    return(NULL)
  }

  trans <- scale_obj$trans %||% scale_obj$scale$trans
  transform <- trans$transform %||% trans$transform_df

  if (is.null(transform)) {
    return(NULL)
  }

  transform
}

apply_fboxplot_scale_transform <- function(data, scales, orientation) {
  transform <- get_fboxplot_scale_transform(scales, orientation)
  if (is.null(transform)) {
    return(data)
  }

  cols <- if (orientation == "x") {
    intersect(c("y", "ymin", "ymax"), names(data))
  } else {
    intersect(c("x", "xmin", "xmax"), names(data))
  }

  for (col in cols) {
    data[[col]] <- transform(data[[col]])
  }

  data
}

warn_unsupported_fboxplot_params <- function(
  staplewidth,
  staple.colour,
  staple.color,
  staple.linetype,
  staple.linewidth,
  notch,
  notchwidth,
  varwidth,
  outlier.shape,
  outlier.stroke
) {
  if (
    !identical(staplewidth, 0) ||
      !is.null(staple.colour) ||
      !is.null(staple.color) ||
      !is.null(staple.linetype) ||
      !is.null(staple.linewidth)
  ) {
    cli::cli_warn(
      "{.arg staple*} parameters are accepted for compatibility but currently ignored by {.fn geom_fboxplot}."
    )
  }
  if (!identical(notch, FALSE) || !identical(notchwidth, 0.5)) {
    cli::cli_warn(
      "{.arg notch} and {.arg notchwidth} are accepted for compatibility but currently ignored by {.fn geom_fboxplot}."
    )
  }
  if (!identical(varwidth, FALSE)) {
    cli::cli_warn(
      "{.arg varwidth} is accepted for compatibility but currently ignored by {.fn geom_fboxplot}."
    )
  }
  if (!is.null(outlier.shape) || !identical(outlier.stroke, 0.5)) {
    cli::cli_warn(
      "{.arg outlier.shape} and {.arg outlier.stroke} are accepted for compatibility but ignored because functional outliers are drawn as curves."
    )
  }
}

#' @export
#' @rdname ggfboxplot
#' @usage NULL
#' @format NULL
StatFboxplot <- ggplot2::ggproto(
  "StatFboxplot",
  ggplot2::Stat,
  required_aes = c("tf"),
  retransform = FALSE,
  setup_params = function(data, params) {
    params$orientation <- resolve_fboxplot_orientation(params$orientation)
    params
  },
  compute_panel = function(
    data,
    scales,
    coef = 1.5,
    na.rm = FALSE,
    orientation = "x",
    depth = "MBD",
    depth_fn = NULL,
    fence_fn = NULL,
    central = 0.5,
    arg = NULL,
    use_group_aes = FALSE
  ) {
    if (!"tf" %in% names(data) || !is_tf(data$tf)) {
      cli::cli_abort(
        "{.fn stat_fboxplot} requires a {.code tf} aesthetic mapping to a {.cls tf} object."
      )
    }

    if (
      !is.numeric(central) ||
        length(central) != 1 ||
        is.na(central) ||
        central <= 0 ||
        central > 1
    ) {
      cli::cli_abort(
        "{.arg central} must be a single number in {.code (0, 1]}."
      )
    }

    tf_vec <- data$tf
    if (isTRUE(na.rm)) {
      keep <- !is.na(tf_vec)
      data <- data[keep, , drop = FALSE]
      tf_vec <- tf_vec[keep]
    }

    if (nrow(data) == 0) {
      return(data.frame())
    }

    arg <- arg %||% default_fboxplot_arg(tf_vec)

    stat_groups <- build_fboxplot_groups(data, use_group_aes = use_group_aes)
    pieces <- vector("list", length(levels(stat_groups)))

    for (idx in seq_along(levels(stat_groups))) {
      level <- levels(stat_groups)[idx]
      take <- stat_groups == level
      group_data <- data[take, , drop = FALSE]
      group_tf <- group_data$tf

      prepared <- prepare_fboxplot_curves(
        tf_vec = group_tf,
        data = group_data,
        arg = arg
      )
      if (is.null(prepared)) {
        next
      }

      group_tf <- prepared$tf_vec
      group_data <- prepared$data
      curve_mat <- prepared$curve_mat

      depth_values <- compute_fboxplot_depth(
        x = group_tf,
        arg = arg,
        depth = depth,
        depth_fn = depth_fn
      )

      rank_idx <- order(depth_values, decreasing = TRUE, na.last = NA)
      if (length(rank_idx) == 0) {
        next
      }

      n_central <- max(1L, ceiling(length(rank_idx) * central))
      central_idx <- rank_idx[seq_len(n_central)]
      median_idx <- rank_idx[1]

      central_mat <- curve_mat[central_idx, , drop = FALSE]
      central_lower <- apply(central_mat, 2, min, na.rm = TRUE)
      central_upper <- apply(central_mat, 2, max, na.rm = TRUE)
      median_curve <- as.numeric(curve_mat[median_idx, ])

      fence <- compute_fboxplot_fence(
        tf_vec = group_tf,
        arg = arg,
        depth = depth_values,
        median = median_curve,
        central_lower = central_lower,
        central_upper = central_upper,
        coef = coef,
        central = central,
        fence_fn = fence_fn
      )

      exemplar <- group_data[rep(1, length(arg)), , drop = FALSE]
      exemplar$tf <- NULL
      exemplar$group <- NULL

      next_group <- idx * 100000L
      box_piece <- cbind(
        band_df(
          x = arg,
          lower = central_lower,
          upper = central_upper,
          stat_group = idx,
          orientation = orientation
        ),
        exemplar
      )
      median_piece <- cbind(
        curve_df(
          x = arg,
          y = median_curve,
          component = "median",
          stat_group = idx,
          path_group = next_group + 1L,
          orientation = orientation
        ),
        exemplar
      )
      whisker_lower_piece <- cbind(
        curve_df(
          x = arg,
          y = fence$lower,
          component = "whisker_lower",
          stat_group = idx,
          path_group = next_group + 2L,
          orientation = orientation
        ),
        exemplar
      )
      whisker_upper_piece <- cbind(
        curve_df(
          x = arg,
          y = fence$upper,
          component = "whisker_upper",
          stat_group = idx,
          path_group = next_group + 3L,
          orientation = orientation
        ),
        exemplar
      )

      outlier_idx <- which(fence$outliers)
      outlier_piece <- if (length(outlier_idx) > 0) {
        outlier_frames <- vector("list", length(outlier_idx))
        for (j in seq_along(outlier_idx)) {
          row_id <- outlier_idx[j]
          curve_data <- cbind(
            curve_df(
              x = arg,
              y = as.numeric(curve_mat[row_id, ]),
              component = "outlier",
              stat_group = idx,
              path_group = next_group + 3L + j,
              orientation = orientation
            ),
            exemplar
          )
          outlier_frames[[j]] <- curve_data
        }
        dplyr::bind_rows(outlier_frames)
      } else {
        NULL
      }

      pieces[[idx]] <- dplyr::bind_rows(
        box_piece,
        median_piece,
        whisker_lower_piece,
        whisker_upper_piece,
        outlier_piece
      )
    }

    pieces <- dplyr::bind_rows(pieces)
    apply_fboxplot_scale_transform(
      pieces,
      scales = scales,
      orientation = orientation
    )
  }
)

#' @export
#' @rdname ggfboxplot
stat_fboxplot <- function(
  mapping = NULL,
  data = NULL,
  geom = "fboxplot",
  position = "identity",
  ...,
  orientation = NA,
  coef = 1.5,
  na.rm = FALSE,
  show.legend = NA,
  inherit.aes = TRUE,
  depth = "MBD",
  depth_fn = NULL,
  fence_fn = NULL,
  central = 0.5,
  arg = NULL
) {
  normalized <- normalize_fboxplot_mapping(mapping)

  ggplot2::layer(
    stat = StatFboxplot,
    data = data,
    mapping = normalized$mapping,
    geom = geom,
    position = position,
    show.legend = show.legend,
    inherit.aes = inherit.aes,
    params = list(
      orientation = orientation,
      coef = coef,
      na.rm = na.rm,
      depth = depth,
      depth_fn = depth_fn,
      fence_fn = fence_fn,
      central = central,
      arg = arg,
      use_group_aes = normalized$use_group_aes,
      ...
    )
  )
}

#' @export
#' @rdname ggfboxplot
#' @usage NULL
#' @format NULL
GeomFboxplot <- ggplot2::ggproto(
  "GeomFboxplot",
  ggplot2::Geom,
  required_aes = character(0),
  optional_aes = c("tf", "tf_x", "tf_y", "tf_ymin", "tf_ymax"),
  default_aes = ggplot2::aes(
    colour = "#333333",
    fill = "grey80",
    linewidth = 0.5,
    linetype = 1,
    alpha = 0.35
  ),
  draw_panel = function(
    self,
    data,
    panel_params,
    coord,
    outliers = TRUE,
    outlier.colour = NULL,
    outlier.color = NULL,
    outlier.fill = NULL,
    outlier.shape = NULL,
    outlier.size = NULL,
    outlier.stroke = 0.5,
    outlier.alpha = NULL,
    whisker.colour = NULL,
    whisker.color = NULL,
    whisker.linetype = NULL,
    whisker.linewidth = NULL,
    staple.colour = NULL,
    staple.color = NULL,
    staple.linetype = NULL,
    staple.linewidth = NULL,
    median.colour = NULL,
    median.color = NULL,
    median.linetype = NULL,
    median.linewidth = NULL,
    box.colour = NULL,
    box.color = NULL,
    box.linetype = NULL,
    box.linewidth = NULL,
    notch = FALSE,
    notchwidth = 0.5,
    staplewidth = 0,
    varwidth = FALSE
  ) {
    warn_unsupported_fboxplot_params(
      staplewidth = staplewidth,
      staple.colour = staple.colour,
      staple.color = staple.color,
      staple.linetype = staple.linetype,
      staple.linewidth = staple.linewidth,
      notch = notch,
      notchwidth = notchwidth,
      varwidth = varwidth,
      outlier.shape = outlier.shape,
      outlier.stroke = outlier.stroke
    )

    grobs <- list()
    stat_groups <- unique(data$.fbox_stat_group)

    for (stat_group in stat_groups) {
      stat_data <- data[data$.fbox_stat_group == stat_group, , drop = FALSE]
      group_colour <- default_fboxplot_colour(stat_data)
      group_fill <- default_fboxplot_fill(
        stat_data,
        fallback_colour = group_colour
      )
      whisker_colour <- whisker.color %||% whisker.colour %||% group_colour
      median_colour <- median.color %||%
        median.colour %||%
        if (has_user_fboxplot_colour(stat_data)) group_colour else "black"
      outlier_colour <- outlier.color %||% outlier.colour %||% group_colour
      outlier_linewidth <- outlier.size %||% stat_data$linewidth[1]

      box_data <- stat_data[stat_data$.fbox_component == "box", , drop = FALSE]
      if (nrow(box_data) > 0) {
        box_poly <- make_box_polygon(box_data)
        box_poly <- cbind(
          box_poly,
          data.frame(
            colour = NA,
            fill = default_fboxplot_fill(
              box_data,
              fallback_colour = group_colour
            ),
            linewidth = 0,
            linetype = 0,
            alpha = box_data$alpha[1]
          )
        )
        grobs[[length(grobs) + 1]] <- GeomPolygon$draw_panel(
          box_poly,
          panel_params,
          coord
        )
      }

      whisker_data <- stat_data[
        stat_data$.fbox_component %in% c("whisker_lower", "whisker_upper"),
        ,
        drop = FALSE
      ]
      if (nrow(whisker_data) > 0) {
        whisker_data <- prepare_component_defaults(
          whisker_data,
          colour = whisker_colour,
          linewidth = whisker.linewidth %||% whisker_data$linewidth[1],
          linetype = whisker.linetype %||% 2
        )
        grobs[[length(grobs) + 1]] <- GeomPath$draw_panel(
          whisker_data,
          panel_params,
          coord
        )
      }

      if (isTRUE(outliers)) {
        outlier_data <- stat_data[
          stat_data$.fbox_component == "outlier",
          ,
          drop = FALSE
        ]
        if (nrow(outlier_data) > 0) {
          outlier_data <- prepare_component_defaults(
            outlier_data,
            colour = outlier_colour,
            fill = outlier.fill,
            linewidth = outlier_linewidth,
            alpha = outlier.alpha %||% outlier_data$alpha[1]
          )
          grobs[[length(grobs) + 1]] <- GeomPath$draw_panel(
            outlier_data,
            panel_params,
            coord
          )
        }
      }

      median_data <- stat_data[
        stat_data$.fbox_component == "median",
        ,
        drop = FALSE
      ]
      if (nrow(median_data) > 0) {
        median_data <- prepare_component_defaults(
          median_data,
          colour = median_colour,
          linewidth = median.linewidth %||% (median_data$linewidth[1] * 1.75),
          linetype = median.linetype %||% median_data$linetype[1],
          alpha = 1
        )
        grobs[[length(grobs) + 1]] <- GeomPath$draw_panel(
          median_data,
          panel_params,
          coord
        )
      }
    }

    do.call(grid::grobTree, grobs)
  },
  draw_key = ggplot2::draw_key_polygon
)

#' @export
#' @rdname ggfboxplot
geom_fboxplot <- function(
  mapping = NULL,
  data = NULL,
  stat = "fboxplot",
  position = "identity",
  ...,
  outliers = TRUE,
  outlier.colour = NULL,
  outlier.color = NULL,
  outlier.fill = NULL,
  outlier.shape = NULL,
  outlier.size = NULL,
  outlier.stroke = 0.5,
  outlier.alpha = NULL,
  whisker.colour = NULL,
  whisker.color = NULL,
  whisker.linetype = NULL,
  whisker.linewidth = NULL,
  staple.colour = NULL,
  staple.color = NULL,
  staple.linetype = NULL,
  staple.linewidth = NULL,
  median.colour = NULL,
  median.color = NULL,
  median.linetype = NULL,
  median.linewidth = NULL,
  box.colour = NULL,
  box.color = NULL,
  box.linetype = NULL,
  box.linewidth = NULL,
  notch = FALSE,
  notchwidth = 0.5,
  staplewidth = 0,
  varwidth = FALSE,
  na.rm = FALSE,
  orientation = NA,
  show.legend = NA,
  inherit.aes = TRUE,
  depth = "MBD",
  depth_fn = NULL,
  fence_fn = NULL,
  central = 0.5,
  arg = NULL
) {
  normalized <- normalize_fboxplot_mapping(mapping)

  params <- canonicalise_fboxplot_colour_args(list(
    outliers = outliers,
    outlier.colour = outlier.colour,
    outlier.color = outlier.color,
    outlier.fill = outlier.fill,
    outlier.shape = outlier.shape,
    outlier.size = outlier.size,
    outlier.stroke = outlier.stroke,
    outlier.alpha = outlier.alpha,
    whisker.colour = whisker.colour,
    whisker.color = whisker.color,
    whisker.linetype = whisker.linetype,
    whisker.linewidth = whisker.linewidth,
    staple.colour = staple.colour,
    staple.color = staple.color,
    staple.linetype = staple.linetype,
    staple.linewidth = staple.linewidth,
    median.colour = median.colour,
    median.color = median.color,
    median.linetype = median.linetype,
    median.linewidth = median.linewidth,
    box.colour = box.colour,
    box.color = box.color,
    box.linetype = box.linetype,
    box.linewidth = box.linewidth,
    notch = notch,
    notchwidth = notchwidth,
    staplewidth = staplewidth,
    varwidth = varwidth,
    na.rm = na.rm,
    orientation = orientation,
    depth = depth,
    depth_fn = depth_fn,
    fence_fn = fence_fn,
    central = central,
    arg = arg,
    use_group_aes = normalized$use_group_aes
  ))

  ggplot2::layer(
    stat = stat,
    data = data,
    mapping = normalized$mapping,
    geom = GeomFboxplot,
    position = position,
    show.legend = show.legend,
    inherit.aes = inherit.aes,
    params = c(params, list(...))
  )
}

#' Autoplot and autolayer methods for `tf` objects
#'
#' Convenient plotting methods for `tf` objects. `autoplot()` creates a
#' complete spaghetti plot, `autolayer()` creates a layer that can be
#' added to an existing [ggplot2::ggplot()] or [tf_ggplot()].
#'
#' @param object a `tf` object
#' @param ... passed to [geom_line()]
#' @returns A [tf_ggplot()] object for `autoplot()`, a [ggplot2::layer()] object for `autolayer()`.
#' @examples
#' \donttest{
#' library(ggplot2)
#' f <- tf_rgp(5)
#' autoplot(f)
#' ggplot() + autolayer(f)
#' tf_ggplot() + autolayer(f)
#' }
#' @name autoplot.tf
#' @family tidyfun visualization
NULL

#' @export
#' @rdname autoplot.tf
autoplot.tf <- function(object, ...) {
  data <- tibble::tibble(tf = object)
  tf_ggplot(data, aes(tf = .data$tf)) + geom_line(...)
}

#' @export
#' @rdname autoplot.tf
autolayer.tf <- function(object, ...) {
  # geom_spaghetti works with plain ggplot(); when added to tf_ggplot() it is
  # automatically translated to geom_line(aes(tf = ...)) by +.tf_ggplot.
  data <- tibble::tibble(y = object)
  geom_spaghetti(
    mapping = aes(y = .data$y),
    data = data,
    ...,
    inherit.aes = FALSE
  )
}

#!/usr/bin/env Rscript

# Visual Test Runner for tf_ggplot
# This script helps run visual tests and organize the output plots

# Load required packages
library(testthat)
if (!require(ggplot2, quietly = TRUE)) {
  stop("ggplot2 is required for visual tests")
}

# Source the package functions (assuming we're in package root)
if (file.exists("R")) {
  source_files <- list.files("R", pattern = "\\.R$", full.names = TRUE)
  invisible(lapply(source_files, source))
} else {
  cat(
    "Warning: R/ directory not found. Make sure tf_ggplot functions are loaded.\n"
  )
}

# Create output directory structure
setup_visual_test_environment <- function() {
  base_dir <- file.path("tests", "visual-output")

  # Create subdirectories for organization
  dirs <- c(
    file.path(base_dir, "plots"), # Individual plot files
    file.path(base_dir, "comparisons"), # Side-by-side comparisons
    file.path(base_dir, "reports") # HTML reports
  )

  for (dir in dirs) {
    if (!dir.exists(dir)) {
      dir.create(dir, recursive = TRUE)
      cat("Created directory:", dir, "\n")
    }
  }

  return(base_dir)
}

# Enhanced plot saving with metadata
save_visual_test_plot <- function(
  plot,
  name,
  category = "general",
  width = 8,
  height = 6,
  dpi = 150,
  description = NULL
) {
  base_dir <- setup_visual_test_environment()

  # Save the plot
  filename <- paste0(
    sprintf(
      "%02d",
      match(
        category,
        c("basic", "advanced", "edge-cases", "comparisons", "general")
      )
    ),
    "_",
    name,
    ".png"
  )
  filepath <- file.path(base_dir, "plots", filename)

  tryCatch(
    {
      ggsave(filepath, plot, width = width, height = height, dpi = dpi)

      # Save metadata
      metadata <- list(
        filename = filename,
        category = category,
        description = description %||% name,
        created = Sys.time(),
        dimensions = c(width = width, height = height),
        dpi = dpi
      )

      metadata_file <- file.path(
        base_dir,
        "plots",
        paste0(tools::file_path_sans_ext(filename), "_meta.rds")
      )
      saveRDS(metadata, metadata_file)

      cat("✓ Saved:", filename, "\n")
      return(filepath)
    },
    error = function(e) {
      cat("✗ Failed to save", filename, ":", e$message, "\n")
      return(NULL)
    }
  )
}

# Create HTML report with all plots
create_visual_report <- function() {
  base_dir <- setup_visual_test_environment()
  plots_dir <- file.path(base_dir, "plots")

  # Get all plot files
  plot_files <- list.files(plots_dir, pattern = "\\.png$", full.names = FALSE)
  plot_files <- sort(plot_files)

  if (length(plot_files) == 0) {
    cat("No plots found to include in report.\n")
    return(NULL)
  }

  # Create HTML content
  html_content <- c(
    "<!DOCTYPE html>",
    "<html>",
    "<head>",
    "<title>tf_ggplot Visual Test Report</title>",
    "<style>",
    "body { font-family: Arial, sans-serif; margin: 40px; }",
    ".plot-container { margin: 20px 0; border: 1px solid #ddd; padding: 20px; }",
    ".plot-title { font-size: 18px; font-weight: bold; margin-bottom: 10px; }",
    ".plot-description { color: #666; margin-bottom: 15px; }",
    ".plot-image { max-width: 100%; height: auto; border: 1px solid #ccc; }",
    ".metadata { font-size: 12px; color: #888; margin-top: 10px; }",
    ".category-header { font-size: 24px; color: #333; margin: 30px 0 15px 0; border-bottom: 2px solid #333; }",
    "</style>",
    "</head>",
    "<body>",
    "<h1>tf_ggplot Visual Test Report</h1>",
    paste0("<p>Generated on: ", Sys.time(), "</p>"),
    paste0("<p>Total plots: ", length(plot_files), "</p>")
  )

  # Group plots by category
  categories <- list()
  for (file in plot_files) {
    # Extract category from filename prefix
    category_num <- substr(file, 1, 2)
    category_name <- switch(
      category_num,
      "01" = "Basic Functionality",
      "02" = "Basic Functionality",
      "03" = "Advanced Features",
      "04" = "Advanced Features",
      "05" = "Advanced Features",
      "06" = "Advanced Features",
      "07" = "Comparisons",
      "08" = "Comparisons",
      "09" = "Geom Types",
      "10" = "Geom Types",
      "11" = "Geom Types",
      "12" = "Geom Types",
      "13" = "Geom Types",
      "14" = "Edge Cases",
      "15" = "Edge Cases",
      "16" = "Edge Cases",
      "17" = "Edge Cases",
      "General"
    )

    if (is.null(categories[[category_name]])) {
      categories[[category_name]] <- character(0)
    }
    categories[[category_name]] <- c(categories[[category_name]], file)
  }

  # Add plots to HTML, grouped by category
  for (category in names(categories)) {
    html_content <- c(
      html_content,
      paste0('<div class="category-header">', category, '</div>')
    )

    for (file in categories[[category]]) {
      # Try to load metadata
      meta_file <- file.path(
        plots_dir,
        paste0(tools::file_path_sans_ext(file), "_meta.rds")
      )

      description <- "No description available"
      metadata_info <- ""

      if (file.exists(meta_file)) {
        tryCatch(
          {
            meta <- readRDS(meta_file)
            description <- meta$description
            metadata_info <- paste0(
              "Size: ",
              meta$dimensions["width"],
              "×",
              meta$dimensions["height"],
              " | DPI: ",
              meta$dpi,
              " | Created: ",
              format(meta$created, "%Y-%m-%d %H:%M")
            )
          },
          error = function(e) {
            # Ignore metadata errors
          }
        )
      }

      plot_html <- c(
        '<div class="plot-container">',
        paste0(
          '<div class="plot-title">',
          tools::file_path_sans_ext(file),
          '</div>'
        ),
        paste0('<div class="plot-description">', description, '</div>'),
        paste0(
          '<img src="plots/',
          file,
          '" class="plot-image" alt="',
          file,
          '">'
        ),
        paste0('<div class="metadata">', metadata_info, '</div>'),
        '</div>'
      )

      html_content <- c(html_content, plot_html)
    }
  }

  # Close HTML
  html_content <- c(html_content, "</body>", "</html>")

  # Write report
  report_file <- file.path(base_dir, "reports", "visual_test_report.html")
  writeLines(html_content, report_file)

  cat("✓ Visual report created:", report_file, "\n")
  cat("Open in browser to view all test plots.\n")

  return(report_file)
}

# Main function to run all visual tests
run_tf_ggplot_visual_tests <- function(
  interactive_display = FALSE,
  create_report = TRUE,
  open_report = interactive()
) {
  cat("=== tf_ggplot Visual Test Suite ===\n\n")

  # Setup environment
  base_dir <- setup_visual_test_environment()

  # Override the save function in the test environment
  assign(
    "save_test_plot",
    function(
      plot,
      filename,
      width = 8,
      height = 6,
      display = interactive_display,
      save_plot = TRUE
    ) {
      if (save_plot) {
        save_visual_test_plot(plot, filename, width = width, height = height)
      }
      if (display) {
        print(plot)
      }
      invisible(plot)
    },
    envir = .GlobalEnv
  )

  # Run the visual tests
  cat("Running visual tests...\n")
  tryCatch(
    {
      test_file(file.path("tests", "testthat", "test-tf-ggplot-visual.R"))
      cat("✓ All visual tests completed successfully!\n\n")
    },
    error = function(e) {
      cat("✗ Error in visual tests:", e$message, "\n\n")
      return(FALSE)
    }
  )

  # Create HTML report
  if (create_report) {
    cat("Creating visual report...\n")
    report_file <- create_visual_report()

    if (open_report && !is.null(report_file) && file.exists(report_file)) {
      if (.Platform$OS.type == "windows") {
        shell.exec(report_file)
      } else {
        system(paste("open", shQuote(report_file)), ignore.stderr = TRUE)
      }
    }
  }

  cat("\nVisual testing complete!\n")
  cat("Results in:", base_dir, "\n")

  return(TRUE)
}

# Quick comparison function
compare_with_current <- function() {
  cat("Creating comparison plots: tf_ggplot vs current geom_spaghetti...\n")

  # This would create side-by-side comparisons
  # Implementation depends on having both systems available

  cat("Comparison plots would be created here.\n")
}

# Print usage information
print_usage <- function() {
  cat("\n=== tf_ggplot Visual Test Runner ===\n\n")
  cat("Usage:\n")
  cat("  source('tests/visual-test-runner.R')\n")
  cat("  run_tf_ggplot_visual_tests()    # Run all tests, create report\n")
  cat(
    "  create_visual_report()          # Create HTML report from existing plots\n"
  )
  cat("  compare_with_current()          # Compare with geom_spaghetti\n\n")
  cat("Options:\n")
  cat("  interactive_display = TRUE      # Show plots as they're created\n")
  cat("  create_report = FALSE           # Skip HTML report generation\n")
  cat("  open_report = FALSE             # Don't auto-open the report\n\n")
  cat("Output directories:\n")
  cat("  tests/visual-output/plots/      # Individual plot files\n")
  cat("  tests/visual-output/reports/    # HTML reports\n")
  cat("  tests/visual-output/comparisons/ # Side-by-side comparisons\n\n")
}

# Show usage when sourced
if (interactive()) {
  print_usage()
}

# Quick Visual Check Script for tf_ggplot Development
# Run this to quickly test and view tf_ggplot functionality during development

# Load required libraries
library(ggplot2)

# Source tf_ggplot functions (adjust path as needed)
# source("R/tf-ggplot.R")  # Uncomment when implemented

# Quick test data generator
quick_test_data <- function(n_funcs = 3, pattern = "random") {
  set.seed(123)
  data <- data.frame(
    id = seq_len(n_funcs),
    group = factor(sample(c("A", "B"), n_funcs, replace = TRUE))
  )

  arg_vals <- seq(0, 2 * pi, length.out = 51)

  if (pattern == "random") {
    data$func <- tf_rgp(n_funcs, arg = arg_vals)
  } else if (pattern == "trig") {
    func_vals <- list(
      sin(arg_vals),
      cos(arg_vals),
      sin(2 * arg_vals)
    )[seq_len(n_funcs)]
    data$func <- tfd(func_vals, arg = arg_vals)
  } else if (pattern == "poly") {
    func_vals <- list(
      arg_vals,
      arg_vals^2,
      arg_vals^3
    )[seq_len(n_funcs)]
    data$func <- tfd(func_vals, arg = arg_vals)
  }

  return(data)
}

# Quick visual tests - run these individually during development
test_basic_line <- function() {
  cat("Testing: Basic line plot\n")
  data <- quick_test_data(3, "trig")

  # Test current approach
  p1 <- ggplot(data, aes(y = func, color = group)) +
    geom_spaghetti() +
    labs(title = "Current: geom_spaghetti") +
    theme_minimal()

  # Test new approach (will fail until implemented)
  tryCatch(
    {
      p2 <- tf_ggplot(data, aes(tf = func, color = group)) +
        geom_line() +
        labs(title = "New: tf_ggplot + geom_line") +
        theme_minimal()

      print(p1)
      Sys.sleep(2) # Brief pause between plots
      print(p2)

      cat("✓ Basic line plot test passed\n")
      return(list(current = p1, new = p2))
    },
    error = function(e) {
      cat("✗ tf_ggplot not yet implemented:", e$message, "\n")
      print(p1)
      cat("Showing current geom_spaghetti for reference\n")
      return(list(current = p1, new = NULL))
    }
  )
}

test_ribbon_plot <- function() {
  cat("Testing: Ribbon/confidence band plot\n")

  # Create confidence band data
  set.seed(456)
  data <- data.frame(method = c("A", "B"))
  arg_vals <- seq(0, 5, length.out = 26)

  mean_vals1 <- 2 * exp(-arg_vals / 2) * cos(arg_vals)
  mean_vals2 <- 1.5 * exp(-arg_vals / 3) * sin(arg_vals * 1.2)

  sd1 <- rep(0.3, 26)
  sd2 <- rep(0.25, 26)

  data$mean_func <- tfd(list(mean_vals1, mean_vals2), arg = arg_vals)
  data$lower_func <- tfd(
    list(mean_vals1 - 1.96 * sd1, mean_vals2 - 1.96 * sd2),
    arg = arg_vals
  )
  data$upper_func <- tfd(
    list(mean_vals1 + 1.96 * sd1, mean_vals2 + 1.96 * sd2),
    arg = arg_vals
  )

  tryCatch(
    {
      p <- tf_ggplot(data, aes(color = method, fill = method)) +
        geom_ribbon(
          aes(tf_ymin = lower_func, tf_ymax = upper_func),
          alpha = 0.3
        ) +
        geom_line(aes(tf = mean_func), size = 1) +
        labs(title = "tf_ggplot: Confidence Bands") +
        theme_minimal()

      print(p)
      cat("✓ Ribbon plot test passed\n")
      return(p)
    },
    error = function(e) {
      cat("✗ tf_ggplot ribbon test failed:", e$message, "\n")

      # Show what it should look like with manual transformation
      cat("This is what the ribbon plot should look like...\n")

      # Manual transformation for reference
      data_long <- tf_unnest(data, mean_func)
      data_bands <- tf_unnest(data, c(lower_func, upper_func))

      p_ref <- ggplot() +
        geom_ribbon(
          data = data_bands,
          aes(x = arg, ymin = lower_func, ymax = upper_func, fill = method),
          alpha = 0.3
        ) +
        geom_line(
          data = data_long,
          aes(x = arg, y = mean_func, color = method),
          size = 1
        ) +
        labs(title = "Reference: Manual transformation", x = "x", y = "f(x)") +
        theme_minimal()

      print(p_ref)
      return(p_ref)
    }
  )
}

test_faceting <- function() {
  cat("Testing: Faceted plot\n")

  data <- expand.grid(
    subject = 1:4,
    condition = c("Control", "Treatment")
  )
  data$group <- factor(rep(c("A", "B"), 4))

  # Create different patterns by condition
  set.seed(789)
  arg_vals <- seq(0, 10, length.out = 21)
  func_vals <- map(seq_len(nrow(data)), function(i) {
    base <- sin(arg_vals * 0.5) + rnorm(21, 0, 0.1)
    if (data$condition[i] == "Treatment") {
      base <- base * 1.5 + 0.5
    }
    base
  })

  data$func <- tfd(func_vals, arg = arg_vals)

  tryCatch(
    {
      p <- tf_ggplot(data, aes(tf = func, color = group)) +
        geom_line(alpha = 0.7) +
        facet_wrap(~condition) +
        labs(title = "tf_ggplot: Faceted Plot") +
        theme_minimal()

      print(p)
      cat("✓ Faceting test passed\n")
      return(p)
    },
    error = function(e) {
      cat("✗ tf_ggplot faceting test failed:", e$message, "\n")
      return(NULL)
    }
  )
}

test_multiple_aesthetics <- function() {
  cat("Testing: Multiple aesthetics\n")

  data <- expand.grid(
    method = c("A", "B", "C"),
    difficulty = c("Easy", "Hard")
  )
  data$performance <- runif(nrow(data))

  set.seed(101112)
  arg_vals <- seq(0, 10, length.out = 26)
  func_vals <- map(seq_len(nrow(data)), function(i) {
    base <- data$performance[i] * (1 - exp(-arg_vals / 3))
    if (data$difficulty[i] == "Hard") base <- base * 0.7
    base + rnorm(26, 0, 0.02)
  })

  data$learning_curve <- tfd(func_vals, arg = arg_vals)

  tryCatch(
    {
      p <- tf_ggplot(
        data,
        aes(
          tf = learning_curve,
          color = method,
          linetype = difficulty,
          alpha = performance
        )
      ) +
        geom_line(size = 1) +
        scale_alpha_continuous(range = c(0.4, 1.0)) +
        labs(title = "tf_ggplot: Multiple Aesthetics") +
        theme_minimal()

      print(p)
      cat("✓ Multiple aesthetics test passed\n")
      return(p)
    },
    error = function(e) {
      cat("✗ tf_ggplot multiple aesthetics test failed:", e$message, "\n")
      return(NULL)
    }
  )
}

# Interactive test runner
run_interactive_tests <- function() {
  cat("\n=== tf_ggplot Interactive Visual Tests ===\n\n")

  tests <- list(
    "1" = list(name = "Basic Line Plot", func = test_basic_line),
    "2" = list(name = "Ribbon/Confidence Bands", func = test_ribbon_plot),
    "3" = list(name = "Faceted Plot", func = test_faceting),
    "4" = list(name = "Multiple Aesthetics", func = test_multiple_aesthetics),
    "a" = list(name = "Run All Tests", func = function() {
      test_basic_line()
      test_ribbon_plot()
      test_faceting()
      test_multiple_aesthetics()
    })
  )

  repeat {
    cat("\nAvailable tests:\n")
    for (key in names(tests)) {
      cat(sprintf("  %s) %s\n", key, tests[[key]]$name))
    }
    cat("  q) Quit\n\n")

    choice <- readline("Choose a test (1-4, a, q): ")

    if (choice == "q") {
      cat("Goodbye!\n")
      break
    }

    if (choice %in% names(tests)) {
      cat("\n", rep("=", 50), "\n")
      result <- tests[[choice]]$func()
      cat(rep("=", 50), "\n")

      if (!is.null(result)) {
        cat("Press Enter to continue...")
        readline()
      }
    } else {
      cat("Invalid choice. Please try again.\n")
    }
  }
}

# Development helpers
check_tf_functions <- function() {
  cat("Checking for required tf functions...\n")

  required_funcs <- c("tf_rgp", "tfd", "tf_unnest", "tf_evaluations", "tf_arg")
  available <- sapply(required_funcs, exists)

  for (i in seq_along(required_funcs)) {
    status <- if (available[i]) "✓" else "✗"
    cat(sprintf("%s %s\n", status, required_funcs[i]))
  }

  if (all(available)) {
    cat("\n✓ All required tf functions are available.\n")
  } else {
    cat("\n✗ Some tf functions are missing. Load the tf/tidyfun packages.\n")
  }

  return(all(available))
}

check_ggplot_functions <- function() {
  cat("Checking for ggplot2 functions...\n")

  required_funcs <- c("ggplot", "aes", "geom_line", "geom_point", "geom_ribbon")
  available <- sapply(required_funcs, exists)

  for (i in seq_along(required_funcs)) {
    status <- if (available[i]) "✓" else "✗"
    cat(sprintf("%s %s\n", status, required_funcs[i]))
  }

  return(all(available))
}

# Print startup message
cat("\n=== tf_ggplot Quick Visual Check ===\n")
cat("Available functions:\n")
cat("  test_basic_line()         # Test basic line plots\n")
cat("  test_ribbon_plot()        # Test ribbon/confidence bands\n")
cat("  test_faceting()           # Test faceted plots\n")
cat("  test_multiple_aesthetics() # Test complex aesthetics\n")
cat("  run_interactive_tests()   # Interactive test menu\n")
cat("  check_tf_functions()      # Check tf package availability\n")
cat("  check_ggplot_functions()  # Check ggplot2 availability\n\n")

# Run basic checks
if (interactive()) {
  check_tf_functions()
  check_ggplot_functions()

  cat("\nReady for testing! Try run_interactive_tests() to start.\n")
}

# clean up globals for R CMD check
utils::globalVariables(c(
  ".",
  "..fill",
  "..label",
  "..order",
  "..order_by_value",
  "..row",
  "..x",
  "..y",
  "V1",
  "V2",
  "id_num",
  "tick_hi",
  "tick_lo",
  "tick_pos",
  "ticks",
  "value",
  "xmax",
  "id",
  ".row_id",
  ".row_id_"
))

#' Spaghetti plots for `tf` objects
#'
#' Plots a line for each entry of a `tf`-object.
#' `geom_spaghetti` draws a line through each function, and `geom_meatballs`
#' adds points for the evaluated grid values.
#'
#' "Flipped" aesthetics are not implemented for these geoms.
#'
#' @section `y` aesthetic:
#'   Mandatory. Used to designate a column of class `tf` to be visualized.
#' @examples
#' set.seed(1221)
#' data <- data.frame(col = sample(gl(5, 2)))
#' data$f <- tf_rgp(10)
#' data$fi <- tf_jiggle(data$f)
#' data$fb <- tfb(data$f)
#' library(ggplot2)
#' ggplot(data, aes(y = f, color = tf_depth(f))) +
#'   geom_spaghetti()
#' ggplot(data, aes(y = fi, shape = col, color = col)) +
#'   geom_meatballs()
#' ggplot(data, aes(y = fi)) +
#'   geom_meatballs(spaghetti = FALSE) +
#'   facet_wrap(~col)
#' @returns A [ggplot2::layer()] object for use in a ggplot.
#' @name ggspaghetti
#' @family tidyfun visualization
#' @import ggplot2
#' @seealso [geom_cappelini()] for glyph plots, [gglasagna()] for heatmaps.
NULL

#' @export
is.finite.tf <- function(x) {
  map_lgl(tf_evaluations(x), \(x) all(is.finite(x) | !is.na(x)))
}

#' @export
scale_type.tf <- function(x) "identity"

#' @export
#' @rdname ggspaghetti
#' @family tidyfun visualization
#' @usage NULL
#' @format NULL
StatTf <- ggproto(
  "StatTf",
  Stat,
  required_aes = "y",
  setup_params = function(data, params) {
    if (is.null(params$arg)) {
      params$arg <- list(tf_arg(pull(data, y)))
    }
    params
  },
  compute_layer = function(self, data, params, layout) {
    if (!is_tf(data$y)) {
      cli::cli_abort(
        "{.arg y} must be a {.cls tf} object, not {.obj_type_friendly {data$y}}."
      )
    }
    tf_eval <- suppressMessages(
      data |>
        mutate(y____id = names(y) %||% seq_along(y)) |>
        tf_unnest(y, arg = params$arg, names_sep = "____")
    ) |>
      select(-group) |>
      rename(group = y____id, x = y____arg, y = y____value)
    tf_eval
  },
  # need this so arg, spaghetti gets recognized as valid parameters
  # because layer() only checks compute_panel & compute_group
  compute_panel = function(self, data, scales, arg, spaghetti) {
    Stat$compute_panel(self, data, scales)
  }
)

#' @export
#' @rdname ggspaghetti
#' @family tidyfun visualization
#' @inheritParams ggplot2::stat_identity
#' @param na.rm remove NAs? defaults to `TRUE`
stat_tf <- function(
  mapping = NULL,
  data = NULL,
  geom = "spaghetti",
  position = "identity",
  na.rm = TRUE,
  show.legend = NA,
  inherit.aes = TRUE,
  arg = NULL,
  ...
) {
  ggplot2::layer(
    stat = StatTf,
    data = data,
    mapping = mapping,
    geom = geom,
    position = position,
    show.legend = show.legend,
    inherit.aes = inherit.aes,
    params = list(na.rm = na.rm, arg = arg, ...)
  )
}

# geom --------------------------------------------------------------------

#' @export
#' @rdname ggspaghetti
#' @family tidyfun visualization
#' @format NULL
#' @param arg where to evaluate the functions in `y`; defaults to the object's
#'   default evaluation grid.
geom_spaghetti <- function(
  mapping = NULL,
  data = NULL,
  position = "identity",
  na.rm = TRUE,
  show.legend = NA,
  inherit.aes = TRUE,
  arg = NULL,
  ...
) {
  ggplot2::layer(
    stat = StatTf,
    data = data,
    mapping = mapping,
    geom = "spaghetti",
    position = position,
    show.legend = show.legend,
    inherit.aes = inherit.aes,
    params = list(na.rm = na.rm, arg = arg, ...)
  )
}
#' @export
#' @rdname ggspaghetti
#' @family tidyfun visualization
#' @usage NULL
#' @format NULL
GeomSpaghetti <- ggplot2::ggproto(
  "GeomSpaghetti",
  ggplot2::Geom,
  default_aes = ggplot2::aes(
    colour = "black",
    linewidth = 0.5,
    linetype = 1,
    alpha = 0.5
  ),
  draw_group = function(data, panel_params, coord) {
    GeomLine$draw_panel(data, panel_params, coord)
  },
  draw_key = GeomLine$draw_key,
  required_aes = c("x", "y", "group"),
  setup_data = function(data, params) {
    GeomLine$setup_data(data, params)
  },
  setup_params = function(data, params) {
    # TODO: implement proper "orientation" - see extending ggplot vignette
    params$flipped_aes <- FALSE
    params
  }
)

#-------------------------------------------------------------------------------

#' @export
#' @rdname ggspaghetti
#' @family tidyfun visualization
#' @format NULL
#' @importFrom grid gList
#' @param spaghetti plot noodles along with meatballs? defaults to TRUE.
geom_meatballs <- function(
  mapping = NULL,
  data = NULL,
  position = "identity",
  na.rm = TRUE,
  show.legend = NA,
  inherit.aes = TRUE,
  arg = NULL,
  spaghetti = TRUE,
  ...
) {
  ggplot2::layer(
    stat = StatTf,
    data = data,
    mapping = mapping,
    geom = "meatballs",
    position = position,
    show.legend = show.legend,
    inherit.aes = inherit.aes,
    params = list(na.rm = na.rm, arg = arg, spaghetti = spaghetti, ...)
  )
}
#' @export
#' @rdname ggspaghetti
#' @family tidyfun visualization
#' @usage NULL
#' @format NULL
GeomMeatballs <- ggplot2::ggproto(
  "GeomMeatballs",
  ggplot2::Geom,
  default_aes = ggplot2::aes(
    colour = "black",
    linewidth = 0.5,
    size = 0.5,
    linetype = 1,
    alpha = 0.5,
    shape = 19,
    fill = NA,
    stroke = 0.5
  ),
  draw_group = function(data, panel_params, coord, spaghetti = TRUE) {
    grid::gList(
      if (spaghetti) GeomLine$draw_panel(data, panel_params, coord),
      GeomPoint$draw_panel(data, panel_params, coord)
    )
  },
  draw_key = function(data, params, size) {
    grid::grobTree(
      if (params$spaghetti) draw_key_path(data, params, size),
      draw_key_point(data, params, size)
    )
  },
  required_aes = c("x", "y", "group"),
  setup_params = function(data, params) {
    # TODO: implement proper "orientation" - see extending ggplot vignette
    params$flipped_aes <- FALSE
    params
  },
  setup_data = function(data, params) {
    GeomLine$setup_data(data, params)
  }
)

d <- tibble(a = tf_rgp(3), b = tf_rgp(3))

test_that("tf_evaluate.data.frame basically works", {
  da <- d$a
  expect_identical(
    tf_evaluate(da)[["a"]],
    da[["a"]][, tf_arg(d$a), matrix = FALSE]
  )
})

test_that("tf_evaluate.data.frame interface works", {
  expect_identical(tf_evaluate(d), tf_evaluate(d, a, b))
  expect_identical(tf_evaluate(d, a), tf_evaluate(d, -b))
  expect_identical(
    tf_evaluate(d, a, arg = seq(0, 1, length.out = 11))[["a"]],
    d[["a"]][, seq(0, 1, length.out = 11), matrix = FALSE]
  )
  expect_identical(
    tf_evaluate(d, b, arg = seq(0, 1, length.out = 11)),
    tf_evaluate(d, arg = seq(0, 1, length.out = 11), b)
  )
})

---
title: "tf Vectors and Operations"
author: "Jeff Goldsmith, Fabian Scheipl"
date: "`r Sys.Date()`"
output: 
    rmarkdown::html_vignette:
      fig_width: 7
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{tf Vectors and Operations}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "##",
  fig.width = 6,
  fig.height = 4,
  dpi = 72,
  fig.retina = 1,
  out.width = "90%"
)

library("tidyverse")
library("viridisLite")

theme_set(theme_minimal() + theme(legend.position = "bottom"))

options(
  ggplot2.continuous.colour = "viridis",
  ggplot2.continuous.fill = "viridis"
)
scale_colour_discrete <- scale_colour_viridis_d
scale_fill_discrete <- scale_fill_viridis_d

library("tidyfun")

pal_5 <- viridis(7)[-(1:2)]
set.seed(1221)
```

This vignette introduces the `tf` class, as well as the `tfd` and `tfb` subclasses, and focuses on vectors of this class. It also illustrates operations for `tf` vectors. 

# `tf`-Class: Definition

##  `tf`-class

**`tf`** is a new data type for (vectors of) functional data: 

- an abstract superclass for functional data in 2 forms:
    - as (argument, value)-tuples: subclass **`tfd`**, also irregular or sparse
    - or in basis representation: subclass **`tfb`** represents each observed function as a weighted sum
    of a fixed dictionary of known "basis functions".
    
- basically, a `list` of numeric vectors  
  (... since `list`s work well as columns of data frames ...)

- with additional attributes that define *function-like* behavior:
    - how to **evaluate** the given "functions" for new arguments
    - their **domain**: the range of valid argument values 

- `S3` based

## Example Data

First we extract a `tf` vector from the `tidyfun::dti_df` dataset containing fractional anisotropy tract profiles for the corpus callosum (`cca`). When printed, `tf` vectors show the first few `arg` and `value` pairs for each subject.  

```{r load-dti}
data("dti_df")

cca <- dti_df$cca
cca
```

We also extract a simple 5-element vector of functions on a regular grid:

```{r ex-def}
cca_five <- cca[1:5, seq(0, 1, length.out = 93), interpolate = TRUE]
rownames(cca_five) <- LETTERS[1:5]
cca_five <- tfd(cca_five, signif = 2)
cca_five
```

For illustration, we plot the vector `cca_five` below. 

```{r, ex-fig}
plot(cca_five, xlim = c(-0.15, 1), col = pal_5)
text(x = -0.1, y = cca_five[, 0.07], labels = names(cca_five), col = pal_5)
```

## **`tf`** subclass: **`tfd`**

**`tfd`** objects contain "raw" functional data: 

 - represented as a list of **`evaluations`** $f_i(t)|_{t=t'}$ and corresponding **`arg`**ument vector(s) $t'$
 - has a **`domain`**:  the range of valid **`arg`**s.

```{r}
cca_five |>
  tf_evaluations() |>
  str()
cca_five |>
  tf_arg() |>
  str()
cca_five |> tf_domain()
```

- each **`tfd`**-vector contains an **`evaluator`** function that defines how to inter-/extrapolate `evaluations` between `arg`s

```{r}
tf_evaluator(cca_five) |> str()
tf_evaluator(cca_five) <- tf_approx_spline
```

- **`tfd`** has two subclasses: one for regular data with a common grid and one for irregular or sparse data. The `cca` data are irregular (values are missing for some subjects at some arguments) but the example below more clearly illustrates support for sparse and irregular data using CD4 cell counts from a longitudinal study included in **`refund`**.

```{r}
cd4_vec <- tfd(refund::cd4)

cd4_vec[1:2]
cd4_vec[1:2] |>
  tf_arg() |>
  str()
cd4_vec[1:20] |> plot(pch = "x", col = viridis(20))
```

## **`tf`** subclass: **`tfb`**

Functional data in basis representation: 

 - represented as a list of **`coefficients`** and a common **`basis_matrix`** of basis function evaluations on a vector of `arg`-values.
 - contains a **`basis`** function that defines how to evaluate the basis functions for new **`arg`**s and how to differentiate or integrate it.
- (internal) flavors: 
    - `tfb_spline`: uses **`mgcv`**-spline bases 
    - `tfb_fpc`: uses functional principal components 
- significant memory and time savings:

```{r}
refund::DTI$cca |>
  object.size() |>
  print(units = "Kb")
cca |>
  object.size() |>
  print(units = "Kb")
cca |>
  tfb(verbose = FALSE) |>
  object.size() |>
  print(units = "Kb")
```

### **`tfb_spline`**: spline basis

- default for `tfb()`
- accepts all arguments of **`mgcv`**'s `s()`-syntax: basis type `bs`, basis dimension `k`, penalty order `m`, etc...
- also does non-Gaussian fits: `family` argument 
    - all exponential families
    - but also: $t$-distribution, ZI-Poisson, Beta, ... 

```{r, message = TRUE}
cca_five_b <- cca_five |> tfb()
cca_five_b[1:2]
cca_five[1:2] |> tfb(bs = "tp", k = 55)

# functions represent ratios in (0,1), so a Beta-distribution is more appropriate:
cca_five[1:2] |>
  tfb(bs = "ps", m = c(2, 1), family = mgcv::betar(link = "cloglog"))
```

### Penalization: 

**Function-specific (default), none**, prespecified (`sp`), or global: 

```{r}
layout(t(1:2))
cca_five |> plot()
cca_five_b |> plot(col = "red")
cca_five |>
  tfb(k = 35, penalized = FALSE) |>
  lines(col = "blue")
cca_five |>
  tfb(sp = 0.001) |>
  lines(col = "orange")
```

Right plot shows smoothing with function-specific penalization in red, without penalization in blue, 
and with manually set strong smoothing (`sp` $\to 0$) in orange.


**"Global" smoothing**:  

1. estimate smoothing parameters for subsample (~10\%) of curves
2. apply geometric mean of estimated smoothing parameters to smooth *all* curves

**Advantages:**

- (much) faster than optimizing penalization for each curve
- should scale well for largish datasets

**Disadvantages**

- no real borrowing of information across curves (very sparse or functional fragment data, e.g.)
- still requires more observations than basis functions *per curve*
- subsample could miss small subgroups with different roughness, over-/undersmooth parts of the data, see below.

```{r, echo = FALSE}
set.seed(1212)
raw <- c(
  tf_rgp(5, scale = 0.2, nugget = 0.05, arg = 101L) - 5,
  tf_rgp(5, scale = 0.02, nugget = 0.05, arg = 101L),
  tf_rgp(5, scale = 0.002, nugget = 0.05, arg = 101L) + 5
)
```

Dataset with heterogeneous roughness:
```{r}
layout(t(1:3))
clrs <- scales::alpha(sample(viridis(15)), 0.5)
plot(raw, main = "raw", col = clrs)
plot(tfb(raw, k = 55), main = "separate", col = clrs)
plot(tfb(raw, k = 55, global = TRUE), main = "global", col = clrs)
```

###  **`tfb`** FPC-based

- uses first few eigenfunctions computed from a simple unregularized (weighted) SVD of the data matrix by default
- corresponding FPC basis and mean function saved as `tfd`-object
- observed functions are linear combinations of those.
- amount of "smoothing" can be controlled (roughly!) by setting the 
minimal *percentage of variance explained* `pve` 

```{r}
cca_five_fpc <- cca_five |> tfb_fpc(pve = 0.999)
cca_five_fpc

cca_five_fpc_lowrank <- cca_five |> tfb_fpc(pve = 0.6)
cca_five_fpc_lowrank
```

```{r}
layout(t(1:2))
cca_five |> plot()
cca_five_fpc |> plot(col = "red", ylab = "tfb_fpc(cca_five)")
cca_five_fpc_lowrank |> lines(col = "blue", lty = 2)
```

`tfb_fpc` is currently only implemented for data on identical 
(but possibly non-equidistant) grids. The **`{refunder}`** `rfr_fpca`-functions 
provide FPCA methods appropriate for highly irregular and sparse data and regularized/smoothed FPCA.


# `tf`-Class: Methods

**`tidyfun`** implements almost all types of operations that are available for conventional
numerical or logical vectors for `tf`-vectors as well, so you can:

### subset & subassign:

```{r}
cca_five[1:2]
cca_five[1:2] <- cca_five[2:1]
cca_five
```

### compare & compute:

```{r, echo = FALSE}
n_cca_five <- names(cca_five)
cca_five <- unname(cca_five)
```

```{r}
cca_five[1] + cca_five[1] == 2 * cca_five[1]
log(exp(cca_five[2])) == cca_five[2]
(cca_five - (2:-2)) != cca_five
```

```{r, echo = FALSE}
names(cca_five) <- n_cca_five
```

### summarize across a vector of functions: 

Compute functional summaries like mean functions, functional standard deviations or variances or functional data depths over a vector of functional data:
```{r}
c(mean = mean(cca_five), sd = sd(cca_five))

tf_depth(cca_five) ## Modified Band-2 Depth (a la Sun/Genton/Nychka, 2012), others to come.
median(cca_five) == cca_five[which.max(tf_depth(cca_five))]
summary(cca_five)
```

### summarize each function over its domain: 

Compute summaries for each function like its mean or extreme values, quantiles, etc.
```{r}
tf_fmean(cca_five) # mean of each function's evaluations
tf_fmax(cca_five) # max of each function's evaluations
# 25%-tile of each f(t) for t > .5:
tf_fwise(cca_five, \(x) quantile(x$value[x$arg > 0.5], prob = 0.25)) |> unlist()
```
`tf_fwise` can be used to define custom statistics for each function that can depend on both its `value` and its `arg`.


In addition, **`tidyfun`** provides methods for operations that are specific for functional data:

## Methods for "functional" operations

### evaluate:

`tf`-objects have a special `[`-operator: Its second argument specifies 
`arg`ument values at which to evaluate the functions and has some additional options, 
so it's easy to get point values for `tf` objects, in `matrix` or `data.frame` formats: 

```{r, warning  = FALSE}
cca_five[1:2, seq(0, 1, length.out = 3)]
cca_five["B", seq(0, 0.15, length.out = 3), interpolate = FALSE]
cca_five[1:2, seq(0, 1, length.out = 7), matrix = FALSE] |> str()
```


### (simple, local) smoothing

```{r}
layout(t(1:3))
cca_five |> plot(alpha = 0.2, ylab = "lowess")
cca_five |>
  tf_smooth("lowess") |>
  lines(col = pal_5)

cca_five |> plot(alpha = 0.2, ylab = "rolling median (k=5)")
cca_five |>
  tf_smooth("rollmedian", k = 5) |>
  lines(col = pal_5)

cca_five |> plot(alpha = 0.2, ylab = "Savitzky-Golay (quartic, 11 steps)")
cca_five |>
  tf_smooth("savgol", fl = 11) |>
  lines(col = pal_5)
```

### differentiate & integrate:

```{r}
layout(t(1:3))
cca_five |> plot(col = pal_5)
cca_five |>
  tf_smooth() |>
  tf_derive() |>
  plot(col = pal_5, ylab = "tf_derive(tf_smooth(cca_five))")
cca_five |>
  tf_integrate(definite = FALSE) |>
  plot(col = pal_5)
```

```{r}
cca_five |> tf_integrate()
```

### query

**`tidyfun`** makes it easy to find (ranges of) `arg`uments $t$ satisfying a condition on `value` $f(t)$ (and `arg`ument $t$):

```{r}
cca_five |> tf_anywhere(value > 0.65)
cca_five[1:2] |> tf_where(value > 0.6, "all")
cca_five[2] |> tf_where(value > 0.6, "range")
cca_five |> tf_where(value > 0.6 & arg > 0.5, "first")
```

### zoom & query

```{r, ex-fig2}
cca_five |> plot(xlim = c(-0.15, 1), col = pal_5, lwd = 2)
text(x = -0.1, y = cca_five[, 0.07], labels = names(cca_five), col = pal_5, cex = 1.5)
median(cca_five) |> lines(col = pal_5[3], lwd = 4)
```

```{r}
# where are the first maxima of these functions?
cca_five |> tf_where(value == max(value), "first")

# where are the first maxima of the later part (t > .5) of these functions?
cca_five[c("A", "D")] |>
  tf_zoom(0.5, 1) |>
  tf_where(value == max(value), "first")

# which f_i(t) are below the functional median anywhere for 0.2 < t < 0.6?
# (t() needed here so we're comparing column vectors to column vectors...)
cca_five |>
  tf_zoom(0.2, 0.6) |>
  tf_anywhere(value <= t(median(cca_five)[, arg]))
```

---
title: "Converting to & from `tf`"
author: "Jeff Goldsmith, Fabian Scheipl"
date: "`r Sys.Date()`"
output: 
    rmarkdown::html_vignette:
      fig_width: 7
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{Converting to & from `tf`}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "##",
  fig.width = 6,
  fig.height = 4,
  dpi = 72,
  fig.retina = 1,
  out.width = "90%"
)

library(tidyverse)
library(viridisLite)
library(lme4)
library(gridExtra)

theme_set(theme_minimal() + theme(legend.position = "bottom"))

options(
  ggplot2.continuous.colour = "viridis",
  ggplot2.continuous.fill = "viridis"
)
scale_colour_discrete <- scale_colour_viridis_d
scale_fill_discrete <- scale_fill_viridis_d

library("tidyfun")

pal_5 <- viridis(7)[-(1:2)]
set.seed(1221)
```

Functional data have often been stored in matrices or data frames. Although these structures have sufficed for some purposes, they are cumbersome or impossible to use with modern tools for data wrangling. 

In this vignette, we illustrate how to convert data from common structures to `tf` objects. Throughout, functional data vectors are stored as columns in a data frame to facilitate subsequent wrangling and analysis. 

# Conversion from matrices

One of the most common structures for storing functional data has been a matrix. Especially when subjects are observed over the same (regular or irregular) grid, it is natural to observations on a subject in rows (or columns) of a matrix. Matrices, however, are difficult to wrangle along with data in a data frame, leading to confusing and easy-to-break subsetting across several objects. 

In the following examples, we'll use `tfd` to get a `tf` vector from matrices. The `tfd` function expects data to be organized so that each row is the functional observation for a single subject. It's possible to focus only on the resulting `tf` vector, but in keeping with the broader goals of **`tidyfun`** we'll add these as columns to a data frame.

The `DTI` data in the **`refund`** package has been a popular example in functional data analysis. In the code below, we create a data frame (or `tibble`) containing scalar covariates, and then add columns for the `cca` and `rcst` track profiles. This code was used to create the `tidyfun::dti_df` dataset included in the package.

```{r}
dti_df <- tibble(
  id = refund::DTI$ID,
  visit = refund::DTI$visit,
  sex = refund::DTI$sex,
  case = factor(ifelse(refund::DTI$case, "MS", "control"))
)

dti_df$cca <- tfd(refund::DTI$cca, arg = seq(0, 1, length.out = 93))
dti_df$rcst <- tfd(refund::DTI$rcst, arg = seq(0, 1, length.out = 55))
```

In `tfd`, the first argument is a matrix; `arg` defines the grid over which functions are observed. The output of `tfd` is a vector, which we include in the `dti_df` data frame. 

```{r}
dti_df
```

Finally, we'll make a quick spaghetti plot to illustrate that the complete functional data is included in each `tf` column. 

```{r}
dti_df |>
  tf_ggplot(aes(tf = cca, col = case, alpha = 0.2 + 0.4 * (case == "control"))) +
  geom_line() +
  facet_wrap(~sex) +
  scale_alpha(guide = "none", range = c(0.2, 0.4))
```

We'll repeat the same basic process using a second, and probably even-more-perennial, functional data example: the Canadian weather data in the **`fda`** package. Here, functional data are stored in a three-dimensional array, with dimensions corresponding to day, station, and outcome (temperature, precipitation, and log10 precipitation).

In the following, we first create a `tibble` with scalar covariates, then use `tfd` to create functional data vectors, and finally include the resulting vectors in the dataframe. In this case, our `arg`s are days of the year, and we use `tf_smooth` to smooth the precipitation outcome. Because the original data matrices record the different observations in the columns instead of the rows, we have to use their transpose in the call to `tfd`:

```{r}
canada <- tibble(
  place = fda::CanadianWeather$place,
  region = fda::CanadianWeather$region,
  lat = fda::CanadianWeather$coordinates[, 1],
  lon = -fda::CanadianWeather$coordinates[, 2]
) |>
  mutate(
    temp = t(fda::CanadianWeather$dailyAv[, , 1]) |>
      tfd(arg = 1:365),
    precipl10 = t(fda::CanadianWeather$dailyAv[, , 3]) |>
      tfd(arg = 1:365) |>
      tf_smooth()
  )
```

The resulting data frame is shown below.

```{r}
canada
```

A plot containing both functional observations is shown below. 

```{r}
temp_panel <- canada |>
  tf_ggplot(aes(tf = temp, color = region)) +
  geom_line()

precip_panel <- canada |>
  tf_ggplot(aes(tf = precipl10, color = region)) +
  geom_line()

gridExtra::grid.arrange(temp_panel, precip_panel, nrow = 1)
```


# Conversion to `tf` from a data frame

### ... in "long" format

"Long" format data frames containing functional data include columns containing a subject identifier, the functional argument, and the value each subject's function takes at each argument. There are also often (but not always) non-functional covariates that are repeated within a subject. For data in this form, we use `tf_nest` to produce a data frame containing a single row for each subject.

A first example is the `sleepstudy` data from the **`lme4`** package, which is a nice example from longitudinal data analysis. This includes columns for `Subject`, `Days`, and `Reaction` -- which correspond to the subject, argument, and value. 

```{r}
data("sleepstudy", package = "lme4")
sleepstudy <- as_tibble(sleepstudy)

sleepstudy
```

We create `sleepstudy_tf` by nesting `Reaction` within subjects. The result is a data frame containing a single row for each curve (one per `Subject`) that holds two columns: the `Reaction` function and the `Subject` ID: 

```{r}
sleepstudy_tf <- sleepstudy |>
  tf_nest(Reaction, .id = Subject, .arg = Days)

sleepstudy_tf
```

We'll make a quick plot to show the result.

```{r}
sleepstudy_tf |>
  tf_ggplot(aes(tf = Reaction)) +
  geom_line()
```

Alternatively, for this simple example that does not contain any additional time-varying or time-constant covariates besides the values that define the functions themselves, we could have simply done:

```{r}
tibble(
  Subject = unique(sleepstudy$Subject), 
  Reaction = tfd(sleepstudy, id = "Subject", arg = "Days", value = "Reaction")
)
```

A second example uses the `ALA::fev1` dataset. **`ALA`** is not on CRAN, so we do not show an install command here. The code below is illustrative only and is not evaluated when the vignette is built.

In this dataset, both `height` and `logFEV1` are observed at multiple ages for each child; that is, there are two functions observed simultaneously, over a shared argument. We can use `tf_nest` to create a dataframe with a single row for each subject, which includes both non-functional covariates (like age and height at baseline), and functional observations `logFEV1` and `height`.

```{r, eval = FALSE}
ALA::fev1 |>
  group_by(id) |>
  mutate(n_obs = n()) |>
  filter(n_obs > 1) |>
  ungroup() |>
  tf_nest(logFEV1, height, .arg = age) |>
  glimpse()
```

### ... in "wide" format

In some cases functional data are stored in "wide" format, meaning that there are separate columns for each argument, and values are stored in these columns. In this case, `tf_gather` can be use to collapse across columns to produce a function for each subject.

The example below again uses the `refund::DTI` dataset. We use `tf_gather` to transfer the `cca` observations from a matrix column (with `NA`s) into a column of irregularly observed functions (`tfd_irreg`). 

```{r}
dti_df <- refund::DTI |>
  janitor::clean_names() |>
  select(-starts_with("rcst")) |>
  glimpse()

dti_df |>
  tf_gather(starts_with("cca")) |>
  glimpse()
```


# Changing representation with `tf_rebase`

Sometimes you need to make different `tf` objects compatible -- for example,
to combine raw observations with a basis representation, or to ensure two
functional data vectors are expressed on the same grid or in the same basis.
`tf_rebase` re-expresses one `tf` object in the representation of another:

```{r}
# reload the tidyfun version of the DTI data
data(dti_df, package = "tidyfun")

# raw functional data
cca_raw <- dti_df$cca[1:5]
cca_raw

# represent in a spline basis
cca_basis <- tfb(dti_df$cca[1:5], k = 25)
cca_basis

# re-express the raw data in the same basis representation
cca_rebased <- tf_rebase(cca_raw, basis_from = cca_basis)
cca_rebased

# or convert a spline-based representation to a grid-based one for a specific grid:
tf_rebase(cca_basis, basis_from = cca_raw)
```

This is useful when you want to ensure that operations between `tf` objects
(e.g., addition, comparison) use a common representation, or when you want
to convert between `tfd` and `tfb` representations while matching a specific
basis configuration. It is required for many operations that would otherwise not
be well-defined.


# Splitting and combining functions

`tf_split` separates each function into fragments defined on sub-intervals
of its domain, and `tf_combine` joins fragments back together. This is
useful for analyzing specific parts of a function separately or for
stitching together functional observations from different sources.

```{r}
# split CCA profiles at their midpoint
cca_halves <- tf_split(dti_df$cca[1:10], splits = 0.5)

# result is a list of tf vectors, one per segment
cca_halves[[1]]
cca_halves[[2]]

# recombine
cca_recombined <- tf_combine(cca_halves[[1]], cca_halves[[2]])
cca_recombined 
```


# Conversion from `fda` objects

The **`fda`** package represents functional data as `fd` objects (basis function
coefficients + basis definition). **`tf`** can convert these directly using
`tfb_spline`, which re-expresses the `fd` basis in `tf`'s spline framework.
This also works for `fdSmooth` objects returned by `fda::smooth.basis`.

```{r}
# create an fd object from the Canadian weather data
weather_basis <- fda::create.fourier.basis(c(0, 365), nbasis = 65)
weather_fd <- fda::smooth.basis(
  argvals = 1:365,
  y = fda::CanadianWeather$dailyAv[, , 1],
  fdParobj = weather_basis
)

# convert fdSmooth to tfb
weather_tf <- tfb_spline(weather_fd)
weather_tf[1:3]
```

The resulting `tfb` object can then be used with all **`tidyfun`** tools:

```{r}
tibble(
  place = fda::CanadianWeather$place,
  region = fda::CanadianWeather$region,
  temp = weather_tf
) |>
  tf_ggplot(aes(tf = temp, color = region)) +
  geom_line(alpha = 0.5)
```


# Reversing the conversion

**`tidyfun`** includes a wide range of tools for exploratory analysis and visualization, but many analysis approaches require data to be stored in more traditional formats. Several functions are available to aid in this conversion.

## Conversion from `tf` to data frames

The functions `tf_unnest` and `tf_spread` reverse the operations in `tf_nest` and `tf_gather`, respectively -- that is, they take a data frame with a functional observation and produce long or wide data frames. We'll illustrate these with the `sleepstudy_tf` data set. 

First, to produce a long-format data frame, one can use `tf_unnest`:

```{r}
sleepstudy_tf |>
  tf_unnest(cols = Reaction) |>
  glimpse()
```

To produce a wide-format data frame, one can use `tf_spread`:

```{r}
sleepstudy_tf |>
  tf_spread() |>
  glimpse()
```

## Converting back to a matrix or data frame

To convert `tf` vector to a matrix with each row containing the function evaluations for one function, use `as.matrix`:

```{r}
reaction_matrix <- sleepstudy_tf |> pull(Reaction) |> as.matrix() 
 
head(reaction_matrix)

# argument values of input data saved in `arg`-attribute:
attr(reaction_matrix, "arg")
```

To convert a `tf` vector to a standalone data frame with `"id"`,`"arg"`,`"value"`-columns, use `as.data.frame()` with `unnest = TRUE`:

```{r}
sleepstudy_tf |> pull(Reaction) |> 
  as.data.frame(unnest = TRUE) |>
  head()
```

---
title: "Developer Notes"
author: "Jeff Goldsmith"
date: "`r Sys.Date()`"
output: 
    rmarkdown::html_vignette:
      fig_width: 7
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{Developer Notes}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "##",
  fig.width = 6,
  fig.height = 4,
  dpi = 72,
  fig.retina = 1,
  out.width = "90%"
)

library(tidyverse)
library(viridisLite)

theme_set(theme_minimal() + theme(legend.position = "bottom"))

options(
  ggplot2.continuous.colour = "viridis",
  ggplot2.continuous.fill = "viridis"
)
scale_colour_discrete <- scale_colour_viridis_d
scale_fill_discrete <- scale_fill_viridis_d

library("tidyfun")
data(chf_df, package = "tidyfun")
data(dti_df, package = "tidyfun")

pal_5 <- viridis(7)[-(1:2)]
set.seed(1221)
```


This article is written for package authors and maintainers who want to extend the **`tidyfun`**/**`tf`** ecosystem. End users can usually skip it.

**`tidyfun`** is intended to make user interactions with functional data easier. The package builds on package **`tf`**, which defines a data class (`tf`) so that vectors of functional observations are available, similar to vectors of class `numeric` or `character`^[This class and its subclasses and methods live in a separate package without **`tidyverse`** dependencies in order to keep maintenance simpler and to support users who prefer not to depend on the **`tidyverse`**.]

Such classes make it possible to store functional data alongside other variables in a single dataframe; by extension, tools for data manipulation from the **`tidyverse`** and elsewhere can be used with datasets in which one or more variable is functional. 

Many basic analyses are available in **`tidyfun`** -- it's possible to compute mean functions and other summary statistics of `tf` vectors, to expand observations using a spline basis or using functional principal components analysis. With tools like `group_by` and `summarize`, these can be powerful tools for data exploration. Other analysis approaches are included in the **`refunder`** package, and more will be added to that package over time.

There is an active research community in FDA, and there is constant development of new methods for analysis. Our hope is that the data structures in **`tidyfun`** will be useful as others as new approaches are implemented, and this page is intended to provide some advice and guidance for those implementations. 


## Why use **`tidyfun`**

There are start-up costs to using a new data class when writing code for new methods, and there should be reasons to adopt such a class. We see several benefits to using **`tidyfun`**:

* Compatibility with a dataframe-centric approach to analysis
* Supporting tools for data manipulation
* Plotting in `ggplot` and base R

Together, these make it possible to analyze functional data in a pipeline (import, exploratory analysis, visualization, and formal analysis) that is similar to those used for scalar variables. 

These advantages are user-facing -- they are intended to make things easier when analyzing datasets that include functional observations. Because new methods for functional data typically involve working with "raw" observations (numeric vectors or matrices), implementing methods for **`tidyfun`** will require some consideration of user interfaces, input objects, and data transformations. We believe the benefits are worth this effort.


## Thoughts on design

**`tidyfun`** will be most effective in new methods that are intended to be part of an analysis pipeline. As a starting point, we suggest addressing the following questions:

* How will users' data be organized? What will input dataframes look like? What other variables will exist in addition to `tf` columns? 
* What is a natural user interface for my function? How should users specify functional data columns? 
* What will users expect to obtain from my function? Are there parallels with other modeling strategies? How should I organize output so that it integrates with a dataframe-centric analysis approach? 
* Should functions support `tfd`, `tfb`, or both? Does the output class matter? 
* Are there any supporting tools I should include? Things like `plot` or `predict`, which can be used to summarize results?

At best, users will have a seamless experience across data exploration, modeling, and understanding results; **`tidyfun`** is intended to encourage that. 


## Using **`tidyfun`** in new functions

We anticipate that new methods for functional data will use raw numeric values (in vectors or matrices) for estimation and / or inference. Tools for converting `tf` vectors to matrices or other formats are available, and useful in this context. In general, we have used the following structure for functions that perform analyses:

* Input objects are dataframes containing `tf` vectors
* `tf` vectors are converted to matrices (or numeric vectors)
* The estimation procedure is implemented using these matrices
* Results are formatted as `tf` vectors as appropriate and returned to the user

This pipeline shifts the burden for data conversion from the user to the function author, and in doing so maintains seamlessness for the user. 

As an example, the `refunder::rfr_fpca` function for functional principal components analysis has two main inputs: a dataframe containing one or more `tf` variables, and the name of the variable to decompose using FPCA. Internally, the `tf` vector is converted to a matrix for estimation. The function returns a list of elements relevant for FPCA, and includes estimated functions as a `tfb` vector. There is also a `predict` method, so that FPCA expansions of new data using the estimated basis can be easily obtained by users. 


## Keep us informed!

If you build on **`tidyfun`**, please let us know by opening an [issue](https://github.com/tidyfun/tidyfun/issues) or sending suggestions.

---
title: "Data Wrangling"
author: "Jeff Goldsmith, Fabian Scheipl"
date: "`r Sys.Date()`"
output: 
    rmarkdown::html_vignette:
      fig_width: 7
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{Data Wrangling}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "##",
  fig.width = 6,
  fig.height = 4,
  dpi = 72,
  fig.retina = 1,
  out.width = "90%"
)

library(tidyverse)
library(viridisLite)

theme_set(theme_minimal() + theme(legend.position = "bottom"))

options(
  ggplot2.continuous.colour = "viridis",
  ggplot2.continuous.fill = "viridis"
)
scale_colour_discrete <- scale_colour_viridis_d
scale_fill_discrete <- scale_fill_viridis_d

library("tidyfun")

pal_5 <- viridis(7)[-(1:2)]
set.seed(1221)
```

# Data manipulation using **`tidyfun`**

The goal of **`tidyfun`** is to provide accessible and well-documented software that **makes functional data analysis in `R` easy**. In this vignette, we explore some aspects of data manipulation that are possible using **`tidyfun`** workflows with `tf` vectors, emphasizing compatibility with the **`tidyverse`**.

Other vignettes have examined the **`tfd`** & **`tfb`** data types, and how to convert common formats for functional data (e.g. matrices, long- and wide-format data frames, **`fda`** objects) in these new data types. Because our goal is "tidy" data manipulation for functional data analysis, the result of data conversion processes has been a data frame in which a column contains the functional data of interest. This vignette starts from that point.

Throughout, we make use of some visualization tools -- these are explained in more detail in the [visualization](https://tidyfun.github.io/tidyfun/articles/x04_Visualization.html) vignette.


# Example datasets

The datasets used in this vignette are the `tidyfun::chf_df` and `tidyfun::dti_df` dataset. The first contains minute-by-minute observations of log activity counts (stored as a `tfd` vector called `activity`) over seven days for each of 47 subjects with congestive heart failure. In addition to `id` and `activity`, we observe several covariates. 

```{r view_chf}
data(chf_df)

chf_df
```

A quick plot of the first 5 curves:

```{r plot_chf}
chf_df |>
  slice(1:5) |>
  tf_ggplot(aes(tf = activity)) +
  geom_line(alpha = 0.1)
```

The `tidyfun::dti_df` contains fractional anisotropy (FA) tract profiles for the corpus callosum (cca) and the right corticospinal tract (rcst), along with several covariates. 

```{r view_dti}
data(dti_df)

dti_df
```

A quick plot of the `cca` tract profiles is below.

```{r plot_dti}
dti_df |>
  tf_ggplot(aes(tf = cca)) +
  geom_line(alpha = 0.05)
```

# Existing `tidyverse` functions

Dataframes using **`tidyfun`** to store functional observations can be manipulated using tools from **`dplyr`**, including `select` and `filter`:

```{r}
chf_df |>
  select(id, day, activity) |>
  filter(day == "Mon") |>
  tf_ggplot(aes(tf = activity)) +
  geom_line(alpha = 0.05)
```

Operations using `group_by` and `summarize` also work -- let's look at some
daily averages of these activity profiles:
```{r}
chf_df |>
  group_by(day) |>
  summarize(mean_act = mean(activity)) |>
  tf_ggplot(aes(tf = mean_act, color = day)) +
  geom_line()
```

One can `mutate` functional observations -- here we exponentiate the log activity counts to obtain original recordings:

```{r}
chf_df |>
  slice(1:5) |>
  mutate(exp_act = exp(activity)) |>
  tf_ggplot(aes(tf = exp_act)) +
  geom_line(alpha = 0.2)
```

Functions for data manipulation from **`tidyr`** are also supported. We illustrate by using `pivot_wider` to create new `tfd`-columns containing the activity profiles for each day of the week:

```{r}
chf_df |>
  select(id, day, activity) |>
  pivot_wider(
    names_from = day,
    values_from = activity
  )
```

(Note that this has made the data less "tidy" and is therefore not generally recommended, but may be useful in some cases).

It's also possible to join datasets based on non-functional keys. To illustrate, we'll first create a pair of datasets:

```{r}
monday_df <- chf_df |>
  filter(day == "Mon") |>
  select(id, monday_act = activity)
friday_df <- chf_df |>
  filter(day == "Fri") |>
  select(id, friday_act = activity)
```

These can be joined using the `id` variable as a key (and then tidied using `pivot_longer`):

```{r}
monday_df |>
  left_join(friday_df, by = "id") |>
  pivot_longer(monday_act:friday_act, names_to = "day", values_to = "activity")
```

Similar tidying can be done for the DTI data -- let's look at average RCST tract values for
gender and case status: 

```{r}
dti_df |>
  group_by(case, sex) |>
  summarize(mean_rcst = mean(rcst, na.rm = TRUE)) |>
  tf_ggplot(aes(tf = mean_rcst, color = case)) +
  geom_line(linewidth = 2) +
  facet_grid(~sex)
```


# `tf` helper functions in tidy workflows

Some **`dplyr`** functions are useful in conjunction with **`tf`** helper functions. For example, one might use `filter` with `tf_anywhere()` to filter based on the values of observed functions:

```{r}
like_to_move_it_move_it <- chf_df |> filter(tf_anywhere(activity, value > 9))
glimpse(like_to_move_it_move_it)

like_to_move_it_move_it |>
  tf_ggplot(aes(tf = activity, colour = id)) +
  geom_line()
```

A second example of this functionality in the DTI data is below.

```{r}
dti_df |>
  filter(tf_anywhere(cca, value < 0.26)) |>
  tf_ggplot(aes(tf = cca)) +
  geom_line()
```

The existing `mutate` function can be combined with several `tf` helpers, including `tf_smooth()`, `tf_zoom()`, and `tf_derive()`.

One can smooth existing observations using `tf_smooth`:

```{r}
chf_df |>
  filter(id == 1) |>
  mutate(smooth_act = tf_smooth(activity)) |>
  tf_ggplot(aes(tf = smooth_act)) +
  geom_line()
```

This can be combined with previous steps, like `group_by` and `summarize`, to build intuition through descriptive plots and summaries:

```{r}
chf_df |>
  group_by(day) |>
  summarize(mean_act = mean(activity)) |>
  mutate(smooth_mean = tf_smooth(mean_act)) |>
  tf_ggplot(aes(color = day)) +
  geom_line(aes(tf = mean_act), alpha = 0.2) +
  geom_line(aes(tf = smooth_mean), linewidth = 2)
```

One can also extract observations over a subset of the full domain using `tf_zoom`:

```{r}
chf_df |>
  filter(id == 1) |>
  mutate(daytime_act = tf_zoom(activity, 360, 1200)) |>
  tf_ggplot(aes(tf = daytime_act)) +
  geom_line(alpha = 0.2)
```

We can also convert from `tfd` to `tfb` inside a `mutate` statement as part of a data processing pipeline:

```{r}
dti_df <- dti_df |> mutate(cca_tfb = tfb(cca, k = 25))
```

It's also possible to compute derivatives as part of a processing pipeline:
```{r}
dti_df |>
  slice(1:10) |>
  mutate(
    cca_raw_deriv = tf_derive(cca),
    cca_tfb_deriv = tf_derive(cca_tfb)
  ) |>
  tf_ggplot() +
  geom_line(aes(tf = cca_raw_deriv), alpha = 0.3, linewidth = 0.3, color = "blue") +
  geom_line(aes(tf = cca_tfb_deriv), alpha = 0.3, linewidth = 0.3, color = "red") +
  ylab("d/dt f(t)")
```

# Working with `data.table`

**`tidyfun`** functional data objects work within **`data.table`** as well.

However, there is one specific known caveat when calculating the mean of a `tf` vector within a **`data.table`** context:
By default, **`data.table`** will use optimized routines for summary statistics, which may cause `mean` to dispatch to **`data.table`**'s own optimized implementation,
leading to unexpected behavior for `tf` vectors.

To ensure the correct mean calculation for `tf` vectors, there are two solutions:

- Disable the optimization for the summary step and let `mean()` dispatch normally:

```{r, eval = FALSE}
withr::with_options(
  list(data.table.optimize = 0),
  data.table::as.data.table(chf_df)[, list(mean_act = mean(activity)), by = day]
)
```

This caveat applies specifically to `mean()` and is the only known issue. 

---
title: "Curve Registration: Practical Guide and Pitfalls"
output:
  rmarkdown::html_vignette:
    code_folding: hide
vignette: >
  %\VignetteIndexEntry{Curve Registration: Practical Guide and Pitfalls}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  cache = TRUE,
  collapse = TRUE,
  comment = "#>",
  dpi = 72,
  fig.retina = 1
)
library(tf)
oldpar <- par(no.readonly = TRUE)
par(las = 1)
alpha_palette <- c(
  "#00000080", "#DF536B80", "#61D04F80", "#2297E680", "#28E2E580",
  "#CD0BBC80", "#F5C71080", "#9E9E9E80", "#00000080", "#DF536B80"
)
have_srvf <- rlang::is_installed("fdasrvf")
have_cc <- TRUE
```

This vignette provides access to the registration workflows from **`tf`** and
places them in the **`tidyfun`** site for convenience. It is aimed at users who
need to understand templates, warps, and diagnostics rather than at casual end
users exploring the core wrangling and plotting features.

## Registration: Template, Warp, Domain

Registration aligns functions in time by reducing phase variation (timing differences - "horizontal" variability) while preserving amplitude variation (shape/height differences - "vertical" variability).

Use registration if:

- Curves have broadly similar shapes that are shifted, stretched or compressed relative to each other
    - and your analysis does not care about these timing differences 
    - or you want to analyze these timing differences separately from shape differences.
- You can interpret both the **target template** and the resulting warps in domain terms.

Do not use registration as a (default) preprocessing step if:

- Curves come from clearly different shape classes/clusters (see below).
- Timing differences are meaningful for your question and/or cannot be viewed in isolation from shape differences.
- The time deformation from registration creates implausible subject-specific timelines.

**Core problem:**

Registration models observed curves $x_i(t)$ as *time-deformed versions of a template* $m(s)$, where the time deformation is given by a monotone warping function \(h_i(s)\):

\[
x_i(t) \approx m(h_i^{-1}(t))
\]

Key interpretation:

- \(h_i\) maps aligned/system time to observed time.
- \(h_i^{-1}\) maps observed time to aligned time.
- Slope of \(h_i^{-1}\) greater than 1 indicates local compression in observed time (events happen faster), slope less than 1 indicates local dilation (events happen slower).

**Core workflow in `tf`:**

```{r, eval=FALSE, echo = TRUE, class.source = "fold-show"}
# One-shot registration (returns tf_registration object):
reg <- tf_register(x, method = "...")
tf_aligned(reg)   # registered/aligned curves
tf_inv_warps(reg)     # estimated inverse warping functions (observed -> aligned time)
tf_template(reg)  # template used
summary(reg)      # alignment diagnostics
plot(reg)          # 3-panel diagnostic plot

# Or step by step:
warps <- tf_estimate_warps(x, method = "...")
x_registered <- tf_align(x, warps)
```

### What Is the Template?

Template choice is the most important modeling choice in registration: every estimated warp is relative to that template.

Default template behavior in `tf_register()` / `tf_estimate_warps()`:

| Method | Default template behavior | How to override |
|:--|:--|:--|
| `srvf` | Karcher-type mean shape^[The Karcher mean (also called Frechet mean) generalizes arithmetic means to general spaces. In this case, it is a centroid in the amplitude quotient space (functions modulo reparameterization), computed iteratively using the elastic (Fisher-Rao) distance rather than pointwise averaging.] estimated iteratively by **`fdasrvf`** | pass `template = ...` |
| `cc` | arithmetic mean curve (estimated iteratively) | pass `template = ...` |
| `affine` | arithmetic mean curve | pass `template = ...` |
| `landmark` | column-wise mean of landmark locations | pass `template_landmarks = ...` |

Practical rules:

1. If you need interpretable warpings relative to a known reference, supply an explicit template.
2. Template and data need to be on the same domain and grid for `srvf` and `cc`.
3. Always *visualize* raw curves and the template together before trusting warps.

4. Under strong phase variation, the pointwise mean can be a poor template -- see [Unsuitable template](#unsuitable-template) in the Pitfalls section for an example and remedy.

### What is a Warping Function?

`tf_estimate_warps()` returns forward warps \(h_i\) (aligned -> observed time), while `tf_inv_warps(reg)` returns the inverse warps \(h_i^{-1}\) (observed -> aligned time) that are directly used for alignment. These are the natural functions to inspect and plot -- they show how each curve's observed timepoints map to aligned "system" time.

- Warping functions close to the identity line imply there is little phase correction.
- Local slopes different from 1 imply local time dilation ($\tfrac{d}{dt} h_i^{-1}(t) < 1$: $s$ increases more slowly than $t$) or compression ($\tfrac{d}{dt} h_i^{-1}(t) > 1$: $s$ increases more quickly than $t$) of the observed function relative to the template function.
- Large departures from identity or frequent slope changes and crossings of the identity line without clear alignment gain indicate over-warping risk.

#### Domain-Preserving vs Non-domain-preserving Warps

This distinction is fundamental for interpretation and edge behavior:

- **Domain-preserving**: warps map domain endpoints to themselves. Registration using the `srvf`, `cc`, and `landmark` methods in **`tf`** is domain preserving.
- **Non-domain-preserving**: warps can shift or scale outside (or into a subset of) the observed domain. Only available with the `affine` method in **`tf`**.

Implications:

1. Non-domain-preserving warps can produce boundary `NA`s after `tf_align()` because parts of the aligned timeline have no observed support.
2. Domain-preserving warps avoid these boundary losses but can still overfit timing.
3. Method choice is therefore not only about alignment quality; it is also about what time transformation is scientifically admissible.

The figure below illustrates how warping functions compose with a template to produce shifted/deformed curves. Note that this shows the *forward simulation* view, $m(h(s))$: applying a warp to generate an observed curve from the template. In practice, registration works in the *inverse* direction -- estimating $h_i^{-1}$ to map an observed curve back to system time (i.e., to *align* it to a template).

```{r warp-illustration, echo = FALSE, fig.width = 6, fig.height = 6}
# Synthetic illustration of different warp types
s <- seq(0, 1, length.out = 201)

# Template: 
template_synth <- tfd(exp(-((s - 0.45)^2) / (2 * 0.08^2)) + cos(2*pi*s) + s, arg = s)

# Warp 1: affine shift + scale (not domain-preserving -- extends beyond [0,1])
h_shift <- tfd(0.8 * s + 0.1, arg = s)

# Warp 2: smooth diffeomorphism (domain-preserving)
h_smooth_raw <- s - 0.25 * sin(pi * s)
h_smooth <- tfd(
  (h_smooth_raw - h_smooth_raw[1]) / (h_smooth_raw[length(s)] - h_smooth_raw[1]),
  arg = s)

# Warp 3: wiggly diffeomorphism (domain-preserving)
h_wiggly_raw <- .1 * pbeta(s, .1, .1) + .5 * pbeta(s, 25, 25) + 
  .2 * pbeta(s, 4, 20) + .2 * pbeta(s, 100, 10)
h_wiggly <- tfd(h_wiggly_raw, arg = s)  

# Compose template with warps: m(h(s))
warped_shift <- tf_warp(template_synth, h_shift) |> suppressWarnings()
warped_smooth <- tf_warp(template_synth, h_smooth)
warped_wiggly <- tf_warp(template_synth, h_wiggly)

warp_cols <- c("#DF536B", "#2297E6", "#61D04F")

layout(matrix(1:9, 3, 3))
par(mar = c(4, 4, 3, 1))

warp_names <- c("Affine (shift & scale)", "Smooth elastic", "Wiggly elastic")
warps <- list(h_shift, h_smooth, h_wiggly)
warped <- list(warped_shift, warped_smooth, warped_wiggly)

# Column 1: blank, template, blank
plot.new()
plot(template_synth, lwd = 2.5,
     main = "Template m(s)", xlab = "s", ylab = "m(s)")
plot.new()

# Column 2: one warp per row
for (i in 1:3) {
  plot(s, s, type = "l", lty = 3, lwd = 1,
       main = paste0("h(s): ", warp_names[[i]]),
       xlab = "s", ylab = "h(s)", ylim = c(-0.05, 1.15))
  do.call(graphics::lines, list(warps[[i]], col = warp_cols[[i]], lwd = 2.5))
}

# Column 3: template (dashed) + warped curve per row
for (i in 1:3) {
  plot(template_synth, lwd = 2, lty = 3,
       main = paste0("x(t) = m(h(s)): ", warp_names[[i]]),
       xlab = "t", ylab = "x(t)")
  do.call(graphics::lines, list(warped[[i]], col = warp_cols[[i]], lwd = 2.5))
}
```

- Note that the affine shift warp is NOT domain-preserving (it shrinks the domain down to [0.1, 0.9]), while the smooth and wiggly elastic warps ARE domain-preserving (they start at 0 and end at 1).
- The *flatter* the warping function, the more it compresses time and the *steeper* the more it dilates the reference "system" time of the template.
- Where the warping function $h(s)$ is above the identity line, the template $m(s)$ is *ahead* of the warped function $x(t) = m(h(s))$ , where it is below, the template is *behind* the warped function.

## Example: Shift Registration with Explicit Template 

This minimal example uses simple shift registration, i.e. the warping model assumes $h_i(s) = s + b_i$ (no elastic deformation of the domain, simply shifts the functions' arguments). 
We pass an explicit template to make the target shape transparent:

```{r quickstart, class.source = "fold-show"}
pinch_small <- pinch[1:10]
template_affine <- pinch[7] |> tf_smooth(f= .2)

reg_shift <- tf_register(
  pinch_small,
  method = "affine",
  template = template_affine,
  type = "shift"
)
reg_shift
summary(reg_shift)
pinch_inv_warp_shift <- tf_inv_warps(reg_shift)
pinch_small_reg <- tf_aligned(reg_shift)
pinch_template_shift <- tf_template(reg_shift)
```

```{r quickstart-plot, echo = FALSE, class.source = "fold-show", fig.width = 6, fig.height = 3}
# plot(reg_shift) gives a compact diagnostic; below reuses the extracted
# components for a customized version of the same view:
layout(t(1:3))
plot(
  pinch_small,
  main = "Original Curves",
  xlab = "Observed Time t [sec]",
  ylab = "Force [N]",
  lwd = 1.5,
  col = alpha_palette
)
lines(pinch_template_shift, lwd = 2, lty = 3, col = rgb(0,0,.5))
plot(
  pinch_inv_warp_shift,
  points = FALSE,
  main = "Inverse Warping Functions",
  xlab = "Observed Time t [sec]",
  ylab = "Registered Time s [sec]",
  lwd = 1.5,
  col = alpha_palette
)
abline(0, 1, lty = 3)
plot(
  pinch_small_reg,
  main = "Registered Curves",
  xlab = "Registered Time s [sec]",
  ylab = "Force [N]",
  lwd = 1.5,
  col = alpha_palette, 
  points = FALSE
)
lines(pinch_template_shift, lwd = 2, lty = 3, col = rgb(0,0,.5))
```

Quick success checks:

- The template is plausible for the population you are aligning.
- Major features (e.g., peaks) are more aligned after registration.
- Inverse warps are not excessively far from the identity line for this problem.
- Boundary `NA`s are limited and scientifically acceptable.
- Resulting timing shifts are domain-plausible.

## Choosing a Registration Method

|   | `srvf` (default) | `cc` | `affine` | `landmark` |
|:--|:--|:--|:--|:--|
| **Use when** | Smooth, non-linear timing differences with shared shape family | You want continuous-criterion alignment and can tune criterion | Mostly shift/scale timing variability; need interpretable warps | Reliable, repeated landmarks are available across curves |
| **Avoid when** | Very noisy data, sparse grids, highly heterogeneous shapes | Complex phase-amplitude interactions or severe irregularity | Strong non-linear timing deformation | Ambiguous/noisy landmarks or mismatched counts |
| **Template handling** | Implicit Karcher-type mean unless `template` supplied | Arithmetic mean unless `template` supplied | Arithmetic mean unless `template` supplied | Uses `template_landmarks` (default: column means) |
| **Key arguments** | `method = "srvf"`, `lambda` | `method = "cc"`, `crit = 1/2`, `nbasis`, `lambda`, `conv`, `iterlim` | `method = "affine"`, `type`, bounds | `method = "landmark"`, `landmarks`, `template_landmarks` |
| **Input grid** | regular only | regular only | regular + **irregular** | regular + **irregular** |
| **Noise robustness** | Low (pre-smooth if noisy) | High (most stable) | Moderate | Moderate; outlier-robust |
| **Typical speed** | Fast (but O(n^2^) in grid) | Moderate | Fast | Very fast |
| **Typical failure signs** | Unstable warps, over-warping, inconsistent reruns | Sensitive to criterion choice, weak alignment gains | Residual misalignment of local features, boundary `NA`s under stronger shifts/scales | Forced/broken alignments from bad landmarks |
| **First fallback** | Pre-smooth inputs; try `affine` or `landmark` | Change criterion, basis dimension and/or amount of penalization; try `srvf` or `affine` | Set stricter/looser bounds; upgrade to `srvf` | Check appropriateness of landmarks; use `srvf` or `affine` |

The `srvf` and `cc` methods require regular grids (`tfd_reg` or `tfb`). The `affine` and `landmark` methods also accept irregular grids (`tfd_irreg`).

### Evidence-Based Method Selection

Based on [benchmarks](https://tidyfun.github.io/sim-registration/) across 15 data-generating processes, 3 noise levels, and 5 methods:

1. **Reliable landmarks available?** -> Use `landmark`. Fastest, simplest warps, and most robust to outlier contamination (warp MISE degrades only ~1.6x at 30% contamination vs 2.6-3.7x for iterative methods).
2. **Mostly shift/scale variation?** -> Use `affine`. Fast, interpretable, but produces boundary `NA`s.
3. **Noisy data?** -> Use `cc` (criterion 2) as a stable default -- the most noise-robust method in our benchmarks. Or use `srvf` after pre-smoothing: `tfb(x, k = 25) |> tf_register(method = "srvf")`.
4. **Suspected outlier contamination (>10%)?** -> Use `landmark` if possible. Otherwise, remove outliers before registration.
5. **Clean data, complex phase variation?** -> Use `srvf`. Best overall warp recovery on clean data with domain-preserving warps, but rankings shift with noise level and template shape.
6. **Always compare at least 2 methods** and inspect diagnostics (see [Diagnostics Workflow](#practical-diagnostics-workflow) below). No single method dominates across all conditions.

See [Theoretical Background](#theoretical-background) below for method-specific mathematical details.

## Pitfalls and Likely Failure Modes

#### Structured shape variation

**Issue:** Registered curves still form distinct shape clusters.

**Remedies:** Stratify first; register within clusters; compare cluster-specific results.

Registration assumes a single shared template. If curves come from different shape families, a single registration will try to compromise and may distort curves from both groups.

#### Noise dominates phase signal

**Issue:** Warps become jagged, wiggly or highly variable across near-identical curves.

**Remedies:** Smooth inputs (more) first; reduce basis flexibility; compare before/after sensitivity.

In our benchmarks^[For full details on the benchmark design and results, see [tidyfun.github.io/sim-registration](https://tidyfun.github.io/sim-registration/)], CC methods (`method = "cc"`) were the most noise-robust, while SRVF degraded most sharply under noise -- because SRSFs involve numerical derivatives that amplify observation noise. Pre-smoothing SRVF inputs with `tfb(x, k = 25)` before registration reduced warp error (measured as warp MISE) by 50-70% under moderate noise. Recipe: `x_smooth <- tfb(x, k = 25); reg <- tf_register(x_smooth, method = "srvf")`.

#### Sparse or irregular grids

**Issue:** Registration quality may be affected strongly by interpolation grid choices.

**Remedies:** Re-evaluate on a fine, common grid; run sensitivity checks on grid density.

The `srvf` and `cc` methods require regular grids (`tfd_reg`); interpolate sparse or irregular data to a common regular grid first. The `affine` and `landmark` methods also accept irregular grids (`tfd_irreg`). Grid density can affect results -- try at least two grid resolutions to check stability. SRVF performance varies substantially with grid resolution due to numerical differentiation: grid sizes around 100 points are a robust default. Finer grids (>200 points) can degrade SRVF warp recovery under noise by amplifying derivative artifacts (warp MISE roughly doubles at grid=201 vs 101 under moderate noise). CC, affine, and landmark methods are largely grid-insensitive.

#### Boundary and overlap artifacts

**Issue:** Large `NA` regions after unwarping, especially for aggressive affine shifts/scales.

**Remedies:** Inspect boundary behavior explicitly; narrow affine bounds via `shift_range`/`scale_range`.

This issue is specific to the `affine` method, which is the only non-domain-preserving method in `tf`. The `srvf`, `cc`, and `landmark` methods preserve domain endpoints by construction and do not produce boundary `NA`s.

#### Over-warping

**Issue:** Extreme timing distortions with little gain in feature alignment.

**Remedies:** Prefer simpler methods (`affine`), reduce flexibility, or avoid registration for that subset.

This is primarily a risk with flexible methods (`srvf`, `cc`). Affine warps are constrained by design and rarely over-warp. Comparing results across methods can help detect over-warping: if a flexible method's warps are much more variable than an affine baseline without a clear improvement in alignment, over-warping may be occurring.

#### Lambda penalization

**Issue:** You want to constrain warp flexibility, but are unsure which `lambda` value to use.

**Remedies:** For `cc`, `lambda` in range `1e-4` to `0.01` is a reasonable starting point; higher values pull warps toward the identity, reducing over-warping at the cost of alignment precision. For `srvf`, lambda penalization has inconsistent effects across DGPs and noise levels in our benchmarks -- prefer pre-smoothing inputs (e.g., `tfb(x, k = 25)`) over lambda tuning. Note that optimal lambda values are problem-specific; the ranges above are derived from oracle (ex-post) analysis and have not been validated via cross-validation.

#### Unsuitable template

**Issue:** Registration produces poor alignment or increases variability because the default template does not represent the common curve shape.

**Remedies:** Inspect the template; supply a robust alternative such as the MBD median.

The default template for `affine` and `cc` registration is the pointwise arithmetic mean (re-estimated iteratively). When phase variation is large -- curve features are spread far apart in time -- the pointwise mean gets smeared out and no longer resembles any individual curve. A robust alternative is to use the most *central* observed curve, e.g. the curve with the highest modified band depth (MBD, see `tf_depth()`).

```{r template-choice, fig.width = 6, fig.height = 4}
# Gaussian bumps with large shifts:
s <- seq(-4, 6, length.out = 201)
mus <- c(-2, -1, 0, 1, 2)
bumps <- tfd(t(sapply(mus, \(mu) dnorm(s, mu, sd = 0.5))), arg = s)

# Pointwise mean is smeared -- not a good template:
bumps_mean <- mean(bumps)
# MBD median picks the most central observed curve:
bumps_median <- median(bumps, depth = "MBD")

# Register with default (mean) vs MBD median template:
reg_mean <- tf_register(bumps, method = "affine", type = "shift",
                        template = bumps_mean)
reg_median <- tf_register(bumps, method = "affine", type = "shift",
                          template = bumps_median)

layout(matrix(1:6, 2, 3, byrow = TRUE))
plot(bumps, col = alpha_palette, lwd = 1.5,
     main = "Original + mean template")
lines(bumps_mean, lwd = 3, lty = 2)
plot(tf_inv_warps(reg_mean), col = alpha_palette, lwd = 1.5, points = FALSE,
     main = "Inverse warps (mean template)",
     xlab = "Observed time t", ylab = "Aligned time s")
abline(0, 1, lty = 3)
plot(tf_aligned(reg_mean), col = alpha_palette, lwd = 1.5, points = FALSE,
     main = "Aligned to mean")
lines(bumps_mean, lwd = 3, lty = 2)


plot(bumps, col = alpha_palette, lwd = 1.5,
     main = "Original + MBD median template")
lines(bumps_median, lwd = 3, lty = 2, col = "red3")
plot(tf_inv_warps(reg_median), col = alpha_palette, lwd = 1.5, points = FALSE,
     main = "Inverse warps (median template)",
     xlab = "Observed time t", ylab = "Aligned time s")
abline(0, 1, lty = 3)
plot(tf_aligned(reg_median), col = alpha_palette, lwd = 1.5, points = FALSE,
     main = "Aligned to MBD median")
lines(bumps_median, lwd = 3, lty = 2, col = "red3")
```

The pointwise mean (top row, dashed) is smeared and not representative of any individual curve's shape. Registration toward it produces poor alignment. The MBD median (bottom row, dashed red) is the most central observed curve -- it has the correct shape, and registration aligns the peaks well. When you expect strong phase variation, **inspect the default template** and consider supplying a suitable custom template like `template = median(x)`.

## Practical Diagnostics Workflow

After any registration run, use `summary()` and `plot()` for a quick assessment, then dig deeper as needed:

1. **`summary(reg)`**: Check amplitude variance reduction (should be positive), aggregated warp deviations (how far from identity?), min and max warp slopes (any extreme local distortions?), and domain coverage loss (for affine warps).
2. **`plot(reg)`**: Three-panel view of original curves + template, inverse warping functions + identity line, and aligned curves + template. Check that features are more aligned and warps are plausible.
3. **Quantify alignment** of specific features you care about (e.g., spread of peak locations before vs after).
4. **Compare methods**: Re-run with at least one alternative method. If results disagree strongly, treat conclusions as low confidence.
5. **Check stability**: Re-run after modest smoothing or grid density change.

We demonstrate this workflow below using the `pinch` data.

### Step 1: Register and inspect summary

```{r diagnostics, class.source = "fold-show"}
x <- pinch[1:10]

# Register with affine shift warps:
reg_aff <- tf_register(x, method = "affine", type = "shift_scale")

# summary() gives a quick quantitative overview:
summary(reg_aff)
```

Key things to check in the summary:

- **Amplitude variance reduction** near 100% -- registration captures most of the variability. A negative value would mean registration made things *worse* (see the [Unsuitable template](#unsuitable-template) section above for an example of when this can happen).
- **Warp deviations** are moderate (well below 0.5, in most cases) -- the aggregated timing corrections per function are not too extreme.
- **Warp slopes**: For non-linear methods, slopes that veer far from 1 indicate strong local time dilations or compressions. 
- **Domain coverage loss** shows how much of the original domain is lost per curve after alignment -- relevant only for affine (non-domain-preserving) warps.

### Step 2: Visual inspection via `plot()`

```{r diagnostics-plot, fig.width = 6, fig.height = 3, class.source = "fold-show"}
# plot() provides a 3-panel diagnostic:
plot(reg_aff)
```

Check that (1) the template (dashed) looks representative of the original curve shapes, (2) inverse warping functions are not too far from the identity and have no extremely flat or steep segments, and (3) aligned curves show better feature alignment to the template.

### Step 3: Quantify alignment of specific features

For this example:
```{r diagnostics-features, class.source = "fold-show"}
# Are global peak locations more concentrated after alignment?
peak_before <- tf_where(x, value == max(value)) |> as.numeric()
peak_after_aff <- tf_where(tf_aligned(reg_aff), value == max(value)) |> as.numeric()

data.frame(
  metric = c("sd_peak_before", "sd_peak_after_affine"),
  value = c(sd(peak_before), sd(peak_after_aff))
)
```

### Step 4: Compare with an alternative method

```{r diagnostics-optional-srvf, eval = have_srvf, include = have_srvf, class.source = "fold-show"}
# Compare with SRVF (non-linear warps):
reg_srvf <- tf_register(x, method = "srvf")

# Side-by-side summaries:
summary(reg_srvf)
```

```{r diagnostics-optional-srvf-compare, eval = have_srvf, include = have_srvf, class.source = "fold-show"}
peak_after_srvf <- tf_where(tf_aligned(reg_srvf), value == max(value)) |> as.numeric()

data.frame(
  metric = c("sd_peak_before", "sd_peak_after_affine", "sd_peak_after_srvf"),
  value = c(sd(peak_before), sd(peak_after_aff), sd(peak_after_srvf))
)
```

```{r diagnostics-optional-srvf-plot, eval = have_srvf, include = have_srvf, fig.width = 6, fig.height = 4}
layout(matrix(1:6, 2, 3, byrow = TRUE))
# Affine:
plot(x, main = "Original", col = alpha_palette, lwd = 1.5)
lines(tf_template(reg_aff), lwd = 2, lty = 2)
plot(tf_inv_warps(reg_aff), main = "Affine Inverse Warps",
     col = alpha_palette, lwd = 1.5,  points = FALSE,
     xlab = "Observed time t", ylab = "Aligned time s")
abline(0, 1, lty = 3)
plot(tf_aligned(reg_aff), main = "Affine Aligned",
     col = alpha_palette, lwd = 1.5, points = FALSE)
lines(tf_template(reg_aff), lwd = 2, lty = 2)
# SRVF:
plot(x, main = "Original", col = alpha_palette, lwd = 1.5)
lines(tf_template(reg_srvf), lwd = 2, lty = 2)
plot(tf_inv_warps(reg_srvf), main = "SRVF Inverse Warps",
     col = alpha_palette, lwd = 1.5,   points = FALSE,
     xlab = "Observed time t", ylab = "Aligned time s")
abline(0, 1, lty = 3)
plot(tf_aligned(reg_srvf), main = "SRVF Aligned",
     col = alpha_palette, lwd = 1.5, points = FALSE)
lines(tf_template(reg_srvf), lwd = 2, lty = 2)
```

If different methods disagree strongly and diagnostics are unstable, treat conclusions as low confidence until you resolve data representation and model-choice sensitivity.

## Theoretical Background 101

Problem is to find an *optimal* deformation of the function's domain:

\[
\text{Find } h_i(s) \text{ so that }  d\left(x_i(h_i(s)), m(s)\right) \to \min
\]

- \(x_i(t)\): observed curve for subject \(i\) at observed times \(t\).
- \(m(s)\): the "template" function you register against.
- \(h_i(s)\): warping function mapping aligned time \(s\) ("system time")  to observed time \(t\) ("wallclock time") for the $i$-th curve. These are returned by `tf_estimate_warps()` (called internally by `tf_register()`), and registration methods differ in the representations and constraints for these (see below).
- \(h_i^{-1}(t)\): inverse warping (observed time $\to$ system time) used by `tf_align()` to align observed curves,  can be inspected/extracted via `tf_inv_warps(<tf_registration>)`.
- $d(\cdot, \cdot)$: distance or dissimilarity to minimize (e.g., $L_2$ distance of derivatives $\int (x'_i(h_i(s))h'_i(s) - m'(s))^2 ds)$).

Key references: [Marron et al. (2015)](https://doi.org/10.1214/15-EJS1030) provide an overview of amplitude and phase variation in FDA. Ramsay & Silverman (2005, *Functional Data Analysis*, Ch. 7--8) cover landmark and continuous registration. Srivastava & Klassen (2016, *Functional and Shape Data Analysis*) develop the elastic/SRVF framework in depth.

### SRVF Framework

Represents each curve via its *square root velocity function* (SRVF), $q(t) = \dot{x}(t) / \sqrt{|\dot{x}(t)|}$, and aligns curves by minimizing $L_2$ distances between (aligned) SRVFs. This corresponds to minimizing an *elastic* distance metric between functions modulo reparameterization (i.e. under "warping") and transforms the non-linear alignment problem into a simpler optimization on a Hilbert sphere^[more specifically: the positive orthant of such a sphere, which is the space of SRVFs of warping functions]. See [Srivastava et al. (2011)](https://doi.org/10.48550/arXiv.1103.3817) and [Tucker et al. (2013)](https://doi.org/10.1016/j.csda.2012.12.001) for the `fdasrvf` implementation. 

- **Optimization:** Dynamic programming over the space of diffeomorphisms^[smooth, monotone, one-to-one functions, basically.] (via `fdasrvf`).
- **Template:** Karcher mean on the shape manifold (iterative; see footnote in the [template table][What Is the Template?] above), or user-supplied.
- **Domain-preserving:** Yes -- warps satisfy $h(t_\min) = t_\min$ and $h(t_\max) = t_\max$.
- **Key arguments:** `method = "srvf"`. Pass `template` to override the Karcher mean. Control warp flexibility via `lambda` (default is `0` for unrestricted/unpenalized warps) and `penalty_method` (`"roughness"` (default), `"geodesic"`, or `"norm"`).
- **Strengths:** On clean or low-noise data, SRVF often achieves the best warp recovery for complex phase variation. Makes no assumptions about warping function shape (except domain preservation and monotonicity). Rankings shift with noise level and template shape, so always compare with at least one alternative.
- **Weaknesses:** Sensitive to noise because SRSFs involve numerical derivatives that amplify observation noise. Pre-smoothing inputs (e.g., `tfb(x, k = 25)`) before registration reduces warp error (measured as warp MISE) by 50-70% under moderate noise. Also grid-sensitive: avoid grids >200 points on noisy data (see [Sparse or irregular grids](#sparse-or-irregular-grids)). Shows the most variable performance under template estimation compared to other methods.

### CC (Continuous Registration Criterion)

Estimates smooth monotone warping functions by maximizing the integrated cross-correlation between the derivative of each registered curve and the first eigenfunction of the derivatives' covariance, which represents the dominant common shape of the sample (criterion `crit=2`, the default) or by minimizing the squared differences between aligned function and template (criterion `crit=1`, not recommended). In current `tf`, this is implemented as a tf-native dense-grid optimizer with monotone spline warps rather than the older `fda::register.fd()` backend. See [Ramsay & Li (1998, JRSS-B)](https://doi.org/10.1111/1467-9868.00129) and Ramsay & Silverman (2005, *Functional Data Analysis*, Ch. 8).

- **Optimization:** tf-native dense-grid optimization of a continuous registration criterion with monotone spline warps.
- **Template:** Arithmetic mean, re-estimated over `max_iter` iterations (default 3), or user-supplied.
- **Domain-preserving:** Yes.
- **Key arguments:** `method = "cc"`. Use `crit = 2` (default) to maximize the proportion of variance explained by the first eigenfunction of the registered sample, or `crit = 1` to minimize integrated squared differences. Control warp flexibility via B-spline basis dimension of the warping functions `nbasis` (default is `6`), their wiggliness via penalty parameter `lambda` (default is `0` for no penalization), and optimizer tolerances via `conv` and `iterlim`. In our experience, `crit = 1` without penalization tends to be considerably less reliable than `crit = 2` or penalized variants -- the unpenalized L2 criterion can produce strongly distorted warps.
- **Strengths:** The most noise-robust and grid-insensitive method in our benchmarks. Tends to be the most stable method under template estimation. Benefits from penalization (`lambda > 0`) in most conditions.
- **Weaknesses:** Results are sensitive to the criterion choice and the spline basis configuration.

### Affine (Shift / Scale)

Models warps as linear functions $h(s) = a \cdot s + b$, with three sub-types:

- `type = "shift"` (default): $a = 1$, only horizontal translation ($b$ free).
- `type = "scale"`: only uniform time scaling ($a$ free, $b$ derived from centering).
- `type = "shift_scale"`: both parameters free.

Each curve is aligned independently via bounded L-BFGS-B optimization of the L2 distance to the template. See Ramsay & Silverman (2005, Ch. 7) and [Wang & Gasser (1997)](https://doi.org/10.1214/aos/1069362747) for context on shift/scale alignment models.

- **Template:** Arithmetic mean, or user-supplied.
- **Domain-preserving:** **No** -- warps can shift or scale outside the observed domain, producing boundary `NA`s after `tf_align()`.
- **Input grid:** Supports both regular and irregular grids (`tfd_reg` and `tfd_irreg`).
- **Key arguments:** `method = "affine"`, `type`, `shift_range` and `scale_range` to set upper and lower limits for time shifts $b_i$ and time scales $a_i$.
- **Strengths:** Very fast (typically an order of magnitude or more faster than SRVF), highly interpretable parameters.
- **Weaknesses:** Cannot capture non-linear timing deformation. Results are often shortened functions with substantial leading and/or trailing NAs (under strong shifts/scales).

### Landmark

Constructs piecewise linear warps by mapping user-specified landmark positions to target positions. No continuous optimization is performed -- the warp is fully determined by the landmark correspondence. See Kneip & Gasser (1992, Annals of Statistics; [pdf](https://projecteuclid.org/journals/annals-of-statistics/volume-20/issue-3/Statistical-Tools-to-Analyze-Data-Representing-a-Sample-of-Curves/10.1214/aos/1176348769.pdf)) and Ramsay & Silverman (2005, Ch. 7).

- **Template:** Defined by column-wise means of the landmark time points matrix, or user-supplied `template_landmarks`.
- **Domain-preserving:** Yes -- domain endpoints are used as boundary anchors.
- **Input grid:** Supports both regular and irregular grids (`tfd_reg` and `tfd_irreg`).
- **Key arguments:** `method = "landmark"`, `landmarks` (required: $n \times k$ matrix of landmark positions), `template_landmarks` (optional target positions).
- **Strengths:** Fast and deterministic. In our benchmarks, landmark registration was most robust to outlier contamination: warp MISE degraded ~1.6x at 30% contamination, compared to 2.6-3.7x for iterative methods.
- **Weaknesses:** Requires the ability to identify reliable, consistent landmarks across all curves. Fragile when features are ambiguous and often too simplistic if curves have few identifiable features (e.g., only one peak per curve).
- **Landmark detection:** `tf_landmarks_extrema()` can automatically detect local maxima, minima, or zero crossings and cluster them across curves - usually better to pre-smooth noisy inputs with `tf_smooth()` or `tfb` before calling this function. Results will typically be better for user-defined landmarks that are based on domain knowledge.

## Worked Examples

### Pinch Data: Cross-method Comparison

```{r pinch-compare, class.source = "fold-show"}
pinch_small <- pinch[1:10]

reg_aff <- tf_register(pinch_small, method = "affine", type = "shift_scale")
inv_warp_aff <- tf_inv_warps(reg_aff)

# tf_landmarks_extrema needs smoothed inputs,
#     otherwise it tends to detect lots of spurious features:
pinch_small |> tf_landmarks_extrema(which = "max") |> head()
#     ... so in this case, we simply use the global maximum for each curve:
(peak_locs <- pinch_small |> tf_where(value == max(value)) |> unlist() |> as.matrix())
reg_lm <- tf_register(pinch_small, method = "landmark", landmarks = peak_locs)
inv_warp_lm <- tf_inv_warps(reg_lm)

# ... but using more than one peak location as the only landmark will get us
# better alignment here - use times where curve first & last exceeds 3 as well:
pinch_landmarks  <- cbind(
  start = pinch_small |> tf_where(value > 3, return = "first") |> unlist(),
  peak  = peak_locs,
  end   = pinch_small |> tf_where(value > 3, return = "last") |> unlist()
)
pinch_landmarks |> head()
reg_lm2 <- tf_register(pinch_small, method = "landmark", landmarks = pinch_landmarks)
inv_warp_lm2 <- tf_inv_warps(reg_lm2)
```


```{r pinch-compare-plot, fig.width = 6, fig.height = 6}
layout(t(matrix(1:12, 4, 3)))
par(cex.main = 0.8)
plot.new()
plot(inv_warp_aff, main = "Affine Inverse Warps", ylab = "",
     points = FALSE, col = alpha_palette, lwd = 1.5)
abline(0, 1, lty = 3)
plot(tf_aligned(reg_aff), main = "Affine Registered",
     col = alpha_palette, lwd = 1.5, points = FALSE)
plot(tf_aligned(reg_aff), main = "Affine Registered",
     type = "lasagna", col = hcl.colors(12, rev = TRUE))

plot(pinch_small, main = "Original",
     col = alpha_palette, lwd = 1.5)
plot(inv_warp_lm, main = "Landmark Inverse Warps \n (Peak only)", ylab = "",
     points = FALSE, col = alpha_palette, lwd = 1.5)
abline(0, 1, lty = 3)
plot(tf_aligned(reg_lm), main = "Landmark Registered\n (Peak only)", col = alpha_palette, lwd = 1.5)
plot(tf_aligned(reg_lm), main = "Landmark Registered\n (Peak only)", type = "lasagna", col = hcl.colors(12, rev = TRUE))

plot.new()
plot(inv_warp_lm2, main = "Landmark Inverse Warps \n (Start + Peak + End)", ylab = "",
     points = FALSE, col = alpha_palette, lwd = 1.5)
abline(0, 1, lty = 3)
plot(tf_aligned(reg_lm2), main = "Landmark Registered\n (Start + Peak + End)", col = alpha_palette, lwd = 1.5)
plot(tf_aligned(reg_lm2), main = "Landmark Registered\n (Start + Peak + End)", type = "lasagna", col = hcl.colors(12, rev = TRUE))
```

Note that affine registration with `shift+scale` warps produces some boundary `NA`s because the warps are not domain-preserving, but it does a good job aligning the peaks. Landmark registration with only the peak locations does a decent job aligning the peaks without producing `NA`s, but it does not align the start and end of the curves well. Landmark registration with start + peak + end landmarks does a good job aligning all three features without producing `NA`s. 

```{r pinch-compare-srvf-cc, eval = have_srvf && have_cc, class.source = "fold-show"}
reg_srvf <- tf_register(pinch_small, method = "srvf")
inv_warp_srvf <- tf_inv_warps(reg_srvf)
reg_cc_unpen <- tf_register(pinch_small, method = "cc", max_iter = 10, nbasis = 10, crit = 1)
inv_warp_cc_unpen <- tf_inv_warps(reg_cc_unpen)
reg_cc_pen <- tf_register(pinch_small, method = "cc", lambda = 1e-4, max_iter = 20)
inv_warp_cc_pen <- tf_inv_warps(reg_cc_pen)
```

SRVF and CC both use non-linear warps and tend to be able to align more complex phase variation than the `affine` method, but their warps are more complex and potentially harder to interpret, and registration can go off the rails more easily (as we can see here with the unpenalized `cc` registration plotted below, check the huge maximal warp slope and its poor amplitude variance reduction...) --
using `summary(<tf_registration>)` for quick quantitative diagnostics:
```{r pinch-compare-srvf-cc-summary, eval = have_srvf && have_cc, class.source = "fold-show"}
summary(reg_cc_unpen)
# ... ouch! max slope almost 100 and only 40% amplitude variance reduction ....

summary(reg_srvf)
summary(reg_cc_pen)
```

```{r pinch-compare-srvf-cc-plot, eval = have_srvf && have_cc, fig.width = 6, fig.height = 6}
layout(t(matrix(1:12, 4, 3)))
par(cex.main = 0.8)
plot.new()
plot(inv_warp_srvf, main = "SRVF Inverse Warps", ylab = "",
     points = FALSE, col = alpha_palette, lwd = 1.5)
abline(0, 1, lty = 3)
plot(tf_aligned(reg_srvf), main = "SRVF Registered",
     col = alpha_palette, lwd = 1.5, points = FALSE)
plot(tf_aligned(reg_srvf), main = "SRVF Registered",
     type = "lasagna", col = hcl.colors(12, rev = TRUE))

plot(pinch_small, main = "Original",
     col = alpha_palette, lwd = 1.5)
plot(inv_warp_cc_unpen, main = "CC Inverse Warps (unpen.)", ylab = "",
     points = FALSE, col = alpha_palette, lwd = 1.5)
abline(0, 1, lty = 3)
plot(tf_aligned(reg_cc_unpen), main = "CC Registered (unpen.)",
     col = alpha_palette, lwd = 1.5, points = FALSE)
plot(tf_aligned(reg_cc_unpen), main = "CC Registered (unpen.)",
     type = "lasagna", col = hcl.colors(12, rev = TRUE))

plot.new()
plot(inv_warp_cc_pen, main = expression("CC Inverse Warps (" * lambda * " = 1e-4)"),
     ylab = "", points = FALSE, col = alpha_palette, lwd = 1.5)
abline(0, 1, lty = 3)
plot(tf_aligned(reg_cc_pen), main = expression("CC Registered (" * lambda * " = 1e-4)"),
     col = alpha_palette, lwd = 1.5, points = FALSE)
plot(tf_aligned(reg_cc_pen), main = expression("CC Registered (" * lambda * " = 1e-4)"),
     type = "lasagna", col = hcl.colors(12, rev = TRUE))
```

### Growth Data: Deeper Real-data Workflow

The Berkeley growth data contains height measurements for 39 boys and 54 girls aged 1--18. Growth velocity curves (first derivatives of height) show a prominent pubertal growth spurt whose timing varies substantially between individuals -- a natural target for registration. We use the subset of girls^[.. because aligning female *and* male growth curves to the same template does not make sense -- they show different patterns!] from this dataset to illustrate how data representation, penalization, and landmark choice affect registration quality.

```{r growth-prep, class.source = "fold-show"}
growth <- tf::growth |> dplyr::filter(gender == "female")

# Raw velocity via finite differences -- noisy, only 30 midpoints from 31 measurements:
growth$raw_vel <- tf_derive(growth$height)

# Smooth velocity via spline representation on a much denser grid, then derive analytically:
growth$smooth_vel <- tfb(
  growth$height,
  k = 15,
  bs = "tp",
  arg = seq(1, 18, l = 80),
  global = TRUE
  # family = gaussian(link = "log") # ensures positivity of velocity
) |>
  tf_derive()
```

#### Data representation matters

The raw finite-difference velocity estimates are jagged (30 irregularly spaced points from the original measurement grid). SRVF registration on such noisy inputs tends to produce wiggly warps that chase noise rather than genuine phase variation. Converting to a smooth representation first gives SRVF cleaner input and much better results.

```{r growth-raw-vs-smooth, eval = have_srvf, include = have_srvf, class.source = "fold-show"}
# SRVF on raw (noisy) velocity:
reg_raw_obj <- tf_register(growth$raw_vel, method = "srvf")
inv_warp_raw <- tf_inv_warps(reg_raw_obj)
reg_raw <- tf_aligned(reg_raw_obj)

# SRVF on smooth velocity:
reg_smooth_obj <- tf_register(growth$smooth_vel, method = "srvf")
inv_warp_smooth <- tf_inv_warps(reg_smooth_obj)
reg_smooth <- tf_aligned(reg_smooth_obj)
```

```{r growth-raw-vs-smooth-plot, eval = have_srvf, include = have_srvf, fig.width = 6, fig.height = 4}
reg_brks <- range(c(tf_evaluations(reg_raw), tf_evaluations(reg_smooth))) |>
  (\(x) seq(x[1], x[2], l = 13))()

layout(t(matrix(1:8, 4, 2)))
par(cex.main = 0.8)

plot(growth$raw_vel, main = "Raw Velocity (finite diff.)",
     xlab = "Age [years]", ylab = "cm/year", lwd = 0.8, ylim = c(0, 30))
plot(inv_warp_raw, points = FALSE, main = "SRVF Inverse Warps (raw)",
     xlab = "Chronological Age", ylab = "Registered Age", lwd = 0.8)
abline(0, 1, lty = 3)
plot(reg_raw, main = "SRVF Registered (raw)",
     xlab = "Registered Age [years]", ylab = "cm/year", lwd = 0.8, ylim = c(0, 30))
plot(reg_raw, main = "SRVF Registered (raw)",
     type = "lasagna", col = hcl.colors(12, rev = TRUE), breaks = reg_brks)

plot(growth$smooth_vel, main = "Smooth Velocity (tfb deriv.)",
     xlab = "Age [years]", ylab = "cm/year", lwd = 0.8, ylim = c(0, 30))
plot(inv_warp_smooth, points = FALSE, main = "SRVF Inverse Warps (smooth)",
     xlab = "Chronological Age", ylab = "Registered Age", lwd = 0.8)
abline(0, 1, lty = 3)
plot(reg_smooth, main = "SRVF Registered (smooth)",
     xlab = "Registered Age [years]", ylab = "cm/year", lwd = 0.8, ylim = c(0, 30))
plot(reg_smooth, main = "SRVF Registered (smooth)",
     type = "lasagna", col = hcl.colors(12, rev = TRUE), breaks = reg_brks)
```

The top row shows extreme, biologically implausible warps (i.e. mapping chronological ages 5-8 to "registered" ages < 2) with local wiggles that are more likely to reflect noise in the raw velocity estimates rather than genuine timing differences. The bottom row, using smoother velocity curves, produces simpler and less extreme warps that capture the pubertal timing variation -- the main feature of interest.

#### Penalization

Even on noisy inputs, SRVF penalization (`lambda > 0`) can help suppress over-warping. Higher `lambda` pulls warps toward the identity (no warping), trading less strict alignment for smoother and more subtle  warps.

```{r growth-penalization, eval = have_srvf, include = have_srvf, class.source = "fold-show"}
# Penalized SRVF on the raw velocity:
reg_raw_pen_obj <- tf_register(growth$raw_vel, method = "srvf", lambda = 0.1)
inv_warp_raw_pen <- tf_inv_warps(reg_raw_pen_obj)
reg_raw_pen <- tf_aligned(reg_raw_pen_obj)
```

```{r growth-penalization-plot, eval = have_srvf, include = have_srvf, fig.width = 6, fig.height = 3}
layout(t(matrix(1:4, 4, 1)))
par(cex.main = 0.8)
plot(growth$raw_vel, main = "Raw Velocity",
     xlab = "Age [years]", ylab = "cm/year", lwd = 0.8)
plot(inv_warp_raw_pen, points = FALSE,
     main = expression("Penalized Warps (raw, " * lambda * " = 0.1)"),
     xlab = "Chronological Age", ylab = "Registered Age", lwd = 0.8)
abline(0, 1, lty = 3)
plot(reg_raw_pen,
     main = expression("Penalized Registered (raw, " * lambda * " = 0.1)"),
     xlab = "Registered Age [years]", ylab = "cm/year", lwd = 0.8)
plot(reg_raw_pen, main = "Penalized Registered (raw)",
     type = "lasagna", col = hcl.colors(12, rev = TRUE), breaks = reg_brks)
```

Penalization smooths out the worst warp artifacts, and the result here looks more reasonable than the result for the (oversmoothed) `tfb`-inputs above. 
That said, in most cases the best strategy is likely to be to (carefully!) pre-smooth, and then to use penalization only if the warps still look too flexible.

#### Landmark registration

Landmark registration avoids continuous optimization entirely -- it constructs piecewise linear warps from a (user-specified) correspondence of curve features. For growth velocity, natural landmarks are the *end of rapid infant growth* (velocity drops a lot below initial velocity), *pre-pubertal trough* (minimum velocity before the puberty growth spurt), the *pubertal peak* (maximum velocity during puberty). We identify these by searching for the minimum in age 5--(peak-1) and the maximum in age 8--17 on each smooth velocity curve.

```{r growth-landmarks, class.source = "fold-show"}
# End of rapid infant growth: 
#    less than 2/3 of max early childhood growth (1-3) velocity before age 5
growth_slows <- growth$smooth_vel |> tf_zoom(begin = 1, end = 5) |> 
  tf_where(value < 0.66 * max(value[1:10]), return = "first") |> unlist()

# Pubertal peak: maximum velocity in age 8--17
growth_peaks <- growth$smooth_vel |> tf_zoom(begin = 8, end = 17) |> 
  tf_where(value == max(value)) |> unlist()

# Pre-pubertal trough: minimum velocity between age 5 and 1 year before the peak
growth_troughs <- growth$smooth_vel |> tf_zoom(begin = 5, end = growth_peaks - 1) |> 
  tf_where(value == min(value)) |> unlist() 

# Build landmark matrix and register:
growth_lm <- cbind(slowdown = growth_slows, trough = growth_troughs, peak = growth_peaks)
reg_lm_obj <- tf_register(growth$smooth_vel, method = "landmark", landmarks = growth_lm)
inv_warp_lm <- tf_inv_warps(reg_lm_obj)
reg_lm <- tf_aligned(reg_lm_obj)
```

```{r growth-landmarks-plot, fig.width = 6, fig.height = 3}
if (!exists("reg_brks")) {
  reg_brks <- range(c(tf_evaluations(growth$smooth_vel), tf_evaluations(reg_lm))) |>
    (\(x) seq(x[1], x[2], l = 13))()
}
layout(t(1:4))
plot(growth$smooth_vel, main = "Smooth Velocity",
     xlab = "Age [years]", ylab = "cm/year", lwd = 0.8)
plot(inv_warp_lm, points = FALSE, main = "Landmark Inverse Warps",
     xlab = "Chronological Age", ylab = "Registered Age", lwd = 0.8)
abline(0, 1, lty = 3)
plot(reg_lm, main = "Landmark Registered",
     xlab = "Registered Age [years]", ylab = "cm/year", lwd = 0.8)
plot(reg_lm, main = "Landmark Registered",
     type = "lasagna", col = hcl.colors(12, rev = TRUE), breaks = reg_brks)
```

The piecewise linear warps align the pubertal peaks well, with the pubertal trough and infant growth landmarks anchoring the earlier phases. Compared to SRVF, the warps are simpler and fully interpretable -- each segment directly corresponds to a physiological period (early childhood, late childhood, pubertal acceleration, post-peak deceleration). The trade-off is that landmark registration cannot adapt between landmarks (alignment between successive landmarks depends entirely on those two anchor points) and that landmark definition is often somewhat arbitrary (c.f. definition of "end of rapid infant growth" above...).

```{r restore-par, include = FALSE}
par(oldpar)
```

## References

- Kneip, A., & Gasser, T. (1992). Statistical tools to analyze data representing a sample of curves. *Annals of Statistics*, 20(3), 1266--1305. DOI 10.1214/aos/1176348794 ([pdf](https://projecteuclid.org/journals/annals-of-statistics/volume-20/issue-3/Statistical-Tools-to-Analyze-Data-Representing-a-Sample-of-Curves/10.1214/aos/1176348769.pdf))
- Marron, J.S., Ramsay, J.O., Sangalli, L.M., & Srivastava, A. (2015). Functional data analysis of amplitude and phase variation. *Statistical Science*, 30(4), 468--484. [doi:10.1214/15-EJS1030](https://doi.org/10.1214/15-EJS1030)
- Ramsay, J.O., & Li, X. (1998). Curve registration. *Journal of the Royal Statistical Society: Series B*, 60(2), 351--363. [doi:10.1111/1467-9868.00129](https://doi.org/10.1111/1467-9868.00129)
- Ramsay, J.O., & Silverman, B.W. (2005). *Functional Data Analysis* (2nd ed.). Springer.
- Srivastava, A., & Klassen, E.P. (2016). *Functional and Shape Data Analysis*. Springer.
- Srivastava, A., Wu, W., Kurtek, S., Klassen, E., & Marron, J.S. (2011). Registration of functional data using Fisher-Rao metric. [arXiv:1103.3817](https://doi.org/10.48550/arXiv.1103.3817)
- Tucker, J.D., Wu, W., & Srivastava, A. (2013). Generative models for functional data using phase and amplitude separation. *Computational Statistics & Data Analysis*, 61, 50--66. [doi:10.1016/j.csda.2012.12.001](https://doi.org/10.1016/j.csda.2012.12.001)
- Wang, K., & Gasser, T. (1997). Alignment of curves by dynamic time warping. *Annals of Statistics*, 25(3), 1251--1276. [doi:10.1214/aos/1069362747](https://doi.org/10.1214/aos/1069362747)

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tf-ggplot.R
\name{build_tf_layer_data}
\alias{build_tf_layer_data}
\title{Build long-format data and new mapping for a single tf layer}
\usage{
build_tf_layer_data(
  layer_info,
  plot_tf_aes,
  scalar_col_map,
  layer_idx,
  enriched_data,
  user_arg,
  interpolate
)
}
\arguments{
\item{layer_info}{List from \code{all_layers} with \code{layer}, \code{parsed_aes}, \code{is_tf_layer}}

\item{plot_tf_aes}{tf aesthetics from the plot level (used when \code{inherit.aes = TRUE})}

\item{scalar_col_map}{Named list: key -> column name in enriched_data}

\item{layer_idx}{Integer index of this layer (for keying layer-level scalar aes)}

\item{enriched_data}{Data frame with original data + pre-evaluated scalar tf columns}

\item{user_arg}{Optional evaluation grid (overrides natural grid)}

\item{interpolate}{Whether to interpolate tf objects to \code{arg}}
}
\value{
\code{NULL} when no effective tf aes, otherwise \code{list(long_data, new_mapping)}
}
\description{
Evaluates the layer's tf aesthetics independently on their own natural grid
(or the user-specified grid). Called by \code{\link[=finalize_tf_ggplot]{finalize_tf_ggplot()}} per layer.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tf-ggplot.R
\name{finalize_tf_ggplot}
\alias{finalize_tf_ggplot}
\title{Finalize tf_ggplot by processing all tf layers independently}
\usage{
finalize_tf_ggplot(tf_plot)
}
\arguments{
\item{tf_plot}{A tf_ggplot object with accumulated layers}
}
\value{
Regular ggplot object with all layers properly transformed
}
\description{
Each layer is evaluated on its own natural argument grid (Option B architecture).
The base ggplot has \code{data = NULL}; each layer receives its own long-format data.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/autoplot.R
\name{autoplot.tf}
\alias{autoplot.tf}
\alias{autolayer.tf}
\title{Autoplot and autolayer methods for \code{tf} objects}
\usage{
\method{autoplot}{tf}(object, ...)

\method{autolayer}{tf}(object, ...)
}
\arguments{
\item{object}{a \code{tf} object}

\item{...}{passed to \code{\link[ggplot2:geom_path]{ggplot2::geom_line()}}}
}
\value{
A \code{\link[=tf_ggplot]{tf_ggplot()}} object for \code{autoplot()}, a \code{\link[ggplot2:layer]{ggplot2::layer()}} object for \code{autolayer()}.
}
\description{
Convenient plotting methods for \code{tf} objects. \code{autoplot()} creates a
complete spaghetti plot, \code{autolayer()} creates a layer that can be
added to an existing \code{\link[ggplot2:ggplot]{ggplot2::ggplot()}} or \code{\link[=tf_ggplot]{tf_ggplot()}}.
}
\examples{
\donttest{
library(ggplot2)
f <- tf_rgp(5)
autoplot(f)
ggplot() + autolayer(f)
tf_ggplot() + autolayer(f)
}
}
\seealso{
Other tidyfun visualization: 
\code{\link{ggcapellini}},
\code{\link{gglasagna}()},
\code{\link{ggspaghetti}}
}
\concept{tidyfun visualization}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/datasets.R
\docType{data}
\name{dti_df}
\alias{dti_df}
\title{Diffusion tensor imaging data}
\format{
A tibble with 382 rows and 6 columns:
\describe{
\item{id}{(numeric) Subject identifier.}
\item{visit}{(integer) Visit number.}
\item{sex}{(factor) \code{"male"} or \code{"female"}.}
\item{case}{(factor) \code{"control"} or \code{"MS"} (multiple sclerosis status).}
\item{cca}{(\code{tfd_irreg}) FA tract profiles for the corpus callosum
(up to 93 evaluation points, domain 0--1).}
\item{rcst}{(\code{tfd_irreg}) FA tract profiles for the right corticospinal
tract (up to 55 evaluation points, domain 0--1).}
}
}
\source{
Data are from the Johns Hopkins University and the Kennedy-Krieger Institute.
Also available in a different format as \link[refund:DTI]{refund::DTI}.
}
\usage{
dti_df
}
\description{
Fractional anisotropy (FA) tract profiles for the corpus callosum (\code{cca})
and the right corticospinal tract (\code{rcst}) from a diffusion tensor imaging
(DTI) study of multiple sclerosis patients and healthy controls.

The original data in \link[refund:DTI]{refund::DTI} include additional variables (\code{pasat},
\code{Nscans}, \code{visit.time}) that were not carried over here.
}
\details{
If you use this data as an example in written work, please include the
following acknowledgment: \emph{"The MRI/DTI data were collected at Johns Hopkins
University and the Kennedy-Krieger Institute."}
}
\examples{
dti_df

library(ggplot2)
dti_df |>
  dplyr::filter(visit == 1) |>
  tf_ggplot(aes(tf = cca, color = case)) +
  geom_line(alpha = 0.3)
}
\references{
Goldsmith, J., Bobb, J., Crainiceanu, C., Caffo, B., and Reich, D. (2011).
Penalized Functional Regression. \emph{Journal of Computational and Graphical
Statistics}, 20(4), 830--851. \doi{10.1198/jcgs.2010.10007}

Goldsmith, J., Crainiceanu, C., Caffo, B., and Reich, D. (2012).
Longitudinal Penalized Functional Regression for Cognitive Outcomes on
Neuronal Tract Measurements. \emph{Journal of the Royal Statistical Society:
Series C}, 61(3), 453--469. \doi{10.1111/j.1467-9876.2011.01031.x}
}
\seealso{
\link{chf_df} for another example dataset,
\code{vignette("x04_Visualization", package = "tidyfun")} for usage examples.

Other tidyfun datasets: 
\code{\link{chf_df}}
}
\concept{tidyfun datasets}
\keyword{data}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tf-ggplot.R
\name{tf_ggplot}
\alias{tf_ggplot}
\title{Create a tf-aware ggplot}
\usage{
tf_ggplot(data = NULL, mapping = aes(), ..., arg = NULL, interpolate = TRUE)
}
\arguments{
\item{data}{Default dataset to use for plot. If not provided, must be supplied
in each layer added to the plot.}

\item{mapping}{Default list of aesthetic mappings to use for plot. Can include
tf-specific aesthetics like \code{tf}, \code{tf_x}, \code{tf_y}, \code{tf_ymin}, \code{tf_ymax}.}

\item{...}{Other arguments passed to ggplot2 functions.}

\item{arg}{Optional. Evaluation grid for tf objects. A numeric vector of arg
values, or a single integer specifying the desired grid length (resolved to
an equidistant grid over the tf domain). If \code{NULL} (default), uses the
natural grid of the tf objects.}

\item{interpolate}{Logical. Should tf objects be interpolated to the evaluation
grid? Defaults to TRUE.}
}
\value{
A tf_ggplot object that inherits from ggplot
}
\description{
\code{tf_ggplot()} creates a ggplot object that can handle tf (functional data) aesthetics.
It works similarly to \code{ggplot()} but automatically transforms tf objects into
long-format data suitable for standard ggplot2 geoms.
}
\details{
\code{tf_ggplot} supports the following tf-specific aesthetics:
\itemize{
\item \code{tf}: Maps a \code{tf} object to \code{y} aesthetic (shorthand for \code{tf_y})
\item \code{tf_x}: Maps a \code{tf} object to \code{x} aesthetic
\item \code{tf_y}: Maps a \code{tf} object to \code{y} aesthetic
\item \code{tf_ymin}: Maps a tf object to ymin aesthetic (for ribbons)
\item \code{tf_ymax}: Maps a tf object to ymax aesthetic (for ribbons)
}

When tf aesthetics are used, the data is automatically transformed:
\itemize{
\item tf objects are evaluated on a common grid
\item Each function becomes multiple rows (one per evaluation point)
\item Group identifiers are created to maintain function identity
\item Non-tf columns are replicated appropriately
}
}
\examples{
# Basic usage
data <- data.frame(
id = 1:10,
group = sample(c("A", "B"), 10, replace = TRUE)
)
data$f <- tf_rgp(10)

# Method 1: tf aesthetic in constructor
tf_ggplot(data, ggplot2::aes(tf = f, color = group)) + ggplot2::geom_line()

# Method 2: tf aesthetic in geom (equivalent)
tf_ggplot(data) + ggplot2::geom_line(ggplot2::aes(tf = f, color = group))

# Confidence bands
tf_ggplot(data) +
  ggplot2::geom_ribbon(
    ggplot2::aes(tf_ymin = mean(f) - sd(f), tf_ymax = mean(f) + sd(f)),
    alpha = 0.3
  ) +
  ggplot2::geom_line(ggplot2::aes(tf = mean(f)))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tf-ggplot.R
\name{parse_tf_aesthetics}
\alias{parse_tf_aesthetics}
\title{Parse aesthetic mappings to separate tf and regular aesthetics}
\usage{
parse_tf_aesthetics(mapping, data = NULL)
}
\arguments{
\item{mapping}{An aesthetic mapping created with \code{\link[ggplot2:aes]{ggplot2::aes()}}.}

\item{data}{Data frame to evaluate expressions against.}
}
\value{
A list with components \code{tf_aes}, \code{scalar_tf_aes}, and \code{regular_aes}.
}
\description{
Parse aesthetic mappings to separate tf and regular aesthetics
}
\examples{
parse_tf_aesthetics(ggplot2::aes(tf = f, color = group))
parse_tf_aesthetics(ggplot2::aes(x = x, y = y))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tf-ggplot.R
\name{print.tf_ggplot}
\alias{print.tf_ggplot}
\title{Print method for tf_ggplot}
\usage{
\method{print}{tf_ggplot}(x, ...)
}
\arguments{
\item{x}{A tf_ggplot object}

\item{...}{Additional arguments}
}
\value{
\code{x}, invisibly. Called for its side effect of printing the plot.
}
\description{
Print method for tf_ggplot
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tidyr.R
\name{tf_nest}
\alias{tf_nest}
\title{Turn "long" tables into tidy data frames with \code{tf}-objects}
\usage{
tf_nest(
  data,
  ...,
  .id = "id",
  .arg = "arg",
  domain = NULL,
  evaluator = "tf_approx_linear"
)
}
\arguments{
\item{data}{a data frame}

\item{...}{A selection of columns. If empty, all variables except the
\code{.id} and \code{.arg} columns are selected. You can supply bare variable names,
select all variables between \code{x} and \code{z} with \code{x:z}, exclude \code{y} with \code{-y}.
For more options, see the \code{\link[dplyr:select]{dplyr::select()}} documentation.}

\item{.id}{the (bare or quoted) name of the column defining the different
observations. Defaults to "id".}

\item{.arg}{the (bare or quoted) name of the column defining the \code{arg}-values
of the observed functions. Defaults to "arg".}

\item{domain}{optional. Range of possible \code{arg}-values. See \code{\link[tf:tfd]{tf::tfd()}} for details.}

\item{evaluator}{optional. A function accepting arguments x, arg, evaluations. See \code{\link[tf:tfd]{tf::tfd()}} for details.}
}
\value{
a data frame with (at least) \code{.id} and \code{tfd} columns
}
\description{
Similar in spirit to \code{\link[tidyr:nest]{tidyr::nest()}}. This turns tables in "long" format,
where one column (\code{.id}) defines the unit of observation, one column (\code{.arg})
defines the evaluation grids of the functional observations, and other columns (\code{...})
define the values of the functions at those points into a (much shorter) table containing
\code{tfd}-objects. All other variables are checked for constancy over \code{.id} and
appended as well.
}
\details{
\code{domain} and \code{evaluator} can be specified as lists or vectors
if you are nesting multiple functional data columns with different properties.
Because this interface captures evaluator names as text, supply the evaluator
as a string rather than a bare function name.
}
\examples{
d <- dplyr::tibble(id = rep(1:3, each = 5), arg = rep(1:5, 3), value = rnorm(15))
tf_nest(d, .id = id, .arg = arg)
}
\seealso{
\code{\link[tf:tfd]{tf::tfd()}} for details on \code{domain} and \code{evaluator}.

Other tidyfun data wrangling functions: 
\code{\link{tf_evaluate.data.frame}()},
\code{\link{tf_gather}()},
\code{\link{tf_spread}()},
\code{\link{tf_unnest}()}
}
\concept{tidyfun data wrangling functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tf-ggplot.R
\name{is_tf_ggplot}
\alias{is_tf_ggplot}
\title{Check if object is a tf_ggplot}
\usage{
is_tf_ggplot(x)
}
\arguments{
\item{x}{Object to test}
}
\value{
\code{TRUE} if \code{x} inherits from \code{"tf_ggplot"}, \code{FALSE} otherwise.
}
\description{
Check if object is a tf_ggplot
}
\examples{
p <- tf_ggplot(data.frame(x = 1))
is_tf_ggplot(p)
is_tf_ggplot(ggplot2::ggplot())
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tf-ggplot.R
\name{+.tf_ggplot}
\alias{+.tf_ggplot}
\title{Add layers to tf_ggplot objects}
\usage{
\method{+}{tf_ggplot}(e1, e2)
}
\arguments{
\item{e1}{A tf_ggplot object}

\item{e2}{A ggplot2 layer, scale, theme, etc.}
}
\value{
A modified \code{tf_ggplot} object.
}
\description{
Add layers to tf_ggplot objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pillar.R
\name{type_sum.tf}
\alias{type_sum.tf}
\alias{obj_sum.tf}
\alias{pillar_shaft.tf}
\title{Format tidy functional data for tibbles}
\usage{
\method{type_sum}{tf}(x, ...)

\method{obj_sum}{tf}(x)

\method{pillar_shaft}{tf}(x, ...)
}
\arguments{
\item{x}{object of class tf}

\item{...}{ignored}
}
\description{
Summarize tidy functional data for tibble
}
\details{
see \code{\link[pillar:type_sum]{pillar::type_sum()}}
}
\concept{tidyfun print}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tf-ggplot.R
\name{translate_old_tf_layer}
\alias{translate_old_tf_layer}
\title{Translate a geom_spaghetti/geom_meatballs/geom_errorband layer for tf_ggplot}
\usage{
translate_old_tf_layer(layer, e1)
}
\arguments{
\item{layer}{A LayerInstance using StatTf or StatErrorband}

\item{e1}{The tf_ggplot object (may be modified to clean up plot-level y)}
}
\value{
A list with \code{plot_obj} (possibly modified e1) and \code{layers} (list of
new layers to add)
}
\description{
Remaps old-style y/ymin/ymax tf aesthetics to tf/tf_ymin/tf_ymax and
substitutes the appropriate standard ggplot2 geom.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tf-ggplot.R
\name{add_tf_aes_to_mapping}
\alias{add_tf_aes_to_mapping}
\title{Add a tf aesthetic to an ggplot2 mapping object}
\usage{
add_tf_aes_to_mapping(mapping, aes_name, val_col, arg_col, id_col)
}
\description{
Add a tf aesthetic to an ggplot2 mapping object
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/evaluate-dataframe.R
\name{tf_evaluate.data.frame}
\alias{tf_evaluate.data.frame}
\title{Evaluate \code{tf}s inside a \code{data.frame}}
\usage{
\method{tf_evaluate}{data.frame}(object, ..., arg)
}
\arguments{
\item{object}{a \code{data.frame}-like object with \code{tf} columns.}

\item{...}{optional: a selection of \code{tf}-columns. If empty, all \code{tf}-variables
in the data frame are selected. You can supply bare variable names,
select all variables between \code{x} and \code{z} with \code{x:z}, exclude \code{y} with \code{-y}.
For more options, see the \code{\link[dplyr:select]{dplyr::select()}} documentation.}

\item{arg}{optional evaluation grid (vector or list of vectors).
Defaults to \code{tf_arg(object)}.}
}
\value{
Replaces \code{tf}-columns with list columns of
smaller \code{data.frames} containing the functions' arguments (\code{arg}) and
evaluations (\code{value}) and returns the modified nested dataframe.
}
\description{
Evaluate \code{tf}s inside a \code{data.frame}
}
\details{
The \code{arg}-argument of \code{tf_evaluate.data.frame} method can be a
list of \code{arg}-vectors or -lists used as the \code{arg} argument for the
\code{\link[tf:tf_evaluate]{tf::tf_evaluate()}}-method for the respective \code{tf}-columns in \code{object}.
\code{...} is not used for a \code{tf}-\code{object}, but a second unnamed argument to these
methods will be interpreted as \code{arg}.
}
\examples{
d <- dplyr::tibble(id = 1:3)
d$f <- tf_rgp(3, 11L)
str(tf_evaluate(d))
}
\seealso{
Other tidyfun data wrangling functions: 
\code{\link{tf_gather}()},
\code{\link{tf_nest}()},
\code{\link{tf_spread}()},
\code{\link{tf_unnest}()}
}
\concept{tidyfun data wrangling functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/datasets.R
\docType{data}
\name{chf_df}
\alias{chf_df}
\title{Congestive heart failure accelerometry data}
\format{
A tibble with 329 rows and 8 columns:
\describe{
\item{id}{(numeric) Subject identifier.}
\item{gender}{(character) \code{"Male"} or \code{"Female"}.}
\item{age}{(numeric) Age in years.}
\item{bmi}{(numeric) Body mass index.}
\item{event_week}{(numeric) Week of cardiac event.}
\item{event_type}{(character) Type of cardiac event.}
\item{day}{(ordered factor) Day of the week (\code{Mon} < \code{Tue} < ... < \code{Sun}).}
\item{activity}{(\code{tfd_reg}) Minute-by-minute activity counts over
a 24-hour period (arg domain 1--1440).}
}
}
\source{
Data are from a study of physical activity in CHF patients conducted by
Huang et al. The original data are publicly available; see the paper for
download details.
}
\usage{
chf_df
}
\description{
Activity data from a study of congestive heart failure (CHF) patients. Data
were originally presented in Huang et al. (2019); these data are publicly
available, with download information in the paper.
}
\examples{
chf_df

library(ggplot2)
chf_df |>
  dplyr::filter(id \%in\% 1:5) |>
  gglasagna(activity, order_by = mean)
}
\references{
Huang, L., Bai, J., Ivanescu, A., Harris, T., Maurer, M., Green, P., and
Zipunnikov, V. (2019). Multilevel Matrix-Variate Analysis and its
Application to Accelerometry-Measured Physical Activity in Clinical
Populations. \emph{Journal of the American Statistical Association}, 114(526),
553--564. \doi{10.1080/01621459.2018.1482750}
}
\seealso{
\link{dti_df} for another example dataset,
\code{vignette("x04_Visualization", package = "tidyfun")} for usage examples.

Other tidyfun datasets: 
\code{\link{dti_df}}
}
\concept{tidyfun datasets}
\keyword{data}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tidyr.R
\name{tf_gather}
\alias{tf_gather}
\title{Gather all columns representing functional measurements into a \code{tfd}-object}
\usage{
tf_gather(
  data,
  ...,
  key = ".tfd",
  arg = NULL,
  domain = NULL,
  evaluator = tf_approx_linear
)
}
\arguments{
\item{data}{a data frame -- note that \code{dplyr} does not handle matrix columns well,
if \code{data} contains more than one of those, \code{tf_gather} will fail...}

\item{...}{A selection of columns to collect as a \code{tfd} object. Each column
represents measurements of a functional variable at a specific \code{arg}-val.
Can also be the name of a matrix-valued column, but see above.
If empty, all variables are selected. You can supply bare variable names,
select all variables between x and z with x:z, exclude y with -y. For more
options, see the \code{\link[dplyr:select]{dplyr::select()}} documentation.}

\item{key}{the name of the created \code{tfd}-column. Defaults to \code{".tfd"}, but
the function will try to guess the name based on the column names of the
gathered columns in this case. If a common prefix of all column names
is found, this is used instead. You also get a message about this.}

\item{arg}{optional. Argument values for the functions. If not provided, will be guessed from the column names as well.
See also \code{\link[tf:tfd]{tf::tfd()}}.}

\item{domain}{optional. Range of possible \code{arg}-values. See \code{\link[tf:tfd]{tf::tfd()}} for details.}

\item{evaluator}{optional. A function accepting arguments x, arg, evaluations. See \code{\link[tf:tfd]{tf::tfd()}} for details.}
}
\value{
a modified \code{data.frame} with a \code{tfd} column replacing the \code{...}.
}
\description{
Similar in spirit to \code{\link[tidyr:gather]{tidyr::gather()}}, but does NOT put the values in the
gathered columns into one very long "value"-column while labeling the different
original columns in a very long "key"-column -- instead it creates a \code{tfd}-column
containing the functional measurements of the columns given in \code{...}.
}
\examples{
d <- tf_spread(growth[1:5, ]) |> dplyr::mutate(id = 1:dplyr::n())
dplyr::glimpse(d)
# tidyselect syntax for column selection:
tf_gather(d, starts_with("height"))
tf_gather(d, height_1:height_18)
tf_gather(d, -gender, -id)
# custom key name and arg values:
tf_gather(d, starts_with("height"), key = "height")
tf_gather(d, starts_with("height"), arg = seq(0, 1, length.out = 31))
}
\seealso{
\code{\link[dplyr:select]{dplyr::select()}}

Other tidyfun data wrangling functions: 
\code{\link{tf_evaluate.data.frame}()},
\code{\link{tf_nest}()},
\code{\link{tf_spread}()},
\code{\link{tf_unnest}()}
}
\concept{tidyfun data wrangling functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom-spaghetti.R
\docType{data}
\name{ggspaghetti}
\alias{ggspaghetti}
\alias{StatTf}
\alias{stat_tf}
\alias{geom_spaghetti}
\alias{GeomSpaghetti}
\alias{geom_meatballs}
\alias{GeomMeatballs}
\title{Spaghetti plots for \code{tf} objects}
\usage{
stat_tf(
  mapping = NULL,
  data = NULL,
  geom = "spaghetti",
  position = "identity",
  na.rm = TRUE,
  show.legend = NA,
  inherit.aes = TRUE,
  arg = NULL,
  ...
)

geom_spaghetti(
  mapping = NULL,
  data = NULL,
  position = "identity",
  na.rm = TRUE,
  show.legend = NA,
  inherit.aes = TRUE,
  arg = NULL,
  ...
)

geom_meatballs(
  mapping = NULL,
  data = NULL,
  position = "identity",
  na.rm = TRUE,
  show.legend = NA,
  inherit.aes = TRUE,
  arg = NULL,
  spaghetti = TRUE,
  ...
)
}
\arguments{
\item{mapping}{Set of aesthetic mappings created by \code{\link[ggplot2:aes]{aes()}}. If specified and
\code{inherit.aes = TRUE} (the default), it is combined with the default mapping
at the top level of the plot. You must supply \code{mapping} if there is no plot
mapping.}

\item{data}{The data to be displayed in this layer. There are three
options:

If \code{NULL}, the default, the data is inherited from the plot
data as specified in the call to \code{\link[ggplot2:ggplot]{ggplot()}}.

A \code{data.frame}, or other object, will override the plot
data. All objects will be fortified to produce a data frame. See
\code{\link[ggplot2:fortify]{fortify()}} for which variables will be created.

A \code{function} will be called with a single argument,
the plot data. The return value must be a \code{data.frame}, and
will be used as the layer data. A \code{function} can be created
from a \code{formula} (e.g. \code{~ head(.x, 10)}).}

\item{geom}{The geometric object to use to display the data for this layer.
When using a \verb{stat_*()} function to construct a layer, the \code{geom} argument
can be used to override the default coupling between stats and geoms. The
\code{geom} argument accepts the following:
\itemize{
\item A \code{Geom} ggproto subclass, for example \code{GeomPoint}.
\item A string naming the geom. To give the geom as a string, strip the
function name of the \code{geom_} prefix. For example, to use \code{geom_point()},
give the geom as \code{"point"}.
\item For more information and other ways to specify the geom, see the
\link[ggplot2:layer_geoms]{layer geom} documentation.
}}

\item{position}{A position adjustment to use on the data for this layer. This
can be used in various ways, including to prevent overplotting and
improving the display. The \code{position} argument accepts the following:
\itemize{
\item The result of calling a position function, such as \code{position_jitter()}.
This method allows for passing extra arguments to the position.
\item A string naming the position adjustment. To give the position as a
string, strip the function name of the \code{position_} prefix. For example,
to use \code{position_jitter()}, give the position as \code{"jitter"}.
\item For more information and other ways to specify the position, see the
\link[ggplot2:layer_positions]{layer position} documentation.
}}

\item{na.rm}{remove NAs? defaults to \code{TRUE}}

\item{show.legend}{logical. Should this layer be included in the legends?
\code{NA}, the default, includes if any aesthetics are mapped.
\code{FALSE} never includes, and \code{TRUE} always includes.
It can also be a named logical vector to finely select the aesthetics to
display. To include legend keys for all levels, even
when no data exists, use \code{TRUE}.  If \code{NA}, all levels are shown in legend,
but unobserved levels are omitted.}

\item{inherit.aes}{If \code{FALSE}, overrides the default aesthetics,
rather than combining with them. This is most useful for helper functions
that define both data and aesthetics and shouldn't inherit behaviour from
the default plot specification, e.g. \code{\link[ggplot2:annotation_borders]{annotation_borders()}}.}

\item{arg}{where to evaluate the functions in \code{y}; defaults to the object's
default evaluation grid.}

\item{...}{Other arguments passed on to \code{\link[ggplot2:layer]{layer()}}'s \code{params} argument. These
arguments broadly fall into one of 4 categories below. Notably, further
arguments to the \code{position} argument, or aesthetics that are required
can \emph{not} be passed through \code{...}. Unknown arguments that are not part
of the 4 categories below are ignored.
\itemize{
\item Static aesthetics that are not mapped to a scale, but are at a fixed
value and apply to the layer as a whole. For example, \code{colour = "red"}
or \code{linewidth = 3}. The geom's documentation has an \strong{Aesthetics}
section that lists the available options. The 'required' aesthetics
cannot be passed on to the \code{params}. Please note that while passing
unmapped aesthetics as vectors is technically possible, the order and
required length is not guaranteed to be parallel to the input data.
\item When constructing a layer using
a \verb{stat_*()} function, the \code{...} argument can be used to pass on
parameters to the \code{geom} part of the layer. An example of this is
\code{stat_density(geom = "area", outline.type = "both")}. The geom's
documentation lists which parameters it can accept.
\item Inversely, when constructing a layer using a
\verb{geom_*()} function, the \code{...} argument can be used to pass on parameters
to the \code{stat} part of the layer. An example of this is
\code{geom_area(stat = "density", adjust = 0.5)}. The stat's documentation
lists which parameters it can accept.
\item The \code{key_glyph} argument of \code{\link[ggplot2:layer]{layer()}} may also be passed on through
\code{...}. This can be one of the functions described as
\link[ggplot2:draw_key]{key glyphs}, to change the display of the layer in the legend.
}}

\item{spaghetti}{plot noodles along with meatballs? defaults to TRUE.}
}
\value{
A \code{\link[ggplot2:layer]{ggplot2::layer()}} object for use in a ggplot.
}
\description{
Plots a line for each entry of a \code{tf}-object.
\code{geom_spaghetti} draws a line through each function, and \code{geom_meatballs}
adds points for the evaluated grid values.
}
\details{
"Flipped" aesthetics are not implemented for these geoms.
}
\section{\code{y} aesthetic}{

Mandatory. Used to designate a column of class \code{tf} to be visualized.
}

\examples{
set.seed(1221)
data <- data.frame(col = sample(gl(5, 2)))
data$f <- tf_rgp(10)
data$fi <- tf_jiggle(data$f)
data$fb <- tfb(data$f)
library(ggplot2)
ggplot(data, aes(y = f, color = tf_depth(f))) +
  geom_spaghetti()
ggplot(data, aes(y = fi, shape = col, color = col)) +
  geom_meatballs()
ggplot(data, aes(y = fi)) +
  geom_meatballs(spaghetti = FALSE) +
  facet_wrap(~col)
}
\seealso{
\code{\link[=geom_cappelini]{geom_cappelini()}} for glyph plots, \code{\link[=gglasagna]{gglasagna()}} for heatmaps.

Other tidyfun visualization: 
\code{\link{autoplot.tf}()},
\code{\link{ggcapellini}},
\code{\link{gglasagna}()}

Other tidyfun visualization: 
\code{\link{autoplot.tf}()},
\code{\link{ggcapellini}},
\code{\link{gglasagna}()}

Other tidyfun visualization: 
\code{\link{autoplot.tf}()},
\code{\link{ggcapellini}},
\code{\link{gglasagna}()}

Other tidyfun visualization: 
\code{\link{autoplot.tf}()},
\code{\link{ggcapellini}},
\code{\link{gglasagna}()}

Other tidyfun visualization: 
\code{\link{autoplot.tf}()},
\code{\link{ggcapellini}},
\code{\link{gglasagna}()}

Other tidyfun visualization: 
\code{\link{autoplot.tf}()},
\code{\link{ggcapellini}},
\code{\link{gglasagna}()}

Other tidyfun visualization: 
\code{\link{autoplot.tf}()},
\code{\link{ggcapellini}},
\code{\link{gglasagna}()}
}
\concept{tidyfun visualization}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tidyr.R
\name{tf_unnest}
\alias{tf_unnest}
\alias{tf_unnest.tf}
\alias{tf_unnest.data.frame}
\title{Turn (data frames with) \code{tf}-objects / list columns into "long" tables.}
\usage{
tf_unnest(data, cols, arg, interpolate = TRUE, ...)

\method{tf_unnest}{tf}(data, cols, arg, interpolate = TRUE, ...)

\method{tf_unnest}{data.frame}(
  data,
  cols,
  arg,
  interpolate = TRUE,
  keep_empty = FALSE,
  ptype = NULL,
  names_sep = "_",
  names_repair = "check_unique",
  ...
)
}
\arguments{
\item{data}{a data.frame or a \code{tf}-object}

\item{cols}{<\code{\link[tidyr:tidyr_tidy_select]{tidy-select}}> List-columns to unnest.

When selecting multiple columns, values from the same row will be recycled
to their common size.}

\item{arg}{optional values for the \code{arg} argument of
\code{\link[tf:tf_evaluate]{tf::tf_evaluate()}}}

\item{interpolate}{return function values for \code{arg}-values not on original grid?
Defaults to \code{TRUE}.}

\item{...}{not used currently}

\item{keep_empty}{By default, you get one row of output for each element
of the list that you are unchopping/unnesting. This means that if there's a
size-0 element (like \code{NULL} or an empty data frame or vector), then that
entire row will be dropped from the output. If you want to preserve all
rows, use \code{keep_empty = TRUE} to replace size-0 elements with a single row
of missing values.}

\item{ptype}{Optionally, a named list of column name-prototype pairs to
coerce \code{cols} to, overriding the default that will be guessed from
combining the individual values. Alternatively, a single empty ptype
can be supplied, which will be applied to all \code{cols}.}

\item{names_sep}{If \code{NULL}, the default, the outer names will come from the
inner names. If a string, the outer names will be formed by pasting
together the outer and the inner column names, separated by \code{names_sep}.}

\item{names_repair}{Used to check that output data frame has valid
names. Must be one of the following options:
\itemize{
\item \verb{"minimal}": no name repair or checks, beyond basic existence,
\item \verb{"unique}": make sure names are unique and not empty,
\item \verb{"check_unique}": (the default), no name repair, but check they are unique,
\item \verb{"universal}": make the names unique and syntactic
\item a function: apply custom name repair.
\item \link[tidyr]{tidyr_legacy}: use the name repair from tidyr 0.8.
\item a formula: a purrr-style anonymous function (see \code{\link[rlang:as_function]{rlang::as_function()}})
}

See \code{\link[vctrs:vec_as_names]{vctrs::vec_as_names()}} for more details on these terms and the
strategies used to enforce them.}
}
\value{
a "long" data frame with \code{tf}-columns expanded into \verb{arg, value}-
columns.
}
\description{
Similar in spirit to \code{\link[tidyr:unnest]{tidyr::unnest()}}, the reverse of \code{\link[=tf_nest]{tf_nest()}}.
The \code{tf}-method simply turns a single \code{tfd} or \code{tfb} vector into a "long" \code{\link[tibble:tibble]{tibble::tibble()}}.
}
\details{
\itemize{
\item Caution: this uses slightly different defaults for names of unnested columns
than \code{tidyr::unnest()}.
\item For \code{data.frames}, include an ID column with a unique row identifier before
unnesting. Without it, arg-value pairs cannot be matched back to their
original functions after unnesting.
}
}
\examples{
d <- dplyr::tibble(id = 1:3)
d$f <- tf_rgp(3, 11L)
tf_unnest(d, f)
}
\seealso{
\code{\link[=tf_evaluate.data.frame]{tf_evaluate.data.frame()}}

Other tidyfun data wrangling functions: 
\code{\link{tf_evaluate.data.frame}()},
\code{\link{tf_gather}()},
\code{\link{tf_nest}()},
\code{\link{tf_spread}()}
}
\concept{tidyfun data wrangling functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tf-ggplot.R
\name{ggplot_build.tf_ggplot}
\alias{ggplot_build.tf_ggplot}
\title{ggplot_build method for tf_ggplot}
\usage{
\method{ggplot_build}{tf_ggplot}(plot, ...)
}
\arguments{
\item{plot}{A tf_ggplot object}

\item{...}{Additional arguments passed through from \code{\link[ggplot2:ggplot_build]{ggplot2::ggplot_build()}}.}
}
\value{
A built ggplot object (class \code{ggplot_built}).
}
\description{
ggplot_build method for tf_ggplot
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gglasagna.R
\name{gglasagna}
\alias{gglasagna}
\title{Lasagna plots for \code{tf}s using \code{ggplot2}}
\usage{
gglasagna(
  data,
  tf,
  order = NULL,
  label = NULL,
  arg = NULL,
  order_by = NULL,
  order_ticks = TRUE
)
}
\arguments{
\item{data}{A data frame containing the \code{tf} column to visualize.}

\item{tf}{bare name of the \code{tf} column to visualize}

\item{order}{(optional) bare name of a column in \code{data} to define vertical
order of lasagna layers.}

\item{label}{(optional) bare name of a column in \code{data} to define labels for
lasagna layers. Defaults to names of \code{y}, if present, or row numbers.}

\item{arg}{\code{arg} to evaluate \code{y} on}

\item{order_by}{a function applied to each row in \code{y[, arg]} that must
return a scalar value to define the order of lasagna layers.}

\item{order_ticks}{add horizontal lines indicating borders between levels of
\code{order} (if it is a discrete variable) and labels for its levels?
Defaults to TRUE. Supply a named list to override tick appearance, including
label styling, line type, alpha, rotation, and label placement.
Disable this when faceting; the tick annotations are not designed for faceted layouts.}
}
\value{
a \code{ggplot2} object
}
\description{
Lasagna plots show one color bar for each function.
}
\details{
The vertical order of the lasagna layers is \strong{increasing} in
\itemize{
\item \code{order} (if provided),
\item the values returned by \code{order_by} (if provided),
\item and the row number of the observations.
}

i.e., lowest values are on top so that by default the first layer is the first
observation in \code{data} and the vertical order of the layers is the
ordering of observations obtained by \code{dplyr::arrange(data, order, order_by(value), row_number())}.
}
\examples{
\donttest{
library(ggplot2)
set.seed(1221)
data <- expand.grid(group = factor(1:5), rep = 1:10)
data <- dplyr::mutate(data,
  id = paste(group, rep, sep = "-"),
  f = tf_rgp(50),
  fb = tfb(f)
)

gglasagna(data, f, label = id)
gglasagna(data, fb, label = id, order = group)
# order is lowest first / on top by default
gglasagna(data, f, label = id, order = tf_depth(f))
gglasagna(data, f, label = id, order_by = dplyr::first) +
  facet_wrap(~group, scales = "free")
# order of layers is by "order_by" within "order":
gglasagna(data, fb, label = id, order = group, order_by = dplyr::first)
}
}
\seealso{
Other tidyfun visualization: 
\code{\link{autoplot.tf}()},
\code{\link{ggcapellini}},
\code{\link{ggspaghetti}}
}
\concept{tidyfun visualization}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tidyr.R
\name{tf_spread}
\alias{tf_spread}
\title{Spread a \code{tf}-column into many columns representing the
function evaluations.}
\usage{
tf_spread(data, value, arg, sep = "_", interpolate = FALSE)
}
\arguments{
\item{data}{a data frame with at least one \code{tf}-column}

\item{value}{the name of the \code{tf}-column to 'spread'/evaluate.
You can supply bare variable names etc., see the \code{\link[dplyr:select]{dplyr::select()}}
documentation. Also works without this if there's only one \code{tf} in \code{data},
see examples.}

\item{arg}{(Semi-)optional. A vector of \code{arg}-values on which to evaluate the
functions. If not provided, uses the default \code{arg}s. Should be
specified for \code{tf_irreg}, otherwise \emph{all} observed gridpoints are used for
\emph{every} function.}

\item{sep}{separating character used to create column names for the new columns,
defaults to \code{"_"} for column names "<name of the \code{tf}>_<\code{arg}-value>".
Set to NULL to get column names that only contain the \code{arg}-value.}

\item{interpolate}{\code{interpolate}-argument for evaluating the functional data.
Defaults to FALSE, i.e., \code{tfd}s are \emph{not} inter/extrapolated on unobserved
\code{arg}-values.}
}
\value{
a wider dataframe with the \code{tf}-column spread out into many columns
each containing the functional measurements for one \code{arg}-value.
}
\description{
Similar in spirit to \code{\link[tidyr:spread]{tidyr::spread()}}, but does NOT shorten,
just widens the data frame  -- a \code{tf}-column is spread out into many columns
containing the functional measurements.
}
\examples{
d <- dplyr::tibble(g = 1:3)
d$f <- tf_rgp(3, 11L)
tf_spread(d, f)
tf_spread(d, -g)
tf_spread(d)
}
\seealso{
Other tidyfun data wrangling functions: 
\code{\link{tf_evaluate.data.frame}()},
\code{\link{tf_gather}()},
\code{\link{tf_nest}()},
\code{\link{tf_unnest}()}
}
\concept{tidyfun data wrangling functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tidyfun-package.R
\docType{package}
\name{tidyfun-package}
\alias{tidyfun}
\alias{tidyfun-package}
\title{tidyfun: Tidy Functional Data Wrangling and Visualization}
\description{
\code{tidyfun} makes data wrangling and exploratory analysis for functional data
in \code{R} easier.\cr\cr
The \code{tf} package defines the core functional data classes and methods.
\code{tidyfun} builds on those objects with:
\itemize{
\item data-frame workflows for functional data: \code{\link[=tf_gather]{tf_gather()}} etc.
\item graphics helpers for \code{tf} objects and tidy functional data frames: \code{\link[=geom_spaghetti]{geom_spaghetti()}}, \code{\link[=gglasagna]{gglasagna()}}
\item documentation and examples that show how to use \code{tf} objects in tidy workflows
\cr\cr
\cr
Start with \code{vignette(package = "tidyfun")} for the article index.
}
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/tidyfun/tidyfun}
  \item \url{https://tidyfun.github.io/tidyfun/}
  \item Report bugs at \url{https://github.com/tidyfun/tidyfun/issues}
}

}
\author{
\strong{Maintainer}: Fabian Scheipl \email{fabian.scheipl@googlemail.com} (\href{https://orcid.org/0000-0001-8172-3603}{ORCID}) [copyright holder]

Authors:
\itemize{
  \item Jeff Goldsmith
  \item Julia Wrobel (\href{https://orcid.org/0000-0001-6783-1421}{ORCID})
}

Other contributors:
\itemize{
  \item Maximilian Mücke (\href{https://orcid.org/0009-0000-9432-9795}{ORCID}) [contributor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom-fboxplot.R
\docType{data}
\name{ggfboxplot}
\alias{ggfboxplot}
\alias{StatFboxplot}
\alias{stat_fboxplot}
\alias{GeomFboxplot}
\alias{geom_fboxplot}
\title{Functional boxplots for \code{tf} objects}
\usage{
stat_fboxplot(
  mapping = NULL,
  data = NULL,
  geom = "fboxplot",
  position = "identity",
  ...,
  orientation = NA,
  coef = 1.5,
  na.rm = FALSE,
  show.legend = NA,
  inherit.aes = TRUE,
  depth = "MBD",
  depth_fn = NULL,
  fence_fn = NULL,
  central = 0.5,
  arg = NULL
)

geom_fboxplot(
  mapping = NULL,
  data = NULL,
  stat = "fboxplot",
  position = "identity",
  ...,
  outliers = TRUE,
  outlier.colour = NULL,
  outlier.color = NULL,
  outlier.fill = NULL,
  outlier.shape = NULL,
  outlier.size = NULL,
  outlier.stroke = 0.5,
  outlier.alpha = NULL,
  whisker.colour = NULL,
  whisker.color = NULL,
  whisker.linetype = NULL,
  whisker.linewidth = NULL,
  staple.colour = NULL,
  staple.color = NULL,
  staple.linetype = NULL,
  staple.linewidth = NULL,
  median.colour = NULL,
  median.color = NULL,
  median.linetype = NULL,
  median.linewidth = NULL,
  box.colour = NULL,
  box.color = NULL,
  box.linetype = NULL,
  box.linewidth = NULL,
  notch = FALSE,
  notchwidth = 0.5,
  staplewidth = 0,
  varwidth = FALSE,
  na.rm = FALSE,
  orientation = NA,
  show.legend = NA,
  inherit.aes = TRUE,
  depth = "MBD",
  depth_fn = NULL,
  fence_fn = NULL,
  central = 0.5,
  arg = NULL
)
}
\arguments{
\item{mapping, data, position, show.legend, inherit.aes, na.rm, orientation, ...}{Passed on to \code{\link[ggplot2:layer]{ggplot2::layer()}}.}

\item{coef}{Inflation factor for the central envelope used to define outer fences, defaults to 1.5.}

\item{depth}{Character scalar naming the built-in depth measure passed to
\code{\link[tf:tf_depth]{tf::tf_depth()}} when \code{depth_fn} is \code{NULL}.}

\item{depth_fn}{Optional custom depth function. Must return one numeric depth
value per function.}

\item{fence_fn}{Optional custom fence function. Must return a list with
elements \code{lower}, \code{upper}, and \code{outliers}.}

\item{central}{Fraction of deepest curves used to construct the central
envelope. Defaults to \code{0.5}.}

\item{arg}{Optional evaluation grid used for depth calculation, envelopes, and
drawing. Defaults to the natural grid of the mapped \code{tf}.}

\item{stat, geom}{Use the functional boxplot stat/geom.}

\item{outliers}{Should outlying curves be drawn?}

\item{outlier.colour, outlier.color, outlier.fill, outlier.shape, outlier.size}{Styling parameters for outlier curves.}

\item{outlier.stroke}{Accepted for interface compatibility with
\code{\link[ggplot2:geom_boxplot]{ggplot2::geom_boxplot()}} but ignored because functional outliers are drawn
as curves.}

\item{outlier.alpha}{Alpha used for outlier curves.}

\item{whisker.colour, whisker.color, whisker.linetype, whisker.linewidth}{Styling
parameters for fence lines.}

\item{staple.colour, staple.color, staple.linetype, staple.linewidth}{Accepted for
interface compatibility with \code{\link[ggplot2:geom_boxplot]{ggplot2::geom_boxplot()}} but currently unused.}

\item{median.colour, median.color, median.linetype, median.linewidth}{Styling
parameters for the median curve.}

\item{box.colour, box.color, box.linetype, box.linewidth}{Styling parameters for
the central band outline.}

\item{notch, notchwidth, staplewidth, varwidth}{Accepted for interface
compatibility with \code{\link[ggplot2:geom_boxplot]{ggplot2::geom_boxplot()}} but currently unused.}
}
\value{
A ggplot2 layer.
}
\description{
Draw functional boxplots based on functional depth rankings.
}
\details{
\code{stat_fboxplot()} computes a median curve (thick line), a central region
envelope (filled ribbon), functional fences (dashed lines), and optional
outlying curves (defined as "exceeds the fences somewhere", plotted as solid lines)
from a \code{tf} column. \code{geom_fboxplot()} draws these summaries as a band plus
line layers.

The interface intentionally follows \code{\link[ggplot2:geom_boxplot]{ggplot2::stat_boxplot()}} and
\code{\link[ggplot2:geom_boxplot]{ggplot2::geom_boxplot()}} where this is meaningful for functional data.
Use \code{aes(tf = f)} to map a \code{tf} column. Separate functional boxplots inside a
panel are defined by \code{group}, \code{colour}/\code{color}, or \code{fill}; separate panels are
handled through facetting. Note that only \code{color} or \code{fill} need to be
specified explicitly, as the other will be automatically mapped to the same
variable if not provided.
}
\examples{
\donttest{
library(ggplot2)
set.seed(1312)
data <- data.frame(id = 1:50, grp = rep(c("A", "B"), each = 25))
data$f <- tf_rgp(50) + 5 * (data$grp == "A")

tf_ggplot(data, aes(tf = f, fill = grp)) + # same as `colour = grp` here!
  geom_fboxplot(alpha = 0.3)

tf_ggplot(data, aes(tf = f)) +
  geom_fboxplot(orientation = "y")
}
}
\keyword{datasets}
